#include <stdio.h>
#include <string.h>
#include <sicstus/sicstus.h>

#include "ltlc.h"

void ltlc_modelcheck(SP_term_ref closure,
                     SP_term_ref startnodes,
                     SP_term_ref predicate,
                     SP_term_ref update,
                     SP_term_ref get_transids_pred,
                     SP_term_ref get_enabled_actions_pred,
                     SP_term_ref weak_fairness_args,
                     SP_term_ref strong_fairness_args,
                     SP_term_ref main_formula,
                     SP_term_ref result);

static const int callback_predicate_arity = 5;
static const int update_predicate_arity = 5;
static const int get_transition_ids_predicate_arity = 3;
static const int get_enabled_actions_predicate_arity = 2;

static int init_predicates(ltl_formula *f, SP_term_ref callback, SP_term_ref update,
                           SP_term_ref get_transids_pred, SP_term_ref get_enabled_actions_pred);
static void init_fairness_all(ltl_formula *f, SP_term_ref weak_fairness_all, SP_term_ref strong_fairness_all);
static int lookup_predicate(SP_term_ref reference, int arity, const char *desc, SP_pred_ref *result);
static int pl_convert_closure(SP_term_ref list, ltl_formula *f);
static int check_name_and_arity(const char *name, int must, const char *functor, int is, int *failure);
static int get_integer_arg(const char *msg, SP_term_ref term, int arg, int *value);
static int get_pform_ref(const char *msg, SP_term_ref term, ltl_pform_ref *ref);
static int get_ref_arg(const char *msg, SP_term_ref term, int arg, ltl_pform_ref *sref);
SP_term_ref *prolog_list_to_array(const SP_term_ref list, int *size);
static void create_prolog_solution(SP_term_ref result, ltl_formula *f, modelitem *path, int length, int loop_entry);
static void delete_found_scc(found_scc_type *found_scc);
static void get_start_nodes(ltl_formula *f,SP_term_ref list, ltlnode ***nodes, int *size);
static void process_tarjan_result(ltl_formula *f,found_scc_type *found_scc,SP_term_ref result);
static void start_modelcheck(ltl_formula *f,SP_term_ref startnodes, SP_term_ref result);
static SP_term_ref create_loop_indicator(int loop_entry, int is_loop);
static SP_term_ref get_node_id(ltlnode *node);

void ltlc_modelcheck(SP_term_ref closure,
                     SP_term_ref startnodes,
                     SP_term_ref predicate,
                     SP_term_ref update,
                     SP_term_ref get_transids_pred,
                     SP_term_ref get_enabled_actions_pred,
                     SP_term_ref weak_fairness_args,
                     SP_term_ref strong_fairness_args,
                     SP_term_ref main_formula,
                     SP_term_ref result) {

  init_mallocs_counter();
  ltl_formula f;

  IFLTLTIMING( f.callback_calls = 0;
               f.callback_time = 0;
               clock_t timer_start = clock();
               f.timer_current = timer_start;
               f.rejected_sccs = 0;
               f.fairness_callback_calls = 0;
               f.fairness_callback_time = 0;
               f.fairness_checking_time = 0;);

  IFFAIRNESSTIMING(f.cpp_time_spent = 0;
                   f.weak_fairness_time = 0;
                   f.strong_fairness_time = 0;);

  if (init_predicates(&f, predicate, update, get_transids_pred, get_enabled_actions_pred)
      && pl_convert_closure(closure, &f) 
      && !get_pform_ref("main", main_formula, &(f.mainref))) {

    init_fairness_all(&f,weak_fairness_args,strong_fairness_args);
    f.weak_fairness_arguments = weak_fairness_args;
    f.strong_fairness_arguments = strong_fairness_args;
    start_modelcheck(&f,startnodes,result);
  
  }
  
  IFLTLTIMING( clock_t total_time = clock() - timer_start;
               printf("LTL model checking: total time %lims, %i callbacks needed %lims, netto %lims.\n",
                      1000 * total_time / CLOCKS_PER_SEC, 
                      f.callback_calls,
                      1000 * f.callback_time / CLOCKS_PER_SEC,
                      1000 * (total_time  - f.callback_time - f.fairness_checking_time) / CLOCKS_PER_SEC););
  if (f.check_found_scc_for_fairness_activated) {
     IFLTLTIMING(  printf("LTL model checking (fairness): %i strongly connected components were rejected.\n", f.rejected_sccs);
                   printf("LTL model checking (fairness): total fairness checking time %lims, %i callbacks needed %lims, netto %lims.\n",
                      1000 * f.fairness_checking_time / CLOCKS_PER_SEC,
                      f.fairness_callback_calls,
                      1000 * f.fairness_callback_time / CLOCKS_PER_SEC,
                      1000 * (f.fairness_checking_time - f.fairness_callback_time)/CLOCKS_PER_SEC););
     IFFAIRNESSTIMING(printf("LTL model checking (fairness): time spent on C++ code %lims\n", 1000 * f.cpp_time_spent / CLOCKS_PER_SEC););
     IFFAIRNESSTIMING(printf("LTL model checking (fairness): time spent on checking weak fairness %lims\n", 1000 * f.weak_fairness_time / CLOCKS_PER_SEC););
     IFFAIRNESSTIMING(printf("LTL model checking (fairness): time spent on checking strong fairness %lims\n", 1000 * f.strong_fairness_time / CLOCKS_PER_SEC););
  }

  delete_formula(&f);

  /* check if no memory leak is present */
  ltl_check_memory();
}

/* Trivial formula: true or false, we have two cases
 * a) false: is always false
 * b) true: is true if a start node exists. A counter-example consists of the start node
 */
static void check_trivial_formula(ltl_formula *f,ltlnode **nodes,int nodenumber, SP_term_ref result) {
  if (f->mainref.ref != -1) {
    fprintf(stderr,"Assertion error: called trivial formula code for non-trivial formula\n");
    exit(1);
  }
  int formula_is_true = f->mainref.signum;
  if (formula_is_true && nodenumber>0) {
    /* We have a model: just the first node
     * We manually construct the Prolog term:
     *   model([atom(NodeId,[],none)],no_loop)
     */
    /* Prolog term: atom(NodeId,[],none) */
    SP_term_ref none = SP_new_term_ref();
    SP_cons_functor(none,SP_atom_from_string("none"),0);
    SP_term_ref emptylist = SP_new_term_ref();
    SP_term_ref atom = SP_new_term_ref();
    SP_cons_functor(atom,SP_atom_from_string("atom"),
                    3,get_node_id(nodes[0]),emptylist,none);
    /* Prolog term: [atom(NodeId,[],none)] */
    SP_term_ref path = SP_new_term_ref();
    SP_cons_list(path,atom,emptylist);
    /* Prolog term: model([atom(NodeId,[],none)],no_loop) */
    SP_term_ref loop = create_loop_indicator(-1,0);
    SP_cons_functor(result,SP_atom_from_string("model"),2,path,loop);
  } else {
    /* formula "false" or no node exists, return "none",
     * we do not use create_prolog_solution() to prevent the result "incomplete"
     */
    SP_cons_functor(result, SP_atom_from_string("none"), 0);
  }
}

static void start_modelcheck(ltl_formula *f,SP_term_ref startnodes, SP_term_ref result) {
  ltlnode **nodes;
  int nodenumber;
  size_t tarjan_used_mem = 0;
  found_scc_type *found_scc = 0;

  IFLTLDEBUG(ltlmsg("init formula\n"););
  init_formula(f);
  IFLTLDEBUG(ltlmsg("size of node: %zi bytes\n", sizeof(ltlnode));
             ltlmsg("size of atom: %zi bytes\n", sizeof(ltlatom)+get_bitfield_memsize(f->closure_size));
             ltlmsg("size of node trans: %zi bytes\n", sizeof(ltlnodetrans)+get_bitfield_memsize(f->number_of_tps));
             ltlmsg("size of atom trans: %zi bytes\n", sizeof(ltlatom*)););
  IFLTLDEBUG(ltlmsg("get first node\n"););
    
  /*if (is_safety_property(&f,&mainref)) {
    printf("LTL model checking: Formula is a safety property (this has no effect yet)\n");
    }*/

  get_start_nodes(f,startnodes,&nodes,&nodenumber);
  IFLTLDEBUG(ltlmsg("call start_tarjan\n"););
  if (f->closure_size>0) {
    if (start_tarjan(f, nodes, nodenumber, &tarjan_used_mem, &found_scc)) {
        process_tarjan_result(f,found_scc,result);
    } else if (f->canceled) {
      SP_cons_functor(result, SP_atom_from_string("canceled"), 0);
    }
  } else {
    check_trivial_formula(f,nodes,nodenumber,result);
  }
  
  if (found_scc) {
      delete_found_scc(found_scc);
      ltl_free(found_scc);
  }

  printf("LTL model checking: memory usage approx. %zi KiB, ",
         (ltl_formula_memory_usage(f) + tarjan_used_mem) / 1024);
  printf("%i atoms and %i transitions generated\n",
         f->number_of_atoms, f->number_of_atomtrans);
  ltl_free(nodes);
}

static void process_tarjan_result(ltl_formula *f,found_scc_type *found_scc,SP_term_ref result) {
    IFLTLDEBUG(if (found_scc)
                 ltlmsg("tarjan finished, solution found (path size %i, scc size %i).\n",
                        found_scc->number_of_path_elements,
                        found_scc->number_of_scc_atoms);
               else ltlmsg("tarjan finished, no solution found\n");

               ltlmsg("print graph\n");
               ltl_print_graph(f, found_scc, -1); );
    //ltl_snapshot(f);      
    //ltl_print_prolog_graph(f,found_scc,0,0);

    modelitem *model = 0;
    int ce_length = 0;
    int is_deadlock_path = 0;
    int loop_entry = -3;
    int loop_length;
      
    if (found_scc) {
      IFLTLDEBUG( ltlmsg("find path through SCC.\n"); );
      get_path_through_scc(f, found_scc, &model, &ce_length, &is_deadlock_path, &loop_length);
      IFLTLDEBUG( ltlmsg("delete found SCC.\n"); );
      int shorter = shorten_path(f,model,ce_length);
      IFLTLDEBUG( if(shorter<0) {
          ltlmsg("counter-example has not been shortened\n");
        } else {
          ltlmsg("model shortened from size %i to %i\n",
                 ce_length, shorter); } );
      if (shorter>=0) {
        /* the path has been shortened, thus it cannot be a loop
           and it cannot end in a deadlock */
        loop_entry = -1;
        ce_length = shorter;
      } else if (is_deadlock_path) {
        /* or it is a deadlock path wich neither can be a path */
        loop_entry = -2;
      } else {
        /* this is a loop, we calculate the entry point of the loop by
           using the length of the loop.
        */
        loop_entry = ce_length - loop_length;
      }
    }

    IFLTLDEBUG( ltlmsg("create prolog solution.\n"); );
    create_prolog_solution(result, f, model, ce_length, loop_entry);
    IFLTLDEBUG( ltlmsg("prolog solution created.\n"); );
    if (model) ltl_free(model);
}

void get_start_nodes(ltl_formula *f,SP_term_ref list, ltlnode ***nodes, int *size) {
  SP_term_ref *terms = prolog_list_to_array(list, size);
  int i;
  ltlnode **array;
  if (size>0) {
    array = (ltlnode **) ltl_malloc(sizeof(ltlnode*)* (*size));
    //print_number_of_open_mallocs("get_start_nodes 1");
    for (i=0; i<(*size); i++) {
      get_node_by_term(&f->nstore, terms[i], &(array[i]));
    }
    //print_number_of_open_mallocs("get_start_nodes 2");
    *nodes = array;
  } else {
    *nodes = 0;
  }
  ltl_free(terms);
}

static SP_term_ref create_list_for_bitfield(bitfield *bitfield, int size) {
  SP_term_ref tail;
  int i;
  tail = SP_new_term_ref();
  for (i=size-1; i>=0; i--) {
    SP_term_ref list = SP_new_term_ref();
    SP_term_ref head = SP_new_term_ref();
    SP_put_integer(head, getbit(bitfield, i) ? 1 : 0);
    SP_cons_list(list, head, tail);
    tail = list;
  }
  return tail;
}

static SP_term_ref get_node_id(ltlnode *node) {
  SP_term_ref result;
  switch (node->type) {
  case term_node: result = node->id.ref; break;
  case int_node: result = SP_new_term_ref(); SP_put_integer(result, node->id.id); break;
  default: fprintf(stderr, "fatal: unexpected node type\n"); exit(1);
  }
  return result;
}

/*
 creates a prolog term atom(NodeId,Evaluation,TransId), where
   - 'NodeId' is the node id to which the 'atom' belongs to in the state space,
   - 'Evaluation' is the evaluation of the subformulae of the closure of the chcked LTL formula
   - 'TransId' is the transition id to the successor atom (node) in the state space; 
     if 'atom' happens to have no successor atom, then 'TransId' will be denoted by 'none'
 */
static SP_term_ref create_prolog_atom(ltl_formula *f, modelitem *src, int loop_entry, modelitem *after) {
  SP_term_ref atomrep = SP_new_term_ref();
  SP_atom name = SP_atom_from_string("atom");
  SP_term_ref bits = create_list_for_bitfield(get_atom_evaluation(src->atom), f->closure_size);
  SP_cons_functor(atomrep, name, 3, 
                  get_node_id(src->atom->node),
                  bits,
                  find_transition_operation_with_transition_id(f, src, loop_entry, after));
  return atomrep;
}

static SP_term_ref create_atom_list(ltl_formula *f, modelitem *items, int size, int is_loop, int loop_entry) {
  SP_term_ref tail;
  int i,startindex = size-1;
  modelitem *next_item;
  tail = SP_new_term_ref();
  int dest_is_loop_entry;
  for (i=startindex; i>=0; i--) {
    dest_is_loop_entry = 0;
    if (i<startindex) {
      next_item = &items[i+1];
    } else {
      /* the last atom is reached */
      if (is_loop) {
        /* in case of a loop the loop entry is the next atom */
        next_item = &items[loop_entry];
        dest_is_loop_entry = 1;
      } else {
        /* there is no state after this node */
        next_item = 0;
      }
    }

    SP_term_ref list = SP_new_term_ref();
    SP_term_ref head = create_prolog_atom(f, &items[i], dest_is_loop_entry, next_item);
    SP_cons_list(list, head, tail);
    tail = list;
  }
  return tail;
}

/* create a Prolog term:
 - no_loop, if loop_entry is -1, or
 - deadlock, if loop_entry is -2, or
 - loop(loop_entry), if loop_entry is positive
*/
static SP_term_ref create_loop_indicator(int loop_entry, int is_loop) {
  SP_term_ref loop = SP_new_term_ref();
  SP_term_ref entry;
  if (is_loop) {
    entry = SP_new_term_ref();
    SP_put_integer(entry, (long) loop_entry);
    SP_cons_functor(loop, SP_atom_from_string("loop"), 1, entry);
  } else {
    switch (loop_entry) {
    case -1: 
      SP_put_atom(loop, SP_atom_from_string("no_loop"));
      break;
    case -2:
      SP_put_atom(loop, SP_atom_from_string("deadlock"));
      break;
    default:
      fprintf(stderr, "invalid value for loop_entry: %i, exiting.\n", loop_entry);
      exit(1);
    }
  }
  return loop;
}

void create_prolog_solution(SP_term_ref result, ltl_formula *f, modelitem *model, int ce_length, int loop_entry) {
  int is_loop = loop_entry >= 0;
  if (model) {
    SP_term_ref path = create_atom_list(f,model, ce_length, is_loop, loop_entry);
    SP_term_ref loop = create_loop_indicator(loop_entry, is_loop);
    SP_cons_functor(result, SP_atom_from_string("model"), 2, path, loop);
  } else {
    const char *name;
    if (f->unexplored_nodes)
      name = "incomplete";
    else if (f->canceled)
      name = "canceled";
    else
      name = "none";
    SP_cons_functor(result, SP_atom_from_string(name), 0);
  }
}

void delete_found_scc(found_scc_type *found_scc) {
  if (found_scc) {
    ltl_free(found_scc->scc);
    ltl_free(found_scc->path);
    found_scc->scc = 0;
    found_scc->path = 0;
  }
}

static SP_term_ref create_proposition_list(SP_term_ref *formulas, int start, int size) {
  SP_term_ref list = SP_new_term_ref();
  SP_term_ref head,tail;
  int i,ok;

  for (i=size-1; i>=0; i--) {
    head = SP_new_term_ref();
    tail = list;
    list = SP_new_term_ref();
    ok = SP_get_arg(1, formulas[start+i], head);
    ok &= SP_cons_list(list, head, tail);
    if (!ok) {
      fprintf(stderr, "Illegal state: could create list of propositions.\n");
      exit(1);
    }
  }

  return list;
}

typedef enum {pmode_ap, pmode_tp, pmode_next, pmode_nextnot, pmode_yesterday, pmode_yesterdaynot, pmode_generic} pmode;

static int checkmode(pmode *mode, pmode newmode, int *failure, const char *desc) {
  if (!(*failure)) {
    if (*mode <= newmode)
      *mode = newmode;
    else
      *failure = ltl_raise("wrong position of %s in closure", desc);
  }
  return !(*failure);
}

int pl_convert_closure(SP_term_ref list, ltl_formula *f) {
  SP_term_ref *formulas;
  SP_atom functor;
  ltl_part_formula *res;
  int i,arity,number_of_nextnots,number_of_yesterdaynots;

  const char *name;
  pmode mode;
  int failure = 0;

  res = 0;
  formulas = prolog_list_to_array(list, &(f->closure_size));
  IFLTLDEBUG( ltldeb_inc(); ltlmsg("closure size is %i\n", f->closure_size); );
  f->number_of_aps = 0;
  f->number_of_tps = 0;
  f->number_of_nexts = 0;
  f->number_of_yesterdays = 0;
  f->number_of_generics = 0;
  number_of_nextnots = 0;
  number_of_yesterdaynots = 0;
  if (formulas) {
    res = ltl_malloc(sizeof(ltl_part_formula) * f->closure_size);
    if (res) {
      mode = pmode_ap;
      for (i=0; i<f->closure_size && (!failure); i++) {
        if (SP_get_functor(formulas[i], &functor, &arity)) {
          name = SP_string_from_atom(functor);

          if (check_name_and_arity("ap", 1, name, arity, &failure)) {
            if(checkmode(&mode, pmode_ap, &failure, "atomic proposition")) {
              res[i].op = ltlap;
              f->number_of_aps++;
            }
          }
          else if (check_name_and_arity("tp", 1, name, arity, &failure)) {
            if(checkmode(&mode, pmode_tp, &failure, "transition proposition")) {
              res[i].op = ltltp;
              f->number_of_tps++;
            }
          }
          else if (check_name_and_arity("next", 1, name, arity, &failure)) {
            if (checkmode(&mode, pmode_next, &failure, "next")) {
              res[i].op = ltlnext;
              failure = get_integer_arg("next", formulas[i], 1, &(res[i].args.case_next));
              f->number_of_nexts++;
            }
          }
          else if (check_name_and_arity("nextnot", 1, name, arity, &failure)) {
            if (checkmode(&mode, pmode_nextnot, &failure, "nextnot")) {
              res[i].op = ltlnextnot;
              failure = get_integer_arg("nextnot", formulas[i], 1, &(res[i].args.case_nextnot));
              number_of_nextnots++;
            }
          }
          else if (check_name_and_arity("yesterday", 1, name, arity, &failure)) {
            if (checkmode(&mode, pmode_yesterday, &failure, "yesterday")) {
              res[i].op = ltlyesterday;
              failure = get_integer_arg("yesterday", formulas[i], 1, &(res[i].args.case_yesterday));
              f->number_of_yesterdays++;
            }
          }
          else if (check_name_and_arity("yesterdaynot", 1, name, arity, &failure)) {
            if (checkmode(&mode, pmode_yesterdaynot, &failure, "yesterdaynot")) {
              res[i].op = ltlyesterdaynot;
              failure = get_integer_arg("yesterdaynot", formulas[i], 1, &(res[i].args.case_yesterdaynot));
              number_of_yesterdaynots++;
            }
          }
          else if (check_name_and_arity("or", 2, name, arity, &failure)) {
            if (checkmode(&mode, pmode_generic, &failure, "or")) {
              res[i].op = ltlor;
              failure = get_ref_arg("or", formulas[i], 1, &(res[i].args.case_or.a));
              if (!failure)
                failure = get_ref_arg("or", formulas[i], 2, &(res[i].args.case_or.b));
              f->number_of_generics++;
            }
          }
          else if (check_name_and_arity("until", 3, name, arity, &failure)) {
            if (checkmode(&mode, pmode_generic, &failure, "until")) {
              res[i].op = ltluntil;
              failure = get_ref_arg("until", formulas[i], 1, &(res[i].args.case_until.inv));
              if (!failure)
                failure = get_ref_arg("until", formulas[i], 2, &(res[i].args.case_until.goal));
              if (!failure)
                failure = get_integer_arg("until", formulas[i], 3, &(res[i].args.case_until.next));
              f->number_of_generics++;
            }
          }
          else if (check_name_and_arity("since", 3, name, arity, &failure)) {
            if (checkmode(&mode, pmode_generic, &failure, "since")) {
              res[i].op = ltlsince;
              failure = get_ref_arg("since", formulas[i], 1, &(res[i].args.case_since.inv));
              if (!failure)
                failure = get_ref_arg("since", formulas[i], 2, &(res[i].args.case_since.goal));
              if (!failure)
                failure = get_integer_arg("since", formulas[i], 3, &(res[i].args.case_since.yest));
              f->number_of_generics++;
            }
          }
          else failure = ltl_raise("unexpected term %s/%i", name, arity);
        }
        else failure = ltl_raise("SP_get_functor() failed");
      }
    }
    else failure = ltl_raise("malloc() returned NULL in pl_convert_closure().");
  }
  if (!failure) {
    f->index_tps = f->number_of_aps;
    f->index_nexts = f->index_tps + f->number_of_tps;
    f->index_nextnots = f->index_nexts + f->number_of_nexts;
    f->index_yesterdays = f->index_nextnots + f->number_of_nexts;
    f->index_yesterdaynots = f->index_yesterdays + f->number_of_yesterdays;
    f->index_generics = f->index_yesterdaynots + f->number_of_yesterdays;

    f->ap_list = create_proposition_list(formulas, 0, f->number_of_aps);
    f->tp_list = create_proposition_list(formulas, f->index_tps, f->number_of_tps);
  }

  ltl_free(formulas);
  
  if (!failure)
    failure = !ltl_assert(f->number_of_nexts == number_of_nextnots,
                          "next and nextnot do not correspondent");
  if (!failure) 
    failure = !ltl_assert(f->number_of_yesterdays == number_of_yesterdaynots,
                          "yesterday and yesterdaynot do not correspondent");
  if (failure && res) {
    ltl_free(res);
    res = 0;
  }
  f->closure = res;
  IFLTLDEBUG( ltldeb_dec(); );
  return failure ? 0 : 1;
}

int check_name_and_arity(const char *name, int must, const char *functor, int is, int *failure) {
  if (strcmp(name, functor)==0) {
    if (must != is) {
      *failure = ltl_raise("wrong arity %i of '%s' (expected %i)", is, functor, must);
    }
    return must == is;
  } else {
    return 0;
  }
}


int get_integer_arg(const char *msg, SP_term_ref term, int arg, int *value) {
  SP_term_ref ref;
  int failure = 0;
  long v;
  ref = SP_new_term_ref();
  if (SP_get_arg(arg, term, ref)) {
    if (SP_get_integer(ref,&v)) {
      *value = (int) v;
    } else {
      failure = ltl_raise("failed to convert arg %i to number for: %s", arg, msg);
    }
  } else {
    failure = ltl_raise("not able to get arg %i for: %s", arg, msg);
  }
  return failure;
}

static int get_ref_signum(SP_atom sgn, unsigned int *signum) {
  const char *str = SP_string_from_atom(sgn);
  int failure = 0;
  if(strcmp("pos", str)==0) {
    *signum = 1;
  } else if (strcmp("neg", str)==0) {
    *signum = 0;
  } else {
    failure = ltl_raise("signum '%s' not valid", str);
  }
  return failure;
}

static int get_pform_ref(const char *msg, SP_term_ref term, ltl_pform_ref *ref) {
  SP_atom name;
  int arity;
  int failure = 0;
  if (SP_get_functor(term, &name, &arity)) {
    if (arity == 1) {
      failure = get_ref_signum(name, &(ref->signum));
      if (!failure) {
        failure = get_integer_arg(msg, term, 1, &(ref->ref));
      }
    } else {
      failure = ltl_raise("wrong arity (%i, but 1 expected) of closure reference argument for %s",
                          arity, msg);
    }
  } else {
    failure = ltl_raise("not able to get functor of closure reference argument for %s", msg);
  }
  return failure;
}

static int get_ref_arg(const char *msg, SP_term_ref term, int arg, ltl_pform_ref *sref) {
  SP_term_ref ref = SP_new_term_ref();
  int failure;
  if (SP_get_arg(arg, term, ref)) {
    failure = get_pform_ref(msg, ref, sref);
  } else {
    failure = ltl_raise("not able to get arg %i for %s", arg, msg);
  }
  return failure;
}


SP_term_ref *prolog_list_to_array(SP_term_ref list, int *sz) {
  SP_term_ref tail,elem,*array = 0;
  int i;

  tail = SP_new_term_ref();
  elem = SP_new_term_ref();

  SP_put_term(tail,list);
  int size = 0;
  while (SP_get_list(tail,elem,tail)) size++;

  if (size>0) {
    array = (SP_term_ref *) ltl_malloc(sizeof(SP_term_ref) * size);
    if (array) {
      SP_put_term(tail,list);
      for (i=0; i<size; i++) {
        SP_get_list(tail,elem,tail);
        array[i] = SP_new_term_ref();
        SP_put_term(array[i],elem);
      }
    } else {
      fprintf(stderr, "malloc() returned NULL when converting Prolog list to array.\n");
      exit(1);
    }
  }
  *sz = size;
  return array;
}

int option_all(SP_term_ref option, const char* opt_name) {
  SP_atom functor;
  int arity, res;
  SP_get_functor(option, &functor, &arity);
  const char* name = SP_string_from_atom(functor);
  if (strcmp("all",name) == 0) {
     res = 1;
  } else if (strcmp(".",name) == 0 || strcmp("[]",name) == 0) {
     res = 0;
  } else {
    ltl_raise("error: option value %s for %s unknown.\n",name, opt_name);
    res = 0;  
  }
  return res;
}

static void init_fairness_all(ltl_formula *f, SP_term_ref weak_fairness_args, SP_term_ref strong_fairness_args) {
   f->weak_fairness_all_operations = option_all(weak_fairness_args, "all");
   f->strong_fairness_all_operations = option_all(strong_fairness_args, "all");
}

static int init_predicates(ltl_formula *f, SP_term_ref callback, SP_term_ref update,
                           SP_term_ref get_transids_pred, SP_term_ref get_enabled_actions_pred) {
  int status;
  status = lookup_predicate(callback, callback_predicate_arity, "callback", 
                            &(f->callback_predicate));
  if (status != 1) {
    if (status == 2) {
      ltl_raise("callback predicate to explore state space is mandatory");
    }
    return 0;
  }
  status = lookup_predicate(update, update_predicate_arity, "update", &(f->update_predicate));
  if (status == 0) return 0;
  f->update_predicate_activated = status==1;

  status = lookup_predicate(get_transids_pred,get_transition_ids_predicate_arity, "SCC check", &(f->get_transition_ids_predicate));
  if (status == 0) return 0;
  f->check_found_scc_for_fairness_activated = status==1;

  status = lookup_predicate(get_enabled_actions_pred,get_enabled_actions_predicate_arity, "Fairness Check", &(f->get_enabled_actions_predicate));

  return 1;
}


void writeterm(SP_term_ref term) {
  SP_pred_ref print = SP_predicate("write",1,"user");
  SP_pred_ref nl = SP_predicate("nl",0,"user");
  SP_pred_ref flush = SP_predicate("flush_output",0,"user");
  SP_query(print,term);
  SP_query(nl);
  SP_query(flush);
}

void writestring(const char *str) {
  SP_term_ref atom = SP_new_term_ref();
  SP_put_atom(atom, SP_atom_from_string(str));
  writeterm(atom);
}

static int has_module_pred_form(SP_term_ref term, SP_atom *module, SP_atom *predicate) {
  SP_atom functor;
  SP_term_ref module_term = SP_new_term_ref();
  SP_term_ref predicate_term = SP_new_term_ref();
  int arity;
  const char *str;
  if (SP_is_compound(term)) {
    SP_get_functor(term, &functor, &arity);
    str = SP_string_from_atom(functor);
    if (arity==2 && strcmp(str,":")==0) {
      SP_get_arg(1, term, module_term);
      SP_get_arg(2, term, predicate_term);
      if (SP_is_atom(module_term) && SP_is_atom(predicate_term)) {
        SP_get_atom(module_term, module);
        SP_get_atom(predicate_term, predicate);
        return 1;
      }
    }
  }
  return 0;
}

static int is_none(SP_term_ref term) {
  SP_atom functor;
  if (SP_is_atom(term)) {
    SP_get_atom(term, &functor);
    return strcmp(SP_string_from_atom(functor),"none") == 0;
  }
  return 0;
}

static int lookup_predicate(SP_term_ref reference, int arity, const char *desc, SP_pred_ref *result) {
  SP_pred_ref call;
  const char *module_str, *predicate_str;
  SP_atom module, predicate;
  if (has_module_pred_form(reference, &module, &predicate)) {
      call = SP_pred(predicate, arity, module);
      if (call) {
        *result = call;
        return 1;
      } else {
        module_str = SP_string_from_atom(module);
        predicate_str = SP_string_from_atom(predicate);
        ltl_raise("error: %s predicate %s:%s/%i not found.", desc, module_str, predicate_str, arity);
        return 0;
      }
  } else if (is_none(reference)) {
    return 2;
  } else {
    SP_get_functor(reference, &module, &arity);
    module_str = SP_string_from_atom(module);
    ltl_raise("error: expected :/2 for %s predicate, but was %s/%i.", desc, module_str, arity);
    return 0;
  }
}
