#include <stdio.h>
#include <stdlib.h>

#include "ltlc.h"

static ltlnode *get_from_array(nodestore *st, unsigned int id);
static ltlnode *get_from_list(nodestore_item **list, SP_term_ref id);
static void nodearray_init(void*);
static char *printterm(SP_term_ref);

#ifndef DEBUG
#define DEBUG 0
#endif

void init_nodestore(nodestore *st) {
  if (DEBUG) printf("nodestore: init_nodestore called.\n");
  size_t unit = sizeof(ltlnode);
  init_array(&st->neg, unit, nodearray_init, "nodestore_neg");
  init_array(&st->pos, unit, nodearray_init, "nodestore_pos");
  st->list = 0;
  st->offset = -1;
  st->size = 0;
}

void nodearray_init(void *elem) {
  ((ltlnode *) elem)->number_of_atoms = node_uninitialised;
}

void delete_nodestore(nodestore *st) {
  nodestore_item *iter,*next;

  if (DEBUG) printf("nodestore: init_nodestore called, nodestore size=%u.\n", st->size);

  delete_array(&st->pos);
  delete_array(&st->neg);
  
  iter=st->list;
  st->list = 0;
  while (iter) {
    next = iter->next;
    iter->next = 0;
    ltl_free(iter);
    iter = next;
  }

  st->offset = -1;
}

int get_node_by_term(nodestore *st, SP_term_ref id, ltlnode **node) {
  ltl_nodetype type;
  ltl_nodeid nodeid;
  long long_nodeid;
  if (DEBUG>2) printf("nodestore: get_node_by_term(%i) called.\n", id);
  if (SP_is_integer(id)) {
    type = int_node;
    if (SP_get_integer(id, &long_nodeid)) {
      /* TODO: handle overflow */
      nodeid.id = (unsigned int) long_nodeid;
    } else {
      fprintf(stderr, "SP_get_integer failed!\n");
      exit(1);
    }
  } else {
    type = term_node;
    nodeid.ref = SP_new_term_ref();
    SP_put_term(nodeid.ref, id);
  }
  return get_node_by_id(st, type, nodeid, node);
}

int get_node_by_id(nodestore *st, ltl_nodetype type, ltl_nodeid id, ltlnode **node) {
  int node_is_new;
  if (DEBUG>2) printf("nodestore: get_node_by_id(%s, %u) called.\n", 
                      type==int_node?"int":"term", 
                      type==int_node?id.id:id.ref);
  if (type == int_node) {
    *node = get_from_array(st, id.id);
  } else {
    *node = get_from_list(&st->list, id.ref);
  }
  node_is_new = (*node)->number_of_atoms == node_uninitialised;
  if (node_is_new) {
    st->size++;
    (*node)->number_of_atoms = node_not_opened;
    (*node)->type = type;
    (*node)->id = id;
  }
  if (DEBUG>2) printf("nodestore: get_node returns %s node.\n", (node_is_new ? "new" : "old"));
  return node_is_new;
}

ltlnode *get_from_array(nodestore *st, unsigned int nodeid) {
  ltlarray *array;
  if (DEBUG>2) printf("nodestore:   get_from_array(%u) called\n", nodeid);
  if (st->offset == -1) {
    st->offset = (nodeid >> array_pagesize_log2) << array_pagesize_log2;
    if (DEBUG>1) printf("nodestore:   get_node_by_id: first call, offset set to %li.\n", st->offset);
  }
  if (nodeid >= st->offset) {
    if (DEBUG>2) printf("nodestore:   get_node_by_id: using positive array.\n");
    array = &st->pos;
    nodeid = nodeid - ((unsigned int) st->offset);
  } else {
    if (DEBUG>2) printf("nodestore:   get_node_by_id: using negative array.\n");
    array = &st->neg;
    nodeid = ((unsigned int) st->offset) - nodeid;
  }
  if (assure_arraysize(array, nodeid+1)) {
    return (ltlnode *) get_array_element(array, nodeid);
  } else {
    return 0;
  }
}

ltlnode *get_from_list(nodestore_item **list, SP_term_ref id) {
  nodestore_item *iter;
  int counter = 0;
  if (DEBUG>1) printf("nodestore:   get_from_list(%s) called\n", printterm(id));

  for (iter = *list; iter; iter = iter->next) {
    counter++;
    if (DEBUG>2) printf("nodestore:     comparing with element %s.\n", printterm(iter->id));
    if (SP_unify(id, iter->id)) {
      if (DEBUG>2) printf("nodestore:     same element found.\n");  
      return &iter->node;
    }
  }
  if (DEBUG>2) printf("nodestore:     from_list: element %i not found, %i element(s) already in list.\n", id, counter);
  
  nodestore_item *new = (nodestore_item *) ltl_malloc(sizeof(nodestore_item));
  if (new) {
    new->id = SP_new_term_ref();
    SP_put_term(new->id,id);
    if (DEBUG>2) printf("nodestore:     from_list: insert new reference %i.\n", new->id);  
    new->node.number_of_atoms = node_uninitialised;
    new->next = *list;
    *list = new;
  } else {
    fprintf(stderr, "error: malloc(nodestore_item) returned NULL.\n");
    exit(1);
  }
  return new==0 ? 0 : &new->node;
}

/* introduces memory leak!! */
char *printterm(SP_term_ref ref) {
  SP_atom functor;
  int arity,l;
  char *str;
  if (SP_get_functor(ref, &functor, &arity)) {
    l = SP_atom_length(functor);
    str = (char*) ltl_malloc(sizeof(char) * (l+10));
    sprintf(str, "%s/%i", SP_string_from_atom(functor), arity);
    return str;
  } else {
    printf("*no compound term*");
    return "";
  }
}

size_t ltl_nodestore_memory_usage(nodestore *st) {
  size_t mem = sizeof(nodestore);
  nodestore_item *i;
  mem += ltl_array_memory_usage(&st->pos);
  mem += ltl_array_memory_usage(&st->neg);
  for (i=st->list; i; i=i->next) {
    mem += sizeof(nodestore_item);
  }
  return mem;
}

void print_prolog_node(FILE *file, ltlnode *node) {
  SP_atom atom;
  int arity;
  if (node->type == int_node) {
    fprintf(file, "%u", node->id.id);
  } else {
    SP_get_functor(node->id.ref, &atom, &arity);
    fprintf(file, "'%s/%i'", SP_string_from_atom(atom), arity);
  }
}

static void print_wrapped_prolog_node(FILE *file, ltlnode *node) {
  fputs("node(", file);
  print_prolog_node(file, node);
  fprintf(file, ",%i", node->number_of_atoms);
  fputs(").\n", file);
}

static void print_prolog_nodearray(FILE *file, ltlarray *la) {
  ltlnode *node;
  unsigned int i;
  for (i=0; i<la->size; i++) {
    node = (ltlnode *) get_array_element(la,i);
    if (node->number_of_atoms != node_uninitialised) {
      print_wrapped_prolog_node(file, node);
    }
  }
}

void print_prolog_nodestore(FILE *file, nodestore *nstore) {
  nodestore_item *item;
  for (item = nstore->list; item; item = item->next) {
    print_wrapped_prolog_node(file, &item->node);
  }
  fprintf(file, "/* node(Id,Number of atoms) */\n");
  print_prolog_nodearray(file, &nstore->neg);
  print_prolog_nodearray(file, &nstore->pos);
}
