/*  
    Title: A Profiler for SICStus Prolog
    Author: Philip Höfges
    Heinrich-Heine-University Düsseldorf
    Lehrstuhl für Softwaretechnik und Programmiersprachen
    Prof. Dr. Michael Leuschel, Sebastian Krings
*/

#include <sys/resource.h>
#include <map>
#include <vector>
#include <set>
#include <sicstus/sicstus.h>
#include <stdio.h>
#include <iostream>
#include <sys/time.h>
#include <cassert>
#include <fstream>

using namespace std;

// four maps in order to save times in all four cases
map <SP_atom,vector <long long int> > entry_success;
map <SP_atom,vector <long long int> > entry_fail;
map <SP_atom,vector <long long int> > redo_success;
map <SP_atom,vector <long long int> > redo_fail;

/*
addtional maps
started: time since last co-routine call
stopped: time outtime of all co-routines
recursion_depth: save the depth of recursion
stop_depth: save the depth of co-routines in order just to stop/start the right timers
from_entry: to differ between entry and redo
*/
map <SP_atom,long long int> started;
map <SP_atom,long long int> stopped;
map <SP_atom,int> recursion_depth;
map <SP_atom,int> stop_depth;
map <SP_atom,bool> from_entry;

vector <long long int> get_list(SP_atom);


// check if there is already a recursion running
int check_recursion(SP_atom ProgrammPoint)
{
    if(recursion_depth.count(ProgrammPoint))
    {
        return recursion_depth[ProgrammPoint];
    }
    else
    {
        return 0;
    }
}

extern "C"
{

//--------------------------------------------------------------------------------------
// time functions
//--------------------------------------------------------------------------------------

    // get time
    long long int clock_microseconds(void)
    {
        struct timeval tmpgt;
        gettimeofday(&tmpgt, NULL);
        return  (long long int)(tmpgt.tv_sec * 10000000) + (long long int)(tmpgt.tv_usec);
    }

    // restart alle timers because of end of couroutines
    void restart_timers(void)
    {
        long long int Act_Time = clock_microseconds();
        for(map<SP_atom,long long int>::iterator i=started.begin(); i!=started.end(); i++)
        {
            stop_depth[i->first] -= 1;

            if(stop_depth[i->first] == 1)
            {
                started[i->first] = Act_Time;
            }
        }
    }

    // stop all timers because of start of coroutines
    void stop_timers(void)
    {
        long long int Act_Time = clock_microseconds();
        for(map<SP_atom,long long int>::iterator i=started.begin(); i!=started.end(); i++)
        {
            if(stop_depth[i->first] == 1)
            {
                long long int Old_Time = started[i->first];
                long long int Time = Act_Time - Old_Time;
                stopped[i->first] += Time;
            }
            stop_depth[i->first] += 1;
        } 
    }

    // clear all maps to oversee much data
    void clear_timers(void)
    {
        for(map<SP_atom,long long int>::iterator i=started.begin(); i!=started.end(); i++)
        {
            entry_success.clear();
            entry_fail.clear();
            redo_fail.clear();
            redo_success.clear();
        }
    }

//--------------------------------------------------------------------------------------
// handle functions - central connection to SICStus
//--------------------------------------------------------------------------------------

    // handle the case entry: set recursion_depth, started and save time
    void handle_entry(SP_atom ProgrammPoint)
    {
        if(check_recursion(ProgrammPoint))
        {
            recursion_depth[ProgrammPoint] += 1;
        }
        else
        {
            recursion_depth[ProgrammPoint] = 1;
            stop_depth[ProgrammPoint] = 1;
            long long int Time = clock_microseconds();
            started[ProgrammPoint] = Time;
            stopped[ProgrammPoint] = 0;
            from_entry[ProgrammPoint] = true;
        }
    }

    // handle the case success: get last noticed time and save it, update recursion_depth
    void handle_success(SP_atom ProgrammPoint)
    {
        recursion_depth[ProgrammPoint] -= 1;
        long long int Act_Time = clock_microseconds();
        long long int Old_Time = started[ProgrammPoint];
        long long int Before_Co = stopped[ProgrammPoint];
        long long int Time = Act_Time - Old_Time + Before_Co;
        if(recursion_depth[ProgrammPoint] == 0)
        {
            if(from_entry[ProgrammPoint])
            {
                entry_success[ProgrammPoint].push_back(Time);
            }
            else
            {
                redo_success[ProgrammPoint].push_back(Time);
            }
        }
   }

    // handle the case redo: update recursion_depth, update started, save time at redo
    void handle_redo(SP_atom ProgrammPoint)
    {
        recursion_depth[ProgrammPoint] += 1;
        long long int Time = clock_microseconds();
        started[ProgrammPoint] = Time;
        stopped[ProgrammPoint] = 0;
        from_entry[ProgrammPoint] = false;
    }

    // handle the case fail: update recursion_depth, calculate time
    void handle_fail(SP_atom ProgrammPoint)
    {
        recursion_depth[ProgrammPoint] -= 1;
        long long int Act_Time = clock_microseconds();
        long long int Old_Time = started[ProgrammPoint];
        long long int Before_Co = stopped[ProgrammPoint];
        long long int Time = Act_Time - Old_Time + Before_Co;
        if(recursion_depth[ProgrammPoint] == 0)
        {
            if(from_entry[ProgrammPoint])
            {
                entry_fail[ProgrammPoint].push_back(Time);
            }
            else
            {
                redo_fail[ProgrammPoint].push_back(Time);
            }
        }
    }


//--------------------------------------------------------------------------------------
// calculate informations for output (single)
//--------------------------------------------------------------------------------------

    // get number of call for one of the cases
    long long int get_calls(SP_atom ProgrammPoint, int Sort)
    {
        switch(Sort)
        {
            case 1: return entry_success[ProgrammPoint].size(); break;
            case 2: return entry_fail[ProgrammPoint].size(); break;
            case 3: return redo_success[ProgrammPoint].size(); break;
            case 4: return redo_fail[ProgrammPoint].size(); break;
            default: return -1;
        }
    }

    // get sum of all times for one of the cases
    long long int get_sum(SP_atom ProgrammPoint, int Sort)
    {
        long long int sum = 0;
        vector<long long int> l;

        switch(Sort)
        {
            case 1: l = entry_success[ProgrammPoint]; break;
            case 2: l = entry_fail[ProgrammPoint]; break;
            case 3: l = redo_success[ProgrammPoint]; break;
            case 4: l = redo_fail[ProgrammPoint]; break;
        }
        for(vector<long long int>::iterator i=l.begin(); i!=l.end(); i++)
        {
            sum += *i;
        }
        return sum;
    }

    // get the average time for one of the cases
    long long int get_average(SP_atom ProgrammPoint, int Sort)
    {
        int calls = get_calls(ProgrammPoint, Sort);
        int sum = get_sum(ProgrammPoint, Sort);
        if(calls != 0)
        {
            return sum / calls;
        }
        else
        {
            return 0;
        }
    }

    // get maximum of one of the cases
    long long int get_max(SP_atom ProgrammPoint, int Sort)
    {
        vector<long long int> l;
        switch(Sort)
        {
            case 1: l = entry_success[ProgrammPoint]; break;
            case 2: l = entry_fail[ProgrammPoint]; break;
            case 3: l = redo_success[ProgrammPoint]; break;
            case 4: l = redo_fail[ProgrammPoint]; break;
        }
        if(!l.size())
        {
            return 0;
        }
        long long int max = l.front();
        for(vector<long long int>::iterator i=l.begin(); i!=l.end(); i++)
        {
            if(max<(*i))
            {
                max = (*i);
            }
        }
        return max;
    }

    // get minimum of one of the cases
    long long int get_min(SP_atom ProgrammPoint, int Sort)
    {
        vector<long long int> l;
        switch(Sort)
        {
            case 1: l = entry_success[ProgrammPoint]; break;
            case 2: l = entry_fail[ProgrammPoint]; break;
            case 3: l = redo_success[ProgrammPoint]; break;
            case 4: l = redo_fail[ProgrammPoint]; break;
        }
        if(!l.size())
        {
            return 0;
        }
        long long int min = l.front();
        for(vector<long long int>::iterator i=l.begin(); i!=l.end(); i++)
        {
            if(min>(*i))
            {
                min = (*i);
            }
        }
        return min;
    }

//--------------------------------------------------------------------------------------
// calculate informations for output (total)
//--------------------------------------------------------------------------------------

    // calculate the calls of one function
    long long int get_total_calls(SP_atom ProgrammPoint)
    {
        return entry_success[ProgrammPoint].size() + entry_fail[ProgrammPoint].size() + redo_success[ProgrammPoint].size() + redo_fail[ProgrammPoint].size();
    }

    // sum of all times
    long long int get_total_sum(SP_atom ProgrammPoint)
    {
        return get_sum(ProgrammPoint,1) + get_sum(ProgrammPoint,2) + get_sum(ProgrammPoint,3) + get_sum(ProgrammPoint,4);
    }

    // get average of all times
    long long int get_total_average(SP_atom ProgrammPoint)
    {
        int calls = get_total_calls(ProgrammPoint);
        int sum = get_total_sum(ProgrammPoint);
        if(calls != 0)
        {
            return sum / calls;
        }
        else
        {
            return 0;
        }
    }

    // get the maximum of all times
    long long int get_total_max(SP_atom ProgrammPoint)
    {
        long long int max = get_max(ProgrammPoint,1);
        if(get_max(ProgrammPoint,2)>max)
        {
            max = get_max(ProgrammPoint,2);
        }
        if(get_max(ProgrammPoint,3)>max)
        {
            max = get_max(ProgrammPoint,3);
        }
        if(get_max(ProgrammPoint,4)>max)
        {
            max = get_max(ProgrammPoint,4);
        }
        return max;
    }

    // get the minimum of all times
    long long int get_total_min(SP_atom ProgrammPoint)
    {
        long long int min = get_min(ProgrammPoint,1);
        if(get_min(ProgrammPoint,2)<min && get_min(ProgrammPoint,2)!=0)
        {
            min = get_min(ProgrammPoint,2);
        }
        if(get_min(ProgrammPoint,3)<min && get_min(ProgrammPoint,3)!=0)
        {
            min = get_min(ProgrammPoint,3);
        }
        if(get_min(ProgrammPoint,4)<min && get_min(ProgrammPoint,4)!=0)
        {
            min = get_min(ProgrammPoint,4);
        }
        return min;
    }

//--------------------------------------------------------------------------------------
// output functions
//--------------------------------------------------------------------------------------

    struct classcomp
    {
        bool operator() (pair<SP_atom,long long int> P1, pair<SP_atom,long long int> P2) { return P1.second <= P2.second;}
    };

    // write information on screen
    void pen(void)
    {
        /*sort set by total sum time*/
        set<pair<SP_atom,long long int>,classcomp > sorted;

        /*fill sorted with ProgrammPoints and sums*/
        for(map<SP_atom,long long int>::iterator k=started.begin(); k!=started.end(); k++)
        {
            SP_atom ProgrammPoint = k->first;
            long long int sum = get_total_sum(ProgrammPoint);
            sorted.insert(pair<SP_atom, long long int>(ProgrammPoint,sum));
        }

        printf("\nstart of measurement - all times count in microseconds\n\n");
        for(set<pair<SP_atom,long long int>,classcomp >::iterator i=sorted.begin(); i!=sorted.end(); i++)
        {
            SP_atom PP = i->first;
            vector <long long int> vec = get_list(PP);

            // save ProgrammPoint and list of results in their own lists

            printf("name: %s\n", SP_string_from_atom(PP));
            printf("total ---------- calls: %lld, sum: %lld, average: %lld, min: %lld, max: %lld\n\n", vec[0],vec[1],vec[2],vec[3],vec[4]);
            if(vec[5]!=0)
            {
                printf("entry_success -- calls: %lld, sum: %lld, average: %lld, min: %lld, max: %lld\n", vec[5],vec[6],vec[7],vec[8],vec[9]);
            }

            if(vec[10]!=0)
            {
                printf("entry_fail ----- calls: %lld, sum: %lld, average: %lld, min: %lld, max: %lld\n", vec[10],vec[11],vec[12],vec[13],vec[14]);
            }

            if(vec[15]!=0)
            {
                printf("redo_success --- calls: %lld, sum: %lld, average: %lld, min: %lld, max: %lld\n", vec[15],vec[16],vec[17],vec[18],vec[19]);
            }

            if(vec[20]!=0)
            {
                printf("redo_fail ------ calls: %lld, sum: %lld, average: %lld, min: %lld, max: %lld\n", vec[20],vec[21],vec[22],vec[23],vec[24]);
            }
            printf("\n");
        }
        printf("\nend of measurement\n\n");
    }

    // create prolog list
    SP_term_ref built_list(void)
    {
        SP_term_ref tailo = SP_new_term_ref();
        for(map<SP_atom,long long int>::iterator i=started.begin(); i!=started.end(); i++)
        {
            SP_atom PP = i->first;
            vector <long long int> vec = get_list(PP);

            SP_term_ref tail = SP_new_term_ref();
            for(int j=vec.size()-1;j>=0;j--)
            {
                SP_term_ref list = SP_new_term_ref();
                SP_term_ref e = SP_new_term_ref();
                SP_put_integer(e,vec[j]);
                int rc = SP_cons_list(list,e,tail);
                assert(rc);
                tail = list;
            }
            SP_term_ref atom = SP_new_term_ref();
            SP_put_atom(atom,PP);
            SP_term_ref list = SP_new_term_ref();
            int rc = SP_cons_list(list,atom,tail);
            assert(rc);

            SP_term_ref listo = SP_new_term_ref();
            SP_cons_list(listo,list,tailo);
            tailo = listo;
        }

        return tailo;
    }

    // write information into .csv file
    void file(void)
    {
        FILE *datei;
        datei = fopen("output.csv", "w");

        fprintf(datei, "Predicate,Program_Point,Calls,Sum,Average,Min,Max\n");

        for(map<SP_atom,long long int>::iterator i=started.begin(); i!=started.end(); i++)
        {
            SP_atom ProgrammPoint = i->first;
            vector <long long int> vect = get_list(ProgrammPoint);
            for(size_t j=0;j<vect.size();j++)
            {
        if((vect[j]==0) && (j%5==0))
        {
            break;
        }
        else
        {
                    switch(j)
                    {
                        /*set name and programmpoint*/
                        case 0: fprintf(datei, "%s,total,",SP_string_from_atom(ProgrammPoint)); break;
                        case 5: fprintf(datei, "%s,entry_success,",SP_string_from_atom(ProgrammPoint)); break;
                        case 10: fprintf(datei, "%s,entry_fail,",SP_string_from_atom(ProgrammPoint)); break;
                        case 15: fprintf(datei, "%s,redo_success,",SP_string_from_atom(ProgrammPoint)); break;
                        case 20: fprintf(datei, "%s,redo_fail,",SP_string_from_atom(ProgrammPoint)); break;
                    }
                    fprintf(datei, "%lld,", vect[j]);
                    switch(j)
                    {
                        /*newline at the end of a block*/
                        case 4: fprintf(datei, "\n"); break;
                        case 9: fprintf(datei, "\n"); break;
                        case 14: fprintf(datei, "\n"); break;
                        case 19: fprintf(datei, "\n"); break;
                        case 24: fprintf(datei, "\n"); break;
            }
                }
            }
        }
    printf("file creation (output.csv) successful\n");
        fclose(datei);
    }
}

    // create list of results
    vector <long long int> get_list(SP_atom PP)
    {
        vector <long long int> result_here;

        // find the values
        long long int total_calls = get_total_calls(PP);
        long long int total_sum = get_total_sum(PP);
        long long int total_average = get_total_average(PP);
        long long int total_min = get_total_min(PP);
        long long int total_max = get_total_max(PP);

        long long int entry_success_calls = get_calls(PP,1);
        long long int entry_success_sum = get_sum(PP,1);
        long long int entry_success_average = get_average(PP,1);
        long long int entry_success_min = get_min(PP,1);
        long long int entry_success_max = get_max(PP,1);

        long long int entry_fail_calls = get_calls(PP,2);
        long long int entry_fail_sum = get_sum(PP,2);
        long long int entry_fail_average = get_average(PP,2);
        long long int entry_fail_min = get_min(PP,2);
        long long int entry_fail_max = get_max(PP,2);

        long long int redo_success_calls = get_calls(PP,3);
        long long int redo_success_sum = get_sum(PP,3);
        long long int redo_success_average = get_average(PP,3);
        long long int redo_success_min = get_min(PP,3);
        long long int redo_success_max = get_max(PP,3);

        long long int redo_fail_calls = get_calls(PP,4);
        long long int redo_fail_sum = get_sum(PP,4);
        long long int redo_fail_average = get_average(PP,4);
        long long int redo_fail_min = get_min(PP,4);
        long long int redo_fail_max = get_max(PP,4);

        result_here.push_back(total_calls);
        result_here.push_back(total_sum);
        result_here.push_back(total_average);
        result_here.push_back(total_min);
        result_here.push_back(total_max);

        result_here.push_back(entry_success_calls);
        result_here.push_back(entry_success_sum);
        result_here.push_back(entry_success_average);
        result_here.push_back(entry_success_min);
        result_here.push_back(entry_success_max);

        result_here.push_back(entry_fail_calls);
        result_here.push_back(entry_fail_sum);
        result_here.push_back(entry_fail_average);
        result_here.push_back(entry_fail_min);
        result_here.push_back(entry_fail_max);

        result_here.push_back(redo_success_calls);
        result_here.push_back(redo_success_sum);
        result_here.push_back(redo_success_average);
        result_here.push_back(redo_success_min);
        result_here.push_back(redo_success_max);

        result_here.push_back(redo_fail_calls);
        result_here.push_back(redo_fail_sum);
        result_here.push_back(redo_fail_average);
        result_here.push_back(redo_fail_min);
        result_here.push_back(redo_fail_max);

        return result_here;
}
