#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#include <sicstus/sicstus.h>
#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/hashtrie.h"
#include "../../../common/hash.h"

#include "mbalancer.h"
#include "mprint.h"
#include "mbeat.h"



static int get_maximum_queue(void);
static void dump_statistics(int worker, int qs, zmsg_t *msg);
static void dump_multi_queued(void);

static volatile int *queues = NULL;
static void *stats_recv = NULL;
static void *local_printer;

extern volatile int nmbr_workers;
static int queuethresh;

static pthread_t queue_printer;
static pthread_mutex_t queue_lock;
// we need to use a mutex because the queues array may be re-allocated
// since we have a print thread, it may use the old, invalid pointer
static void *print_queues(void *arg);
static void *work_balancer(void *arg);

extern int checked_states;
static int *multi_queued;


static void *send_work;
static wQueue *work_q = NULL;

static int empty_workers = 0;
static int request_pending = 0;

static int h_stats (void);
static int get_first_empty_queue(void);
static void send_request(int id, int maxpackages);
static int send_workitem(int worker);
static int h_work(void);
static void hash_and_enqueue(char *term, int len, wQueue *q);
static void put_assertions(int number_of_assertions);
static void add_worker(const int id);
static void send_wpmsg_to_worker(zmsg_t *work, int worker);
static void set_q_size(int worker, int size);
static zmsg_t *build_wpmsg_with_digest(char *term, int len, char *digest);
static char *build_term(int i, int len);

/* logic 
 * the work socket only receives following types of messages:
 * OHAI <ID>   - enables routing to a specific worker
 * <work data> - the work transfer itself
 * ACK <ID>    - acknowledges a work transfer
 *
 * we know exactly when to look for a message on this socket
 * - we add a worker --> it says OHAI
 * - we initiate a work transfer --> it will send work data
 * - we forward a message --> it (hopefully) says ACK
 *
 * These messages should be seen fairly seldom.
 * Hence, we poll the socket only when we expect a message.
 *
 */



void start_work_balancer(pthread_t *thread) {
    pthread_create(thread, NULL, work_balancer, NULL);
}

void set_empty_workers(int amount) {
    empty_workers = amount;
}


extern volatile int running;

static void *work_balancer(void *arg) {
    while(running) {
        h_stats();
        int i;
        for (i = 0; i < 24 && zsocket_poll(send_work, 0) == true; i++) {
            h_work();
        }
        if (empty_workers && !request_pending) {
            // find full worker
            if (is_empty(work_q)) {
                int worker = get_maximum_queue();
                //DEBUG_PRINT(DEBUG_NORMAL,,,":id %d :longest-queue %d :queuethresh %d", worker, queues[worker], queuethresh);
                if (worker > 0 && queues[worker] > 2 * queuethresh) {
                    DEBUG_PRINT(DEBUG_NORMAL,,,":requesting-from %d :empty-workers %d", worker, empty_workers);
                    send_request(worker, 5 * queuethresh);
                    DEBUG_PRINT(DEBUG_NORMAL,,,":requested-from %d :empty-workers %d", worker, empty_workers);
                }
            } else {
                int worker = get_first_empty_queue();
                send_workitem(worker);
                DEBUG_PRINT(DEBUG_NORMAL,,,":sent-queueitem-to %d :empty-workers %d", worker, empty_workers);
            }
        }
    }
    pthread_exit(NULL);
}



/* interface */
void init_work_balancer(zctx_t *ctx, zmq_pollitem_t *poller, int qsize, int port_start) {
    local_printer = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(local_printer, "inproc://printer");
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_work_balancer :max-queuesize %d :port-start %d", qsize, port_start);
    queuethresh = qsize;
    queues = (int*) calloc(nmbr_workers + 1, sizeof(int));
    assert(queues != NULL);

    multi_queued = (int*) calloc(nmbr_workers + 1, sizeof(int));
    assert(multi_queued);
    int i;
    for (i = 1; i <= nmbr_workers; i++) {
        queues[i] = -1;
        multi_queued[i] = 0;
    }
    assert(stats_recv == NULL);

    stats_recv = zsocket_new(ctx, ZMQ_PULL);
    zsocket_bind (stats_recv, "tcp://*:%d", port_start + 4);
    int rc = pthread_mutex_init(&queue_lock, NULL);
    assert(rc == 0);

    *poller = (zmq_pollitem_t) { stats_recv, 0, ZMQ_POLLIN };
    pthread_create(&queue_printer, NULL, print_queues, (void*) ctx);



    /* former mwork setup */

    send_work = zsocket_new (ctx, ZMQ_ROUTER);
    zsocket_bind (send_work, "tcp://*:%d", port_start + 2);
    *poller = (zmq_pollitem_t) { send_work, 0, ZMQ_POLLIN };

    work_q = init_queue(0);


    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_work_balancer");

}

void teardown_work_balancer(zctx_t *ctx) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :teardown_work_balancer");
    assert(stats_recv != NULL);
    zsocket_destroy(ctx, stats_recv);
    pthread_cancel(queue_printer);
    pthread_mutex_destroy(&queue_lock);
    dump_multi_queued();
    free(multi_queued);

    /* former teardown mwork code */

    zsocket_destroy(ctx, send_work);
    teardown_queue(&work_q);

    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :teardown_work_balancer");
    zsocket_disconnect(local_printer, "inproc://printer");
    zsocket_destroy(ctx, local_printer);
}

void start_work(char *fwp, int fwp_len) {
    DEBUG_PRINT(DEBUG_VERBOSE,,,":enter :start_work");
    hash_and_enqueue(fwp, fwp_len, work_q);
    // only used for model checking
    DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :start_work");
}


void start_work_assertion(int number_of_assertions) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :start_work_assertion");
    put_assertions(number_of_assertions);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :start_work_assertion");
}

static int h_stats (void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :h_stats");
    int i;
    const int max = nmbr_workers | 5;
    // 5 is an arbitrary constant just in case nmbr_workers is zero or one
    for (i = 0; i < max; i++) {
        if (zsocket_poll(stats_recv, 0) == false) {
            break;
        }
        zmsg_t *msg = zmsg_recv(stats_recv);

        int worker;
        int queuesize;

        char *tmp = zmsg_popstr(msg);
        int rc = sscanf(tmp, "%d", &worker);
        assert(rc == 1);
        free(tmp);

        //set_worker_alive(worker);

        if (zmsg_size(msg) > 0) {
            // statistics
            tmp = zmsg_popstr(msg);
            int rc = sscanf(tmp, "%d", &queuesize);
            assert(rc == 1);
            queues[worker] = queuesize;
            if (queuesize == 0) {
                empty_workers++;
                DEBUG_PRINT(DEBUG_NORMAL,,,":empty-worker %d :empty-workers %d", worker, empty_workers);
            }

            if (zmsg_size(msg) > 0) {
                dump_statistics(worker, queuesize, msg);
            } else {
                multi_queued[worker]++;
            }
        }
        
        zmsg_destroy(&msg);
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :h_stats");
    return 0;
}


static int h_work(void) {

    zmsg_t *msg = zmsg_recv(send_work);

    zframe_t *from = zmsg_pop(msg);
    zframe_destroy(&from);

    if (zframe_size(zmsg_last(msg)) <= TRIE_HASHLEN) { // can't be a wp
        char *str = zframe_strdup(zmsg_first(msg));
        if (0 == strncmp("OHAI ", str, 5)) {
            int id;
            int rc = sscanf(str, "OHAI %d", &id);
            assert(rc == 1);
            DEBUG_PRINT(DEBUG_NORMAL,,,":ohai %d :empty-workers %d", id, empty_workers);
            zmsg_t *msg = zmsg_new();
            zmsg_addstr(msg, "%d", id);
            zmsg_addstr(msg, "OHAI-OK");
            zmsg_send(&msg, send_work);
            zmsg_destroy(&msg);
            add_worker(id);
        } else if (0 == strncmp("ACK ", str, 4)) {
            int id;
            int rc = sscanf(str, "ACK %d", &id);
            assert(rc == 1);
            zmsg_t *answer = zmsg_new();
            zmsg_addstr(answer, "%d", id);
            zmsg_addstr(answer, "ACK");
            zmsg_send(&answer, send_work);
            DEBUG_PRINT(DEBUG_NORMAL,,,":ack %d :empty-workers %d", id, empty_workers);
            request_pending = 0;
        } else {
            assert(0 == strcmp(str, "done."));
            DEBUG_PRINT(DEBUG_NORMAL,,,":empty-transfer :done :empty-workers %d", empty_workers);
            request_pending = 0;
        }

        free(str);
    } else if (msg != NULL) {
        int worker = get_first_empty_queue();

        zstr_send(local_printer, "{:type :Worktransfer :to %d :size %d}\n", worker, zmsg_size(msg));
        // sending a wpmsg to a worker decreases the amount of empty workers
        send_wpmsg_to_worker(msg, worker);
        DEBUG_PRINT(DEBUG_NORMAL,,,":transfer-to %d :empty-workers %d", worker, empty_workers);
    } else {
        zmsg_destroy(&msg);
    }
    return 0;
}



/* logic */




static void add_worker(const int id) {
    /*
     * adding a worker
     * if arrays are big enough, set entries to 0;
     * else
     *   expand both array limits to id;
     *   fill with negative 1s (to avoid work balancing
     *   and to distinguish unready workers from ready ones);
     *   set its entry to 0
     */
    DEBUG_PRINT(DEBUG_VERBOSE,,,":enter :add_worker :id %d", id);
    int tmp = nmbr_workers;
    if (id <= tmp) {
        queues[id] = 0;
        multi_queued[id] = 0;
    } else {
        nmbr_workers = id;

        multi_queued = (int*) realloc(multi_queued, sizeof(int) * (nmbr_workers + 1));
        assert(multi_queued);

        pthread_mutex_lock(&queue_lock);
            queues = (volatile int*) realloc((int*)queues, sizeof(int) * (nmbr_workers + 1));
            assert(queues);

            int i;
            for (i = tmp + 1; i < nmbr_workers; i++) {
                queues[i] = -1;
                multi_queued[i] = 0;
            }
            queues[nmbr_workers] = 0;
            multi_queued[nmbr_workers] = 0;
        pthread_mutex_unlock(&queue_lock);
    }
    empty_workers++;
    //send_workitem(id);
    DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :add_worker");
}






static void hash_and_enqueue(char *term, int len, wQueue *q) {
    char *digest = (char*) malloc(TRIE_HASHLEN);
    hash(term, len, digest);
    unsigned char infos[] = {0};
    trie_put(digest, infos);
    enqueue(q, term, len, digest);
}


static void send_request(int id, int maxpackages) {
    if (request_pending)
        return;
    zmsg_t *request = zmsg_new();
    zmsg_addstr(request, "%d", id);
    zmsg_addstr(request, "give work");
    zmsg_addstr(request, "%d", maxpackages);
    zmsg_send(&request, send_work);
    zmsg_destroy(&request);
    request_pending = 1;
}

static zmsg_t *build_wpmsg(char *term, int len) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :build_wpmsg");
    char digest[TRIE_HASHLEN];
    hash(term, len, digest);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :build_wpmsg");
    return build_wpmsg_with_digest(term, len, digest);
}

static zmsg_t *build_wpmsg_with_digest(char *term, int len, char *digest) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[TRIE_HASHLEN * 2 + 1]; print_readable_digest((unsigned char*)digest, d);,,
                ":enter :build_wpmsg_with_digest :digest :%s", d);
    char *content = (char*) malloc(TRIE_HASHLEN + len);
    assert(content);

    memcpy(content, digest, TRIE_HASHLEN);
    memcpy(content+TRIE_HASHLEN, term, len);
   
    zmsg_t *res = zmsg_new();
    zmsg_addmem(res, content, len+TRIE_HASHLEN);

    free(content);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :build_wpmsg_with_digest");
    return res;
}

static void send_wpmsg_to_worker(zmsg_t *work, int worker) {
    empty_workers--;
    int size = zmsg_size(work);
    set_q_size(worker, size);
    zmsg_pushstr(work, "%d", worker);
    int rc = zmsg_send(&work, send_work);
    assert(rc == 0);
    zmsg_destroy(&work);
}


static int send_workitem(int worker) {
    DEBUG_PRINT(DEBUG_VERBOSE,,,":enter :send_workitem :args {:worker %d}", worker);
    if (!is_empty(work_q)) {
        tCell *cell = dequeue(work_q);
        zmsg_t *msg = build_wpmsg_with_digest(cell->term,
                                              cell->term_length, cell->digest);
        zmsg_pushstr(msg, "0");
        send_wpmsg_to_worker(msg, worker);
        free(cell->term);
        free(cell->digest);
        free(cell);
        DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :send_workitem :retval 1");
        return 1;
    }
    DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :send_workitem :retval 0");
    return 0;
}


static void put_assertions(int number_of_assertions) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :put_assertions");
    int i;
    for (i = 0; i < number_of_assertions; i++) {
        int len_number = snprintf(NULL, 0, "%d", i);
        int len_term = len_number + 11;
        char *term = build_term(i, len_term);
        hash_and_enqueue(term, len_term, work_q);
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :put_assertions");
}




static void *print_queues(void *arg) {
    zctx_t *ctx = (zctx_t*) arg;
    void *printsocket = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(printsocket, "inproc://printer");
    while(1) {
        int i;
        int count = trie_get_checked_states_count();
        zmsg_t *msg = zmsg_new();
        zmsg_addstr(msg, "{:type :queues :multichecked %d :checked %d :known %d :queuesizes [",
                    checked_states - count, count,
                    trie_count_elements());
        pthread_mutex_lock(&queue_lock);
        for (i = 0; i <= nmbr_workers; i++) {
            zmsg_addstr(msg, "%d ", queues[i]);
        }
        pthread_mutex_unlock(&queue_lock);
        zmsg_addstr(msg, "]}\n");
        zmsg_send(&msg, printsocket);
        zmsg_destroy(&msg);
        zstr_send(printsocket, "{:type :debug :empty-workers %d}\n", empty_workers);
        sleep(1);
    }
    zsocket_destroy(ctx, printsocket);
    return NULL;
}



/* helpers */

static void set_q_size(int worker, int size) {
    queues[worker] = size;
}

int get_q_size(int worker) {
    return queues[worker];
}


static void dump_multi_queued(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :dump_multi_queued");
    int i;
    zmsg_t *msg = zmsg_new(); // FIXME 
    zmsg_addstr(msg, "{:type :queuecache :data [ ");
    for (i = 1; i <= nmbr_workers; i++) {
        zmsg_addstr(msg, "%d ", multi_queued[i]);
    }
    zmsg_addstr(msg, "]}\n");
    zmsg_send(&msg, local_printer);
    zmsg_destroy(&msg);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :dump_multi_queued");
}

static void dump_statistics(int worker, int qs, zmsg_t *msg) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :dump_statistics");
    assert(zmsg_size(msg) == 4);
    // digest, tProlog, timestamp, #state
    char *digest = zmsg_popstr(msg);
    char *tProlog = zmsg_popstr(msg);
    assert(strlen(tProlog) >= 5); // should be preformatted
    char *timestamp = zmsg_popstr(msg);
    char *nmbr_state = zmsg_popstr(msg);
    
    char readable[TRIE_HASHLEN * 2 + 1];
    print_readable_digest((unsigned char*) digest, readable);
    zstr_send(local_printer, "{:type :stats :worker-id %d :queuesize %d :tProlog %s :timestamp %s :statenumber %s :digest :%s}\n",
                                 worker, qs, tProlog,
                                 timestamp, nmbr_state, readable);

    free(digest);
    free(tProlog);
    free(timestamp);
    free(nmbr_state);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :dump_statistics");
}

static int get_maximum_queue(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :get_maximum_queue");
    int i;
    int max = 0;
    int n = -1;

    for (i = 1; i <= nmbr_workers; i++) { 
        if (queues[i] > max) {
            max = queues[i];
            n = i;
        }
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :get_maximum_queue");
    return n;
}

static int get_first_empty_queue(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :get_first_empty_queue");
    int i;
    for (i = 1; i <= nmbr_workers; i++) { 
        if (queues[i] == 0) {
            return i;
        }
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :get_first_empty_queue");
    return -1;
}

static char *build_term(int i, int len) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :build_term");
    char *term = (char*) malloc(len);
    int rc = snprintf(term, len, "DScheck_(I%d", i);
    assert(rc == len - 1);
    term[7] = 0; // null seperator
    term[8] = 1; // arity of term 'check' is 1 (overwrites '(')
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :build_term");
    return term;
}
