#include <stdlib.h>
#include <pthread.h>

#include "../include/zmq.h"
#include "../include/czmq.h"

const char *host = NULL;

struct proxy_arg {
    zctx_t *ctx;
    int type_to_worker;
    int type_to_master;
    int port;
};

volatile int running = 1;

void *run_proxy(void *args) {
    struct proxy_arg *arg = (struct proxy_arg*) args;
    int unidirectional = (arg->type_to_worker == ZMQ_PUB ||
                          arg->type_to_worker == ZMQ_PULL);
    void *to_worker = zsocket_new(arg->ctx, arg->type_to_worker);
    void *to_master = zsocket_new(arg->ctx, arg->type_to_master);
    assert(to_worker);
    assert(to_master);

    int rc;
    rc = zsocket_bind(to_worker, "ipc:///tmp/zmq-proxy-%d", arg->port);
    assert(rc == 0);
    rc = zsocket_connect(to_master, "tcp://%s:%d", host, arg->port);
    assert(rc == 0);
    if (arg->type_to_master == ZMQ_SUB) {
        char *filter = "";
        zmq_setsockopt (to_master, ZMQ_SUBSCRIBE, filter, strlen (filter));
    }

    void *recv_zocket;
    void *send_zocket;
    if (arg->type_to_worker == ZMQ_PUB) {
        recv_zocket = to_master;
        send_zocket = to_worker;
    } else {
        assert(arg->type_to_worker == ZMQ_PULL || arg->type_to_worker == ZMQ_REP);
        recv_zocket = to_worker;
        send_zocket = to_master;
    }

    if (unidirectional) {
        // forward
        while (running) {
            if (zsocket_poll(recv_zocket, 1)) {
                zmsg_t *msg = zmsg_recv(recv_zocket);
                zmsg_send(&msg, send_zocket);
                zmsg_destroy(&msg);
            }
        }
    } else {
        // request-reply
        while (running) {
            if (zsocket_poll(recv_zocket, 1)) {
                zmsg_t *req = zmsg_recv(recv_zocket);
                zmsg_send(&req, send_zocket);
                zmsg_destroy(&req);

                zmsg_t *rep = zmsg_recv(send_zocket);
                zmsg_send(&rep, recv_zocket);
                zmsg_destroy(&rep);
            }

        }
    }
    recv_zocket = NULL;
    send_zocket = NULL;

    rc = zsocket_unbind(to_worker, "ipc:///tmp/zmq-proxy-%d", arg->port);
    assert(rc == 0);
    rc = zsocket_disconnect(to_master, "tcp://%s:%d", host, arg->port);
    assert(rc == 0);
    zsocket_destroy(arg->ctx, to_worker);
    zsocket_destroy(arg->ctx, to_master);

    pthread_exit(NULL);
}

void *welcomer(void *args) {
    zctx_t *ctx = (zctx_t*) args;

    void *zocket = zsocket_new(ctx, ZMQ_REP);
    int rc = zsocket_bind(zocket, "ipc:///tmp/zmq-proxy-sync");
    assert(rc == 0);

    while (running) {
        if (zsocket_poll(zocket, 1)) {
            zmsg_t *msg = zmsg_recv(zocket);
            zmsg_destroy(&msg);
            zstr_send(zocket, "go");
        }
    }

    rc = zsocket_unbind(zocket, "ipc:///tmp/zmq-proxy-sync");
    assert(rc == 0);
    zsocket_destroy(ctx, zocket);

    pthread_exit(NULL);
}

int main(int argc, char *argv[]) {
    if (argc != 3) {
        perror("Usage: ./proxy <master IP> <port>");
    }
    host = argv[1];
    int port = atoi(argv[2]);

    zctx_t *ctx = zctx_new();

    void *recv_control = zsocket_new(ctx, ZMQ_SUB);
    zsocket_connect(recv_control,"ipc:///tmp/zmq-proxy-%d", port + 7);
    char *filter = "";
    zmq_setsockopt (recv_control, ZMQ_SUBSCRIBE, filter, strlen (filter));


    pthread_t stats;
    struct proxy_arg stats_arg;
    stats_arg.ctx = ctx;
    stats_arg.type_to_worker = ZMQ_PULL;
    stats_arg.type_to_master = ZMQ_PUSH;
    stats_arg.port = port + 4;
    pthread_create(&stats, NULL, run_proxy, &stats_arg);


    pthread_t hash_collect;
    struct proxy_arg hash_collect_arg;
    hash_collect_arg.ctx = ctx;
    hash_collect_arg.type_to_worker = ZMQ_PULL;
    hash_collect_arg.type_to_master = ZMQ_PUSH;
    hash_collect_arg.port = port + 1;
    pthread_create(&hash_collect, NULL, run_proxy, &hash_collect_arg);

    pthread_t printer;
    struct proxy_arg printer_arg;
    printer_arg.ctx = ctx;
    printer_arg.type_to_worker = ZMQ_PULL;
    printer_arg.type_to_master = ZMQ_PUSH;
    printer_arg.port = port + 6;
    pthread_create(&printer, NULL, run_proxy, &printer_arg);

    pthread_t hash_publish;
    struct proxy_arg hash_publish_arg;
    hash_publish_arg.ctx = ctx;
    hash_publish_arg.type_to_worker = ZMQ_PUB;
    hash_publish_arg.type_to_master = ZMQ_SUB;
    hash_publish_arg.port = port;
    pthread_create(&hash_publish, NULL, run_proxy, &hash_publish_arg);


    pthread_t id_handler;
    struct proxy_arg id_handler_arg;
    id_handler_arg.ctx = ctx;
    id_handler_arg.type_to_worker = ZMQ_REP;
    id_handler_arg.type_to_master = ZMQ_REQ;
    id_handler_arg.port = port + 5;
    pthread_create(&id_handler, NULL, run_proxy, &id_handler_arg);

    pthread_t control;
    struct proxy_arg control_arg;
    control_arg.ctx = ctx;
    control_arg.type_to_worker = ZMQ_PUB;
    control_arg.type_to_master = ZMQ_SUB;
    control_arg.port = port + 7;
    pthread_create(&control, NULL, run_proxy, &control_arg);

    pthread_t welcome_thread;
    pthread_create(&welcome_thread, NULL, welcomer, ctx);

    pthread_detach(stats);
    pthread_detach(hash_collect);
    pthread_detach(printer);
    pthread_detach(hash_publish);
    pthread_detach(id_handler);
    pthread_detach(control);

    zmsg_t *msg = zmsg_recv(recv_control);
    running = 0;
    zmsg_destroy(&msg);

    zsocket_disconnect(recv_control,"ipc:///tmp/zmq-proxy-%d", port + 7);
    zsocket_destroy(ctx, recv_control);
    sleep(1);
    zctx_destroy(&ctx);

//    pthread_cancel(stats);
//    pthread_cancel(hash_collect);
//    pthread_cancel(printer);
    return 0;

}
