//#include <sys/mman.h>
//#include <sys/stat.h>
//#include <fcntl.h>

#include <assert.h>
#include <string.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "../include/zmq.h"
#include "../include/czmq.h"

#include "../common/shm_srv.h"
#include "../common/shm_common.h"
#include "../common/hashtrie.h"
#include "../common/memusage.h"
#include "../common/proxy-connector.h"

void *hash_zocket = NULL;
void *ctrl_zocket = NULL;
void *sync_workers = NULL;

int write_hash(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    int count = 0;
    // FIXME: magic constant 1000
    // hash insertion has top priority
    // polling termination or adding a new worker may take a few milliseconds
    // possible enhancement: heartbeating with the master, master sends number of current workers
    while (count < 1000 && zsocket_poll(hash_zocket, 0) == true) {
        zmsg_t *msg = zmsg_recv(hash_zocket);
        zframe_t *frame;
        while ((frame = zmsg_pop(msg)) != NULL) {
            assert(TRIE_DATALEN == zframe_size(frame));
            char *digest = zframe_strdup(frame);
            trie_put(digest, (unsigned char*) digest + TRIE_HASHLEN);
            free(digest);
            zframe_destroy(&frame);
        }
        zmsg_destroy(&msg);
        count++;
    }
    return 0;
}

int signal_go(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(sync_workers);
    zmsg_destroy(&msg);
    zstr_send(sync_workers, "go");
    return 0;
}

int handle_ctrl(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    char *cmd = zstr_recv(ctrl_zocket);
    if (strcmp(cmd, "TERM") == 0) {
        free(cmd);
        return -1;
    } else if (strcmp(cmd, "MEMORY") == 0) {
        printf("memory allocated: %ld\n", memory_usage_in_kb());
    }
    free(cmd);
    // TODO: invalid message
    return 0;
}


void put_hashes(char *hashes, int len) {
    int i;
    for (i = 0; i < len; i += TRIE_DATALEN) {
        char *d = hashes + i;
        trie_put(d, (unsigned char*) d + TRIE_HASHLEN);
    }
    assert(i == len);
}

int main(int argc, char *argv[]) {
    assert(argc == 3);
    int port = atoi(argv[1]);
    int masternode = atoi(argv[2]);

    struct semap_env *senv = create_semaphores();
    init_shm_common(senv);
    init_trie();

    zctx_t *ctx = zctx_new();
    connect_to_proxy(ctx);

    hash_zocket = zsocket_new(ctx, ZMQ_SUB);
    int rc = zsocket_connect(hash_zocket, "ipc:///tmp/zmq-proxy-%d", port);
    assert(rc == 0);
    char *filter = "";
    zmq_setsockopt(hash_zocket, ZMQ_SUBSCRIBE, filter, strlen(filter));
    ctrl_zocket = zsocket_new(ctx, ZMQ_SUB);
    rc = zsocket_connect(ctrl_zocket, "ipc:///tmp/zmq-proxy-%d", port + 7);
    assert(rc == 0);
    zmq_setsockopt(ctrl_zocket, ZMQ_SUBSCRIBE, filter, strlen(filter));
    printf("Connected.\n");

    sync_workers = zsocket_new(ctx, ZMQ_REP);
    rc = zsocket_bind(sync_workers, "ipc:///tmp/zmq-syncer");
    assert(rc == 0);

    if (!masternode) {
        void *join_zocket = zsocket_new(ctx, ZMQ_REQ);
        rc = zsocket_connect(join_zocket, "ipc:///tmp/zmq-proxy-%d", port + 5);
        assert(rc == 0);

        zstr_send(join_zocket, "hasher");
        zmsg_t *msg = zmsg_recv(join_zocket);
        zframe_t *f = zmsg_pop(msg);
        assert(f);
        int len = zframe_size(f);
        char *hashes = zframe_strdup(f);
        put_hashes(hashes, len);
        free(hashes);
        zframe_destroy(&f);
        zmsg_destroy(&msg);

        zsocket_destroy(ctx, join_zocket);
    }
    
    zmq_pollitem_t hash_poller = (zmq_pollitem_t) {hash_zocket, 0, ZMQ_POLLIN};
    zmq_pollitem_t ctrl_poller = (zmq_pollitem_t) {ctrl_zocket, 0, ZMQ_POLLIN};
    zmq_pollitem_t sync_poller = (zmq_pollitem_t) {sync_workers, 0, ZMQ_POLLIN};

    zloop_t *reactor = zloop_new();
    
    zloop_poller(reactor, &hash_poller, write_hash, NULL);
    zloop_poller(reactor, &sync_poller, signal_go, NULL);
    zloop_poller(reactor, &ctrl_poller, handle_ctrl, NULL);
    zloop_start(reactor);
    zloop_destroy(&reactor);


    zsocket_destroy(ctx, hash_zocket);
    zctx_destroy(&ctx);
    
    shm_srv_cleanup(semaphores);

    shm_cleanup(1);

    return 0;
}


