% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_read_write_info,[get_accessed_vars/4,
                             recompute_while_read_write_info/2]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'ProBs type checker and type inference module.').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(bsyntaxtree).


recompute_while_read_write_info(TExpr,NewTExpr) :-
   (transform_bexpr(b_read_write_info:update_while_info,TExpr,NewTExpr)
     -> true
     ;  add_internal_error('Call failed: ',transform_bexpr(b_read_write_info:update_while_info,TExpr,NewTExpr)),
        NewTExpr=TExpr).

update_while_info(b(WHILE,subst,OldInfo),b(WHILE,subst,NewInfo)) :-
    WHILE = while(_TPred,_TBody,_TInv,_TVar),
    get_accessed_vars(b(WHILE,subst,OldInfo),[],Modifies,Reads),
    (debug:debug_mode(off) -> true
     ; nl,print('UPDATING WHILE with VARIANT: '),translate:print_bexpr(_TVar),nl,
       format(' WHILE Reads : ~w~n   Modifies : ~w~n',[Reads,Modifies])
    ),
    (delete(OldInfo,reads(OldReads),OI1) -> true ; OldReads=[], OI1=OldInfo),
    (delete(OI1,modifies(OldModifies),OI2) -> true ; OldModifies=[], OI2=OI1),
    ((Reads,Modifies)=(OldReads,OldModifies)
      -> NewInfo=OldInfo
      ;  format('***** different from Reads : ~w~n   Modifies : ~w~n',[OldReads,OldModifies]),
         ord_subtract(Reads,OldReads,NewR),
         ord_subtract(OldReads,Reads,DelR),
         format(' added: ~w~n deleted: ~w~n****~n',[NewR,DelR]),
         NewInfo = [modifies(Modifies),reads(Reads)|OI2]
    ).

% TO DO: also update info for operation call !

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* get_accessed_vars/4 computes the variables that are read or written to
   by a substitution.
   get_accessed_vars(+Subst,+LocalVars,-Modified,-Reads):
    Subst is the substitution (as a typed syntax tree)
    LocalVars is a list of (untyped) ids that are local variables.
       Local variables are ignored and not included in the list of read or written
       variables.
    Modified is the list of (untyped) identifiers that are written by the substitution and
    Reads is the list of (untyped) identifiers that are read. */
get_accessed_vars(TSubst,LocalVars,Modified,Reads) :-
    get_accessed_vars1(TSubst,LocalVars,In,Out),
    extract_reads_and_modifies(In,Out,Modified,Reads).
extract_reads_and_modifies((UnsortedMods,UnsortedReads),([],[]),Modified,Reads) :-
    sort(UnsortedMods,Modified),
    sort(UnsortedReads,Reads).
get_accessed_vars1(TSubst,LocalVars,In,Out) :-
    get_texpr_expr(TSubst,Subst),
    ( Subst = operation_call(Operation,Results,Arguments) ->
        get_texpr_info(TSubst,Info),
        def_get_texpr_id(Operation,op(OperationName)), %print(info(OperationName,Info)),nl,
        memberchk(modifies(MVars),Info),
        memberchk(reads(RVars),Info),
        %op_modifies_reads_cache(OperationName,MVarsC,RVarsC) %this is no longer necessary
        ((ground(MVars), ground(RVars)) -> true
          ; add_error(b_read_write_info, 'Cannot obtain modifies/reads info for called operation (probably illegal local call): ',OperationName),
             % TEST 655 used to reach this branch
            MVars=[], RVars=[]
           % this should only happen when local operation called !?!
           % or ensure that operations are processed before we reach operation calls
           % this may be problematic for local operations
        ),
        %print(compute_acc_vars_for_op_call(modifies(MVars),Subst)),nl,trace,
        accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments,In,Out)
    ; get_accessed_vars2(Subst,LocalVars,In,Out) ->
        true
    ; otherwise ->
        add_internal_error('Could not extract information about accessed variables for substitution',
                           get_accessed_vars2(Subst,LocalVars,In,Out)), fail).
accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments) -->
    add_modified_vars2(MVars),
    add_read_vars2(RVars),
    add_modified_vars(Results,LocalVars),
    add_read_vars_l(Arguments,LocalVars).

get_accessed_vars2(skip,_) --> !.
get_accessed_vars2(block(Subst),LocalVars) --> get_accessed_vars1(Subst,LocalVars).
get_accessed_vars2(assertion(A,Subst),R) -->   add_read_vars(A,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assign(LHS,E),R) -->        %Note: LHS can be nested function application f(a)(b) :=
      {maplist(get_lhs_assigned_identifier,LHS,Ids)}, add_modified_vars(Ids,R),
      {convlist(get_lhs_read_expression,LHS,LHS_R), append(LHS_R,E,RE)}, add_read_vars_l(RE,R).
get_accessed_vars2(assign_single_id(Id,E),R) --> 
      add_read_vars_l([E],R), % Note: LHS is a single Identifier: nothing is ever read there
      add_modified_vars([Id],R).
get_accessed_vars2(precondition(P,Subst),R) -->add_read_vars(P,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assertion(Subst),R) -->     get_accessed_vars1(Subst,R).
get_accessed_vars2(choice(Substs),R) -->       get_accessed_vars_for_choice(Substs,R).
get_accessed_vars2(if(Substs),R) -->           get_accessed_vars_for_if(Substs,R).
get_accessed_vars2(select(Whens,Else),R) -->   get_accessed_vars_for_select(Whens,Else,R).
get_accessed_vars2(select(Whens),R) -->        get_accessed_vars_for_select(Whens,none,R).
get_accessed_vars2(case(Expr,Substs,Else),R) -->  get_accessed_vars_for_case(Expr,Substs,Else,R).
get_accessed_vars2(case_or(Exprs,Subst),R) --> add_read_vars_l(Exprs,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(any(Ids,G,Subst),R) -->     append_vars(R,Ids,L),add_read_vars(G,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(let(Ids,G,Subst),R) -->     append_vars(R,Ids,L),add_read_vars(G,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(becomes_element_of(Ids,E),R) --> add_modified_vars(Ids,R),add_read_vars(E,R).
get_accessed_vars2(becomes_such(Ids,E),LocalVars0) -->  add_modified_vars(Ids,LocalVars0),
    {append_primed_vars(Ids,LocalVars0,LocalVars1)}, % all Ids are already marked as accessed, any reference to Id$0 is a read of the old value; do not add Id$0 to the accessed variables
    add_read_vars(E,LocalVars1).
get_accessed_vars2(evb_becomes_such(Ids,E),R) --> add_modified_vars(Ids,R),add_read_vars(E,R).
get_accessed_vars2(sequence(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(parallel(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(while(Cond,Subst,Inv,Variant),R) -->
    get_accessed_vars_for_while(Cond,Subst,Inv,Variant,R).
get_accessed_vars2(var(Ids,Subst),R)         --> append_vars(R,Ids,L),get_accessed_vars1(Subst,L).
% TODO: we need information about which variables are modified by an external function
get_accessed_vars2(external_subst_call(_Name,Exprs),R) --> add_read_vars_l(Exprs,R).

get_accessed_vars_for_while(Cond,Subst,Inv,Variant,Locals) -->
    get_accessed_vars_for_while2(Inv,Locals),
    add_read_vars(Cond,Locals),
    add_read_vars(Variant,Locals),
    get_accessed_vars1(Subst,Locals).
get_accessed_vars_for_while2(Inv,Locals,In,Out) :-
    add_read_vars(Inv,Locals,LI,LO),
    extract_reads_and_modifies(LI,LO,Modified,Reads0),
    maplist(remove_optional_prime,Reads0,Reads),
    add_read_vars2(Reads,In,Inter),
    add_modified_vars2(Modified,Inter,Out).
remove_optional_prime(Id,Orig) :-  atom_concat(Orig,'$0',Id),!.
remove_optional_prime(Id,Id).

get_accessed_vars_for_if(ElseIfs,Locals) -->
    { maplist(split_elseif,ElseIfs,Conds,Subst),
      ( if_has_else_block(Conds) -> HasElse = has_else
      ; otherwise                -> HasElse = no_else) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_elseif(Expr,Cond,Subst) :- get_texpr_expr(Expr,if_elsif(Cond,Subst)).

get_accessed_vars_for_select(Whens,Else,Locals) -->
    { maplist(split_when,Whens,Conds,Subst1),
      ( Else = none ->
          HasElse=no_else, Subst = Subst1
      ; otherwise ->
          HasElse=has_else, Subst = [Else|Subst1]) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_when(Expr,Cond,Subst) :- get_texpr_expr(Expr,select_when(Cond,Subst)).

get_accessed_vars_for_if_when(HasElse,Subst,Locals) -->
    { get_accessed_vars_for_if_when2(HasElse,Subst,Locals,Modified,Reads) },
    add_read_vars2(Reads),
    add_modified_vars2(Modified).

get_accessed_vars_for_if_when2(has_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    % We want to calculate those elements that are modified by some (in the union)
    % but not by all (in the intersection)
    ord_union(Modifies,ModifiedBySome),
    ord_intersection(Modifies,ModifiedByAll),
    ord_subtract(ModifiedBySome,ModifiedByAll,ModifiedByNotAll),
    % we add those values that are not modified by all to the set of read variables
    % because e.g. "IF P THEN x:=1 END" is equivalent to "IF P THEN x:=1 ELSE x:=x END"
    ord_union([ModifiedByNotAll|Reads],ReadBySome).
get_accessed_vars_for_if_when2(no_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    % no else block present, we must assume that each written variable
    % can also be treaten like a read variable (see example in the other condition)
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    ord_union(Modifies,ModifiedBySome),
    ord_union([ModifiedBySome|Reads],ReadBySome).

get_accessed_vars_for_choice(Substs,Locals) -->
    get_accessed_vars_for_if_when(has_else,Substs,Locals).

get_accessed_vars_for_case(Expr,Eithers,Else,Locals) -->
    % we can treat a case statement analogously to an IF-THEN-ELSE statement
    % where an ELSE is present
    % additionally, we have to add all expressions to the read set
    add_read_vars(Expr,Locals),
    { maplist(split_case_or,Eithers,ExprsList,Substs), append(ExprsList,Exprs) },
    add_read_vars_l(Exprs,Locals),
    get_accessed_vars_for_if_when(has_else,[Else|Substs],Locals).
split_case_or(CaseOr,Exprs,Subst) :- get_texpr_expr(CaseOr,case_or(Exprs,Subst)).

% just the same as get_accessed_vars, but with different order of parameters
% to allow call with maplist
get_accessed_vars_x(Locals,Subst,Modifies,Reads) :-
    get_accessed_vars(Subst,Locals,Modifies,Reads).

if_has_else_block(Conds) :-
    last(Conds,Cond),is_truth(Cond).

append_vars(Vars,TIds,NewVars,X,X) :-
    get_texpr_ids(TIds,Ids),
    append(Vars,Ids,NewVars).

add_read_vars_l([],_LocalVars) --> !.
add_read_vars_l([Expr|Rest],LocalVars) -->
    add_read_vars(Expr,LocalVars),
    add_read_vars_l(Rest,LocalVars).

% extract modified identifier from LHS of assign:
get_lhs_assigned_identifier(b(function(LHS,_),_,_),Res) :- !, get_lhs_assigned_identifier(LHS,Res).
get_lhs_assigned_identifier(ID,ID).

% extract part of LHS of assign which reads information:
get_lhs_read_expression(b(function(LHS,Read1),T,I),Res) :- !,
    Res = b(function(LHS,Read1),T,I).
% old version which for f(a) := b only  marked a as read in LHS f(a); now we also mark f
% indeed f(a) := b is equivalent to f := f <+ {a|->b} ; so the old value of f is implicitly read (important for ltsmin symbolic model checking)
%get_lhs_read_expression(b(function(LHS,Read1),_,_),Res) :- !,
%   (get_lhs_read_expression(LHS,Read2) -> create_couple(Read1,Read2,Res)
%     ; Res = Read1).


add_read_vars(Expression,LocalVars) -->
    {find_identifier_uses(Expression,LocalVars,ReadVars)},
    add_read_vars2(ReadVars).
add_read_vars2(Vars,(Modified,InRead),(Modified,OutRead)) :- append(Vars,OutRead,InRead).

add_modified_vars(Ids,LocalVars) -->
    {filter_result_vars(Ids,LocalVars,ModifiedVars)},
    add_modified_vars2(ModifiedVars).
add_modified_vars2(Vars,(InModified,Read),(OutModified,Read)) :-
    (var(Vars) -> add_internal_error('Modified Vars variable: ',add_modified_vars2(Vars,(InModified,Read),(OutModified,Read))) ; true),
    append(Vars,OutModified,InModified).

:- use_module(tools,[remove_all/3]).
filter_result_vars(Ids,LocalVars,ModifiedVars) :-
    % extract_changed_id might fail if the user entered
    % a wrong LHS of an assignment. This should lead to an
    % type error, handled by the assign type checking rule.
    convlist(extract_changed_id,Ids,Names),
    remove_all(Names,LocalVars,ModifiedVars).
extract_changed_id(Id,Name) :-
    get_texpr_id(Id,Name),!.
extract_changed_id(Func,Name) :-
    get_texpr_expr(Func,function(TId,_)), !,
    extract_changed_id(TId,Name). % we can have nested function calls f(x)(y) := z
extract_changed_id(X,Name) :-
    add_internal_error('Could not extract changed id: ',extract_changed_id(X,Name)),fail.

get_accessed_vars_l([],_) --> [].
get_accessed_vars_l([S|Rest],R) --> get_accessed_vars1(S,R),get_accessed_vars_l(Rest,R).



:- use_module(tools,[number_suffix/3]).
:- assert_must_succeed(( b_read_write_info:append_primed_vars([b(identifier(x),integer,[])],[],R), R = ['x$0'])).
append_primed_vars(TIds,R,L) :-
    maplist(add_dollar_zero,TIds,Primed),
    append(Primed,R,L).
:- assert_must_succeed(( b_read_write_info:add_dollar_zero(b(identifier(x),integer,[]),R), R = 'x$0')).
add_dollar_zero(TId,P) :-
    get_texpr_id(TId,Id),
    number_suffix(Id,0,P).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

