% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_show_history,[write_history_to_file/1, write_history_to_file/2,
   write_history_to_user_output/0,write_history_to_user_output/1,
   write_values_to_file/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'Write animator history into a file in B format.').

:- use_module(library(lists)).

:- use_module(state_space,         [transition/4,visited_expression/2,history/1, op_trace_ids/1,
                                    get_action_trace/1,current_expression/2, current_state_id/1]).
:- use_module(specfile,            [state_corresponds_to_initialised_b_machine/1,
                                    expand_to_constants_and_variables/3]).
:- use_module(translate,           [translate_bvalue/2,
                                    translate_event/2]).
:- use_module(bmachine,            [b_get_machine_operation/4]).
:- use_module(bsyntaxtree,         [get_texpr_expr/2]).
:- use_module(kernel_waitflags,    [init_wait_flags/1,ground_wait_flags/1]).
:- use_module(b_state_model_check, [set_up_transition/7]).

:- use_module(error_manager,       [add_internal_error/2]).

:- use_module(preferences).

:- use_module(tools_io).
:- use_module(debug).
:- use_module(store,[lookup_value/3]).

write_history_to_file(Filename) :- 
    write_history_to_file(Filename,[show_init]).
write_history_to_file(Filename,Options) :- 
    debug_println(9,write_history_to_file(Filename,Options)),
    assert_options(Options),
    safe_open_file(Filename,write,_,[alias(histout)]),
    get_preference(expand_avl_upto,Max),
    set_preference(expand_avl_upto,-1), % means no limit
    call_cleanup(write_history, (close(histout),set_preference(expand_avl_upto,Max))).

write_history_to_user_output :-
   write_history_to_user_output([show_init]).
write_history_to_user_output(Options) :-
    assert_options(Options),
    write_history.

assert_options(Options) :-
    retractall(option(_)),
    maplist(assert_option,(Options)).
assert_option(Option) :-
    ( valid_option(Option) -> assert(option(Option))
    ; otherwise            -> add_internal_error('unknown history option',assert_option(Option))).

:- volatile option/1.
:- dynamic option/1.
valid_option(show_init).
valid_option(show_states).
valid_option(show_any_vars).

write_history :- option(show_any_vars),!,
    print('### WARNING: using old, inefficient version of write_history for show_any_vars'),nl,
    old_write_history.
write_history :- 
    get_state_id_trace(StateIds),
    op_trace_ids(OpIds), %print(Tuples),nl,
    reverse(OpIds,ForwardOpIds),
    %print(write(StateIds,ForwardOpIds)),nl,
    write_history_state(StateIds,ForwardOpIds,not_initialized).

% alternate between writing states and operations
write_history_state([],Ops,LastStateInit) :- % we have reached the last state
  (Ops=[] -> true ; add_internal_error('Extra ops: ',write_history_state([],Ops,LastStateInit))).
write_history_state([StateID|T],Ops,LastStateInit) :-  %print(stateid(StateID)),nl, 
    (option(show_states)
     -> write_state_id(StateID,Init)
     ;  (option(show_init), LastStateInit=not_initialized)
         -> write_state_id(StateID,Init)
         ; Init = initialized
    ),
    write_history_op(Ops,T,StateID,Init).

% now write an operation, and then call write_history_state to print the next state
write_history_op([],States,_,_) :- 
  (States=[] -> true ; add_internal_error('Extra states: ',write_history_op([],States,_,_))).
write_history_op([OpID|T],States,StateID,Init) :- %print(op(OpID)),nl,
    (OpID=jump(Dest)
     ->  hwrite('/* JUMPING TO ANOTHER STATE : '), hwrite(Dest), hwrite('*/\n')
   % TO DO: check after jumping whether constants have changed in case states are written out
      ;  transition(StateID,OperationTerm,OpID,_),
         write_operation(OperationTerm,T)
    ),
    write_history_state(States,T,Init).


write_state_id(root,Init) :- !, Init=not_initialized.
write_state_id(ID,Init) :- visited_expression(ID,S), write_state(S,Init),!.
write_state_id(ID,Init) :- add_internal_error('Failed: ',write_state_id(ID,Init)),
  Init=initialized.

write_state(csp_and_b_root,Init) :- !, Init=not_initialized.
write_state(csp_and_b(_CSP,BState),Init) :- !, write_state(BState,Init).
write_state(concrete_constants(CstStore),not_initialized) :- write_store(CstStore,'Constants').
write_state(const_and_vars(_ID,VarStore),initialized) :- write_store(VarStore,'Variables').
write_state([],initialized) :- write_store([],'Variables').
write_state([H|T],initialized) :- write_store([H|T],'Variables').

write_operation(F,_) :- functor(F,'$setup_constants',_),!.
write_operation(F,_) :- functor(F,'$initialise_machine',_),!, 
   ((option(show_states) ; option(show_init)) -> hwrite('/* Initialisation */\n') ; true).
write_operation('$JUMP'(J),_) :- !, hwrite('/* JUMPING TO ANOTHER STATE : '),
  hwrite(J), hwrite('*/\n').
write_operation(Operation,T) :-  write_operation1(Operation),
    (T==[] -> hnl ; otherwise -> hwrite(';\n')).
write_operation1(Operation) :-
   (translate_event(Operation,OpStr) ->
        hwrite(OpStr) ; add_internal_error('Failed: ',translate_event(Operation,_))).

get_state_id_trace(StateIds) :-
    history(History),
    %print(history(History)),nl,
    current_state_id(CurID),
    reverse([CurID|History],StateIds).

    
    
old_write_history :-
    get_trace(Initialisation,Ops,Stores),
    % tools:print_message(get_trace(Initialisation,Ops,Stores)),
    ( option(show_init) -> write_initialisation(Initialisation); true),
    write_trace(Ops,Stores,Initialisation).

get_trace(First,Trace,Rest) :-
    get_action_trace(Tuples), maplist(extract_op,Tuples,Operations),
    history(History),
    current_state_id(CurId),
    reverse([CurId|History],StateIds),
    reverse(Operations,Trace1),
    maplist(extract_store,StateIds,Stores),!,
    find_first_initialised_state(Stores,Trace1,First,Rest,Trace).

extract_op(action(_,B),R) :- !, R=B.
extract_op(J,'$JUMP'(J)). % e.g., inserted by various cbc operations into history

extract_store(Id,store(Store,ConstStore,VarStore)) :-
    visited_expression(Id,Store),
    expand_to_constants_and_variables(Store,ConstStore,VarStore).

find_first_initialised_state([Store|Srest],OpTrace,InitialStore,Rrest,Trace) :-
    Store = store(S,_,_),
    ( state_corresponds_to_initialised_b_machine(S) ->
        Rrest = Srest, OpTrace = Trace, InitialStore = Store
    ; otherwise -> OpTrace=[_|Trest],
        find_first_initialised_state(Srest,Trest,InitialStore,Rrest,Trace)).

write_trace([],Stores,P) :-
   (Stores = [] -> true ; add_internal_error('Extra Stores: ',write_trace([],Stores,P))).
write_trace([Operation|OpRest],[Store|Srest],PrevStore) :-
    list_is_empty(OpRest,IsLastOp),
    (write_operation(Operation,PrevStore,Store,IsLastOp) -> true
      ; add_internal_error('Call failed: ',write_operation(Operation,PrevStore,Store,IsLastOp))),
    ( option(show_states)
      -> (write_variable_store(Store) -> true ; add_internal_error('Call failed: ',write_variable_store(Store)))
      ; true),
    write_trace(OpRest,Srest,Store).

list_is_empty([],true).
list_is_empty([_|_],false).


write_initialisation(store(_,ConstStore,VarStore)) :-
    write_store(ConstStore,'Constants'),
    write_store(VarStore,'Initialisation').
write_variable_store(store(_,_,VarStore)) :-
    write_store(VarStore,'Variables').

write_store(Store,Info) :-
    hwrite('/* '),hwrite(Info),hnl,
    write_store2(Store),
    hwrite('*/\n').
write_store2(Bindings) :- write_store3(Bindings,head).
write_store3([],_).
write_store3([bind(Name,Value)|Srest],Pos) :-
    ( Pos==head -> hwrite('   '); Pos==tail -> hwrite(' & ')),
    hwrite(Name),hwrite(' = '),
    translate_bvalue(Value,ValStr),hwrite(ValStr),hnl,
    write_store3(Srest,tail).

write_operation('$JUMP'(J),_,_,_) :- !, hwrite('/* JUMPING TO ANOTHER STATE : '),
  hwrite(J), hwrite('*/\n').
write_operation(Operation,Prev,Next,IsLastOp) :-
    ( translate_event(Operation,OpStr) ->
        hwrite(OpStr),
        ( option(show_any_vars) ->
            show_any_vars(Operation,Prev,Next)
        ; otherwise -> true),
        ( IsLastOp == true -> hnl ; otherwise -> hwrite(';\n'))
    ; otherwise ->
        add_internal_error('Call failed: ',translate_event(Operation,_))).

show_any_vars(Operation,Prev,Next) :-
    functor(Operation,OpName,_),
    has_any(OpName),!,
    extract_anys(OpName,Operation,Prev,Next,AnyBindings),
    hwrite(':\n'),write_store2(AnyBindings).
show_any_vars(_,_,_).

% THIS is really bad performance wise: avoid recomputing transitions !
% TO DO: only present info also stored in transition
extract_anys(OpName,Operation,store(_,ConstantState,Prev),store(_,_,Next),AnyBindings) :-
    init_wait_flags(WF),
    set_up_transition(OpName,Operation,ConstantState,Prev,Next1,TransInfo,WF),
    unify_states(Next,Next1),
    ground_wait_flags(WF),
    memberchk(path(Path),TransInfo),
    extract_anys2(Path,AnyBindings).
extract_anys2(pre(P),AnyBindings) :- extract_anys2(P,AnyBindings),!.
extract_anys2(any(AnyBindings,_),AnyBindings).

unify_states([],_).
unify_states([bind(Id,V)|R],Store) :-
    lookup_value(Id,Store,V),
    unify_states(R,Store).
    

has_any(OpName) :-
    b_get_machine_operation(OpName,_,_,Body),
    has_any2(Body).
has_any2(TSubst) :-
    get_texpr_expr(TSubst,Subst),
    has_any3(Subst).
has_any3(precondition(_,S)) :- has_any2(S),!.
has_any3(any(_,_,_)).

% catch exception to allow to call write_history directly and write to user_output
hwrite(X) :- on_exception(error(domain_error(stream_or_alias,histout),_),
         write(histout,X),
         write(user_output,X)).
hnl :- on_exception(error(domain_error(stream_or_alias,histout),_),nl(histout),nl(user_output)).

% write all values of variables and constants to a file
write_values_to_file(Filename) :-
    safe_open_file(Filename,write,_,[alias(histout)]),
    get_preference(expand_avl_upto,Max),
    set_preference(expand_avl_upto,-1), % means no limit
    call_cleanup(write_all_values, (close(histout),set_preference(expand_avl_upto,Max))).

write_all_values :-
    current_expression(_,CurStore),
    expand_to_constants_and_variables(CurStore,ConstStore,VarStore),
    ( ConstStore = [] -> H=head
    ; otherwise       -> hwrite('/* Constants */\n'), write_store2(ConstStore),H=tail
    ),
    ( VarStore = [] -> true
    ; otherwise     -> hwrite('/* Variables */\n'), write_store3(VarStore,H)).
