% (c) 2013-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bmachine_static_checks, [static_check_machine/1, toplevel_raw_predicate_sanity_check/2]).

:- use_module(probsrc(module_information)).
:- module_info(group,typechecker).
:- module_info(description,'Static checking of B machines upon construction. Detection of name clashes and uninitalised variables.').

:- use_module(bmachine_structure, [get_section/3]).
:- use_module(b_read_write_info, [get_accessed_vars/4]).
:- use_module(bsyntaxtree, [get_texpr_expr/2,get_texpr_id/2,get_texpr_ids/2]).
:- use_module(error_manager, [add_warning/2, add_warning/3, add_warning/4]).
:- use_module(tools, [ajoin/2]).
:- use_module(library(lists)).

static_check_machine(Machine) :-
    initialises_all_variables(Machine),
    check_name_clashes(Machine).


% ---------------------------------------------
% check for unnatural invariants and properties
% ---------------------------------------------
% the priority of the implication is sometimes surprising
%  a & (x=1) => (y=2) & c   --> top-level symbol is the implication !
% the test is done on the raw predicate before ast_cleanup removes potential typing predicates; see tests 106, 107
toplevel_raw_predicate_sanity_check(invariant,RawPredicate) :-
    check_top_level(RawPredicate,'INVARIANT'),
    fail.
toplevel_raw_predicate_sanity_check(properties,RawPredicate) :-
    check_top_level(RawPredicate,'PROPERTIES'),
    fail.
toplevel_raw_predicate_sanity_check(_,_).

check_top_level(implication(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level implication (=>) in clause: ',SECT).
check_top_level(disjunct(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level disjunction (or) in clause: ',SECT).
check_top_level(equivalence(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level equivalence (<=>) in clause: ',SECT).

    
% ---------------------
% Checks is all variables are initialised by the machine
% ---------------------
initialises_all_variables(Machine) :-
    get_section(initialisation,Machine,Initialisation),
    % check wich variables are read / modified by the initialisation
    get_accessed_vars(Initialisation,[],Modifies,_Reads),
    % get all variables that should be initialised
    get_section(abstract_variables,Machine,AbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    append(AbsVars,ConcVars,TAllVars),
    get_texpr_ids(TAllVars,AllVars),
    % check for each variable, if the initialisation modifies it
    exclude(is_initialised(Modifies),AllVars,Uninitialised),
    % generate a warning if unitialised is not empty
    generate_uninitialised_warning(Uninitialised).

is_initialised(Modifies,Id) :-
    memberchk(Id,Modifies).

generate_uninitialised_warning([]) :- !.
generate_uninitialised_warning(Vars) :-
    ajoin(['Machine may not initialise some of its variables: ', Vars], Msg),
    add_warning(bmachine_static_checks,Msg).


% ---------------------
% Checks if an operations parameter or local variable clashes with a variable
% ---------------------
check_name_clashes(Machine) :-
    get_section(operation_bodies,Machine,Operations),
    % get all variables and constants that might clash
    get_section(abstract_variables,Machine,AbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    get_section(abstract_constants,Machine,AbsCons),
    get_section(concrete_constants,Machine,ConcCons),
    append([AbsVars,ConcVars,AbsCons,ConcCons],AllVarsAndConstants),
    get_texpr_ids(AllVarsAndConstants,AllIds),
    % for each operation, compare the parameter names with existing vars/constants
    maplist(name_clashes(AllIds),Operations).

:- use_module(bsyntaxtree,[map_over_typed_bexpr/2]).

name_clashes(AllIds,Operation) :-
    get_texpr_expr(Operation,operation(IdFull,Results,Params,Subst)),
    get_texpr_id(IdFull,op(Id)),
    get_texpr_ids(Results,ResultIds), get_texpr_ids(Params,ParamIds),
    include(clash_warnings(parameter,AllIds,Id),ParamIds,_ParamsCausingWarning),
    include(clash_warnings(result,AllIds,Id),ResultIds,_ResultsCausingWarning),
    (map_over_typed_bexpr(bmachine_static_checks:local_variable_clash(AllIds,Id),Subst),fail ; true).

:- public local_variable_clash/3.
local_variable_clash(AllIds,Operation,TSubst) :-
   get_texpr_expr(TSubst,Subst),
   introduces_local_variable(Subst,ID),
   %print(check(ID,Operation)),nl,
   clash_local_warnings(AllIds,Operation,ID,TSubst).

% check for constructs which introduced local variables
introduces_local_variable(var(Parameters,_),ID) :-   % currently B-interpreter cannot correctly deal with this in the context of operation_call 
   member(TID,Parameters), get_texpr_id(TID,ID).

clash_warnings(ResultOrParameter,AllIds,OperationId,Name) :-
    member(Name,AllIds), !,
    ajoin(['Operation ', ResultOrParameter, ' "', Name, '" has the same name as a variable or constant in operation "', OperationId,'".'], Msg),
    add_warning(bmachine_static_checks,Msg).

clash_local_warnings(AllIds,OperationId,Name,Pos) :-
    member(Name,AllIds), !,
    ajoin(['Local variable has the same name as a variable or constant in operation ', OperationId,': '], Msg),
    add_warning(bmachine_static_checks,Msg,Name,Pos).

% TODO: this does not and can not work here:
% - Some preconditions are removed (typing only....) during machine simplification
% - Needs to be verified during typechecking
% ---------------------
% Checks if an operations parameter is not typed by a pre condition
% ---------------------
%parameters_without_pre_condition(Machine) :-
%    get_section(operation_bodies,Machine,Operations),
%    maplist(parameters_without_pre_condition_aux,Operations).
%
%parameters_without_pre_condition_aux(Operation) :-trace,
%    get_texpr_expr(Operation,operation(IdFull,_Results,Params,Subst)),
%    get_texpr_id(IdFull,op(Id)),
%    (Params == []
%     -> true % no parameters
%     ;  (get_texpr_expr(Subst,precondition(_,_))
%         -> true % parameters and precondition
%         ;  ajoin(['Operation ', Id, ' has parameters but no pre-condition.'], Msg),
%            add_warning(bmachine_static_checks,Msg)
%          )).