% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(csp_sets,[ %is_a_set/1, % reported as an unnecessary export by infolog
                    %evaluate_set/3,
					evaluate_set/2, 
          % force_evaluate_set/2, % export superfluous (used in haskell_csp.pl)
                    evaluate_closure/2,
                    is_empty_set/2,
                    is_member_set/2,
                    is_subset_of/2, % reported as superfluous by infolog, though predicate called by meta call in haskell_csp.pl (see last two clauses of check_boolean_expression/1 in haskell_csp.pl)
                    is_member_comprehension_set/3,
                    extract_variables_from_generator_list/2,
                    try_get_cardinality_for_wait_flag/2,
                    subsets/2, enum_subset/2,
                    cardinality/2,
                    singleSetElement/3,
                    union_set/3,diff_set/3,inter_set/3,
                    equal_element/2, not_equal_element/2,
                    expand_set_comprehension/3, replicate_expand_set_comprehension/3,
                    expand_symbolic_set/3,
                    big_union/2, big_inter/2,
                    unify_also_patterns/3,
                    closure_expand/2,
                    is_member_set_alsoPat/2
                    %,csp_full_expanded_type/2
                    %,expand_int_value/2
                    ]).

:- use_module(probsrc(module_information)).
:- module_info(group,csp).
:- module_info(description,'Operations on CSP sets.').

/******* SICSTUS libraries *******/
:- use_module(library(lists)).
%:- load_files(library(detcheck), [when(compile_time), if(changed)]).
/******* ---------------- *******/

/*************** PROB modules ****************/
:- use_module(probsrc(tools),[remove_variables/3,flatten/2,exact_member/2]).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(self_check)).
%% :- use_module(probsrc(preferences),[preference/2]).
%% :- use_module(probsrc(debug)).
%-------- CSP modules:
:- use_module(probcspsrc(haskell_csp_analyzer),[is_csp_constructor/1]).
:- use_module(probcspsrc(haskell_csp),
      [is_a_datatype/2, csp_constructor/3, channel/2, dataTypeDef/2,channel_type_list/2,
       evaluate_argument/2,force_evaluate_argument/2,force_evaluate_argument_for_member_check/2,evaluate_int_argument/2,
       check_boolean_expression/1, enumerate_channel_input_value/4,enumerate_datatype_el/5]).
:- use_module(probcspsrc(csp_sequences)).
:- use_module(probcspsrc(haskell_csp_analyzer),[csp_full_type_constructor/3,csp_full_type_constant/2]).

/*************** ----------- ****************/

% --------------------------------------------------------
%                       SETS
% --------------------------------------------------------

% Possible sets:
%  setValue([])  
%  setValue([el1,...,eln])  sorted, without duplicates
%  setFromTo(int(Low),int(Up))
% ... intType,....

:- assert_must_succeed(is_a_set(setFrom(1))).
:- assert_must_succeed(is_a_set(dataType(bool))).
:- assert_must_succeed(is_a_set(dotTupleType(list([int,int])))).
:- assert_must_succeed(is_a_set(setExp(val,int))).
:- assert_must_succeed(is_a_set(closureComp(_,_))).
:- assert_must_succeed(is_a_set('Seq'(setValue([1,2,3])))).

is_a_set(setValue(_)).
is_a_set(closure(_)).
is_a_set(setFromTo(_,_)).
is_a_set(setFrom(_)).
is_a_set(intType).
is_a_set(boolType).
is_a_set(dataType(_)).
is_a_set(dotTupleType(_)).
is_a_set(setExp(_,_)).
is_a_set(closureComp(_,_)).  % right?
is_a_set('Seq'(_)).
%is_a_set(DT) :- dataTypeDef(DT,_).


:- assert_must_succeed(( csp_sets:evaluate_set([int(3)],R), R == setValue([int(3)]) )).

evaluate_set([],R) :- !, R=setValue([]).
evaluate_set([H|T],R) :- !, mnf(haskell_csp:evaluate_argument(H,EH)), evaluate_set(T,ET),
      add_element(EH,ET,R).
evaluate_set(X,_) :- add_internal_error('Internal Error: Could not evaluate: ',evaluate_set(X)),fail.

force_evaluate_set([],R) :- !, R=setValue([]).
force_evaluate_set([H|T],R) :- !, mnf(haskell_csp:force_evaluate_argument(H,EH)), 
      force_evaluate_set(T,ET),
      add_element(EH,ET,R).
force_evaluate_set(X,_) :- add_internal_error('Interal Error: Could not evaluate: ',force_evaluate_set(X)),fail.


/* This implementation of evanluate_set/2 make the Basin_Bank_CSP benchmark's performnance slower. */ 
/*
% Functor can be evaluate_argument/2 or force_evaluate_argument/2
% Clause match is only possible through internal call. CSP-M parser always provide list inside rangeEnum().
% See implementation of haskell_csp: evaluate_set_expression/2.
evaluate_set(L,Set,Functor) :- is_list(L),!,
   evaluate_set(L,setValue([]),Set,Functor).
evaluate_set(X,_,_Functor) :-
   add_internal_error('Internal Error: Could not evaluate: ',evaluate_set(X)),fail.

evaluate_set([],Set,Set,_Functor).
evaluate_set([H|T],Set,R,Functor) :-
   % this variant of defining the argument call is less memory wasteful than the ordinary way (Call =.. [Functor|Args])
   functor(Call,Functor,2),
   arg(1,Call,H),arg(2,Call,EH),
   mnf(haskell_csp:Call),!,
   add_element(EH,Set,Set1),
   evaluate_set(T,Set1,R,Functor).
*/

evaluate_closure(X,closure(RS)) :- (X=tuple(XList) -> true ; X=XList),
    evaluate_set(XList,setValue(RS)).%evaluate_set(XList,setValue(RS),evaluate_argument).

:- use_module(probcspsrc(csp_sequences),[is_empty_list/2]).
:- use_module(probcspsrc(csp_basic)).

:- assert_must_succeed((csp_sets: is_empty_set(setValue([setValue([])]),R), R == false)).
:- assert_must_succeed((csp_sets: is_empty_set(setValue([]),R), R == true)).
:- assert_must_succeed((csp_sets: is_empty_set(setFromTo(3,1),R), R == true)).
:- assert_must_succeed((csp_sets: is_empty_set(closure([tuple(ack),tuple(rec)]),R), R == false)).
:- assert_must_succeed((csp_sets: is_empty_set(setFrom(1),R), R == false)).

:- block is_empty_set(-,?).
is_empty_set(setValue(X),R) :- !, is_empty_list(X,R).
is_empty_set(closure(X),R) :- !, is_empty_list(X,R).
is_empty_set(setFromTo(Low,Up),R) :- !, safe_less_than(Up,Low,R).
is_empty_set(setFrom(_),R) :- !, R=false.
is_empty_set(X,_) :- add_error(csp_sets,'Could not evaluate: ',is_empty_set(X)),fail.


:- assert_must_succeed(( csp_sets:is_member_set(X,setValue([int(1),int(9)])), X=int(9) )).
:- assert_must_fail(( csp_sets:is_member_set(X,setValue([int(1),int(9)])), X=int(3) )).
:- assert_must_succeed((csp_sets: is_member_set(int(10),setFrom(5)))).
:- assert_must_fail((csp_sets: is_member_set(int(3),setFrom(5)))).
:- assert_must_succeed( is_member_set(na_tuple([int(3)]),typeTuple([setFromTo(1,10)])) ).
:- assert_must_succeed( is_member_set(na_tuple([int(3),int(4)]),typeTuple([setFromTo(1,10),setValue([int(1),int(2),int(3),int(4)])])) ).
:- assert_must_succeed( is_member_set(tuple([int(3),int(4)]),dotTupleType([setFromTo(1,10),setValue([int(1),int(2),int(3),int(4)])])) ).

is_member_set(El,S) :- is_member_set2(S,El).

:- block is_member_set2(-,?).
is_member_set2(setValue(Set),El) :- !, blocking_member(El,Set).
is_member_set2(boolType,X) :- !, (X=true;X=false).
is_member_set2(intType,R) :- !, R=int(_).
is_member_set2(setFromTo(Low,Up),R) :- !,  R=int(X),
   is_member_from_to(X,Low,Up).
is_member_set2(setFrom(Low),int(X)) :- !, is_member_from(X,Low).
is_member_set2('Seq'(X),C) :- expand_sequence(C,list(EC)), !, list_elements_member_set(EC,X).
is_member_set2('dotTupleType'(X),T) :- !,(T=tuple(TT) ; T=dotTuple(TT)), l_dot_is_member_set(TT,X). %%%%%% see trace output
is_member_set2('typeTuple'(X),T) :-  !, T=na_tuple(TT), l_is_member_set(TT,X).
is_member_set2(dataType(DT),C) :- is_a_datatype(DT,L),!, % to do: precompute this
   ( (atomic(C),member(constructor(C),L)) ; ( C=record(Cons,Fields),
               csp_constructor(Cons,DT,ArgSubTypes),
               maplist(haskell_csp:get_value_alsoPat,Fields,Fields1), % could be possible that some of the elements are wrapped in in(.) or alsoPat(.,.)
               l_dot_is_member_set(Fields1,ArgSubTypes) )
   ).
is_member_set2(setExp(RangeExpr),C) :- !, is_member_set2(setExp(RangeExpr,[]),C).
is_member_set2(setExp(RangeExpr,GeneratorSet),C) :- !, 
    is_member_comprehension_set(C,RangeExpr,GeneratorSet).
   % print(is_member_comprehension_set(C,Tuple,GeneratorSet)),nl.
  % what about closureComp ?
is_member_set2('Union'(LS),El) :- ground(El),!,is_member_union(LS,El).
is_member_set2('Union'(LS),El) :- !,force_evaluate_argument('Union'(LS),ES), is_member_set2(ES,El).
is_member_set2('Inter'(LS),El) :- !,is_member_inter(LS,El).
is_member_set2(agent_call(Span,F,Par),El) :- !, haskell_csp: unfold_function_call_once(F,Par,Body,Span),
   force_evaluate_argument_for_member_check(Body,R), is_member_set(El,R).
is_member_set2(closure(Cl),El) :- !, closure_expand(Cl,R),is_member_set(El,R).
is_member_set2(S,El) :- haskell_csp: name_type(S,Type),!,is_member_set2(Type,El).
is_member_set2(R,X) :- add_error(csp_sets,'Could not evaluate: ',is_member_set(X,R)),fail.

:- assert_must_fail(csp_sets: is_member_union(setExp(rangeEnum([])),int(1))).

:- block is_member_union(-,?).
is_member_union(LS,El) :- 
   (deconstruct_set_of_sets(LS,H,T) ->
       %print(lazy_Union(El,H,T)),nl,  % args should not be setComp; otherwise we have problem with cut below
       (is_member_set2(H,El) -> true ; is_member_set2('Union'(T),El))
     ; empty_set_of_sets(LS) -> fail
     ; add_error_fail(is_member_set,'Illegal argument: ','Union'(LS))).
:- block is_member_inter(-,?).
is_member_inter(LS,El) :- 
 (deconstruct_set_of_sets(LS,H,T) ->
       is_member_set2(H,El), 
       (empty_set_of_sets(T) -> true ; is_member_set2('Inter'(T),El))
     ; empty_set_of_sets(T) -> add_error(is_member_set2,'Empty set not allowed for Inter(-): ',T)
     ; add_error_fail(is_member_set,'Illegal argument: ','Inter'(LS))).

:- block is_member_from_to(-,-,?),is_member_from_to(-,?,-).
is_member_from_to(X,Low,Up) :- ground(X),!,geq(X,Low), leq(X,Up).
is_member_from_to(X,Low,Up) :- enumerate_csp_int(X,Low,Up).
:- block geq(?,-).
geq(X,Low) :- X>=Low.
:- block leq(?,-).
leq(X,Up) :- X=<Up.

:- block is_member_from(-,?),is_member_from(?,-).
is_member_from(X,Low) :- X >= Low.

:- block blocking_member(?,-).
blocking_member(X,[H|T]) :- 
   (equal_element(X,H) ; blocking_member(X,T)).

:- assert_must_succeed((csp_sets: l_dot_is_member_set([int(1),int(2),int(3)],['dotTupleType'([setFromTo(1,3),setFrom(1),setFromTo(1,3)])]))).
:- assert_must_fail((csp_sets: l_dot_is_member_set([int(1),int(2),int(3)],['dotTupleType'([setFromTo(1,3),setFromTo(1,3),setFrom(14)])]))).
:- assert_must_succeed((csp_sets: l_dot_is_member_set([int(1),int(2),int(3),int(10),int(9)],
		['dotTupleType'([setFromTo(1,3),setFrom(1),setFromTo(1,3)]),intType,setValue([int(1),int(9),int(10)])]))).
:- assert_must_fail((csp_sets: l_dot_is_member_set([int(1),int(2),int(3),int(10),int(11)],
		['dotTupleType'([setFromTo(1,3),setFrom(1),setFromTo(1,3)]),intType,setValue([int(1),int(9),int(10)])]))).
:- assert_must_succeed((csp_sets: l_dot_is_member_set([int(10),int(9),int(1),int(2),int(3)],
		['dotTupleType'([intType,setValue([int(1),int(9),int(10)]),setFromTo(1,3),setFrom(1),setFromTo(1,3)])]))).

l_dot_is_member_set(L,TL) :- 
   unfold_dot_tuples(L,R),
   l_unfold_datatype_dot_tuple(TL,TR),
   l_is_member_set(R,TR).

l_is_member_set(SetList,SetList1) :-
   is_list(SetList),!,
   maplist(is_member_set,SetList,SetList1).
l_is_member_set(L,S) :- add_internal_error('Internal Error: Could not evaluate: ', l_is_member_set(L,S)),fail.

:- assert_must_succeed((csp_sets: unfold_dot_tuples([],[]))).
:- assert_must_succeed((csp_sets: unfold_dot_tuples([int(1),tuple([int(2),int(3)]),int(4)],R), R == [int(1),int(2),int(3),int(4)])).
:- assert_must_succeed((csp_sets: unfold_dot_tuples([int(1),int(2),int(3),int(4)],R), R == [int(1),int(2),int(3),int(4)])).
:- assert_must_succeed((csp_sets: unfold_dot_tuples([tuple([int(1),int(5)]),tuple([int(2),int(3)])],R), R == [int(1),int(5),int(2),int(3)])).
:- assert_must_succeed((csp_sets: unfold_dot_tuples([tuple([int(1),tuple([int(2),int(3)])])],R), R == [int(1),int(2),int(3)])).

unfold_dot_tuples([],[]).
unfold_dot_tuples([H|T],R) :- 
   ((\+var(H),(H = tuple(L) ; H=dotTuple(L))) -> unfold_dot_tuples(L,LRes), append(LRes,R1,R) ; R=[H|R1]), unfold_dot_tuples(T,R1).

:- assert_must_succeed((csp_sets: l_unfold_datatype_dot_tuple([],[]))).
:- assert_must_succeed((csp_sets: l_unfold_datatype_dot_tuple(['dotTupleType'([intType,setValue([int(1),int(9),int(10)])])],R), R == [intType,setValue([int(1),int(9),int(10)])])).
:- assert_must_succeed((csp_sets: l_unfold_datatype_dot_tuple([intType,intType],R), R == [intType,intType])).
:- assert_must_succeed((csp_sets: l_unfold_datatype_dot_tuple(['dotTupleType'([intType,setValue([int(1),int(9),int(10)])]),'dotTupleType'([intType,setValue([int(1),int(9),int(10)])])],R),
                        R == [intType,setValue([int(1),int(9),int(10)]),intType,setValue([int(1),int(9),int(10)])])).

unfold_datatype_dot_tuple(DT,R) :-
   (\+var(DT), DT = 'dotTupleType'(L) -> R=L ; R=[DT]).

l_unfold_datatype_dot_tuple(LDotTypes,Res) :-
   maplist(unfold_datatype_dot_tuple,LDotTypes,R),
   append(R,Res).

list_elements_member_set([],_) :- !.
list_elements_member_set([H|T],Set) :- !, is_member_set2(Set,H), list_elements_member_set(T,Set).
list_elements_member_set(L,S) :- add_internal_error('Internal Error: Could not evaluate: ', list_elements_member_set(L,S)),fail.

:- assert_must_succeed((csp_sets: deconstruct_set_of_sets(setExp(rangeEnum([int(1),int(2),int(3)])),H,T),H==int(1),T == setExp(rangeEnum([int(2),int(3)])))).
:- assert_must_succeed((csp_sets: deconstruct_set_of_sets(setValue([int(1),int(2),int(3)]),H,T),H==int(1),T == setValue([int(2),int(3)]))).
:- assert_must_fail((csp_sets: deconstruct_set_of_sets(setValue([]),_H,_T))).

%deconstruct_set_of_sets(setEnum([H|T]),H,setEnum(T)).
deconstruct_set_of_sets(setExp(rangeEnum([H|T])),H,setExp(rangeEnum(T))).
deconstruct_set_of_sets(setValue(V),H,setValue(T)) :- deconstruct_setValue(V,H,T).
:- block deconstruct_setValue(-,?,?).
deconstruct_setValue([H|T],H,T).

:- assert_must_succeed((csp_sets: empty_set_of_sets(setValue([])))).
empty_set_of_sets(setExp(rangeEnum([]))).
empty_set_of_sets(setValue(X)) :- is_empty_list(X,true).

:- assert_must_succeed(( csp_sets:cardinality(setValue([int(1),int(9)]),R), R==int(2) )).
:- block cardinality(-,?).
cardinality(setValue(S),R) :- my_length(S,C),!,R=int(C).
cardinality(setFromTo(Low,Up),R) :- !,
   R=int(C), when((ground(Low),ground(Up)),compute_from_to_cardinality(Low,Up,C)).
cardinality(setFrom(Low),_) :- !,
   add_error(csp_sets,'Trying to compute cardinality of infinite set: ',set_from(Low)),fail.
cardinality(closure(ChannelList),R) :- !,my_length(Closure,C), R=int(C), when(ground(ChannelList),
               expand_symbolic_set(closure(ChannelList),setValue(Closure),closure_cardinality)).
cardinality(dataType(DT),R) :- !,R=int(C),my_length(DTSet,C),expand_symbolic_set(dataType(DT),setValue(DTSet),datatype_set_cardinality).
cardinality(X,_R) :- add_error(csp_sets,'Could not compute card of: ',X),fail.

:- assert_must_succeed((csp_sets: compute_from_to_cardinality(1,3,C), C == 3)).
:- assert_must_succeed((csp_sets: compute_from_to_cardinality(-1,3,C), C == 5)).
:- assert_must_succeed((csp_sets: compute_from_to_cardinality(3,1,C), C == 0)).
:- assert_must_succeed((csp_sets: compute_from_to_cardinality(-3,1,C), C == 5)).
:- assert_must_succeed((csp_sets: compute_from_to_cardinality(-3,-1,C), C == 3)).

compute_from_to_cardinality(Low,Up,C) :-
   (Up<Low ->
       C is 0
   ; C is Up-Low+1  
   ).

my_length(L,Len) :- my_length_aux(L,0,Len).
:- block my_length_aux(-,?,?).
my_length_aux([],Acc,Acc).
my_length_aux([_|T],Acc,R) :- A1 is Acc+1, my_length_aux(T,A1,R).


try_get_cardinality_for_wait_flag(setValue(S),R) :-  try_get_length(S,C),!,R=C.
try_get_cardinality_for_wait_flag(setFromTo(Low,Up),C) :-
   ground(Low),ground(Up),!,C is Up-Low+1.
try_get_cardinality_for_wait_flag(_,1000).

try_get_length(X,_) :- var(X),!,fail.
try_get_length([],0).
try_get_length([_|T],R) :- try_get_length(T,R1), R is R1+1.

:- block singleSetElement(-,?,?).
singleSetElement(S,El,Span) :- expand_symbolic_set(S,setValue(V),singleSetElement),
    singleSetElement_aux(V,El,Span).

:- block singleSetElement_aux(-,?,?).
singleSetElement_aux([H|T],El,_Span) :- (var(T);T==[]),!,H=El.
singleSetElement_aux(Set,_El,Span) :- 
   add_error(singleSetElement,'This is not a singleton set: ',setValue(Set),Span),fail.

:- assert_must_succeed(( csp_sets:is_subset_of(R,setValue([int(2),int(4)])),
                          R = setValue([int(4)]) )). 
:- assert_must_fail(( csp_sets:is_subset_of(R,setValue([int(2),int(4)])),
                          R = setValue([int(3)]) )). 
:- block is_subset_of(-,?), is_subset_of(?,-).
is_subset_of(X,Y) :- 
      expand_symbolic_set(X,setValue(EX),is_subset_of_x),
      expand_symbolic_set(Y,setValue(EY),is_subset_of_y),
      is_subset_of2(EX,EY).

:- block is_subset_of2(-,?).
is_subset_of2([],_).
is_subset_of2([H|T],S) :- remove_element(H,S,S2), is_subset_of2(T,S2).

:- assert_must_succeed(( csp_sets:subsets(setValue([int(1),int(9)]),R), 
       R==setValue([ setValue([int(1),int(9)]),setValue([int(9)]),setValue([int(1)]),setValue([]) ]) )).

subsets(S,setValue(RS)) :- expand_symbolic_set(S,setValue(ES),subsets),sub2(ES,RS).

:- block sub2(-,?).
sub2([],[setValue([])]).
sub2([El|T],Res) :- sub2(T,TR), sub3(TR,El,Res).
  
:- block sub3(-,?,?).
sub3([],_,[]).
sub3([setValue(S1)|T],El,[setValue([El|S1]),setValue(S1)|ST]) :- sub3(T,El,ST).


:- assert_must_succeed(( csp_sets:enum_subset(setValue([int(1),int(9)]),R), 
       R==setValue([int(9)]) )).
enum_subset(S,setValue(Subset)) :- expand_symbolic_set(S,setValue(ES),enum_subset),
  enum_sub2(ES,Subset).
enum_sub2([],[]).
enum_sub2([El|T],Res) :-
  (Res = [El|T2], enum_sub2(T,T2))
  ; enum_sub2(T,Res).

:- assert_must_succeed(( csp_sets:add_element(int(3),setValue([int(1),int(9)]),R), R==setValue([int(1),int(3),int(9)]) )).
:- assert_must_succeed(( csp_sets:add_element(int(3),setValue([int(1),int(3),int(9)]),R), R==setValue([int(1),int(3),int(9)]) )).
:- assert_must_succeed(( csp_sets:add_element(int(11),setValue([int(1),int(9)]),R), R==setValue([int(1),int(9),int(11)]) )).
:- assert_must_succeed(( csp_sets:add_element(int(1),setValue([int(3),int(9)]),R), R==setValue([int(1),int(3),int(9)]) )).
:- assert_must_succeed(( csp_sets:add_element(setValue([int(3)]),setValue([]),R), R==setValue([setValue([int(3)])]) )).
:- assert_must_succeed(( csp_sets:add_element(boolType,setValue([]),R), R == setValue([setValue([true,false])]))).

:- block add_element(-,?,?), add_element(?,-,?).
add_element(El,S,Res) :- Res = setValue(R), expand_symbolic_set(S,setValue(ES),add_element),
   (is_a_set(El) -> expand_symbolic_set(El,ExEl,add_element) ; ExEl=El), % normalise element before storing in set
   when(ground(ExEl),add_element1(ES,ExEl,R)).
  % when( (/* ground(El),*/ nonvar(ES)), add_element2(ES,El,R)).

:- block add_element1(-,?,?).
%add_element1(T,El,Res) :- print(add_element1(T,El,Res)),nl,fail.
add_element1([],El,[El]).
add_element1([H|T],El,Res) :- when(?=(El,H),(El @=<H -> (El=H -> Res = [El|T] ; Res = [El,H|T])
                                             ; (Res=[H|R2],add_element1(T,El,R2)))).

:- assert_must_succeed(( csp_sets:union_set(setValue([int(3),int(4)]),setValue([int(2),int(9)]),R),
                          R == setValue([int(2),int(3),int(4),int(9)]) )).
:- assert_must_succeed(( csp_sets:union_set(setValue([int(3),int(4)]),setValue([int(4),int(9)]),R),
                          R == setValue([int(3),int(4),int(9)]) )).
:- assert_must_succeed(( csp_sets:union_set(setValue([int(3),int(4)]),setValue([]),R),
                          R == setValue([int(3),int(4)]) )).
:- assert_must_succeed(( csp_sets:union_set(setValue([]),setValue([int(3),int(4)]),R),
                          R == setValue([int(3),int(4)]) )).
:- block union_set(-,?,?), union_set(?,-,?).
union_set(S1,S2,Res) :- Res = setValue(R), expand_symbolic_set(S1,setValue(ES1),union_set1),
  expand_symbolic_set(S2,setValue(ES2),union_set2),
  when(ground((ES1,ES2)),union_add_elements(ES1,ES2,R,none,none))
 % , print(union(S1,S2,Res)),nl
  .

union_add_elements([],R,R,_,_).
union_add_elements([H|T],[],[H|T],PrevH1,_) :- (ground(H) -> check_sorted(union_set,PrevH1,H) ; true).
union_add_elements([H1|T1],[H2|T2],Res,PrevH1,PrevH2) :- %check_sorted(union_set,PrevH1,H1), % unnecessary call
   when((ground(H1),ground(H2)),
         (check_sorted(union_set,PrevH1,H1), check_sorted(union_set,PrevH2,H2),
          (H1=H2 -> Res=[H1|RT],union_add_elements(T1,T2,RT,H1,H1)
                      ;  (H1 @=< H2 ->  Res=[H1|RT], union_add_elements(T1,[H2|T2],RT,H1,none)
                                     ;  Res=[H2|RT], union_add_elements([H1|T1],T2,RT,none,H2)) ))).

check_sorted(Src,PrevH,H) :- ((PrevH=none;PrevH @< H) -> true ; add_error(Src,'CSP set not sorted: ',[PrevH,H])).
 
:- assert_must_succeed(( csp_sets:diff_set(setValue([int(3),int(4)]),setValue([int(2),int(3)]),R),
                          R == setValue([int(4)]) )).
:- assert_must_succeed(( csp_sets:diff_set(setValue([int(3),int(4)]),setValue([int(2),int(5)]),R),
                          R == setValue([int(3),int(4)]) )).
:- assert_must_succeed(( csp_sets:diff_set(setValue([int(3),int(4)]),setValue([int(2),int(3),int(4),int(5),int(9)]),R),
                          R == setValue([]) )).

:- block diff_set(-,?,?), diff_set(?,-,?).
diff_set(S1,S2,Res) :- Res = setValue(R), expand_symbolic_set(S1,setValue(ES1),diff_set1),
  expand_symbolic_set(S2,setValue(ES2),diff_set2),
  when(ground((ES1,ES2)),diff_elements(ES1,ES2,R)).
  %, print(diff(S1,S2,Res,ES1,ES2)),nl.

diff_elements([],_,[]).
diff_elements([H|T],S2,Res) :-
  (remove_element(H,S2,S3)
    -> diff_elements(T,S3,Res)
    ;  (Res=[H|R2], diff_elements(T,S2,R2))
  ).

%:- block remove_element(-,?,?), remove_element(?,-,?).
remove_element(X,[H|T],R) :- 
  (selectchk(X,[H|T],R)
      -> true
  ; equal_element(X,H)
      -> R=T
  ;  (X@>H,  /* diff set assumes that arguments are already evaluated ! */ 
        R=[H|RT], remove_element(X,T,RT))
  ).


:- assert_must_succeed(( csp_sets:inter_set(setValue([int(3),int(4)]),setValue([int(2),int(4)]),R),
                          R == setValue([int(4)]) )). 
:- assert_must_succeed(( csp_sets:inter_set(setValue([int(3),int(4)]),setValue([int(3),int(4)]),R),
                          R == setValue([int(3),int(4)]) )).
:- assert_must_succeed(( csp_sets:inter_set(setValue([int(3),int(4)]),setFromTo(4,5),R),
                          R == setValue([int(4)]) )).
:- assert_must_succeed(( csp_sets:inter_set(setValue([int(2)]),setFromTo(0,1),R),
                          R == setValue([]) )).
:- assert_must_succeed(( csp_sets:inter_set(setFrom(1),setFromTo(3,5),R),
			  R == setValue([int(3),int(4),int(5)]))).
:- assert_must_succeed(( csp_sets:inter_set(setFrom(6),setFromTo(3,5),R),
			  R == setValue([]))).
:- assert_must_succeed(( csp_sets:inter_set(setFrom(6),setValue([]),R),
			  R == setValue([]))).

:- block inter_set(-,?,?), inter_set(?,-,?).
inter_set(S1,S2,Res) :- Res = setValue(R), 
  (S1=setFrom(Low) -> 
	expand_symbolic_set(S2,setValue(ES2),inter_set1),
  	when(ground((Low,ES2)),inter_merge_elements_from(ES2,Low,R))
  ;S2=setFrom(Low) -> 
	expand_symbolic_set(S1,setValue(ES1),inter_set2),
	when(ground((ES1,Low)),inter_merge_elements_from(ES1,Low,R))
  ; otherwise -> 
	expand_symbolic_set(S1,setValue(ES1),inter_set1),
	expand_symbolic_set(S2,setValue(ES2),inter_set2),
	when(ground((ES1,ES2)),inter_merge_elements(ES1,ES2,R))
  ).

inter_merge_elements([],_R,[]).
inter_merge_elements([_|_],[],[]). 
inter_merge_elements([H1|T1],[H2|T2],Res) :- 
  (equal_element(H1,H2)
    -> (Res = [H1|TR], inter_merge_elements(T1,T2,TR))
    ;  (H1@<H2 -> inter_merge_elements(T1,[H2|T2],Res)
               ;  inter_merge_elements([H1|T1],T2,Res)
       )
  ).

inter_merge_elements_from([],_Low,[]).
inter_merge_elements_from([int(H)|T],Low,Res) :- 
   ((Low =< H) -> 
	Res = [int(H)|TR], inter_merge_elements_from(T,Low,TR)
   ; otherwise -> 
	inter_merge_elements_from(T,Low,Res)
   ).

:- assert_must_succeed((X=true, csp_sets: equal_element(true, X))).
:- assert_must_fail((X=false, csp_sets: equal_element(true, X))).
:- assert_must_succeed((X=false, csp_sets: equal_element(false, X))).
:- assert_must_fail((X=true, csp_sets: equal_element(false, X))).
:- assert_must_succeed((R=setFrom(1),csp_sets: equal_element(setFrom(1), R))).
:- assert_must_fail((R=setFrom(2),csp_sets: equal_element(setFrom(1), R))).
:- assert_must_fail((R=intType,csp_sets: equal_element(setFrom(1), R))).
:- assert_must_succeed((R=setFromTo(1,46), csp_sets: equal_element(setFromTo(1,46),R))).
:- assert_must_fail((R=setFromTo(1,2), csp_sets: equal_element(setFromTo(1,46),R))).
:- assert_must_fail((R=setFrom(1), csp_sets: equal_element(setFromTo(1,46),R))).
:- assert_must_succeed((R=setValue([int(1),int(2),int(3)]), csp_sets: equal_element(setFromTo(1,3),R))).
:- assert_must_succeed((R=setValue([]), csp_sets: equal_element(setFromTo(3,1),R))).
:- assert_must_fail((R=setValue([int(1),int(2),int(3)]), csp_sets: equal_element(setFromTo(1,4),R))).
:- assert_must_fail((R=setFrom(10), csp_sets: equal_element(setValue(_X), R))).
:- assert_must_succeed((R=boolType, csp_sets: equal_element(setValue([true,false]),R))).
:- assert_must_succeed((R=setFromTo(1,1), csp_sets: equal_element(setValue([int(1)]),R))).
:- assert_must_succeed((R=tuple([v1,int(0),na_tuple([int(0),int(0)])]),  csp_sets: equal_element(tuple([v1,tuple([int(0),na_tuple([int(0),int(0)])])]),R))).
:- assert_must_succeed((R=tuple([v1,tuple([int(0),na_tuple([int(0),int(0)])])]),  csp_sets: equal_element(tuple([v1,int(0),na_tuple([int(0),int(0)])]),R))).
:- assert_must_succeed((R=record(v1,[int(0),na_tuple([int(0),int(0)])]), csp_sets: equal_element(tuple([v1,tuple([int(0),na_tuple([int(0),int(0)])])]),R))).

equal_element(X,Y) :-
   (var(X);var(Y)),!,X=Y.
equal_element(true,X) :- !, 
   ( X=true  -> true
   ; X=false -> fail
   ; add_error_fail(haskell_csp,'Type error in equality: ',true=X)
   ).
equal_element(false,X) :- !, 
   ( X=false -> true
   ; X=true  -> fail
   ; add_error_fail(haskell_csp,'Type error in equality: ',false=X)
   ).
equal_element(int(X),R) :- !, 
   (R=int(Y) -> X=Y
   ; add_error_fail(haskell_csp,'Type error in equality: ',int(X)=R)
   ).
equal_element(setFrom(X),R) :- !,
   (R=setFrom(Y) -> X=Y
   ; is_a_set(R) -> fail
   ; add_error_fail(haskell_csp,'Type error in equality: ',setFrom(X)=R)
   ).
equal_element(setFromTo(X,Y),R) :- !,
   (R=setFromTo(X2,Y2) -> X=X2,Y=Y2
   ; (R=setValue(_) ;R=setExp(_,_)) -> equal_sets(setFromTo(X,Y),R)
   ; is_a_set(R) -> fail % Set different for setValue and setExp
   ; add_error_fail(haskell_csp,'Type error in equality: ',setFromTo(X,Y)=R)
   ).
equal_element(setValue(X),R) :- !,
  (R=setValue(Y) -> equal_setValue(X,Y)
   ; R=setFrom(_) -> fail % infinite set cannot be equal to finite one
   ; is_a_set(R) -> expand_symbolic_set(R,setValue(ER),equal_element), equal_setValue(X,ER)
   ; add_error_fail(haskell_csp,'Type error in equality: ',setValue(X)=R)
   ).
equal_element(list(X),R) :- !,
   (R=list(Y) -> X=Y
   ;add_error_fail(haskell_csp,'Type error in equality: ',list(X)=R)
   ).
equal_element(na_tuple(X),R) :- !,
   (R=na_tuple(Y) -> X=Y
   ;add_error_fail(haskell_csp,'Type error in equality: ',na_tuple(X)=R)
   ).
equal_element(tuple(X),R) :- !,
   ( R=tuple(Y) -> (X=Y ; equal_interleaved_dot_tuples(X,Y))
   ; R=record(C,A) -> X=[H|T], C=H, (T=A ; equal_interleaved_dot_tuples(T,A))
   ; add_error_fail(haskell_csp,'Type error in equality: ',tuple(X)=R)
   ).
equal_element(dotTuple(X),R) :- !,
   ( R=tuple(Y) -> (X=Y ; equal_interleaved_dot_tuples(X,Y))
   ; R=record(C,A) -> X=[H|T], C=H, (T=A ; equal_interleaved_dot_tuples(T,A))
   ; add_error_fail(haskell_csp,'Type error in equality: ',dotTuple(X)=R)
   ).   
equal_element(record(C,A),R) :-
   get_constructor_type(C,Type),!,
   ((R=record(C2,A2),get_constructor_type(C2,Type)) -> C=C2,(A=A2 ; equal_interleaved_dot_tuples(A,A2)) % missing subtype checks 
   ; R=tuple([H|T]) -> (H=C,(T=A ; equal_interleaved_dot_tuples(A,T)))
   ; (atomic(R),get_constant_type(R,Type)) -> fail
   ; R=agent_call(_,_,_) -> force_evaluate_argument(R,Res),equal_element(record(C,A),Res)
   ; add_error_fail(haskell_csp,'Type error in equality: ',record(C,A)=R)
   ).
equal_element(X,R) :-
    atomic(X),get_constant_type(X,Type),!,
    ((atomic(R),get_constant_type(R,Type)) -> X=R
    ;(R=record(C,_Args),get_constructor_type(C,Type)) -> fail
    ; add_error_fail(haskell_csp,'Type error in equality: ',X=R)
    ).
equal_element(Set,R) :-
    is_a_set(Set),!,
    expand_symbolic_set(Set,ES,equal_element),
    equal_element(ES,R).
% To do: Further Improve this predicate, and check typing
equal_element(X,Y) :- print(equal_element(X,Y)),nl,X=Y.

equal_interleaved_dot_tuples(X,Y) :-
    unify_tuple_elements(X,Y,R,tuple),
    (X=R -> true; unfold_dot_tuples(X,XR),XR=R).

equal_sets(setFromTo(X,Y),R) :-
   ((R=setValue(S),ground((X,Y,S))) ->
       cardinality(setFromTo(X,Y),int(N)),
       length(S,N),
       expand_from_to(X,Y,XYSet),
       diff_elements(XYSet,S,[])
   ;otherwise ->
       expand_symbolic_set(setFromTo(X,Y),ES,equal_element),
       equal_element(ES,R)
   ).
% check if two lists inside setValue are equal; should be sorted !
% as all elements that appear in setValue are normalised we could simply use Prolog Unification: X=Y ?
equal_setValue(X,Y) :-
    (var(X);var(Y)),!,X=Y.
equal_setValue(L1,L2) :-
    maplist(equal_element,L1,L2).

get_constructor_type(C,Type) :- csp_full_type_constructor(C,DT,_ArgTypes),!,Type=DT.
get_constructor_type(C,_) :- add_internal_error(/*get_constructor_type,*/'Internal Error: Unknown record constructor: ',C),fail.
get_constant_type(C,Type) :- csp_full_type_constant(C,DataType),!,Type=DataType.
get_constant_type(C,Type) :- csp_full_type_constructor(C,DataType,_ArgTypes),!,
    % a type constructor is passed as an atomic value; some CSP specs do this (stc.csp of Kharmeh PhD spec)
   Type=constructor(DataType).
get_constant_type(C,_) :- add_internal_error(/*get_constant_type,*/'Internal Error: Unknown constant: ',C),fail.

:- assert_must_fail((X=true, csp_sets: not_equal_element(true, X))).
:- assert_must_succeed((X=false, csp_sets: not_equal_element(true, X))).
:- assert_must_fail((X=false, csp_sets: not_equal_element(false, X))).
:- assert_must_succeed((X=true, csp_sets: not_equal_element(false, X))).
:- assert_must_fail((R=setFrom(1),csp_sets: not_equal_element(setFrom(1), R))).
:- assert_must_succeed((R=setFrom(2),csp_sets: not_equal_element(setFrom(1), R))).
:- assert_must_succeed((R=intType,csp_sets: not_equal_element(setFrom(1), R))).
:- assert_must_fail((R=setFromTo(1,46), csp_sets: not_equal_element(setFromTo(1,46),R))).
:- assert_must_succeed((R=setFromTo(1,2), csp_sets: not_equal_element(setFromTo(1,46),R))).
:- assert_must_succeed((R=setFrom(1), csp_sets: not_equal_element(setFromTo(1,46),R))).
:- assert_must_fail((R=setValue([int(1),int(2),int(3)]), csp_sets: not_equal_element(setFromTo(1,3),R))).
:- assert_must_fail((R=setValue([]), csp_sets: not_equal_element(setFromTo(3,1),R))).
:- assert_must_succeed((R=setValue([int(1),int(2),int(3)]), csp_sets: not_equal_element(setFromTo(1,4),R))).
:- assert_must_succeed((R=setFrom(10), csp_sets: not_equal_element(setValue(_X), R))).
:- assert_must_fail((R=boolType, csp_sets: not_equal_element(setValue([true,false]),R))).
:- assert_must_fail((R=setFromTo(1,1), csp_sets: not_equal_element(setValue([int(1)]),R))).
:- assert_must_succeed((R=tuple([int(1),int(2)]), csp_sets: not_equal_element(tuple([int(1),int(3)]), R))).
:- assert_must_fail((R=tuple([int(1),int(2)]), csp_sets: not_equal_element(tuple([int(1),int(2)]), R))).
:- assert_must_succeed((R=na_tuple([int(1),int(2)]), csp_sets: not_equal_element(na_tuple([int(1),int(3)]), R))).
:- assert_must_fail((R=na_tuple([int(1),int(2)]), csp_sets: not_equal_element(na_tuple([int(1),int(2)]), R))).
:- assert_must_succeed((R=list([int(1),int(2)]), csp_sets: not_equal_element(list([int(1),int(3)]), R))).
:- assert_must_fail((R=list([int(1),int(2)]), csp_sets: not_equal_element(list([int(1),int(2)]), R))).
:- assert_must_succeed((R=record(seq,[int(1),int(2)]), csp_sets: not_equal_element(record(seq,[int(1),int(3)]), R))).
:- assert_must_fail((R=record(seq,[int(1),int(2)]), csp_sets: not_equal_element(record(seq,[int(1),int(2)]), R))).

not_equal_element(X,Y) :- var(X),!,dif(X,Y).
not_equal_element(true,X) :- !,
   (X=false -> true
   ; X=true -> fail
   ; add_error_fail(haskell_csp,'Type error in disequality: ',true\=X)
   ).
not_equal_element(false,X) :- !,
   (X=true -> true
   ; X=false -> fail
   ; add_error_fail(haskell_csp,'Type error in disequality: ',false\=X)
   ).
not_equal_element(int(X),R) :- !,
   (R=int(Y) -> X\=Y 
   ; add_error_fail(haskell_csp,'Type error in disequality: ',int(X)\=R)
   ).
not_equal_element(setFrom(X),R) :- !,
   (R=setFrom(Y) -> X\=Y
   ; is_a_set(R) -> true
   ; add_error_fail(haskell_csp,'Type error in disequality: ',setFrom(X)\=R)
   ).
not_equal_element(setFromTo(X,Y),R) :- !,
   (R=setFrom(X2,Y2) -> (X,Y)\=(X2,Y2)
   ; R=setFrom(_) -> true
   ; is_a_set(R) -> expand_symbolic_set(setFromTo(X,Y),ES,equal_element), not_equal_element(ES,R)
   ; add_error_fail(haskell_csp,'Type error in disequality: ',setFromTo(X,Y)\=R)
   ).
not_equal_element(setValue(X),R) :- !,
   (R=setValue(Y) -> not_equal_setValue(X,Y)
   ; R=setFrom(_) -> true /* infinite set cannot be equal to finite one */
   ; is_a_set(R) -> expand_symbolic_set(R,setValue(ER),equal_element), not_equal_setValue(X,ER)
   ; add_error_fail(haskell_csp,'Type error in disequality: ',setValue(X)\=R)
   ).
not_equal_element(list(X),R) :- !,
   (R=list(Y) -> X\=Y 
   ; add_error_fail(haskell_csp,'Type error in disequality: ',list(X)\=R)
   ).
not_equal_element(na_tuple(X),R) :- !,
   (R=na_tuple(Y) -> X\=Y 
   ; add_error_fail(haskell_csp,'Type error in disequality: ',na_tuple(X)\=R)
   ).
not_equal_element(tuple(X),R) :- !,
   ( R=tuple(Y) -> X\=Y
   ; R=record(C,A) -> X=[H|T], (C\=H ; (T\=A , \+equal_interleaved_dot_tuples(T,A)))
   ; add_error_fail(haskell_csp,'Type error in disequality: ',tuple(X)\=R)
   ).
not_equal_element(record(C,A),R) :-  !,
  (atomic(R) -> true
   ; R=record(C2,A2) -> (C\=C2 ; (A\=A2 , \+equal_interleaved_dot_tuples(A,A2)))
   ; R=tuple([H|T]) -> (H\=C ; (A\=T , \+equal_interleaved_dot_tuples(A,T)))
   ; add_error_fail(haskell_csp,'Type error in disequality: ',record(C,A)\=R)
   ).
not_equal_element(Set,R) :-
   is_a_set(Set),
   expand_symbolic_set(Set,ES,equal_element),
   not_equal_element(ES,R).
not_equal_element(X,R) :- atomic(X),!,
   (atomic(R) -> X\=R
   ; R=record(_,_) -> fail
   ; add_error_fail(haskell_csp,'Type error in disequality: ',X\=R)
   ).
not_equal_element(X,Y) :- dif(X,Y).

% we normalise all elements that appear in setValue; dif or \= is sufficient
not_equal_setValue(X,Y) :- dif(X,Y).

/* expand_symbolic_set */
/*  force expansion of symbolic sets into explicit sets */

/* testing the equality: {x*y | x <- {1,3}, y <- {2,4}} == {2,4,6,12} */
:- assert_must_succeed(( csp_sets:expand_symbolic_set(setExp(rangeEnum(['*'(_x,_y)] ),
 [comprehensionGenerator(_x,setValue([int(1),int(3)])),
  comprehensionGenerator(_y,setValue([int(2),int(4)]))] ),R,test), R = setValue([int(2),int(4),int(6),int(12)]))).
:- assert_must_succeed(( csp_sets:expand_symbolic_set(boolType,R,_X), R == setValue([true,false]))).
/* testing the equality: {(x,y) | x <- {0..100}, y <- { -99..5}, x+y == 100} == {(95,5),(96,4),(97,3),(98,2),(99,1),(100,0)} */
:- assert_must_succeed(( csp_sets:expand_symbolic_set(setExp(rangeEnum([tuplePat([_x,_y])]),
 [comprehensionGenerator(_x,setFromTo(0,100)),
  comprehensionGenerator(_y,setFromTo(-99,5)),comprehensionGuard('=='('+'(_x,_y),'int'(100)))] ),R,test), 
  R = setValue([na_tuple([int(95),int(5)]),na_tuple([int(96),int(4)]),na_tuple([int(97),int(3)]),
                na_tuple([int(98),int(2)]),na_tuple([int(99),int(1)]),na_tuple([int(100),int(0)])]))).

%expand_symbolic_set(X,R) :- print(expand_symbolic_set(X,R)),nl,fail.
expand_symbolic_set(X,R,Context) :- var(X),!,
   add_error(expand_symbolic_set,'Variable argument for expand_symbolic_set, Context: ',Context),
   R=X.
expand_symbolic_set(dataType(T),R,Context) :- !, R=setValue(SET),
   (dataTypeDef(T,Def)
     -> %print(dt(T,Def,SET)),nl,
         expand_datatypedefbody(Def,T,SET,Context)
     ;  add_error(csp_sets,'Could not expand dataType. No datatype definition for: ',T:context(Context)),
        SET=[]
   ).
expand_symbolic_set(closure(X),R,_Context) :- !, closure_expand(X,R).
expand_symbolic_set(setValue(X),R,_Context) :- !, R=setValue(X).
expand_symbolic_set(setFrom(X),R,Context) :- !, add_warning(expand_symbolic_set,'Warning: Tried to expand infinite set: ',setFrom(X):context(Context)),R=setFrom(X).
expand_symbolic_set(setFromTo(Low,Up),R,_Context) :- !,R=setValue(RS),expand_from_to(Low,Up,RS). 
expand_symbolic_set(setExp(RangeExpr,GeneratorList),R,_Context) :- !,
    expand_set_comprehension(RangeExpr,GeneratorList,R).
expand_symbolic_set(listFromTo(X,Y),_,Context) :- 
    add_error(expand_symbolic_set,'Type error; expected set: ',listFromTo(X,Y):context(Context)),fail.
expand_symbolic_set(agent_call(Span,F,Par),R,Context) :- !,
    force_evaluate_argument(agent_call(Span,F,Par),RF),
    expand_symbolic_set(RF,R,Context).
expand_symbolic_set(boolType,R,_Context) :- !, R = setValue([true,false]).
expand_symbolic_set(dotTupleType(L),R,_Context) :- !, % we want to expand some more complicated Types like {0..2}.({0..2},{0..2})    
    %print(expand_symbolic_set(dotTupleType(L))),nl,
    l_unfold_datatype_dot_tuple([dotTupleType(L)],RL),
    findall(Val,haskell_csp: enumerate_channel_input_value2(dotTupleType(RL),Val,_Ch,2,no_loc_info_available),Set),
    R = setValue(Set).
expand_symbolic_set(typeTuple(L),R,_Context) :- !,
     haskell_csp: evaluate_type_list(L,LR),
     findall(Val,haskell_csp: enumerate_channel_input_value2(typeTuple(LR),Val,_Ch,2,no_loc_info_available),Set),
     R = setValue(Set).
expand_symbolic_set(Set,R,Context) :-
   add_error(expand_symbolic_set,'Could not expand set: ',Set:context(Context)),R=Set.

:- block expand_from_to(-,?,?), expand_from_to(?,-,?).   
expand_from_to(X,Y,R) :- expand_from_to2(X,Y,R).
expand_from_to2(X,Y,R) :- X>Y,!, R=[].
expand_from_to2(X,Y,[int(X)|T]) :- X1 is X+1, expand_from_to2(X1,Y,T).

expand_datatypedefbody([],_DT,R,_) :- !, R=[].
expand_datatypedefbody([constructor(C)|T],DT,R,Context) :- !, R=[C|ET],
   expand_datatypedefbody(T,DT,ET,Context).
expand_datatypedefbody([constructorC(Cons,Type)|T],DT,R,Context) :- !,
   (haskell_csp: check_list_for_infinite_set([Type]) ->
           findall(Record,csp_sets:expand_constructor_to_record(Cons,DT,2,Record),L),
	   append(L,RT,R),
	   expand_datatypedefbody(T,DT,RT,Context)
   ; add_error(csp_sets, 'Could not expand infinite datatype body: ',constructorC(Cons,Type):context(Context)),fail
   ).
expand_datatypedefbody(L,_DT,R,Context) :-
  add_internal_error('Internal Error: Could not expand datatype body (potentially infinite): ',L:context(Context)),
  R=[].

expand_constructor_to_record(Cons,DT,MaxRec,Res) :-
  Res=record(Cons,_L),
  enumerate_datatype_el(DT,Res,_Channel,MaxRec,no_loc_info_available).

/* ------------------ */
/* EXPANDING CLOSURES */
/* ------------------ */

/* csp_sets:closure_expand([tuple([outf])],R),print(R),nl */

closure_expand(ListOfEls,setValue(ExpandedList)) :-
  %print(closure_expand(ListOfEls)),nl,
  when(ground(ListOfEls),
   (findall(EEl,(member(El,ListOfEls), 
                 closure_expand_single_element(El,EEl)),EEls),
          %print(expanded(EEls)),nl,
    sort(EEls,ExpandedList)  /* is sort ok wrt to @< used by various set operations?? */
   )).

closure_expand_single_element(tuple([Ch|List]),R) :-
   channel_type_list(Ch,ChannelTypeList),!, R = tuple([Ch|NewList]),
   %print(tuple([Ch|NewList],ChannelTypeList)),nl,
  %% print(gen_expanded_list(ChannelTypeList,List,NewList,Ch)),nl,
   gen_expanded_list(ChannelTypeList,List,NewList,Ch).
closure_expand_single_element(Cons,C) :-
   is_csp_constructor(Cons),!,
   csp_constructor(Cons,DT,_ArgSubTypes),
   %print(csp_constructor(Cons,DT,_ArgSubTypes)),nl,
   C=record(Cons,_Fields),
   enumerate_datatype_el(DT,C,_Ch,2,no_loc_information).
closure_expand_single_element(tuple([Ch|_]),_R) :- !,
   add_error(csp_sets,'Cannot compute closure: This is not a defined channel: ',Ch),
   fail.
closure_expand_single_element(X,_R) :-
   add_error(csp_sets,'Cannot expand closure: ',X),
   fail.

gen_expanded_list([],List,[],Channel) :-
 (List=[] -> 
	true 
	; add_error(csp_sets,'Pattern list too long for channel: ',(List,Channel))
 ).
gen_expanded_list([Type|TT],List,Res,Channel) :-
   (List=[H|LT]  
     -> (is_incomplete_record(H,CompletedH,RecType)
          -> ((LT==[] -> 
		true 
		; add_error(gen_expanded_list,'Incomplete Record at non-tail position:',Channel:H)
	      ),
              enumerate_channel_input_value(dataType(RecType),CompletedH,Channel,no_loc_info_available),
              %%print(enum(Type,CompletedH,Channel)),nl,
              Res = [CompletedH|RT]
             )
           ; Res=[H|RT]
        )
     ;  Res=[NewEl|RT],LT=List,  /* is it ok not to put dot(NewEl) here ? */
        enumerate_channel_input_value(Type,NewEl,Channel,no_loc_info_available)
   ),
   gen_expanded_list(TT,LT,RT,Channel).

:- assert_must_succeed((assert(csp_sets: csp_full_type_constructor(sq,values,[dataType(values), dataType(values)])),
    is_incomplete_record(record(sq,['A']), _R, values), 
    retractall(csp_sets:csp_full_type_constructor(_,_,_)))).
:- assert_must_fail((assert(csp_sets: csp_full_type_constructor(sq,values,[dataType(values), dataType(values)])),
    is_incomplete_record(record(sq,['A','B']), _R, values), 
    retractall(csp_sets:csp_full_type_constructor(_,_,_)))).
is_incomplete_record(record(Constructor,Fields),record(Constructor,FullFields),Type) :- 
   csp_full_type_constructor(Constructor,Type,SubTypes),
   length(SubTypes,NrReqArgs),
   length(Fields,NrFields),
   (NrReqArgs > NrFields
      -> ( %print(record_incomplete(Constructor,Fields,SubTypes)),nl,
          length(FullFields,NrReqArgs),
          append(Fields,_,FullFields)
          %print(completed(record(Constructor,FullFields))),nl
         )
      ;  ((NrReqArgs<NrFields -> add_error(csp_sets,'Too many arguments for record: ',Constructor:Fields) ; true),
          fail)
   ).
/* ------------------ */
/* SET COMPREHENSIONS */
/* ------------------ */

expand_set_comprehension(RangeExpr,GeneratorList,Res) :-%trace,
   %%%% print(expand_set_comprehension(RangeExpr,GeneratorList,Res)),nl,
   get_waitvars_for_generator_list(GeneratorList,WaitVars),
   %print(get_waitvars_for_generator_list(GeneratorList,WaitVars)),nl,
   when(ground(WaitVars), generate_set_comprehension_solutions(RangeExpr,GeneratorList,Res)).

generate_set_comprehension_solutions(RangeExpr,GeneratorList,Res) :-
   treat_generators(GeneratorList,GenVars,Sets,Guard),
   findall(EExpr,get_generators_solution(Guard,GenVars,RangeExpr,Sets,EExpr),Expressions),
   %print(force_evaluate_set(Expressions,Res)),nl,   
   force_evaluate_set(Expressions,Res).%evaluate_set(Expressions,Res,force_evaluate_argument).

get_generators_solution(Guard,GenVars,RangeExpr,Sets,EExpr) :-
   check_boolean_expression(Guard),
   %print(generator_sol(guard(Guard),GenVars,Sets)),nl,
   generator_sol(GenVars,Sets,set), % unifies the variables of the comprehension generator expressions (e.g. x <- {0..10})
   %print(checking_range),nl,
   member_range_expr(RangeExpr,EExpr).

/* not used anymore
% temporary CLPFD will be not used for CSP
check_boolean_expression_set(Guard) :-
  preference(use_clpfd_solver,true),
  arith_boolean_expression(Guard,EvBExpr),!,
  set_clpfd_constraints(EvBExpr).
check_boolean_expression_set(Guard) :- 
  check_boolean_expression(Guard).

arith_boolean_expression(BExpr,EvBExpr) :-
   functor(BExpr,F,2),arg(1,BExpr,Arg1),arg(2,BExpr,Arg2),
   %BExpr =.. [F,Arg1,Arg2],
   (F == '=='; F == '!='; F == '<'; F == '<='; F == '>'; F == '>='),!,
   cspm_compute_arith_expression(Arg1,EArg1),
   cspm_compute_arith_expression(Arg2,EArg2),
   functor(EvBExpr,F,2),arg(1,EvBExpr,EArg1),arg(2,EvBExpr,EArg2).
   %EBExpr=..[F,EArg1,EArg2].

set_clpfd_constraints('=='(X,Y)) :- !,
   clpfd_interface:csp_clpfd_eq(X,Y).
set_clpfd_constraints('!='(X,Y)) :- !,
   clpfd_interface:csp_clpfd_neq(X,Y).
set_clpfd_constraints('<'(X,Y)) :- !,
   clpfd_interface:csp_clpfd_lt(X,Y).
set_clpfd_constraints('<='(X,Y)) :- !,
   clpfd_interface:csp_clpfd_leq(X,Y).
set_clpfd_constraints('>'(X,Y)) :- !,
   clpfd_interface:csp_clpfd_gt(X,Y).
set_clpfd_constraints('>='(X,Y)) :- !,
   clpfd_interface:csp_clpfd_geq(X,Y).
   
   
cspm_compute_arith_expression(Expr,Res) :-
        var(Expr),!,Res=Expr.
cspm_compute_arith_expression('-'(Arg1),Value) :- !,
        cspm_compute_arith_expression(Arg1,SV1),  
        Value = '-'(SV1).
cspm_compute_arith_expression('-'(Arg1,Arg2),Value) :- !,
        cspm_compute_arith_expression(Arg1,SV1),
        cspm_compute_arith_expression(Arg2,SV2),
        Value = '-'(SV1,SV2).
cspm_compute_arith_expression('+'(Arg1,Arg2),Value) :- !,
        cspm_compute_arith_expression(Arg1,SV1),
        cspm_compute_arith_expression(Arg2,SV2),
        Value = '+'(SV1,SV2).
cspm_compute_arith_expression('*'(Arg1,Arg2),Value) :- !,
        cspm_compute_arith_expression(Arg1,SV1),
        cspm_compute_arith_expression(Arg2,SV2),
        Value = '*'(SV1,SV2).
cspm_compute_arith_expression(int(Expr),Expr) :- !.
*/


:- assert_must_succeed((csp_sets: member_range_expr(rangeEnum([int(1),int(2),int(3)]),int(E)), E == 2)).
:- assert_must_fail((csp_sets: member_range_expr(rangeEnum([]),_E))).
:- assert_must_succeed((csp_sets: member_range_expr(rangeClosed(int(1),int(5)),int(E)), E == 3)).
:- assert_must_fail((csp_sets: member_range_expr(rangeClosed(int(3),int(1)),_E))).
% no lazy-evaluation, argument E must be initialized before calling member_range_expr(rangeOpen(int(_)),E) 
:- assert_must_succeed((E = int(30000), csp_sets: member_range_expr(rangeOpen(int(1)),E))). 
:- assert_must_fail((E = int(1), csp_sets: member_range_expr(rangeOpen(int(3)),E))).

member_range_expr(rangeEnum(ExprList),EExpr) :- !,
   /*(preference(use_clpfd_solver,true),nonvar(ExprList) ->
        %print(expr_list_1(ExprList)),nl,
        term_variables(ExprList,Vars),
        %print(csp_clpfd_labeling([ffc,enum],Vars)),nl,
        clpfd_interface: csp_clpfd_labeling([ffc,enum],Vars)
     ;  true
   ),*/
   member(Expr,ExprList),force_evaluate_argument(Expr,EExpr).
member_range_expr(rangeClosed(X,Y),EExpr) :- !,
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   is_member_set(EExpr,setFromTo(EX,EY)).
member_range_expr(rangeOpen(X),EExpr) :- !,
   evaluate_int_argument(X,EX),
   is_member_set(EExpr,setFrom(EX)). % could flounder if Guard not specific enough?!
/* Internal error. CSP-M Parser guarantees that the expression on the left side of | in the parsed comprehension set is one of 
   the rangeEnum(-), rangeClosed(_,_) or rangeOpen(-) predicates. */
member_range_expr(Range,_) :-
  add_internal_error('Internal Error: Illegal range expr in set comprehension: ',Range),fail.

:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

% compute the variables that have to be ground before expanding a set comprehension:
get_waitvars_for_generator_list(GeneratorList,WaitVars) :-
   extract_local_variables_from_generator_list(GeneratorList,LocalVars),
   term_variables(GeneratorList,GVars),
   % Do not wait on local variables, they will never be grounded:
   remove_variables(GVars,LocalVars,WaitVars).
 
% Note: this predicate is also called for the replicated operators ! The expressions can
% be CSPM agents : do not use force_evaluate
replicate_expand_set_comprehension(ExprList,GeneratorList,Res) :-
    %% print(replicate_expand_setComp(ExprList,GeneratorList)),nl,  %%%
   extract_local_variables_from_generator_list(GeneratorList,LocalVars),
   term_variables(GeneratorList,GVars),
   % Do not wait on local variables, they will never be grounded:
   remove_variables(GVars,LocalVars,WaitVars),
   when(ground(WaitVars), generate_replicate_set_comprehension_solutions(ExprList,GeneratorList,Res)).

generate_replicate_set_comprehension_solutions(ExprList,GeneratorList,Res) :-
   treat_generators(GeneratorList,GenVars,Sets,Guard),
   findall(EExpr,get_replicate_generators_solution(Guard,GenVars,ExprList,Sets,EExpr),Expressions),
   evaluate_set(Expressions,Res).%evaluate_set(Expressions,Res,evaluate_argument).

get_replicate_generators_solution(Guard,GenVars,ExprList,Sets,EExpr) :-
   check_boolean_expression(Guard),
   generator_sol(GenVars,Sets,replicated), % unifies the variables of the comprehension generator expressions (e.g. x <- {0..10})
   member(Expr,ExprList),
   evaluate_argument(Expr,EExpr).

generator_sol([],[],_Context).
generator_sol([Pattern|VT],[Set|ST],Context) :-
   (ground(Set) -> true ; print(generator_sol_set_non_ground(Set)),nl), % for nested set comprehension this could actually be non-ground
   translate_pattern(Pattern,TranslPattern),
   % print(evaluated_pattern(TranslPattern,Pattern,Set)),nl,
   (ground(TranslPattern) ->  /* we do not need to enumerate; generator variable already ground */
        force_evaluate_argument_for_member_check(Set,ESet),
        is_member_set_alsoPat(TranslPattern,ESet)
    ;   force_evaluate_argument(Set,EvSet),
        is_member_clpfd(TranslPattern,EvSet,Context)
   ),
   %  print(gen_is_member(TranslPattern,Pattern,Set)),nl,
   generator_sol(VT,ST,Context).

% constraining the variables domains
/*is_member_clpfd(Pat,EvSet,set) :-
  preference(use_clpfd_solver,true),
  simple(Pat),check_intset_type(EvSet),!,
  csp_set_pattern_constraints(EvSet,Pat).

csp_set_pattern_constraints(setFrom(Low),Pat) :- !,
   Up=sup,
   clpfd_interface: csp_clpfd_domain([Pat],Low,Up).
csp_set_pattern_constraints(setFromTo(Low,Up),Pat) :- !,
   clpfd_interface: csp_clpfd_domain([Pat],Low,Up).
csp_set_pattern_constraints(setValue(L),Pat) :- !,
   maplist(expand_int_value,L,LDom),
   clpfd_interface:csp_in_fdset(Pat,LDom).

expand_int_value(int(X),X).

check_intset_type(setFrom(_Low)) :- !.
check_intset_type(setFromTo(_Low,_Up)) :- !.
check_intset_type(setValue(L)) :-
   maplist(functor1(int,1),L),!.

functor1(Name,N,Term) :-
   functor(Term,Name,N).
   
*/

is_member_clpfd(Pat,EvSet,_Context) :-
  expand_symbolic_set(EvSet,ESet,generator_sol),
  is_member_set_alsoPat(Pat,ESet).

is_member_set_alsoPat(TranslPattern,ESet) :-
   ((\+var(TranslPattern),TranslPattern = alsoPat(X,Y)) ->
        is_member_set(X,ESet),
        unify_also_patterns(X,Y)
   ; is_member_set(TranslPattern,ESet)
   ).

unify_also_patterns(X,Y) :-
   unify_also_patterns(X,Y,R),
   evaluate_argument(X,EX),
   evaluate_argument(Y,EY),
   ((EX=EY;EY=R) -> true % both patterns should be equal
   ; add_error_fail(csp_sets, 'Both patterns in the also pattern do not match: ', alsoPat(X,Y))
   ).




%%%%%%%%%%%% Unit Tests for unify_also_patterns/3 %%%%%%%%%%%%%%
:- assert_must_succeed((csp_sets: unify_also_patterns(int(3),int(X),R), X == 3, R == int(3))).
:- assert_must_fail((csp_sets: unify_also_patterns(int(3),int(4),_R))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([X,Y,Z]),tuple([c,int(1),int(2)]),R), X == c, Y == int(1), Z == int(2), R == tuple([c,int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([X,_Y]),tuple([c,int(1),int(2)]),R), X == c, R == tuple([c,tuple([int(1),int(2)])]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([c,int(1),int(2)]),tuple([X,Y,Z]),R), X == c, Y == int(1), Z == int(2), R == tuple([c,int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([c,int(1),int(2)]),tuple([X,_Y]),R), X == c, R == tuple([c,tuple([int(1),int(2)])]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([c,int(1),int(2),int(3)]),tuple([X,_Y]),R), X == c, R == tuple([c,tuple([int(1),int(2),int(3)])]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(record(c,[int(1),int(2)]),record(c,[X,Y]),R), X == int(1), Y == int(2), R == record(c,[int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(record(c,[int(1),int(2)]),tuple([c,X,Y]),R), X == int(1), Y == int(2), R == record(c,[int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([c,int(1),int(2)]),record(c,[X,Y]),R), X == int(1), Y == int(2), R == record(c,[int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(tuple([c,int(1),int(2)]),record(c,[_X]),R), R == record(c,[tuple([int(1),int(2)])]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(record(c,[int(1),int(2)]),Y,R), Y == record(c,[int(1),int(2)]), R == record(c,[int(1),int(2)]))).
:- assert_must_succeed((csp_sets:unify_also_patterns(record(c,[_A]),tuple([c,int(1),na_tuple([int(2),int(3)])]),D), D == record(c,[tuple([int(1),na_tuple([int(2),int(3)])])]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(na_tuple([X,Y,Z]),na_tuple([c,int(1),int(2)]),R), X == c, Y == int(1), Z == int(2), R == na_tuple([c,int(1),int(2)]))).
:- assert_must_succeed((csp_sets: unify_also_patterns(list([X,Y,Z]),list([c,int(1),int(2)]),R), X == c, Y == int(1), Z == int(2), R == list([c,int(1),int(2)]))).
%%%%%%%%%%%% Unit Tests for unify_also_patterns/3 %%%%%%%%%%%%%%

unify_also_patterns(X,Y,R) :- (var(X) ; var(Y)), !, X=Y,R=Y.
unify_also_patterns(int(X),int(Y),int(R)) :- !,int(X)=int(Y),R=X.
unify_also_patterns(tuple(L),Tuple,tuple(R)) :- (Tuple = tuple(L1); Tuple = dotTuple(L1)),!,unify_tuple_elements(L,L1,R,tuple).
unify_also_patterns(dotTuple(L),Tuple,tuple(R)) :- (Tuple = tuple(L1); Tuple = dotTuple(L1)),!,unify_tuple_elements(L,L1,R,tuple).
unify_also_patterns(X,Y,record(CR,LR)) :- 
  ( X = record(CX,LX) -> !, 
  ( Y=tuple([H|T]) -> CX=H,!,unify_tuple_elements([CX|LX],[H|T],R,tuple),R=[CR|LR]
  ; Y=dotTuple([H|T]) -> CX=H,!,unify_tuple_elements([CX|LX],[H|T],R,dotTuple),R=[CR|LR]
  ; Y=record(CY,LY) -> CX=CY,!,unify_tuple_elements([CX|LX],[CY|LY],R,tuple),R=[CR|LR]
  ; atomic(Y) -> fail % in case we are comparing a record with a simple constructor
  ; otherwise -> add_error_fail(unify_also_patterns, 'Could not unify values (inside of set comprehension): ',unify_also_patterns(X,Y))
  )
  ; Y = record(_,_) -> !, unify_also_patterns(Y,X,record(CR,LR))
  ; otherwise -> fail).
unify_also_patterns(na_tuple(L),na_tuple(L1),na_tuple(R)) :- !,unify_tuple_elements(L,L1,R,na_tuple).
unify_also_patterns(list(X),list(Y),list(R)) :- !,if(list(X)=list(Y),R=Y,fail).
%  add_error_fail(unify_also_patterns,'Unification type failure: ', '='(list(X),list(Y)))).
unify_also_patterns(X,Y,_R) :- add_error_fail(csp_sets, 'Could not unify values (inside of set comprehension): ',unify_also_patterns(X,Y)).


%%%%%%%%%%%% Unit Tests for unify_tuple_elements/3 %%%%%%%%%%%%%%
:- assert_must_succeed((csp_sets: unify_tuple_elements([int(1),int(2)],[int(1),int(2)],R,_), R == [int(1),int(2)])).
:- assert_must_succeed((csp_sets: unify_tuple_elements([int(1),int(2),int(3)],[int(1),_X],R,tuple), R == [int(1),tuple([int(2),int(3)])])).
:- assert_must_succeed((csp_sets: unify_tuple_elements([int(1),_X],[int(1),int(2),int(3)],R,tuple), R == [int(1),tuple([int(2),int(3)])])).
:- assert_must_succeed((csp_sets: unify_tuple_elements([int(0),int(1),int(2),int(3)],[int(0),tuple([int(1),int(2),int(3)])],R,tuple), R == [int(0),int(1),int(2),int(3)])).
%%%%%%%%%%%% Unit Tests for unify_tuple_elements/3 %%%%%%%%%%%%%%

unify_tuple_elements([],[],R,_TupleType) :- !,R=[].
unify_tuple_elements([HX|TX],[HY|TY],R,TupleType) :- !,
     unfold_dot_tuples([HX|TX],[HHX|TTX]),unfold_dot_tuples([HY|TY],[HHY|TTY]), %still possible that some tuples() are lurking inside of the dot tuple list 
     ( (TTY = [], var(HHY), TTX \= [], TupleType=tuple) -> unify_to_rest([HHX,TTX], R, TupleType),[HHY]=R
     ; (TTX = [], var(HHX), TTY \= [], TupleType=tuple) -> unify_to_rest([HHY,TTY], R, TupleType),[HHX]=R
     ; otherwise -> csp_tuples: unify_arg2(HHX,HHY,HR,no_loc_info_available), unify_tuple_elements(TTX,TTY,TR,TupleType),R = [HR|TR]).
% we don't need to raise an exception when we cannot unify the tuple elements.
%unify_tuple_elements(X,Y,_R,_TupleType) :- add_error_fail(csp_sets, 'Could not unify values (inside of set comprehension): ', unify_tuple_elements(X,Y)).

unify_to_rest(L,R,Tuple) :-
   flatten(L,FL),
   functor(Term,Tuple,1),arg(1,Term,FL),
   R = [Term].

treat_generators(Generators,Pats,Sets,ResGuard) :-
   treat_generators(Generators,Pats,Sets,true,ResGuard).
   %,print(treat_generators(Generators,Pats,Sets,true,ResGuard)),nl.

treat_generators([],Pats,Sets,Guard,ResGuard) :-
   Pats=[],Sets=[],ResGuard=Guard.
treat_generators([H|T],Pats,Sets,CurGuard,ResGuard) :-
   (H=comprehensionGenerator(Pat,Set) ->
       Pats=[Pat|PatT],Sets=[Set|SetT],
       CurGuard1=CurGuard
   ;H=comprehensionGuard(Guard) ->
       PatT=Pats,SetT=Sets,
       % Choosing the order of the first two arguments does matter. why? (see CSP/ref_becnchmarks/basin_olderog_bank.csp example)
       clever_bool_and(CurGuard,Guard,CurGuard1)
   ;otherwise ->
       add_internal_error('Internal Error: Could not treat Set Comprehension Generator List: ',[H|T]),fail
   ),
   treat_generators(T,PatT,SetT,CurGuard1,ResGuard).     

clever_bool_and(true,X,R) :- !,R=X.
clever_bool_and(X,true,R) :- !,R=X.
clever_bool_and(G1,G2,bool_and(G1,G2)).

:- use_module(probcspsrc(csp_tuples),[is_constructor/3]).
% maybe we should use same code as for compile_head_para

:- assert_must_succeed((csp_sets:l_translate_pattern([emptySet,set([X]),'Set'(setValue([int(1),int(2)])),dotpat([Y,Z,emptySet])],R), 
                 R == [setValue([]),setValue([X]),setValue([int(1),int(2)]),tuple([Y,Z,setValue([])])])).

translate_pattern(V,R) :- var(V),!,R=V.
translate_pattern(dotpat([X|T]),R) :- \+var(X),is_constructor(X,Constructor,_SubTypes),
     l_translate_pattern(T,LT),!, R=record(Constructor,LT).
translate_pattern(dotpat(T),R) :- l_translate_pattern(T,LT),!, R=tuple(LT).
translate_pattern(tuplePat(T),R) :- l_translate_pattern(T,LT),!, R=na_tuple(LT).
translate_pattern(listPat(List),R) :- l_translate_pattern(List,LT),!, R=list(LT).
translate_pattern(singleSetPat(List),R) :- l_translate_pattern(List,LT),!, R=setValue(LT).
translate_pattern(emptySet,R) :- !, R=setValue([]).
translate_pattern(set(List),R) :- l_translate_pattern(List,LT),!, R=setValue(LT).
translate_pattern('Set'(S),R) :- !,R=S.
translate_pattern(appendPattern([H|T]),R) :- T==[],!, translate_pattern(H,R).
translate_pattern(appendPattern([H|T]),R) :- nonvar(H),H=listPat(HH),
   haskell_csp:is_list_skeleton(HH),
   l_translate_pattern(HH,LHH),
   translate_pattern(appendPattern(T),list(LTT)),!, append(LHH,LTT,ResL),R=list(ResL).
% more complicated append patterns
translate_pattern(alsoPattern([X,Y]),R) :- !, translate_pattern(X,XR),translate_pattern(Y,YR),R = alsoPat(XR,YR).
translate_pattern(X,R) :- ground(X),force_evaluate_argument(X,EX),!,R=EX.
translate_pattern(X,X) :-
   add_internal_error('Internal Error: Could not translate pattern: ',X).

l_translate_pattern(Patterns,TranslatedPatterns) :-
   maplist(translate_pattern,Patterns,TranslatedPatterns).

:- assert_must_succeed(( csp_sets:is_member_comprehension_set(int(12), 
              rangeEnum(['*'(_x,_y)]),
              [comprehensionGenerator(_x,setValue([int(1),int(3)])),
               comprehensionGenerator(_y,setValue([int(1),int(2),int(4)]))])  )).           
:- assert_must_fail(( csp_sets:is_member_comprehension_set(int(11), 
              rangeEnum(['*'(_x,_y)]),
              [comprehensionGenerator(_x,setValue([int(1),int(3)])),
               comprehensionGenerator(_y,setValue([int(1),int(2),int(4)]))])  )).       
:- assert_must_succeed(( csp_sets:is_member_comprehension_set(int(X), 
              rangeEnum(['*'(_x,_y)]),
              [comprehensionGenerator(_x,setValue([int(1),int(3)])),
               comprehensionGenerator(_y,setValue([int(1),int(2),int(4)]))]),X=12  )).
/*
:- assert_must_succeed(( csp_sets:is_member_comprehension_set(int(X), 
              rangeEnum([XX,YY]),[comprehensionGenerator(XX,setExp(rangeClosed(int(3),int(4)))),
                       comprehensionGenerator(YY,setExp(rangeClosed('+'(int(XX),int(2)),'+'(int(XX),int(3)))))]),X=7  )).
*/
       
is_member_comprehension_set(X,rangeEnum(ExprList),GeneratorList) :- ExprList=[Expr|TT],TT==[],var(Expr),!,
   get_waitvars_for_generator_list(GeneratorList,WaitVars),
     when(ground(WaitVars),
        (treat_generators(GeneratorList,Vars,Sets,Guard),
        % print(treat_generators(for(X,ExprList),Vars,Sets,Guard)),nl,
         Expr=X,
         check_boolean_expression(Guard),
        % print(checked(Guard,Vars,Sets)),nl,
         generator_sol(Vars,Sets,set))). 
is_member_comprehension_set(X,ExprList,GeneratorList) :-  !, %ExprList = [_,_|_],fail,!,
   /* if more than one element in ExprList:
      we need to expand it; we cannot instantiate the single variable and just check the generators,guards 
      otherwise pending co-routines can occur (e.g., { x , x1 | x<-{1..4}, x1 <-{x+2..x+3} }  )
      Also: currently we cannot check it symbolically if the elment of ExprList is not a variable */
  % print(expanding_comprehension_set(ExprList,GeneratorList)),nl,
   expand_set_comprehension(ExprList,GeneratorList,ExpandedSet),
   is_member_set(X,ExpandedSet).
is_member_comprehension_set(X,T,G) :-
   add_error_fail(is_member_comprehension_set,'Could not evaluate: ', is_member_comprehension_set(X,T,G)).


:- assert_must_succeed(( csp_sets:extract_variables_from_generator_list(
              [comprehensionGenerator(_x,setValue([int(1),int(3)])),
               comprehensionGenerator(_y,setValue([int(1),int(2),int(4)]))],R),
               R == [_x,_y])).   

% TODO: does not extract local quantified variables for nested set comprehensions !
% extract locally quantified variables from a set comprehension generator list
extract_variables_from_generator_list([],R) :- !,R=[].
extract_variables_from_generator_list([comprehensionGuard(_)|T],Res) :- !,
   extract_variables_from_generator_list(T,Res).
extract_variables_from_generator_list([comprehensionGenerator(Var,Set)|T],Res) :- !,
    check_variable(Var),
    extract_variables_from_generator_list(T,TVar),
    term_variables(Var,Vars),
    add_variables(Vars,TVar,Res,Set).
extract_variables_from_generator_list(X,R) :-
   add_internal_error('Not a generator list: ', X),
   R=[].
   
 :- assert_must_succeed(( csp_sets:extract_local_variables_from_generator_list(
              [comprehensionGenerator(_x,setValue([int(1),int(3)])),
               comprehensionGenerator(_y,setValue([int(1),int(2),int(4)]))],R),
               R == [_x,_y])).
% TODO: does not extract local quantified variables for nested set comprehensions !
% extract locally quantified variables from a set comprehension generator list
extract_local_variables_from_generator_list([],R) :- !,R=[].
extract_local_variables_from_generator_list([comprehensionGuard(_)|T],Res) :- !,
   extract_local_variables_from_generator_list(T,Res).
extract_local_variables_from_generator_list([comprehensionGenerator(Var,Set)|T],Res) :- !,
    check_variable(Var),
    extract_local_variables_from_generator_list(T,TVar),
    term_variables(Var,Vars),
    add_variables(Vars,TVar,Res1,Set),
    extract_local_variables_from_set_expression(Set,Res1,Res).
extract_local_variables_from_generator_list(X,R) :-
   add_internal_error('Not a generator list: ', X),
   R=[].

% TODO: are there any operators we are missing !??
extract_local_variables_from_set_expression(X,I,O) :- var(X),!,
  %add_error(extract_local_variables_from_set_expression,'Variable expr. :',X),
  I=O.
extract_local_variables_from_set_expression(Set,InVar,OutVar) :- unary_set_op(Set,A),!,
   extract_local_variables_from_set_expression(A,InVar,OutVar).
extract_local_variables_from_set_expression(Set,InVar,OutVar) :- binary_set_op(Set,A,B),!,
   extract_local_variables_from_set_expression(A,InVar,V1),
   extract_local_variables_from_set_expression(B,V1,OutVar).
extract_local_variables_from_set_expression(setEnum(List),InVar,OutVar) :- !,
   l_extract_local_variables_from_set_expression(List,InVar,OutVar).
extract_local_variables_from_set_expression(closureComp(Generators,Set),In,Out) :- !,
   extract_local_variables_from_generator_list(Generators,GV),
   add_variables(GV,In,Out,Set).
extract_local_variables_from_set_expression(setExp(RangeExpr,Generators),In,Out) :- !,
   extract_local_variables_from_generator_list(Generators,GV),
   add_variables(GV,In,Out,RangeExpr).

extract_local_variables_from_set_expression(_Set,In,Out) :-
  %print(uncovered_set_extract(_Set)),nl,
  Out=In.
% what if we have a set of values, containing e.g. the card operator on setComprehensions !!
% TO DO: maybe propagate local variables up in haskell_csp_analyzer.pl and make available to setComp ??

unary_set_op(builtin_call(X),R) :- unary_set_op(X,R).
unary_set_op('Union'(A),A).
unary_set_op('Inter'(A),A).
binary_set_op(builtin_call(X),A,B) :- binary_set_op(X,A,B).
binary_set_op(union(A,B),A,B).
binary_set_op(diff(A,B),A,B).
binary_set_op(inter(A,B),A,B).

:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([],X,Y), Y == X)).
:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([builtin_call(union(S,setExp(rangeEnum([_i])))), 
		builtin_call(union(_X,_Y))],S,Out), Out == S)).
:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([builtin_call(union(_S,setExp(rangeEnum([I])))), 
		builtin_call(union(_X,_Y))],_i,Out), Out == I)).
:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([builtin_call(union(_S,setExp(rangeEnum([_i])))), 
		builtin_call(inter(X,_Y))],X,Out), Out == X)).
:- assert_must_succeed((csp_sets: csp_sets: extract_local_variables_from_generator_list([comprehensionGenerator(rangeEnum([Y]),setExp(rangeEnum([X]),[comprehensionGenerator(X,setExp(rangeClosed(int(1),int(4))))]))],L), L == [X,Y])).
:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([setEnum([S,Y,_X]), 
		builtin_call('Inter'(Y))],S,Out), Out == S)).
:- assert_must_succeed((csp_sets:l_extract_local_variables_from_set_expression([builtin_call('Union'(_S)), 
		builtin_call('Inter'(Y))],Y,Out), Out == Y)).

l_extract_local_variables_from_set_expression(X,I,O) :- var(X),!,
  add_internal_error(/*l_extract_local_variables_from_set_expression,*/'Variable expr. :',X), I=O.
l_extract_local_variables_from_set_expression([],In,Out) :- !, Out = In.
l_extract_local_variables_from_set_expression([H|T],In,Out) :- !,
  extract_local_variables_from_set_expression(H,In,In2),
  l_extract_local_variables_from_set_expression(T,In2,Out).
l_extract_local_variables_from_set_expression(X,In,Out) :-
  add_internal_error('Unknown expr.: ',X), In=Out.


check_variable(V) :- atomic(V), channel(V,_),!,
  add_error(csp_sets,'Channel name used for local variable: ',V).
check_variable(_).
 
add_variables([],TVar,TVar,_).
add_variables([Var|T],TVar,Res,Set) :-
   (exact_member(Var,TVar)
       -> (add_error(csp_sets,'Variable appears twice in Generator list:',
                       (Var,[comprehensionGenerator(Var,Set)|T])),
                       /* TODO: FIX; this is actually allowed ?? !! */
           TVar1 = TVar)
       ;  TVar1 = [Var|TVar]),
   add_variables(T,TVar1,Res,Set).

 /* --------- */  
 /* BIG UNION */
 /* --------- */ 
 
 
:- assert_must_succeed(( csp_sets:big_union(setValue([setValue([int(3),int(4)]),setValue([int(2),int(9)])]),R),
                          R == setValue([int(2),int(3),int(4),int(9)]) )).

:- block big_union(-,?).
big_union(S1,Res) :- %print(big_union(S1,Res)),nl,
   expand_symbolic_set(S1,setValue(ES1),big_union),
   %%print(big2(ES1)),nl,
   big_union_add(ES1,setValue([]),Res). %, print(big_res(Res)),nl.

big_union_add([],R,R).
big_union_add([H|T],S2,Res) :- union_set(H,S2,S3),
   big_union_add(T,S3,Res).  
   
   
   
 /* --------- */ 
 /* BIG INTER */
 /* --------- */ 
 
 
:- assert_must_succeed(( csp_sets:big_inter(setValue([setValue([int(3),int(4)]),setValue([int(2),int(4)])]),R),
                          R == setValue([int(4)]) )).
:- assert_must_succeed(( csp_sets:big_inter(setValue([setValue([int(3),int(4)]),setValue([int(2),int(4)]),setValue([])]),R),
                          R == setValue([]) )).


:- block big_inter(-,?).
big_inter(S1,Res) :- 
   expand_symbolic_set(S1,setValue(ES1),big_inter),
   (ES1 = [H|T]
     -> big_inter_del(T,H,Res)
     ;  (add_error(csp_sets,'At least one set needed for Inter: ','Inter'(S1)),
         fail)
   ).

big_inter_del([],R,R).
big_inter_del([H|T],S2,Res) :- inter_set(H,S2,S3),
   big_inter_del(T,S3,Res).  
   
