% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(closures,[construct_closure/4, is_closure/4, is_closure_x/5,
                    construct_closure_if_necessary/4,
                    is_non_expanded_closure/1,
                    construct_member_closure/4,
                    construct_not_member_closure/4,
                    construct_complement_closure/3,
                    is_member_closure/5, is_not_member_closure/5,
                    is_not_member_value_closure/3,
                    is_not_member_value_closure_or_integerset/3,
                    is_lambda_value_domain_closure/5, is_lambda_closure/7,
                    select_equality/6,
                    is_special_infinite_closure/3,
                    is_id_closure_over/5,
                    is_full_id_closure/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides various utility functions to analyse ProB closures.').

construct_closure(Parameters, ParameterTypes, Body, Res) :-
    Res = closure(Parameters, ParameterTypes, Body).
   % Res = closure_x(Parameters, ParameterTypes, Body,_).  %% STILL HAS PROBLEMS with delay, e.g. inside b_test_exists !!


% an optimized version of construct_closure, which will try to produce explicit values if possible
construct_closure_if_necessary(_,_,b(falsity,pred,_),Res) :- !, Res=[].
construct_closure_if_necessary([ID], [T1], b(Pred,pred,_), Res) :- %print(s(ID,T1,Pred)),nl,trace,
    construct_unary_closure(Pred,ID,T1,SET),!,
    Res = SET.
construct_closure_if_necessary(Parameters, ParameterTypes, Body, Res) :-
    Res = closure(Parameters, ParameterTypes, Body).
    
:- use_module(b_global_sets,[try_b_type2global_set/2]).
:- use_module(custom_explicit_sets,[try_expand_and_convert_to_avl/2]).
construct_unary_closure(member(b(identifier(ID),T1,_),b(value(SET),set(T1),_)),ID,T1,Res) :- Res=SET.
construct_unary_closure(truth,_,T1,Res) :- try_b_type2global_set(T1,Res).
construct_unary_closure(equal(b(identifier(ID),T1,_),b(value(SET),T1,_)),ID,T1,Res) :- 
    try_expand_and_convert_to_avl([SET],Res).


is_closure(closure_x(Parameters, ParameterTypes, Body, _Exp), Parameters, ParameterTypes, Body).
is_closure(closure(Parameters, ParameterTypes, Body), Parameters, ParameterTypes, Body).


is_closure_x(closure_x(Parameters,ParameterTypes,Body,Exp), Parameters,ParameterTypes,Body,Exp).
is_closure_x(  closure(Parameters,ParameterTypes,Body),      Parameters,ParameterTypes,Body,none).


is_non_expanded_closure(closure(_,_,_)).
is_non_expanded_closure(closure_x(_,_,_,E)) :- nonvar(E).



:- use_module(bsyntaxtree,[create_texpr/4, safe_create_texpr/3]).
% following not useful: construct_member_closure currently always called where the construction is needed 
%construct_member_closure(ID,_Type,ClosureSetExpression,Result) :-
%        nonvar(ClosureSetExpression),ClosureSetExpression = value(S),!,
%        print(construct_member_closure_value(ID,S)),nl, %%
%        Result=S.
construct_member_closure(ID,Type,ClosureSetExpression,Result) :-
        check_result_instantiation(Result,construct_member_closure(ID)),
	    create_texpr(identifier(ID),Type,[],TIdentifier), % used to be [generated]
	    safe_create_texpr(ClosureSetExpression,set(Type),TClosureSet),
	    safe_create_texpr(member(TIdentifier,TClosureSet),pred,TPred),
        construct_closure([ID],[Type],TPred,Result).
construct_not_member_closure(ID,Type,ClosureSetExpression,Result) :-
        check_result_instantiation(Result,construct_not_member_closure(ID)),
	    create_texpr(identifier(ID),Type,[],TIdentifier), % used to be [generated]
	    safe_create_texpr(ClosureSetExpression,set(Type),TClosureSet),
	    safe_create_texpr(not_member(TIdentifier,TClosureSet),pred,TPred),
        construct_closure([ID],[Type],TPred,Result).

:- use_module(error_manager,[add_internal_error/2]).
% check that we do not instantiate result to early (rather than using equal_object)
check_result_instantiation(X,_) :- var(X),!.
check_result_instantiation(closure(_,_,_),_PP) :- !.
check_result_instantiation(X,PP) :-
    add_internal_error('Result already instantiated in incompatible way: ',check_result_instantiation(X,PP)).

is_member_closure([ID],[TYPE],b(PRED,_Pred,_), TYPE,SET) :-
    is_member_closure_aux(PRED, ID,TYPE,SET).

is_member_closure_aux(member(b(identifier(ID),TYPE,_),b(SET,set(TYPE),_)), ID,TYPE,SET).
is_member_closure_aux(subset(b(identifier(ID),TYPE,_),BSET),  ID,TYPE,SET) :- SET = pow_subset(BSET).
% can we also detect pow1_subset ? {x| x/= {} & x<: BSET}

% detect not_member closures + integerset as special not_member_closures
is_not_member_value_closure_or_integerset(global_set(X),TYPE,SET) :- !,
   is_not_member_global_set(X,TYPE,SET).
is_not_member_value_closure_or_integerset(C,TYPE,SET) :- is_not_member_value_closure(C,TYPE,SET).

is_not_member_global_set('INTEGER',integer,[]).
is_not_member_global_set('NATURAL',integer,X) :- 
   custom_explicit_sets:construct_less_equal_closure(-1,X). % {x|x<0}.
is_not_member_global_set('NATURAL1',integer,X) :- 
   custom_explicit_sets:construct_less_equal_closure(0,X). %X = {x|x<1}.

is_not_member_value_closure(closure(Par,T,B),TYPE,SET) :-
   is_not_member_closure(Par,T,B,TYPE,value(SET)).
is_not_member_closure([ID],[TYPE],b(PRED,_Pred,_),TYPE,SET) :-
    is_not_member_closure_aux(PRED,ID,TYPE,SET).

is_not_member_closure_aux(not_member(b(identifier(ID),TYPE,_),b(SET,set(TYPE),_)),ID,TYPE,SET).
is_not_member_closure_aux(not_equal(b(identifier(ID),TYPE,_),ONE),ID,TYPE,SET) :-
  (ONE = b(value(Val),_,_)
     -> custom_explicit_sets:construct_one_element_custom_set(Val,SetVal), SET = value(SetVal)
     ;  SET = set_extension([ONE])).
%is_not_member_closure_aux(PRED,ID,TYPE,SET) :- print(check_not_mem(PRED,ID,TYPE,SET)),nl,fail.

construct_complement_closure(Delta,Type,Closure) :-
        % print(generating_complement_closure(GlobalSet,Delta,Type)),nl,
        construct_not_member_closure('_zzzz_unary',Type,value(Delta),Closure).

     

/* lambda abstractions */
:- use_module(self_check).
:- assert_must_succeed((closures:is_lambda_closure([x,y],[integer,integer],b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))]),b(value(global_set('NATURAL')),set(integer),[])),pred,[]),b(equal(b(identifier(y),integer,[]),b(multiplication(b(identifier(x),integer,[]),b(identifier(x),integer,[])),integer,[])),pred,[])),pred,[]),OtherIDs,OtherTypes,_DOMAINPRED,_Res), OtherIDs=[x], OtherTypes=[integer])).
:- assert_must_fail((closures:is_lambda_closure([x,y],[integer,integer],b(conjunct(b(conjunct(b(member(b(identifier(x),integer,[]),b(value(global_set('NATURAL')),set(integer),[])),pred,[]),b(equal(b(identifier(y),integer,[]),b(multiplication(b(identifier(x),integer,[]),b(identifier(x),integer,[])),integer,[])),pred,[])),pred,[]),b(less(b(identifier(y),integer,[]),b(value(int(10)),integer,[])),pred,[])),pred,[]),_,_,_D,_Res)
)).

:- use_module(bsyntaxtree,[conjunction_to_list/2,conjunct_predicates/2]).
is_lambda_closure(Args,Types,ClosurePred, OtherIDs, OtherTypes, DOMAINPRED,Res) :-
% TO DO: do this more efficiently: if LambdaID occurs in any non-equal predicate : stop searching
% TO DO: check if is_infinite_equality_closure is not a special case of lambda closure ?
        append(OtherTypes,[LambdaType],Types), OtherTypes \= [],
        append(OtherIDs,[LambdaID],Args),
        Res=b(EXPR,LambdaType,EXPRINFO),
        %used to call: b_interpreter:member_conjunct(EQ,ClosurePred,DOMAINPRED), ; but inlined below for efficiency
        select_equality(ClosurePred,LambdaID,EXPR,LambdaType,EXPRINFO,DOMAINPRED),
        !. % avoid backtracking member_conjunct
%        tools:print_bt_message(is_lambda_closure(LambdaID)).
% Note: LAMBDA is usually '_lambda_result_' 

identifier_equality(b(equal(b(LHS,Type,LHSInfo),b(RHS,Type,RHSInfo)),pred,_),ID,Type,EXPR,EXPRINFO) :-
        identifier_equality_aux(LHS,LHSInfo,RHS,RHSInfo,ID,EXPR,EXPRINFO).
identifier_equality_aux(identifier(ID),_,EXPR,EXPRINFO,ID,EXPR,EXPRINFO) :- !.
identifier_equality_aux(EXPR,EXPRINFO,identifier(ID),_,ID,EXPR,EXPRINFO).

% find an equality ID = RHSExpr so that ID does not occur in RHSExpr nor in RestPred
% the identifier should be provided as input (for the cut below)
select_equality(ClosurePred,ID,RHSExpr,Type,Info,RestPred) :-
        conjunction_to_list(ClosurePred,List),
        select(EQ,List,RestList),
        identifier_equality(EQ,ID,Type,RHSExpr,Info),
        !, % once we find a first equality : no need to look for a second one as then does_not_occur in RestPred will always fail !
        does_not_occur_in(ID,b(RHSExpr,Type,Info)),
        conjunct_predicates(RestList,RestPred),
        does_not_occur_in(ID,RestPred).
   


% check whether we have a lambda closure and whether we can compute its domain
is_lambda_value_domain_closure(Args,Types,B,  DomainValue, EXPR) :-
   % tools_printing:print_term_summary(try_is_lambda_domain(Args,Types,B)), %
   is_lambda_closure(Args,Types,B, OtherIDs,OtherTypes, DomainPred, EXPR),!,
   %print(lambda_closure(OtherIDs)), translate:print_bexpr(EXPR),nl,
   construct_closure_if_necessary(OtherIDs,OtherTypes,DomainPred,DomainValue).
   %print(lambda_domain(Args)),nl, (IDs=[_,_|_] -> trace ; true),
   %translate:print_bvalue(DomainValue),nl.

%  LAMBDARES is usually _lambda_result_, LAMBDARES cannot occur in DOMAIN (is value)


  
:- use_module(typing_tools,[is_infinite_type/1]).
/* checking for infinite closures */
is_special_infinite_closure(_Par,T,b(truth,_Pred,_)) :- !,
    member(Type,T), is_infinite_type(Type),!.
is_special_infinite_closure(Par,T,Body) :- %print(check_infinite(Par,T,Body)),nl,
   is_infinite_equality_closure(Par,T,Body),!.
%is_special_infinite_closure(Par,T,Body) :- is_full_id_closure(Par,T,Body,TYPE), is_infinite_type(TYPE).
%is_special_infinite_closure(Par,T,Body) :- is_prj1_closure(Par,T,Body,T1,_T2), is_infinite_type(T1).
%is_special_infinite_closure(Par,T,Body) :- is_prj2_closure(Par,T,Body,_T1,T2), is_infinite_type(T2).
is_special_infinite_closure(Par,T,Body) :-
   is_not_member_closure(Par,T,Body,Type,value(_)), is_infinite_type(Type).

:- use_module(library(lists)).

greater_typing(greater(b(identifier(ID),integer,_),b(VUP,integer,_)),ID,UP) :- is_integer_val(VUP,UP).
greater_typing(greater_equal(b(identifier(ID),integer,_),b(VUP,integer,_)),ID,UP) :- is_integer_val(VUP,UP).
greater_typing(less(b(VUP,integer,_),b(identifier(ID),integer,_)),ID,UP) :- is_integer_val(VUP,UP).
greater_typing(less_equal(b(VUP,integer,_),b(identifier(ID),integer,_)),ID,UP) :- is_integer_val(VUP,UP).

less_typing(less(b(identifier(ID),integer,_),b(VUP,integer,_)),ID,UP) :- is_integer_val(VUP,UP).
less_typing(less_equal(b(identifier(ID),integer,_),b(VUP,integer,_)),ID,UP) :- is_integer_val(VUP,UP).
less_typing(greater(b(VUP,integer,_),b(identifier(ID),integer,_)),ID,UP) :- is_integer_val(VUP,UP).
less_typing(greater_equal(b(VUP,integer,_),b(identifier(ID),integer,_)),ID,UP) :- is_integer_val(VUP,UP).

is_integer_val(integer(UP),UP).
is_integer_val(value(V),UP) :- nonvar(V),V=int(UP).

infinite_set(value(V)) :- !, % most common case, all other clauses do not seem to get covered
   nonvar(V),infinite_value_set(V).
infinite_set(Rel) :- is_relation(Rel,A,B),!,
  (infinite_set(A) -> true ; infinite_set(B)).
infinite_set(cartesian_product(b(A,_,_),b(B,_,_))) :-
  (infinite_set(A) -> true ; infinite_set(B)).
infinite_set(pow_subset(b(A,_,_))) :- infinite_set(A).
infinite_set(pow1_subset(b(A,_,_))) :- infinite_set(A).
infinite_set(fin_subset(b(A,_,_))) :- infinite_set(A).
infinite_set(fin1_subset(b(A,_,_))) :- infinite_set(A).
infinite_set(seq(_)).
infinite_set(seq1(_)).
infinite_set(iseq(b(A,_,_))) :- infinite_set(A).
infinite_set(iseq1(b(A,_,_))) :- infinite_set(A).
infinite_set(perm(b(A,_,_))) :- infinite_set(A).
infinite_set(integer_set(X)) :-
   X='NATURAL' ; X='NATURAL1' ; X='INTEGER'.
infinite_set(string_set).

infinite_value_set(global_set(X)) :- 
   X='NATURAL' ; X='NATURAL1' ; X='INTEGER'.
infinite_value_set(closure(P,T,B)) :-
   T \= [integer], % otherwise we could intersect with NATURAL,...
   custom_explicit_sets:is_infinite_closure(P,T,B).
infinite_value_set(freetype(ID)) :- kernel_freetypes:is_infinite_freetype(ID).

% to do: extend; could be value(infinite_closure)...

is_relation(relations(b(A,_,_),b(B,_,_)),A,B).
is_relation(partial_function(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_function(b(A,_,_),b(B,_,_)),A,B).
is_relation(partial_injection(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_injection(b(A,_,_),b(B,_,_)),A,B).
is_relation(partial_surjection(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_surjection(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_bijection(b(A,_,_),b(B,_,_)),A,B).
is_relation(partial_bijection(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_relation(b(A,_,_),b(B,_,_)),A,B).
is_relation(surjection_relation(b(A,_,_),b(B,_,_)),A,B).
is_relation(total_surjection_relation(b(A,_,_),b(B,_,_)),A,B).


/* Equality closures {x1,x2,...|id=E2}, where id does not occur in E2 and id =xi */
% should cover id, prj1, prj2
% {x,y|y:BOOL & x=f(y) } or %x.(x:NATURAL|Expr(x))
% would not be infinite {x,y|x:BOOL & x=f(g(x)*y)}  , g={FALSE|->0, TRUE|->1}, f = ...
% we assume Well-Definedness

%is_infinite_equality_closure(IDs,TYPES,B) :- IDs = [_,_|_],
%  print(try(IDs,TYPES,B)),nl,fail.
is_infinite_equality_closure(IDs, TYPES, Body) :-  
  IDs = [_,_|_], % at least two variables
  %print(identify_eq(IDs)),nl,
  check_eq_body(Body,[],OutConstrained),
  %print(out(OutConstrained)),nl,
  (member(_ID/infinite,OutConstrained) -> true
    ; contains_infinite_type(IDs,TYPES,OutConstrained)). % , print(infinite(IDs)),nl.
 
contains_infinite_type([ID|IT],[H|T],OutConstrained) :-
  ((is_infinite_type(H),\+ member(ID/_,OutConstrained))
    -> true ; contains_infinite_type(IT,T,OutConstrained)).

:- use_module(b_ast_cleanup,[definitely_not_empty_and_finite/1]).
:- use_module(external_functions,[external_pred_always_true/1]).
check_eq_body(b(B,pred,_),InConstrained,OutConstrained) :- check_eq_body_aux(B,InConstrained,OutConstrained).
check_eq_body_aux(conjunct(A,B),InConstrained,OutConstrained) :- !,
        check_eq_body(A,InConstrained,OutConstrained1),
        check_eq_body(B,OutConstrained1,OutConstrained).
check_eq_body_aux(equal(LHS,RHS), Constrained, [ID/equal|Constrained]) :-
        LHS=b(identifier(ID),_TYPE,_),
        \+ member(ID/_,Constrained), % no constraints on ID so far
        does_not_occur_in(ID,RHS),!. % the equation must have a solution; assuming well-definedness
check_eq_body_aux(equal(LHS,RHS), Constrained, [ID/equal|Constrained]) :- % symmetric to case above
        RHS=b(identifier(ID),_TYPE,_),
        \+ member(ID/_,Constrained),
        does_not_occur_in(ID,LHS),!.
check_eq_body_aux(member(b(identifier(ID),TYPE,_),SET),Constrained,[ID/INFINITE|Constrained]) :-
       \+ member(ID/_,Constrained),
       (is_infinite_type(TYPE)
        -> %check that SET is infinite; otherwise remove from IDs
           SET = b(S,set(TYPE),_),
           % print(check_infinite(S)),nl,
          (infinite_set(S) -> INFINITE=infinite; 
            b_ast_cleanup:definitely_not_empty_and_finite(SET) -> INFINITE = finite)
    ;  %print(finite(TYPE)),nl,
       b_ast_cleanup:definitely_not_empty_and_finite(SET), % otherwise we have no solution
       INFINITE = finite
   ).
check_eq_body_aux(EXPR,Constrained,[ID/infinite|Constrained]) :-
     %% TO DO: store bounds in ID/... list !
     greater_typing(EXPR,ID,_UP),
     \+ member(ID/_,Constrained).
check_eq_body_aux(EXPR,Constrained,[ID/infinite|Constrained]) :-
     less_typing(EXPR,ID,_UP),
     \+ member(ID/_,Constrained).
check_eq_body_aux(truth,Constrained,Constrained).
check_eq_body_aux(external_pred_call(FunName,_Args),Constrained,Constrained) :-
     external_pred_always_true(FunName).


:- use_module(bsyntaxtree,[occurs_in_expr/2]).
does_not_occur_in(ID,EXPR) :- \+ occurs_in_expr(ID,EXPR).



% check if we have a closure of type id(SetValue)

is_id_closure_over([ID1,ID2], [TYPE,TYPE],Body, ID_Domain, Full) :- nonvar(Body), Body=
  b(equal(b(identifier(ID1),TYPE,_),b(identifier(ID2),TYPE,_)),pred,_),!,
  convert_type_to_value(TYPE,ID_Domain), Full=true.
is_id_closure_over(Par,Types,Body,ID_Domain,Full) :- nonvar(Par),nonvar(Body),
  is_member_closure(Par,Types,Body,_,Set), % print(member_closure(Set)),nl,
  nonvar(Set),
  Set = identity(b(VAL,set(_TYPE),_)), 
  nonvar(VAL), VAL=value(ID_Domain),
  (custom_explicit_sets:is_definitely_maximal_set(ID_Domain) -> Full=true ; Full=false).

%:- use_module(kernel_objects,[all_strings/1]).
convert_type_to_value(integer,global_set('INTEGER')).
convert_type_to_value(global(G),global_set(G)).
convert_type_to_value(boolean,BS) :- BS=[pred_true /* bool_true */,pred_false /* bool_false */].  % TO DO: generate AVL ?
convert_type_to_value(string,global_set('STRING')). %  :- all_strings(S).
%convert_type_to_value(Type,closure([x],[Type],TRUTH)) :- ... TO DO



/* Event-B id closure over full Type */

is_full_id_closure(P,T,B) :- is_id_closure_over(P,T,B,_,true).


% currently commented out in is_special_infinite_closure
%is_prj1_closure([ID1,_ID2,RESID],[Type1,Type2,Type1],
%  b(equal(b(identifier(RESID),Type1,_),b(identifier(ID1),Type1,_)),pred,_),Type1,Type2).
%is_prj2_closure([_ID1,ID2,RESID],[Type1,Type2,Type2],
%  b(equal(b(identifier(RESID),Type2,_),b(identifier(ID2),Type2,_)),pred,_),Type1,Type2).

