:- module(input_syntax_tree, [remove_raw_position_info/2
                             ,get_raw_position_info/2
                             ,extract_raw_identifiers/2
                             ,get_definition_name/2
                             ,raw_replace/3
                             ,raw_conjunct/2
                             ,create_fresh_identifier/3
                             ]).


:- use_module(error_manager).

get_raw_position_info(Compound,Info) :-
   compound(Compound),
   functor(Compound,_,Arity), Arity>0,
   arg(1,Compound,Info),
   Info = pos(_,_,_,_,_,_),!.
get_raw_position_info(Compound,Info) :-
   add_internal_error('Could not get position info: ',get_raw_position_info(Compound,Info)),
   Info = unknown.

:- use_module(tools,[safe_functor/4]).

remove_raw_position_info([],Output) :- !,Output=[].
remove_raw_position_info([I|Irest],[O|Orest]) :-
    !,remove_raw_position_info(I,O),
    remove_raw_position_info(Irest,Orest).
remove_raw_position_info(Input,Output) :-
    compound(Input),functor(Input,Functor,Arity),Arity>0,!,
    Arity2 is Arity-1,
    functor(Output,Functor,Arity2),
    remove_raw_position_info2(1,Arity,Input,Output).
remove_raw_position_info(Input,Input).


remove_raw_position_info2(Pos,Arity,Input,Output) :-
    Pos < Arity,!,Pos2 is Pos+1,
    arg(Pos2,Input,IA), arg(Pos,Output,OA),
    remove_raw_position_info(IA,OA),
    remove_raw_position_info2(Pos2,Arity,Input,Output).
remove_raw_position_info2(_Pos,_Arity,_Input,_Output).


extract_raw_identifiers(Ast,Ids) :-
    extract_raw_identifiers2(Ast,[],Unsorted,[]),
    sort(Unsorted,Ids).
extract_raw_identifiers2([],_) --> !,[].
extract_raw_identifiers2([I|Irest],Bound) -->
    !,extract_raw_identifiers2(I,Bound),
    extract_raw_identifiers2(Irest,Bound).
extract_raw_identifiers2(identifier(_,I),Bound) --> !,add_if_free(I,Bound).
extract_raw_identifiers2(definition(_,I,Args),Bound) --> !,
    add_if_free(I,Bound), extract_raw_identifiers2(Args,Bound).
extract_raw_identifiers2(Input,Bound) -->
    {raw_quantifier(Input,Ids,Subs)},!,
    {append(Ids,Bound,NewBound)},
    extract_raw_identifiers2(Subs,NewBound).
extract_raw_identifiers2(Input,Bound) -->
    {compound(Input),functor(Input,_Functor,Arity),Arity>1,!},
    extract_raw_identifiers3(2,Arity,Input,Bound).
extract_raw_identifiers2(_,_) --> [].
add_if_free(Id,Bound,In,Out) :-
    ( memberchk(Id,Bound) -> In=Out
    ; otherwise           -> In=[Id|Out]).

extract_raw_identifiers3(Pos,Arity,Input,Bound) -->
    {Pos=<Arity,!,Pos2 is Pos+1, arg(Pos,Input,Arg)},
    extract_raw_identifiers2(Arg,Bound),
    extract_raw_identifiers3(Pos2,Arity,Input,Bound).
extract_raw_identifiers3(_Pos,_Arity,_Input,_Bound) --> [].

raw_quantifier(forall(_,Ids,P),Ids,[P]).
raw_quantifier(exists(_,Ids,P),Ids,[P]).
raw_quantifier(comprehension_set(_,Ids,P),Ids,[P]).
raw_quantifier(event_b_comprehension_set(_,Ids,E,P),Ids,[E,P]).
raw_quantifier(lambda(Ids,P,E),Ids,[P,E]).
raw_quantifier(general_sum(_,Ids,P,E),Ids,[P,E]).
raw_quantifier(general_product(_,Ids,P,E),Ids,[P,E]).
raw_quantifier(quantified_union(_,Ids,P,E),Ids,[P,E]).
raw_quantifier(quantified_intersection(_,Ids,P,E),Ids,[P,E]).
raw_quantifier(any(_,Ids,P,S),Ids,[P,S]).
raw_quantifier(let(_,Ids,P,S),Ids,[P,S]).
raw_quantifier(var(_,Ids,S),Ids,[S]).
raw_quantifier(recursive_let(_,Id,E),[Id],[E]).

get_definition_name(expression_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(predicate_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(substitution_definition(_Pos,Name,_Args,_Body),Name).

% raw_replace(+InputAST,-Replaces,+OutputAST):
% InputAST:  A raw AST as it comes from the parser (without type information)
% Replaces:  A list of "replace(Id,Expr)" where Id the the identifier which
%            should be replaced by the expression Expr
% OutputAST: A raw AST where all matching identifiers are replaced by their
%            expression
raw_replace(Expr,[],Expr) :- !.
raw_replace(Old,Replaces,New) :- raw_replace2(Old,Replaces,New).
raw_replace2([],_,[]) :- !.
raw_replace2([E|Rest],Replaces,[N|NRest]) :-
    !,raw_replace2(E,Replaces,N),raw_replace2(Rest,Replaces,NRest).
raw_replace2(rewrite_protected(X),_Replaces,rewrite_protected(Y)) :- !,X=Y.
raw_replace2(identifier(_,Old),Replaces,New) :-
    memberchk(replace(Old,New),Replaces),!.
raw_replace2(Expr,Replaces,New) :-
    safe_functor(raw_replace2_expr,Expr,Functor,Arity),
    safe_functor(raw_replace2_new,New,Functor,Arity),
    raw_replace3(Arity,Expr,Replaces,New).
raw_replace3(0,_,_,_) :- !.
raw_replace3(N,Expr,Replaces,NExpr) :-
    arg(N,Expr,Old),
    arg(N,NExpr,New),
    raw_replace2(Old,Replaces,New),
    N2 is N-1,
    raw_replace3(N2,Expr,Replaces,NExpr).

raw_conjunct([],truth(none)).
raw_conjunct([H|T],R) :-
    ( T=[] -> H=R
    ; otherwise ->
        R=conjunct(none,H,RT),
        raw_conjunct(T,RT)
    ).

% create_fresh_identifier(+PreferredName,+Ast,-Name):
%   Generates an identifier that does not occur in Ast
%   PreferredName: An atom with the preferred name of the identifier
%   Ast: An untyped AST
%   Name: An identifier (an atom) that does not occur free in AST,
%      If PreferredName does not occur free in AST, it's 'PreferredName',
%      otherwise it's 'PreferredName_N' where N is a natural number
create_fresh_identifier(PreferredName,Ast,Name) :-
    extract_raw_identifiers(Ast,UsedIds),
    ( memberchk(PreferredName,UsedIds) ->
        % Allready in use, find a fresh one by adding a number
        atom_codes(PreferredName,PreferredCodes),
        create_fresh_identifier_aux(PreferredCodes,UsedIds,1,Name)
    ; otherwise ->
        Name = PreferredName).
create_fresh_identifier_aux(PreferredCodes,UsedIds,I,Name) :-
    number_codes(I,ICodes),append(PreferredCodes,[95|ICodes],Codes),
    atom_codes(NewName,Codes),
    ( memberchk(NewName,UsedIds) ->
        % Allready in use, try again with a higher number
        I2 is I+1,
        create_fresh_identifier_aux(PreferredCodes,UsedIds,I2,Name)
    ; otherwise ->
        Name = NewName).
