% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(kodkod,[apply_kodkod/3, apply_kodkod/4,
                  replace_by_kodkod/3,
                  kodkod_request/5,
                  apply_kodkod_to_counterexample_search/5,
                  kodkod_reset/0]).

:- use_module(library(lists)).
:- use_module(library(avl)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(translate)).
:- use_module(probsrc(predicate_analysis)).
:- use_module(probsrc(btypechecker)).
:- use_module(probsrc(b_global_sets),[b_global_set/1,lookup_global_constant/2]).
:- use_module(probsrc(preferences),[get_preference/2]).

:- use_module(kodkod_translate).
:- use_module(kodkod_tools).
:- use_module(kodkod_rewrite).
:- use_module(kodkod2).
:- use_module(kodkod_typing).
:- use_module(kodkod_process).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,kodkod).
:- module_info(description,'This is the central module of the B to Kodkod translation (still experimental).').

:- dynamic kodkod_error/1.

kodkod_reset :-
    ( stop_kodkod -> true; true).

:- block kodkod_request(?,?,?,?,-).
kodkod_request(Id,Signum,LocalState,State,_LWF) :-
    kodkod_request_intern(Id,Signum,LocalState,State).

apply_kodkod(Identifier,Predicate,NewPredicate) :-
    get_preference(try_kodkod_on_load,true),
    replace_by_kodkod(Identifier,Predicate,NewPredicate),!.
apply_kodkod(_Identifier,Predicate,Predicate).

create_id(Id,Type,b(identifier(Id),Type,[])).

% a version which takes ids and types list:
apply_kodkod(Ids,Types,Predicate,NewPredicate) :- 
    get_preference(try_kodkod_on_load,true),
    maplist(create_id,Ids,Types,Identifiers),
    replace_by_kodkod(Identifiers,Predicate,NewPredicate),!.
apply_kodkod(_Identifier,_Types,Predicate,Predicate).

:- use_module(probsrc(b_expression_sharing),[expand_cse_lazy_lets_if_cse_enabled/2]).
replace_by_kodkod(Constants,Predicate,NewPredicate) :-
    % Remove any CSE lazy_lets as Kodkod translation cannot deal with them yet
    expand_cse_lazy_lets_if_cse_enabled(Predicate,Predicate0),
    % The Predicate may already contain Kodkod calls, replace
    % them by their original B predicate
    undo_kodkod_replacements(Predicate0,Predicate1),
    % first we do a predicate analysis to determine ranges for integers
    try_predicate_analysis_with_global_sets(Predicate1,Constants,TaggedConstants,TaggedPredicate),
    replace_by_kodkod2(TaggedConstants,TaggedPredicate,NewPredicate).
replace_by_kodkod2(Constants,_Predicate,NewPredicate) :-
    % If the static analysis states that there is no possible solution for one
    % of the identifiers, we just say "false"
    somechk(is_inconsistent_expression,Constants),!,
    write('kodkod: Static analysis detected inconsistencies, new predicate is "false"'),nl,
    create_texpr(falsity,pred,[],NewPredicate).
replace_by_kodkod2(Constants,Predicate,NewPredicate) :-
    split_predicates(Predicate,PList),
    % we try to translate each predicate alone
    try_apply_transformation_l(PList,Constants,1,Result),
    print_results(Result),
    % split the result into two lists: the successfull translatable predicates (ToKodkod)
    % and those which were not translatable (StayProb)
    filter_translatable(Result,ToKodkod,StayProb),
    % fail if there are no translatable predicates
    ToKodkod = [_|_],
    % now, translate all translatable predicates in one run
    apply_transformation(Constants,ToKodkod,KPred),
    % the new 
    conjunct_predicates([KPred|StayProb],NewPredicate).

apply_kodkod_to_counterexample_search(Variables,Assumption,Assertion,KPred,Rest) :-
    get_preference(try_kodkod_on_load,true),    
    % The predicates may already contain Kodkod calls, replace
    % them by their original B predicate
    undo_kodkod_replacements(Assumption,Assumption1),
    undo_kodkod_replacements(Assertion,Assertion1),
    % first we do a predicate analysis to determine ranges for integers
    analysis_for_counterexample(Variables,Assumption1,Assertion1,
                                TaggedVariables,TaggedAssumption,TaggedAssertions),

    % analyse assumptions
    split_assumptions(TaggedAssumption,TaggedVariables,
                      TAssumptions,NTAssumptions,AssumptionsComponents),

    % analyse assertions
    translate_assertions(TaggedAssertions,TaggedVariables,TAssumptions,NTAssumptions,
                         UsedAssumptionNumbers,TAssertions,NTAssertions,UsedVariables),

    % if we do not find a translatalbe assertion, fail and do not apply Kodkod at all
    TAssertions = [_|_],

    findall( UP,
             (member(N,UsedAssumptionNumbers),nth1(N,AssumptionsComponents,component(UP,_))),
             UsedAssumptions),
    findall( UP,
             (nth1(N,AssumptionsComponents,component(UP,_)),nonmember(N,UsedAssumptionNumbers)),
             UnusedAssumptions),
    
    conjunct_predicates(TAssertions,TAssertion),
    create_texpr(negation(TAssertion),pred,[],Negation),
    append(UsedAssumptions,[Negation],ToTranslate),
    apply_transformation(UsedVariables,ToTranslate,KPred),

    ( NTAssertions = [] ->
        conjunct_predicates(UnusedAssumptions,Rest)
    ; otherwise ->
        conjunct_predicates(NTAssertions,RestAssertion),
        create_texpr(negate(RestAssertion),pred,[],RestNegated),
        conjunct_predicates([Assumption,RestNegated],Rest)).

split_assumptions(Assumption,Variables,Translatable,NotTranslated,AssumptionsComponents) :-
    predicate_components(Assumption,AssumptionsComponents),
    try_apply_transformation_to_components(AssumptionsComponents,Variables,1,AssumptionResults),
    map_ids_to_assumption(AssumptionResults,Translatable,NotTranslated).

translate_assertions(Assertion,Variables,TAssumptions,NTAssumptions,UsedAssumptions,
                     Translatable,NotTranslated,UsedVariables) :-
    conjunction_to_list(Assertion,Assertions),
    translate_assertions2(Assertions,Variables,TAssumptions,NTAssumptions,Unsorted,
                          Translatable,NotTranslated,UIds),
    sort(Unsorted,UsedAssumptions),
    sort(UIds,Ids),
    findall(V,(member(V,Variables),get_texpr_id(V,I),memberchk(I,Ids)),UsedVariables).
translate_assertions2([],_Variables,_TAssumptions,_NTAssumptions,[],[],[],[]).
translate_assertions2([Assertion|Arest],Variables,TAssumptions,NTAssumptions,
                      UsedAssumptions,Translatable,NotTranslated,Ids) :-
    ( translate_assertion(Assertion,Variables,TAssumptions,NTAssumptions,LUsedAssumptions,LUsedIds) ->
        append(LUsedAssumptions,Urest,UsedAssumptions),
        append(LUsedIds,Irest,Ids),
        Translatable = [Assertion|Trest], NotTranslated = NTrest
    ; otherwise ->
        UsedAssumptions = Urest,
        Ids = Irest,
        Translatable = Trest, NotTranslated = [Assertion|NTrest]),
    translate_assertions2(Arest,Variables,TAssumptions,NTAssumptions,Urest,Trest,NTrest,Irest).
translate_assertion(Assertion,Variables,TAssumptions,NTAssumptions,UsedAssumptions,Ids) :-
    find_identifier_uses(Assertion,[],Ids),
    find_used_predicates(Ids,TAssumptions,NTAssumptions,UsedAssumptions),
    try_apply_transformation(Assertion,Variables,1,R),
    functor(R,ok,_).
find_used_predicates(Ids,TAssumptions,NTAssumptions,UsedAssumptions) :-
    maplist(find_predicate(TAssumptions,NTAssumptions),Ids,AssumptionsL),
    append(AssumptionsL,Assumptions),
    sort(Assumptions,UsedAssumptions).

find_predicate(_TAssumptions,NTAssumptions,Id,_UsedAssumption) :-
    % if the assertion uses an ID that is also used by a non-translatable
    % assumption, we do not translate the assumption
    avl_fetch(Id,NTAssumptions),!,fail.
find_predicate(TAssumptions,_NTAssumptions,Id,UsedAssumptions) :-
    ( avl_fetch(Id,TAssumptions,UsedAssumption) ->
        UsedAssumptions = [UsedAssumption]
    ; otherwise ->
        UsedAssumptions = []).

map_ids_to_assumption(AssumptionResults,AssumptionIds,NotTranslated) :-
    findall( Id-N,
             ( nth1(N,AssumptionResults,ok(Ids,_,_,_)),
                 member(Id,Ids)),
             Ids1),
    list_to_avl(Ids1,AssumptionIds),
    findall( N,
             ( nth1(N,AssumptionResults,failure(Ids,_,_,_,_)),
                 member(Id,Ids)),
             Ids2),
    list_to_avl(Ids2,NotTranslated).
    
analysis_for_counterexample(Variables,Assumption,Assertion,
                            TaggedVariables,TaggedAssumption,TaggedAssertion) :-
    create_texpr(negation(Assertion),pred,[],NAssertion),
    create_texpr(conjunct(Assumption,NAssertion),pred,[],Conj),
    try_predicate_analysis_with_global_sets(Conj,Variables,TaggedVariables,TaggedConj),
    create_texpr(conjunct(TaggedAssumption,TaggedNAssertion),pred,_,TaggedConj),
    create_texpr(negation(TaggedAssertion),pred,_,TaggedNAssertion).

split_predicates(Predicate,Parts) :-
    get_preference(kodkod_for_components,true),!,
    predicate_components(Predicate,Components),
    findall(P,member(component(P,_),Components),Parts).
split_predicates(Predicate,Parts) :-
    conjunction_to_list(Predicate,Parts).

split_simple_equalities(InPredicate,Equalities,Rest) :-
    conjunction_to_list(InPredicate,InPredicates),
    split_list(is_simple_equality,InPredicates,EqualitiesL,RestL),
    conjunct_predicates(EqualitiesL,Equalities),
    conjunct_predicates(RestL,Rest).

is_simple_equality(TExpr) :-
    get_texpr_expr(TExpr,equal(A,B)),
    ( get_texpr_id(A,_),is_base_expression(B)
    ; get_texpr_id(_,B),is_base_expression(A)),!.

is_base_expression(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    is_base_expression2(Expr).
is_base_expression2(identifier(Id)) :- b_global_set(Id),!. % a deferred set
is_base_expression2(identifier(Id)) :- lookup_global_constant(Id,_). % an enumerated set element
is_base_expression2(integer(_)).
is_base_expression2(couple(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(set_extension(L)) :- maplist(is_base_expression,L).
is_base_expression2(value(_)).
is_base_expression2(boolean_true).
is_base_expression2(boolean_false).
is_base_expression2(empty_set).
is_base_expression2(bool_set).
is_base_expression2(add(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(minus(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(unary_minus(A)) :- is_base_expression(A).
is_base_expression2(multiplication(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(div(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(floored_div(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(modulo(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(power_of(A,B)) :- is_base_expression(A),is_base_expression(B).


undo_kodkod_replacements(Old,New) :-
    contains_kodkod_call(Old),!,
    undo_kodkod_replacements2(Old,New).
undo_kodkod_replacements(Old,Old).
undo_kodkod_replacements2(TOld,TNew) :-
    get_texpr_expr(TOld,Old),
    undo_kodkod_replacements3(Old,TOld,TNew).
undo_kodkod_replacements3(conjunct(A,B),_TOld,TNew) :-
    !,create_texpr(conjunct(NewA,NewB),pred,[],TNew),
    undo_kodkod_replacements2(A,NewA),
    undo_kodkod_replacements2(B,NewB).
undo_kodkod_replacements3(kodkod(Id,_),_TOld,Orig) :-
    !,get_original_problem(Id,Orig).
undo_kodkod_replacements3(_Old,TOld,TOld).

contains_kodkod_call(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    contains_kodkod_call2(Expr),!.
contains_kodkod_call2(kodkod(_,_)).
contains_kodkod_call2(conjunct(A,_)) :- contains_kodkod_call(A).
contains_kodkod_call2(conjunct(_,B)) :- contains_kodkod_call(B).


try_predicate_analysis_with_global_sets(Predicate,Constants,TaggedConstants,TaggedPredicate) :-
    catch(predicate_analysis_with_global_sets(Predicate,[],Constants,TaggedConstants,TaggedPredicate),
          failed_analysis(Expr),
          ( write('kodkod fail: predicate analysis failed on expression '),
              translate_bexpression_with_limit(Expr,80,Msg),
              write(Msg),nl,fail)),
    !.
try_predicate_analysis_with_global_sets(_Predicate,_Constants,_TaggedConstants,_TaggedPredicate) :-
    write('kodkod fail: predicate analysis failed.\n'),
    fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- meta_predicate try_call(-,-,0).

try_apply_transformation(Predicate,Constants,Index,Result) :-
    Log = log(Predicate,Result,Index,Ints,Bits),
    try_call( 'applying rewriting rules', Log,
              apply_rewriting_rules(Predicate,Rewritten)),
    try_call( 'extracting used types', Log,
              extract_used_types(Rewritten,Types,Ints,Bits)),
    try_call( 'creating Kodkod atoms for types', Log,
              create_atoms_for_types(Types,Ints,Bits,Typemap,Idmap1,_Valuemap)),
    try_call( 'translate and type Kodkod formula', Log,
              typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,_Idmap,Rewritten,_KExpr)),
    finalize_result(Log).
try_call(_Msg,log(_Predicate,Result,_Index,_,_),_Call) :-
    nonvar(Result),!.
try_call(Msg,log(Predicate,Result,Index,_,_),Call) :-
    ( catch(Call,
            kodkod(T),
            Result = failure(Index,Predicate,Msg,kodkod_exception(T))
            ) -> true
    ; otherwise ->
        Result = failure(Index,Predicate,Msg,failed)
    ).
finalize_result(log(Predicate,Result,Index,Ints,Bits)) :-
    var(Result),!,Result=ok(Index,Predicate,Ints,Bits).
finalize_result(_).


typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,Idmap,Rewritten,KExpr) :-
    % Create the relations for the constants that are actuall used
    restrict_to_used_constants(Constants,Predicate,UsedConstants),
    create_identifier_relations(UsedConstants,Typemap,Bits,Idmap1,Idmap),
    % translate the formula in two steps:
    kodkod_translate(Rewritten,Typemap,Idmap,Formula),
    kodkod_insert_casts(Formula,Typemap,Idmap,KExpr),
    !.

restrict_to_used_constants(Constants,Predicate,UsedConstants) :-
    find_identifier_uses(Predicate,[],UsedIds),
    include(constant_is_used(UsedIds),Constants,UsedConstants).
constant_is_used(UsedIds,C) :-
    get_texpr_id(C,Id),memberchk(Id,UsedIds).    

try_apply_transformation_l([],_Constants,_,[]).
try_apply_transformation_l([Predicate|Prest],Constants,I,[Result|Rrest]) :-
    try_apply_transformation(Predicate,Constants,I,Result),
    I2 is I+1,
    try_apply_transformation_l(Prest,Constants,I2,Rrest).

try_apply_transformation_to_components([],_Vars,_,[]).
try_apply_transformation_to_components([component(Predicate,Ids)|Crest],Vars,I,[Result|Rrest]) :-
    try_apply_transformation(Predicate,Vars,I,Result1),
    Result1 =.. [Functor|Args],
    Result  =.. [Functor,Ids|Args],
    I2 is I+1,
    try_apply_transformation_to_components(Crest,Vars,I2,Rrest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

filter_translatable([],[],[]).
filter_translatable([Res|Rrest],ToKodkod,StayProb) :-
    ( Res = ok(_,P,_,_)      -> ToKodkod=[P|Krest], StayProb=Prest
    ; Res = failure(_,P,_,_) -> ToKodkod=Krest, StayProb=[P|Prest]),
    filter_translatable(Rrest,Krest,Prest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- use_module(probsrc(error_manager),[add_warning/3]).

apply_transformation(AllConstants,PredicateList,Result) :-
    conjunct_predicates(PredicateList,Predicate1),
    (debug:debug_mode(on) -> print('TRANSLATING FOR KODKOD: '),nl,translate:print_bexpr(Predicate1),nl ; true),
    apply_transformation2(AllConstants,Predicate1,Result).

apply_transformation2(AllConstants,Predicate1,Result) :-
    split_simple_equalities(Predicate1,Equalities,Predicate),
    (is_truth(Predicate) -> debug:debug_println('Only equalities'),fail
      ; apply_transformation3(AllConstants,Predicate,Equalities,Result)
       -> (debug:debug_mode(on) -> print('RESULT: '),nl,translate:print_bexpr(Result),nl ; true)
      ; add_warning(kodkod,'Kodkod Transformation Failed: ',AllConstants),fail).

apply_transformation3(AllConstants,Predicate,Equalities,Result) :-
    restrict_to_used_constants(AllConstants,Predicate,Constants),
    apply_rewriting_rules(Predicate,Rewritten),
    extract_used_types(Rewritten,Types,Ints,Bits),
    create_atoms_for_types(Types,Ints,Bits,Typemap,Idmap1,Valuemap),
    % check if any types are used, if not Kodkod would exit with an
    % empty universe exception (or something similar)
    \+ empty_avl(Typemap),
    typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,Idmap,Rewritten,KExpr),
    
    register_problem(Constants,KExpr,Idmap,Typemap,Valuemap,Predicate,ProblemNr),
    create_texpr(kodkod(ProblemNr,Constants),pred,[],KPred),
    conjunct_predicates([Equalities,KPred],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_results([]).
print_results([R|Rrest]) :-
    print_result(R),nl,print_results(Rrest).
print_result(ok(_Ids,Index,P,Ints,Bits)) :-
    print_result(ok(Index,P,Ints,Bits)).
print_result(ok(_Index,P,Ints,Bits)) :-
    write('kodkod ok:   '),print_bexpr_with_limit(P,50),write('  '),
    write('ints: '),write(Ints),write(', intatoms: '),write(Bits).
print_result(failure(_Ids,Index,P,Msg,Reason)) :-
    print_result(failure(Index,P,Msg,Reason)).
print_result(failure(_Index,P,Msg,Reason)) :-
    write('kodkod fail: '),print_bexpr_with_limit(P,50),write('  '),
    write(', reason: '),write_reason(Reason),write(' while '),write(Msg).
write_reason(kodkod_exception(KE)) :-
    write_reason2(KE),!.
write_reason(E) :- write(E).

write_reason2(unsupported_type(T)) :-
    pretty_type(T,PT),
    write('unsupported type '),write(PT).
write_reason2(unsupported_expression(TE)) :-
    get_texpr_expr(TE,T),
    functor(T,Functor,Arity),
    translate_bexpression(TE,PE),
    format('unsupported expression ~w/~w: ~w',[Functor,Arity,PE]).
write_reason2(unsupported_int(Msg)) :-
    write(Msg).
write_reason2(side_condition(Msg,PPExpr)) :-
    format('~w: ~w',[Msg,PPExpr]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pdebug :-
    install_b_portray_hook,
    assert(( user:portray(X) :- is_avl(X),!,portray_avl(X) )).
