% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kodkod_rewrite,[ apply_rewriting_rules/2 ]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,kodkod).
:- module_info(description,'This module contains rewriting rules that are applied during tge translation to Kodkod.').

:- use_module(library(lists)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(kodkod_tools).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% applying re-writing rules to make some statements translatable to Kodkod

apply_rewriting_rules(TExpr,NTExpr) :-
    remove_bt(TExpr,Expr,NExpr,NTExpr),
    get_texpr_info(TExpr,Info),
    apply_rewriting_rules_pre(Expr,Info,IExpr),
    syntaxtransformation(IExpr,Subs,_,NSubs,Expr1),
    maplist(apply_rewriting_rules,Subs,NSubs),
    apply_rewriting_rules_post(Expr1,Info,NExpr).
apply_rewriting_rules_pre(Expr,Info,NExpr) :-
    kodkod_rewrite_pre(Expr,Info,Expr1),!,
    apply_rewriting_rules_pre(Expr1,Info,NExpr).
apply_rewriting_rules_pre(Expr,_,Expr).
apply_rewriting_rules_post(Expr,Info,NExpr) :-
    kodkod_rewrite_post(Expr,Info,Expr1),!,
    apply_rewriting_rules_post(Expr1,Info,NExpr).
apply_rewriting_rules_post(Expr,_,Expr).

kodkod_rewrite_pre(not_equal(A,B),_,negation(Equal)) :- !,
    create_kk_tpred(equal(A,B),Equal).
kodkod_rewrite_pre(member(Elem,Powset),_,Result) :-
    get_texpr_expr(Powset,PExpr),
    get_texpr_info(Elem,ElemInfo),
    kodkod_rewrite_membership(PExpr,Elem,ElemInfo,Result),
    !.
kodkod_rewrite_pre(subset(Subset,Powset),_,forall([Elem],TPre,TCond)) :-
    get_texpr_info(Subset,Info),
    get_texpr_type(Subset,set(ElemType)),
    unique_identifier(TId),
    element_analysis(Info,ElemInfo),
    create_texpr(identifier(TId),ElemType,ElemInfo,Elem),
    get_texpr_expr(Powset,PExpr),
    kodkod_rewrite_membership(PExpr,Elem,ElemInfo,Cond),!,
    create_texpr(member(Elem,Subset),pred,[],TPre),
    create_texpr(Cond,pred,[],TCond).
kodkod_rewrite_pre(size(Seq),_,card(Seq)) :- !.
kodkod_rewrite_post(interval(A,B),Info,comprehension_set([TmpId],Predicate)) :- !,
    element_analysis(Info,ElemInfo),
    unique_identifier(TId),
    create_texpr(identifier(TId),integer,ElemInfo,TmpId),
    create_texpr(less_equal(A,TmpId),pred,[],Left),
    create_texpr(less_equal(TmpId,B),pred,[],Right),
    conjunct_predicates([Left,Right],Predicate).
kodkod_rewrite_post(sequence_extension(Elems),_Info,set_extension(Pairs)) :- !,
    convert_elements_to_pairs(Elems,1,Pairs).
kodkod_rewrite_post(general_sum([TId],Pred,Expr),_Info,general_sum([TId],NPred,Expr)) :-
    get_texpr_id(TId,Id), get_texpr_id(TElem,Id),
    extract_finite_integer_range(TId,[],IRange),
    \+ get_texpr_expr(Pred,member(TElem,_TSet)),!,
    write('rewritten\n'),
    create_texpr(comprehension_set([TId],Pred),set(integer),[analysis([elem(interval):IRange])],
                 TCompSet),
    create_texpr(member(TId,TCompSet),pred,[],NPred).
convert_elements_to_pairs([],_,[]).
convert_elements_to_pairs([Elem|Erest],Index,[Pair|Prest]) :-
    convert_element_to_pair(Elem,Index,Pair),
    I2 is Index + 1,
    convert_elements_to_pairs(Erest,I2,Prest).
convert_element_to_pair(Elem,Index,Pair) :-
    Range = irange(Index,Index),
    create_texpr(integer(Index),integer,[analysis([interval:Range])],TIndex),
    get_texpr_type(Elem,Type),
    get_texpr_info(Elem,EInfo),
    findall(right(Scope):Info,
            ( memberchk(analysis(AInfos),EInfo),member(Scope:Info,AInfos)),
            RInfos),
    create_texpr(couple(TIndex,Elem),couple(integer,Type),[analysis([left(interval):Range|RInfos])],Pair).

kodkod_rewrite_membership(pow_subset(Set),Elem,_,subset(Elem,Set)).
kodkod_rewrite_membership(fin_subset(Set),Elem,_,subset(Elem,Set)).
kodkod_rewrite_membership(relations(A,B),Elem,_,subset(Elem,Cartesian)) :-
    create_kk_tpred(cartesian_product(A,B),Cartesian).
kodkod_rewrite_membership(pow1_subset(Set),Elem,ElemInfo,conjunct(Subset,Exists)) :-
    get_texpr_type(Set,set(Type)),
    create_texpr(identifier('___elem'),Type,[kodkod|ElemInfo],Id),
    create_kk_tpred(subset(Elem,Set),Subset),
    create_kk_tpred(exists([Id],Member),Exists),
    create_kk_tpred(member(Id,Set),Member).
kodkod_rewrite_membership(fin1_subset(Set),Elem,Info,Pred) :-
    kodkod_rewrite_membership(pow1_subset(Set),Elem,Info,Pred).
kodkod_rewrite_membership(interval(A,B),Elem,_,conjunct(Lower,Upper)) :-
    create_kk_tpred(less_equal(A,Elem),Lower),
    create_kk_tpred(less_equal(Elem,B),Upper).
kodkod_rewrite_membership(integer_set('NATURAL'),Elem,_,less_equal(Zero,Elem)) :-
    create_integer_with_analysis(0,Zero).
kodkod_rewrite_membership(integer_set('NATURAL1'),Elem,_,less_equal(One,Elem)) :-
    create_integer_with_analysis(1,One).
create_kk_tpred(Pred,TPred) :-
    create_texpr(Pred,pred,[kodkod],TPred).
create_integer_with_analysis(I,Expr) :-
    create_texpr(integer(I),integer,[analysis([interval:irange(I,I)])],Expr).

element_analysis(SetInfo,[analysis(MemberInfo)]) :-
    findall( S:R,
             (   member(analysis(Analysis), SetInfo),
                 member(elem(S):R,Analysis)),
             MemberInfo).
