% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(ltl_safety,[is_safety_property/2,ltl_mc_safety_property/2,aut_transition/3, aut_finite_state/1, aut_initial_state/1]).

/* SICSTUS Libraries */
:- use_module(library(lists)).
%% :- use_module(library(codesio)).
:- use_module(library(terms)).
:- use_module(library(file_systems)).

/* ProB Libraries */
:- use_module(probsrc(tools),[ajoin/2]).
:- use_module(probsrc(error_manager)).
%% :- use_module(probsrc(debug)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(preferences),[preference/2,set_preference/2,get_preference/2]).
:- use_module(probsrc(system_call),[system_call/5]).

/* ProB LTL Model Checker Modules */
:- use_module(probltlsrc(safety_mc),[start_mc_safety_property/1]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for checking whether an LTL formula is a safety property.').

:- dynamic ap_representation/2, tp_representation/2, tps_occurring/0.

:- dynamic aut_transition/3, aut_finite_state/1, aut_initial_state/1.

clear_automaton_facts :-
    retractall(aut_transition(_,_,_)),
    retractall(aut_initial_state(_)),
    retractall(aut_finite_state(_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% Checking whether an LTL property represents a safety property %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The predicates check whether the parsed LTL property represents a safety property
% by checking the syntactical structure of the formula.
% The idea is borrowed from the following theorem from the paper 
% "Safety, Liveness and Fairness in Temporal Logic" of A. Sistla: 
%% THEOREM: Every propositional formula is a safety formula and if f,g are safety formulae,
%% then so are f & g, f or g, X f, f W g, and G f, where 'W' is the weak-until operator.
%
% Note that the checked formula should be a positive formula. An LTL formula is said to
% be a positive if none of the temporal operators appears in the scope of 'not'.

% Testing various properties for being safety or liveness. Some of 
% the unit tests here represent examples from the paper mentioned above.
% Note that the first argument of is_safety_property/2 should be negated, as 
% this is done after negating the original property that should be model checked (see select_ltl_mode/4 in ltl.pl)
:- assert_must_succeed((F = ap(p),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(deadlock),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = weakuntil(not(ap(alloc)),ap(request)),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = and(weakuntil(not(ap(alloc)),ap(request)),
                                globally(implies(ap(alloc),next(weakuntil(not(ap(alloc)),ap(request)))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(implies(not(ap(request)),or(ap(alloc),or(next(ap(alloc)),next(next(ap(alloc))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(or(not(ap(p)),globally(not(ap(p))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),ap(p3)))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),until(ap(p3),ap(p4))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),until(ap(p3),until(ap(p4),until(ap(p5),until(ap(p6),ap(p7)))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(not(globally(deadlock))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(or(ap(p),and(next(ap(q)),not(next(ap(q)))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = and(globally(or(ap(p),next(globally(ap(q))))),globally(or(ap(p),next(globally(not(ap(q))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(implies(ap(request),finally(ap(alloc)))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(finally(ap(alloc))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(until(ap(request),finally(ap(alloc)))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = implies(globally(finally(ap(send))),globally(finally(receive))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = not(until(ap(p1),not(until(ap(p2),until(ap(p3),ap(p4)))))),
                        is_safety_property_test(not(F),_NF))).

is_safety_property_test(F,NF) :-
  get_preference(use_safety_ltl_model_checker,Val),
  set_preference(use_safety_ltl_model_checker,true),
  is_safety_property(F,NF),
  set_preference(use_safety_ltl_model_checker,Val).  

is_safety_property(not(F),NF) :-
   preference(use_safety_ltl_model_checker,true),
   normalize_to_release_pnf(F,NF),
   is_syntactically_safe(NF).

is_syntactically_safe(deadlock) :- !.
is_syntactically_safe(false) :- !.
is_syntactically_safe(true) :- !.
is_syntactically_safe(ap(_)) :- !.
is_syntactically_safe(action(_)) :- !.
is_syntactically_safe(and(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(or(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(next(F)) :- !,
   is_syntactically_safe(F).
is_syntactically_safe(weakuntil(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(release(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(not(F)) :- !,
   is_atomic_proposition(F).

is_atomic_proposition(F) :-
   F =.. [Functor|_Args],
   memberchk(Functor,[true,false,ap,action]).

% Release PNF (the Release PNF does not imply exponential blow-up of the origin LTL formula)
% f = true | false | a | not a | f1 & f2 | f1 or f2 | X f | f1 U f2 | f1 R f2 
normalize_to_release_pnf(deadlock,deadlock) :- !.
normalize_to_release_pnf(false,false) :- !.
normalize_to_release_pnf(true,true) :- !.
normalize_to_release_pnf(ap(P),ap(P)) :- !.
normalize_to_release_pnf(action(P),action(P)) :- !.
normalize_to_release_pnf(and(F,G),and(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(or(F,G),or(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(implies(F,G),NImplies) :- !,
   normalize_to_release_pnf(or(not(F),G),NImplies).
normalize_to_release_pnf(next(F),next(NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(globally(F),release(false,NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(finally(F),until(true,NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(weakuntil(F,G),weakuntil(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(release(F,G),release(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(until(F,G),until(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
% not rules
normalize_to_release_pnf(not(true),false) :- !.
normalize_to_release_pnf(not(false),true) :- !.
normalize_to_release_pnf(not(and(F,G)),NOr) :- !,
   normalize_to_release_pnf(or(not(F),not(G)),NOr).
normalize_to_release_pnf(not(or(F,G)),NAnd) :- !,
   normalize_to_release_pnf(and(not(F),not(G)),NAnd).
normalize_to_release_pnf(not(next(F)),NNext) :- !,
   normalize_to_release_pnf(next(not(F)),NNext).
normalize_to_release_pnf(not(finally(F)),NGlobally) :- !,
   normalize_to_release_pnf(release(false,not(F)),NGlobally).
normalize_to_release_pnf(not(globally(F)),NUntil) :- !,
   normalize_to_release_pnf(until(true,not(F)),NUntil).
normalize_to_release_pnf(not(until(F,G)),NRelease) :- !,
   normalize_to_release_pnf(release(not(F),not(G)),NRelease).
normalize_to_release_pnf(not(F),NF) :- !,
   (F = not(F1) ->
     normalize_to_release_pnf(F1,NF)
   ; normalize_to_release_pnf(F,NF1),
     NF=not(NF1)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% Predicates for using the LTL2BA Tool %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_ltl2ba_command(LTL2BA) :-
  absolute_file_name(prob_home('/lib/ltl2ba'),LTL2BA).

:- use_module(probsrc(error_manager), [add_error_fail/3]).
ltl_mc_safety_property(Ltl,Result) :-
    translate_to_ltl2ba_syntax(Ltl,Ltl1),
    get_ltl2ba_command(LTL2BACmd),
    ajoin(['"',Ltl1,'"'],LTLFormula),
    % ./ltl2ba -f "formula" -b > buchi.pl %%% TODO: Tool not yet available for Windows
    % note the '-b' option should be set to print the Buchi automaton in PL syntax,
    % otherwise the Buchi automaton will be printed out in promela syntax
    %% system_call(LTL2BACmd, ['-f', LTLFormula, '-b', '>', BuchiPrologFile],Text,JExit),
    system_call(LTL2BACmd, ['-f', LTLFormula, '-b'],TextOut,ErrText,JExit),
    !,
    (JExit=exit(0)
       -> create_ba_pl_file(TextOut,TmpFile),
          load_and_check_safety_property(TmpFile,Result)
        ; name(ET,ErrText),
          add_error_fail(translating_ltl2ba,'Std error: ',ET)
    ).
ltl_mc_safety_property(_Ltl,_Result) :- 
    add_error_fail(translating_ltl2ba,'Be sure that ltl2ba is installed (see Help menu for ProB Tcl/Tk)','').

create_ba_pl_file(Text,Filename) :-
  write_string_to_file('prob_buchi_temp.pl',Filename,Text).

write_string_to_file(BaseName,Filename,Machine) :-
    open(temp(BaseName),write,S,[if_exists(generate_unique_name)]),
    stream_property(S, file_name(Filename)),
    print_string(Machine,S),
    close(S).

print_string([],_).
print_string([C|Rest],S) :- put_code(S,C),print_string(Rest,S).

load_and_check_safety_property(File,Result) :-
  safety_consult_without_redefine_warning(File),
  redefine_automaton_facts,
  start_mc_safety_property(Result).

redefine_automaton_facts :-
  aut_transition(Src,P,Dst),
  redefine_transition(Src,P,Dst),
  fail.
redefine_automaton_facts.

redefine_transition(Src,Predicate,Dst) :-
  retract(aut_transition(Src,Predicate,Dst)),
  redefine_transition_label(Predicate,Proposition),
  assert(aut_transition(Src,Proposition,Dst)).

redefine_transition_label(true,Res) :- !,
  Res = ap(bpred(b(truth,[],pred))).
redefine_transition_label(false,Res) :- !,
  Res = ap(bpred(b(falsity,[],pred))).
redefine_transition_label(AP,ap(Res)) :-
  ap_representation(AP,Res),!.
redefine_transition_label(TP,tp(Res)) :-
  tp_representation(TP,Res),!.
redefine_transition_label(F,Res) :-
  F =.. [Functor|Args],
  maplist(redefine_transition_label,Args,NewArgs),
  Res =.. [Functor|NewArgs].

safety_consult_without_redefine_warning(File) :-
    clear_automaton_facts,
    (file_exists(File) -> true ; add_error_fail(safety_mc,'Automaton PL File does not exist: ',File)),
    print(consult_without_redefine(File)),nl,
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off), 
    prolog_flag(discontiguous_warnings, Old3, off), 
    (load_files(File,[compilation_mode(assert_all)]) %consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    prolog_flag(discontiguous_warnings, _, Old3),
    OK=true.

translate_to_ltl2ba_syntax(Formula,Text) :-
    retractall(ap_representation(_,_)),
    retractall(tp_representation(_,_)),
    retractall(tps_occurring),
    pp_to_ltl2ba_syntax(Formula,Atoms,[]),
    ajoin(Atoms,Text).

pp_to_ltl2ba_syntax(true,['true'|S],S) :- !.
pp_to_ltl2ba_syntax(false,['false'|S],S) :- !.
pp_to_ltl2ba_syntax(deadlock,['deadlock'|S],S) :- !.
pp_to_ltl2ba_syntax(ap(X),[AP|S],S) :- !,
   term_hash(X,Hash),
   ajoin(['ap',Hash],AP),
   (\+ap_representation(AP,_) -> assert(ap_representation(AP,X)); true).
pp_to_ltl2ba_syntax(action(X),[TP|S],S) :- !,
   assert(tps_occurring),
   term_hash(X,Hash),
   ajoin(['tp',Hash],TP),
   (\+tp_representation(TP,_) -> assert(tp_representation(TP,X)); true).
pp_to_ltl2ba_syntax(not(X),['! '|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,T).
pp_to_ltl2ba_syntax(and(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' && '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(or(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' || '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(implies(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' -> '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(next(X),['X ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(globally(X),['[] ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(finally(X),['<> ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(until(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' U '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(release(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' V '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(weakuntil(X,Y),S,T) :- !,
    pp_to_ltl2ba_syntax(release(or(not(X),Y),or(X,Y)),S,T).
