% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% A model checker for safety properties.

% At this point the property intended to be checked has been recognized as a safety property. 
% As a consequence the negation of the property is represented by a finite automaton which 
% accepts all bad prefixes violating the safety property. The algorithm implemented in this module
% traverses on-the-fly the state space and simulates at the same time movement through the automaton
% in regard to the evaluation of the atomic propositions. The algorithm halts when either a finite
% state of the automaton is visited (bad prefix for the property has been found and will be reported as counter-example) 
% or when all nodes of the state space has been visited.

:- module(safety_mc,[start_mc_safety_property/1]).

:- use_module(probsrc(bmachine), [b_parse_machine_predicate/2]).
:- use_module(probsrc(eclipse_interface),[test_boolean_expression_in_node/2]).
:- use_module(probsrc(state_space),[set_max_nr_of_new_impl_trans_nodes/1, visited_expression_id/1, transition/4]).
:- use_module(probsrc(error_manager), [add_error_fail/3, add_warning/2]).
:- use_module(probsrc(preferences),[set_preference/2, get_preference/2]).
:- use_module(probltlsrc(ltl),[init_states/1]).
:- use_module(probltlsrc(ltl_safety),[aut_transition/3, aut_finite_state/1, aut_initial_state/1]).
:- use_module(probsrc(specfile),[animation_mode/1]).
:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(probsrc(tools),[cputime/1]).

:- use_module(library(file_systems)).
:- use_module(library(ordsets)).
:- use_module(library(random)).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides LTL model checking of safety properties.').

:- dynamic visited_pair/2.
% automaton_transition(Src,bpred(Property),Dst)

% model_checker: get_open_node_to_check(DFMODE,NodeID)
% model_checker: set_context_state(ID)
% model_checker: visited_expression(ID,CurState)

% created just for testing
tcltk_parse_b_predicate(String,Tree) :-
  get_preference(bool_expression_as_predicate,Extended),
  % parse also boolean expressions as predicates
  set_preference(bool_expression_as_predicate,true),
  %% atom_codes(String,Codes),
  %% parse_predicate(Codes,Tree),
  b_parse_machine_predicate(String,Tree),
  set_preference(bool_expression_as_predicate,Extended).

start_mc_safety_property(Result) :- %trace,
  prepare_mc_safety_property,
  init_states(StartNodes), % the initial states of the model
  initial_states_mc_safety_property(StartNodes,ProductInitStates),
  cputime(T1),
  (run_mc_safety_property(ProductInitStates,CEPath,Result) ->
    cputime(T2),
    print('Counter-example has been found'),nl,
    print(counter_example(Result,CEPath)),nl
  ; otherwise ->
    cputime(T2),
    get_state_space_statistics(TS,Product),
    print('All nodes visited. No counter-example has been found.'), nl,
    print('Number of states checked: '), print(Product), print(', where '), 
    print('number of states of the transition system is: '), print(TS), nl,
    Result = none
    %% ,print('number of states of the automaton is: '), print(Aut),nl
  ),
  print('% Overall Checking Time: '), D is T2-T1, print(D), print(' ms'),nl.

get_state_space_statistics(TS,Product) :-
  findall((X,Y),visited_pair(X,Y),L),
  debug_println(9,product_states(L)),
  length(L,Product),
  findall(ID,visited_expression_id(ID),IDL),
  length(IDL,TS).

prepare_mc_safety_property :-
  retractall(visited_pair(_,_)).

% getting the initial states i.r.t the safety property
initial_states_mc_safety_property(StartNodes,ProductInitStatesSet) :-
  findall(Init,aut_initial_state(Init),InitStates),
  get_product_init_states(StartNodes,InitStates,ProductInitStates),
  list_to_ord_set(ProductInitStates,ProductInitStatesSet).

get_product_init_states([],_InitStates,[]).
get_product_init_states([StartNode|Nodes],InitStates,Res) :-
  (ltl_safety:tps_occurring -> user:tcltk_compute_options(StartNode,_ActionsAndIDs);true),
  get_init_product(StartNode,InitStates,IStates),
  append(IStates,Res1,Res),
  get_product_init_states(Nodes,InitStates,Res1).

get_init_product(_StartNode,[],[]).
get_init_product(StartNode,[Init|Inits],Res) :-
  findall(Succ,get_successor_state(StartNode,Init,Succ),L),
  append(L,Res1,Res),
  get_init_product(StartNode,Inits,Res1).

get_successor_state(StartNode,Init,(StartNode,Next,ActionID)) :-
  check_for_aut_transition(Init,StartNode,Next,ActionID).

% Checking whether the automaton expresses the trivial property 'true'.
% If so, a counter-example [inital_state] is returned
nontrivial_automaton :-
  aut_transition(P,L,Q),
  (P\='accept_init';L \= ap(bpred(b(truth,[],pred)));Q \='accept_init').

non_empty_automaton :-
  aut_transition(_P,_L,_Q).

run_mc_safety_property(InitStates,CEPath,Result) :-
  ((InitStates == [],non_empty_automaton) ->
    % Note, we assume here that the property is satisfied. Important to add 'fail'...
    add_warning(run_mc_safety_property,'No initial states of the product transition system have been generated.'),fail
  ; \+ aut_finite_state('accept_all') ->
    add_error_fail(run_mc_safety_property,'No accept_all state has been found. The property is possibly not a safety property: ', InitStates)    
  ; otherwise ->
    run_mc_safety_property_l(InitStates,CEPath,CEPath,Result)
  ).

run_mc_safety_property_l(InitStates,CEPath,CEPath,Result) :-
  member((CurID,AutState,ActionID),InitStates),
  (\+nontrivial_automaton ->
    CEPath = [atom(CurID,CurID,none)],
    Result=model([atom(CurID,CurID,none)],no_loop)
  ; otherwise ->
    mc_safety_property_init((CurID,AutState,ActionID),CEPath,CEPath,Result)
  ).
  %% add_error_fail(run_mc_safety_property,'Implementation is missing for more than one init state: ', InitStates).

% at the init state we already checked for next automaton transitions
mc_safety_property_init((CurID,AutState,ActionID),History,HTail,Result) :-
  (transition(CurID,_Op,ActionID,CurID) -> true ; assert(visited_pair(CurID,AutState))),
  mc_safety_property_next_state((CurID,AutState,ActionID),History,HTail,Result).

% model_checker: get_open_node_to_check(DFMODE,NodeID)
mc_safety_property((CurID,AutState,ActionID),History,HTail,Result) :-
  assert(visited_pair(CurID,AutState)),
  debug_println(9,current_history(CurID,History)),
  check_for_aut_transition(AutState,CurID,NextAutState,ActionID),
  mc_safety_property_next_state((CurID,NextAutState,ActionID),History,HTail,Result).

mc_safety_property_next_state((CurID,AutState,_ActionID),History,[atom(CurID,CurID,none)],Result) :-
  aut_finite_state(AutState),
  AutState = 'accept_all',!,
  print(found_error(counter_example_found,CurID,History)),nl,
  Result = model(History,no_loop).
mc_safety_property_next_state((CurID,AutState,ActionID),History,HTail,Result) :-
  user:tcltk_compute_options(CurID,ActionsAndIDs),
  random_permutation(ActionsAndIDs,ActionsAndIDs1),
  %% visited_expression(CurID,CurState),
  %% user: compute_all_transitions(ID,CurState),
  %% model_checker: get_open_node_to_check(depth_first,NodeID),
  HTail = [atom(CurID,CurID,ActionID)|HTail2],
  get_next_state(ActionsAndIDs1,ActionID,DestID),
  \+visited_pair(DestID,AutState),
  (ltl_safety:tps_occurring -> user:tcltk_compute_options(DestID,_DstActionsAndIDs);true),
  mc_safety_property((DestID,AutState,_NewActionID),History,HTail2,Result).

get_next_state(ActionsAndIDs,ActionID,DstID) :-
  member((ActionID,_Term,DstID),ActionsAndIDs).

check_for_aut_transition(AutState,CurID,NextAutState,ActionID) :-
  findall((Labl,Next),aut_transition(AutState,Labl,Next),L),
  random_permutation(L,L1),
  member((Label,NextAutState),L1),
  check_transition(Label,AutState,NextAutState,CurID,ActionID).

check_transition(ap(AP),_AutState,_NextAutState,CurID,_ActionID) :- !,
  (animation_mode(cspm) ->
    (AP = bpred(b(truth,[],pred)) ->
      true
    ; AP = bpred(_) ->
      add_error_fail(check_for_aut_transition,'Atomic proposition not allowed for model checking CSP models: ', ap(AP))
    ; otherwise ->
      ltl:check_ap(AP,CurID)
    )
  ; otherwise ->
    ltl:check_ap(AP,CurID)).
check_transition(tp(TP),_AutState,_NextAutState,CurID,ActionID) :- !,
  transition(CurID,Op,ActionID,DestId),
  ltl:check_transition(TP,Op,CurID,DestId).
check_transition(not(AF),AutState,NextAutState,CurID,ActionID) :- !,
  \+check_transition(AF,AutState,NextAutState,CurID,ActionID).
check_transition(and(AF,AG),AutState,NextAutState,CurID,ActionID) :- !,
  check_transition(AF,AutState,NextAutState,CurID,ActionID),
  check_transition(AG,AutState,NextAutState,CurID,ActionID).
check_transition(or(AF,AG),AutState,NextAutState,CurID,ActionID) :- !,
  (check_transition(AF,AutState,NextAutState,ActionID,CurID) ;
   check_transition(AG,AutState,NextAutState,ActionID,CurID)).
check_transition(Label,AutState,NextAutState,_CurID,_ActionID) :-
  add_error_fail(check_for_aut_transition,'Unknown transition label: ', aut_transition(AutState,Label,NextAutState)).
