% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(pge_algo, [
          %%%% Static Analysis Predicates %%%%
          compute_action_relations_graph/2, disable_graph/1, tcltk_disabling_analysis/1,
          %%%% Model Checker predicates %%%%
          clear_dynamic_predicates_for_pge/0, compute_enabled_transitions_only/5,
          %%%% INFO %%%%
          disabled_operations_stats/1
           ]).

/* SICSTUS Libraries */
:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).
%:- use_module(library(wgraphs)).
:- use_module(library(timeout)).

/* B module sources */
:- use_module(probsrc(bsyntaxtree), [get_texpr_expr/2, conjunct_predicates/2]).
:- use_module(probsrc(translate), [translate_bexpression/2, translate_substitution/2]).

:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,b_operation_preserves_full_invariant/1,b_top_level_operation/1, b_machine_has_constants_or_properties/0]).
:- use_module(probsrc(b_state_model_check), [get_guard/2,get_negated_guard/3]).
:- use_module(probporsrc(static_analysis), [get_conj_inv_predicate/3,dependent_actions/5,catch_and_ignore_well_definedness_error/2]).
:- use_module(probsrc(enabling_analysis), [init_or_op/1, print_enable_table/1,b_get_read_write/5,b_get_read_write/3]).

/* Importing predicates for handling the states of a respective B/Event-B machine */
:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- use_module(probsrc(specfile),[specfile_possible_trans_name/2,prepare_state_for_specfile_trans/2]).
:- use_module(probsrc(state_space),[visited_expression/2,assert_time_out_for_node/3,transition/3,use_no_timeout/1,assert_max_reached_for_node/1]).
:- use_module(probsrc(value_persistance), [save_constants/1,add_new_transitions_to_cache/1]).
:- use_module(probsrc(succeed_max),[reset_max_reached/0,max_reached/0]).
:- use_module(probsrc(store), [empty_state/1]).
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(error_manager),[enter_new_error_scope/1, exit_error_scope/2,specific_event_occurred_at_level/2,clear_all_errors_in_error_scope/1,
                                      is_time_out_result/1,set_error_context/1,clear_error_context/0,time_out_with_enum_warning_for_findall/3]).
:- use_module(probsrc(tools),[cputime/1,print_message/1]).
:- use_module(probsrc(debug)).
:- use_module(probltlsrc(ltl),[init_states/1]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for model checking B and Event-B models using the partial guard evaluation optimisastion.').


%%%%%% Static analysis %%%%%%
% TODO: Move PGE predicates to a new module!!!
:- dynamic disable_graph/1.

compute_action_relations_graph(FindInvViolations,DisableGraph) :-
   (disable_graph(DisableGraph) -> true
   ; otherwise -> enter_new_error_scope(ScopeID),
                  call_cleanup(compute_disable_graph(FindInvViolations,DisableGraph),
                               my_exit_error_scope(ScopeID))

   ).
my_exit_error_scope(ScopeID) :-
  (specific_event_occurred_at_level(ScopeID,Event) ->
   format('### ERROR/ENUMERATION WARING OCCURED : ~w !~n',[Event])),fail.
my_exit_error_scope(ScopeID) :-  %error_manager:print_error_scopes,
  clear_all_errors_in_error_scope(ScopeID),
  exit_error_scope(ScopeID,_).

/* DEAD CODE */
compute_disable_graph(FindInvViolations,DisableGraph) :-
        findall(Action, init_or_op(Action),AllActions),
        print('********** DETERMINE ENABLE/DISABLE GRAPH *************'),nl,
        get_preference(timeout_cbc_analysis,Timeout),
        cputime(T1),
        determine_action_relations(AllActions,FindInvViolations,Timeout,DisableGraph,determine_neighbors),
        debug_println(9,disgraph(DisableGraph)),
        cputime(T2),
        print('ENABLE/DISABLE ANALYSIS Checking Time: '), D is T2-T1, print(D),print(' ms.'),nl,
        !,
        assert(disable_graph(DisableGraph)).

determine_action_relations(AllActions,FindInvViolations,Timeout,DisGraph,Predicate) :-
   vertices_edges_to_wgraph([],[],Graph),
   determine_action_relations_graph(AllActions,AllActions,FindInvViolations,Timeout,Graph,DisGraph,Predicate).

determine_action_relations_graph([],_AllActions,_FindInvViolations,_Timeout,Graph,Graph,_Predicate).
determine_action_relations_graph([Action|Actions],AllActions,FindInvViolations,Timeout,Graph,DisGraph,Predicate) :-
   call(Predicate,Action,AllActions,FindInvViolations,Timeout,Graph,ResGraph),
   determine_action_relations_graph(Actions,AllActions,FindInvViolations,Timeout,ResGraph,DisGraph,Predicate).

determine_neighbors(_Act,[],_FindInvViolations,_Timeout,Graph,ResGraph) :- !, ResGraph=Graph.
determine_neighbors(Act1,[Act2|Actions],FindInvViolations,Timeout,Graph,ResGraph) :-
   ((get_preference(use_pge,true),get_preference(use_pge_dis,true),disables(Act1,Act2,FindInvViolations,Timeout,1)) ->
          debug_println(9,possibly_disables(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
          add_edges(Graph1,[Act2-(Act1-3)],Graph2)
   ; (get_preference(por,sleep_sets),dependent_actions(Act1,Act2,FindInvViolations,Timeout,Result),(Result==independent;Result==syntactic_independent)) ->
          debug_println(9,independent_actions(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
          add_edges(Graph1,[Act2-(Act1-2)],Graph2)
   ; (get_preference(use_pge,true),get_preference(use_pge_keep,true),keeps_enabled_syntactic(Act1,Act2)) ->
          debug_println(9,keeps_enabled_syntactic(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
          add_edges(Graph1,[Act2-(Act1-1)],Graph2)
   ; otherwise ->
          Graph2=Graph
   ),
   determine_neighbors(Act1,Actions,FindInvViolations,Timeout,Graph2,ResGraph).

%------------------------------------
disables('INITIALISATION', OpName2,_FindInvViolations,Timeout,Assert) :-
    b_top_level_operation(OpName2),
    get_guard(OpName2,PosGuard2),
    (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(init,Timeout,[],PosGuard2,_,_,_,_,_R),_Result) ->
      fail
    ; otherwise ->
        % OpName1 always disables OpName2
        (Assert ==1 -> assert(disables('INITIALISATION',OpName2)); true),
        debug_println(9,disables('INITIALISATION', OpName2))
    ).
disables(OpName1, OpName2, FindInvViolations, Timeout,Assert) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    get_guard(OpName1,PosGuard1),
    get_guard(OpName2,PosGuard2),
    %get_conj_inv_predicate([PosGuard1],FindInvViolations,StartPred),
    (FindInvViolations=1 -> StartPred = pred(PosGuard1); StartPred=typing(PosGuard1)),
    (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(StartPred,Timeout,[OpName1],PosGuard2,_,_,_,_,_R),_Result)
        -> fail
    ; otherwise % OpName2 is impossible to be executed after OpName1
        -> (Assert == 1 -> assert(disables(OpName1,OpName2)); true),
           debug_println(9,disables(OpName1, OpName2))
    ).

guaranteed('INITIALISATION', OpName2,_FindInvViolations,Timeout,Assert) :-
    b_top_level_operation(OpName2),
    get_negated_guard(OpName2,_PosGuard2,NegGuard2),
    (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(init,Timeout,[],NegGuard2,_,_,_,_,_R),_Result) ->
      fail
    ; otherwise ->
        % OpName1 always disables OpName2
        (Assert ==1 -> assert(guaranteed('INITIALISATION',OpName2)); true),
        debug_println(9,guaranteed('INITIALISATION', OpName2))
    ).
guaranteed(OpName1, OpName2, FindInvViolations, Timeout,Assert) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    get_guard(OpName1,PosGuard1),
    get_negated_guard(OpName2,_PosGuard2,NegGuard2),
    %get_conj_inv_predicate([PosGuard1],FindInvViolations,StartPred),
    (FindInvViolations=1 -> StartPred = pred(PosGuard1); StartPred=typing(PosGuard1)),
    (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(StartPred,Timeout,[OpName1],NegGuard2,_,_,_,_,_R),_Result)
        -> fail
    ; otherwise % OpName2 is impossible to be executed after OpName1
        -> (Assert == 1 -> assert(guaranteed(OpName1,OpName2)); true),
           debug_println(9,guaranteed(OpName1, OpName2))
    ).

% be careful here: keeps_enabled_syntactic/2 is not the same as keeps_enabled/2!
% keeps_enabled_syntactic/2 is a legal keep relationship with regard to the current constellation
% of both operations in the state space.
% If s -OpName2-> s' & s -OpName1-> t & curState = t & wr(OpName1) /\ rd(OpName2) = {} then t -OpName2-> t'
keeps_enabled_syntactic(OpName1, OpName2) :-
  b_get_read_write(OpName1,_Reads1,Writes1),
  b_get_read_write(OpName2,GReads2,_AReads2,_Reads2,_Writes2),
  \+ord_intersect(Writes1,GReads2),
  assert(keeps(OpName1,OpName2)).

compute_operation_relations_for_pge(FindInvViolations) :-
  (pge_predicates_computed
    -> true
  ;otherwise ->
    print('********** START PGE ANALYSIS *************'),nl,
    get_preference(timeout_cbc_analysis,Timeout),
    cputime(T1),
    enter_new_error_scope(ScopeID),
    call_cleanup(compute_keep_disabled(FindInvViolations,Timeout),
                 my_exit_error_scope(ScopeID)),
    assert_disabled_ops_to_init_states_if_necessary,
    cputime(T2),
    print('********** PGE ANALYSIS FINISHED **********'), nl,
    print('Analysis Checking Time: '), D is T2-T1, print(D),print(' ms.'),nl,
    assert(pge_predicates_computed),
    assert(nr_skipped_guard_evaluations(0)),
    !
  ).

% failure loop
compute_keep_disabled(FindInvViolations,Timeout) :-
  init_or_op(OpName1),
  b_top_level_operation(OpName2),
  b_get_read_write(OpName1,_Reads1,Writes1),
  b_get_read_write(OpName2,GReads2,_AReads2,_Reads2,_Writes2),
  ((OpName1\='INITIALISATION',\+ord_intersect(Writes1,GReads2)) ->
    assert(keeps(OpName1,OpName2)),
    debug_println(9,keeps(OpName1, OpName2))
  ; otherwise ->
    \+disables(OpName1,OpName2,FindInvViolations,Timeout,1),
    guaranteed(OpName1,OpName2,FindInvViolations,Timeout,1)
  ),fail.
compute_keep_disabled(_FindInvViolations,_Timeout).

/*
check_predicate_timeout(Pred,Timeout,State,Result) :-
    time_out(solve_predicate(Pred,S,R),
             Timeout,
             T),
    ( T==success -> Result=R,State=S
    ; otherwise  -> Result=timeout,State=empty).
*/

tcltk_disabling_analysis(list([list(['Origin'|Ops])|Result])) :-
  get_preference(timeout_cbc_analysis,Timeout),
  findall(Op, b_top_level_operation(Op), Ops),
  findall(list([OpName1|DisableList]),
     (init_or_op(OpName1),
      findall(Disables,(b_top_level_operation(OpName2),(disables(OpName1,OpName2,1,Timeout,0) -> Disables=yes; Disables=no)),DisableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

%%%%%% Old PGE Algorithm %%%%%
:- register_event_listener(clear_specification,clear_dynamic_predicates_for_pge,
                            'Partial Guard Evaluation.').

clear_dynamic_predicates_for_pge :-
   retractall(disable_graph(_)), 
   retractall(back_transition(_,_,_)),
   retractall(do_not_evaluate_guard),
   retractall(keeps(_,_)),
   retractall(disables(_,_)),
   retractall(disabled_operations(_,_,_)),
   retractall(guaranteed(_,_)),
   retractall(pge_predicates_computed),
   retractall(nr_skipped_guard_evaluations(_)).

%----------------- Auxiliary predicates for the Partial Guard Evaluation implementation

%:- use_module(library(wgraphs)).
% predicate checks if there is a previous executed action that can disable 
% the current action under considiration
action_disabled(Action,CurID,Neighbors) :-
   get_preference(use_pge_dis,true),
   member((PrevAction-3),Neighbors),
   back_transition(CurID,PrevAction,_PrevId),
   debug_println(9,avoid_computing_guard_of_action(CurID,Action)).

% a sort of 'sleep sets' implementation 
action_leads_to_already_computed_state(Action,CurID,Neighbors,FindDeadlocks) :-
   member((PrevAction-2),Neighbors),
   back_transition(CurID,PrevAction,PrevId),
   transition(PrevId,Action,SuccId),
   transition(SuccId,PrevAction,OtherId),
   % we store here the transition in order to 
   % not cause a non-expected deadlock when skipping the transition evaluation
   (FindDeadlocks==1 ->
        store_transition(CurID,Action,OtherId,_TransId)
   ; otherwise ->
      debug_println(9,action_not_stored(CurID,Action))
   ),
   debug_println(9,operation_skipped_at_state(CurID,Action)).

omit_evaluating_guard(Action,CurID,Neighbors) :-
   get_preference(use_pge_keep,true),
   % get previous actions that cannot disable action Action
   member((PrevAction-1),Neighbors),
   % if PrevAction precedes Action then test if Action was last enabled in the previous state
   back_transition(CurID,PrevAction,PrevId),
   transition(PrevId,Action,OtherId),
   CurID\=OtherId,
   debug_println(9,avoid_computing_guard_of_action_ind(CurID,Action)).

/******************* Adding transitions to state space using the Partial Guard Evaluation optimisation (BEGIN) *******************/

:- use_module(probsrc(state_space),[clear_context_state/0]).

compute_enabled_transitions_only(CurID,CurState,DisableGraph,FindDeadlocks,SleepSetsMode) :-
   user:catch_clpfd_overflow_call_for_state(CurID,pge_algo:compute_enabled_transitions_only2(CurID,CurState,DisableGraph,FindDeadlocks,SleepSetsMode),clear_context_state).

compute_enabled_transitions_only2(CurID,CurState,DisableGraph,FindDeadlocks,SleepSetsMode) :-
    reset_max_reached,
    prepare_state_for_specfile_trans(CurState,PreparedCurState),
    (use_no_timeout(CurID) ->
        add_transitions_fail_loop_pge(CurID,PreparedCurState,DisableGraph,FindDeadlocks,SleepSetsMode,0)
    ;   preferences: preference(time_out,CurTO),
        add_transitions_fail_loop_pge(CurID,PreparedCurState,DisableGraph,FindDeadlocks,SleepSetsMode,CurTO)
    ),
    (max_reached -> assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
    (CurID==root -> save_constants(MaxReached) ; add_new_transitions_to_cache(CurID)).

:- dynamic back_transition/3,do_not_evaluate_guard/0.

add_transitions_fail_loop_pge(CurID,CurState,DisableGraph,FindDeadlocks,SleepSetsMode,CurTO) :-
   specfile_possible_trans_name(CurState,ActionName),
   debug_println(9,add_transition(CurID,ActionName)),
   get_action_neighbors(ActionName, DisableGraph, Neighbors),
   ( action_disabled(ActionName,CurID,Neighbors) ->
       % if ActionName is disabled then add_transitions_fail_loop_pge/5 fails.
       % this means that the computing of the ActionName's guard will be skipped
       fail
   ; otherwise ->
       true % in this case the 'disable' optimisation is switched off
   ),
   %debug_println(9,cannot_be_disabled(CurID,ActionName)),
   (SleepSetsMode==on ->
      \+ action_leads_to_already_computed_state(ActionName,CurID,Neighbors,FindDeadlocks)
   ; otherwise ->
      true
   ),
   %debug_println(9,action_does_not_lead_to_computed_state(CurID,ActionName)),
   (omit_evaluating_guard(ActionName,CurID,Neighbors) ->
      % in case we know that the action is enabled in the current state,
      % we don't have to compute the guard of the action.
        assert(do_not_evaluate_guard),
        %debug_println(9,possibility_of_skipping_guard_evaluation(CurID,ActionName)),
        % do_not_evaluate_guard/0 will be asserted in order to omit the evaluation of the guard (valid only for actions without arguments)
        % see implementation of b_execute_top_level_statement2(precondition(PreCond,Body), LocalState,InState,OutState,WF,Path,OR)
        % in b_interpreter.pl.
        add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId),
        retract(do_not_evaluate_guard)
   ; otherwise ->
        %debug_println(9,no_optimisation_possible(CurID,ActionName)),
        add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId)
   ),
   % keeping track of the incoming transitions for CurID
   % back_transition/3 needed because in Prolog
   % predicates are indexed on their firsts arguments
   (\+back_transition(NewId,ActionName,CurID) ->
       assert(back_transition(NewId,ActionName,CurID))
   ; true),
   fail.
add_transitions_fail_loop_pge(_,_,_,_,_,_) :-
   clear_error_context.

get_action_neighbors(ActionName, DisableGraph, Neighbors) :-
    (neighbors(ActionName, DisableGraph, Neighbors) -> true ; Neighbors=[]).

add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId) :-
    set_error_context(operation(ActionName,CurID)),
    (CurTO > 0
        -> time_out_with_enum_warning_for_findall(user:add_transition(CurID,CurState,ActionName,NewId),CurTO,TimeOutRes),
           check_timeout_result(CurID,ActionName,CurTO,TimeOutRes)
        ;  user:add_transition(CurID,CurState,ActionName,NewId)
    ).

check_timeout_result(CurID,ActionName,CurTO,TimeOutRes) :-
    (is_time_out_result(TimeOutRes)
        -> 
        ( TimeOutRes = virtual_time_out(_) ->
            true
        ;   print_message('TIME OUT: '), print_message(CurID), print_message(CurTO),
          print_message(ActionName)
        ),
    (var(ActionName)
         -> assert_time_out_for_node(CurID,'unspecified_operation',TimeOutRes)
         ;  assert_time_out_for_node(CurID,ActionName,TimeOutRes))
         ;  true
  ).

:- dynamic nr_skipped_guard_evaluations/1.
disabled_operations_stats(ResPGE) :-
  %% retractall(nr_skipped_guard_evaluations(_)),
  %% assert(nr_skipped_guard_evaluations(0)),
  %% disabled_operations_stats_failure_loop,
  nr_skipped_guard_evaluations(NrSkippedEvaluations),
  ResPGE = ['SKIPPED_GUARD_EVALUATIONS: ',NrSkippedEvaluations].

disabled_operations_stats_failure_loop :-
  disabled_operations(_NodeID,DisOps),
  get_number_of_disabled_operations_in_node(DisOps),
  fail.
disabled_operations_stats_failure_loop.

get_number_of_disabled_operations_in_node(DisOps) :-
  length(DisOps,NrDisOps),
  nr_skipped_guard_evaluations(Nr),
  Nr1 is Nr + NrDisOps,
  retract(nr_skipped_guard_evaluations(_)),
  assert(nr_skipped_guard_evaluations(Nr1)).

/******************* Adding transitions to state space using the Partial Guard Evaluation optimisation (END) *******************/

%%%%%% New PGE Algorithm %%%%%
%% a predicate for keeping track of the disabled operations in a state
:- dynamic disabled_operations/3, pge_predicates_computed/0.
%% predicates for saving the relations between operations in B and Event-B machines
:- dynamic disables/2, keeps/2, guaranteed/2.

compute_transitions_pge(CurID,CurState) :- %trace,
   user:catch_clpfd_overflow_call_for_state(CurID,pge_algo:compute_transitions_pge2(CurID,CurState),clear_context_state).

compute_transitions_pge2(CurID,CurState) :-
  reset_max_reached,
  prepare_state_for_specfile_trans(CurState,PreparedCurState),
  (use_no_timeout(CurID) ->
    add_transitions_fail_loop_pge1(CurID,PreparedCurState,0)
  ;otherwise ->
    preferences: preference(time_out,CurTO),
    add_transitions_fail_loop_pge1(CurID,PreparedCurState,CurTO)
  ),
  (max_reached -> assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
  (CurID==root -> save_constants(MaxReached) ; add_new_transitions_to_cache(CurID)).

add_transitions_fail_loop_pge1(CurID,CurState,CurTO) :-
  % TODO: it should be guaranteed that for each visited state there is a fact disabled_opertions(CurID,DisOps)
  (disabled_operations(CurID,DisabledOperations,EnabledOperations) -> true; DisabledOperations = [],EnabledOperations = []),
  % just getting a possible transition name which may be executed from current state
  specfile_possible_trans_name(CurState,ActionName),
  (ord_member(ActionName,DisabledOperations) ->
    % guard of ActionName doesn't have to be tested,
    % we already know that ActionName is disabled in CurID
    retract(nr_skipped_guard_evaluations(NrDisabled)),
    NrDisabled1 is NrDisabled +1,
    assert(nr_skipped_guard_evaluations(NrDisabled1)),
    debug_println(9,skipped_guard_evaluation(CurID,ActionName)),
    fail
  ; otherwise ->
    (ord_member(ActionName,EnabledOperations) -> 
      debug_println(9,guaranteed_enabled(ActionName)),
      assert(do_not_evaluate_guard)
    ; true),
    % we use findall here in order to not call compute_disabled_operations/4 for each representation of ActionName
    findall(NewId,add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId),NewIds),
    compute_and_update_disabled_enabled_operations(NewIds,ActionName,DisabledOperations,EnabledOperations,CurID),
    retractall(do_not_evaluate_guard)
  ), fail.
add_transitions_fail_loop_pge1(CurID,CurState,_) :-
  %print(retract(disabled_operations(CurID,CurState))),nl,
  % forget the information once a state has been fully explored
  (setup_constants_state(CurState) -> true ; retract(disabled_operations(CurID,_,_))),
  debug_println(9,state_has_been_explored(CurID,CurState)),
  clear_error_context.

compute_and_update_disabled_enabled_operations(NewIds,ActionName,DisabledOperations,EnabledOperations,CurID) :-
  (NewIds == [] ->
    % ActionName is disabled in state CurID, add ActionName to DisabledOperations set of CurID
    ord_add_element(DisabledOperations,ActionName,NewDisabledOperations),
    retractall(disabled_operations(CurID,_,_)),
    assert(disabled_operations(CurID,NewDisabledOperations,EnabledOperations))
  ; otherwise ->
    compute_disabled_enabled_operations(ActionName,DisabledOperations,EnabledOperations,NewDisabledOperations,NewEnabledOperations),
    update_disabled_operations_l(NewIds,NewDisabledOperations,NewEnabledOperations)
  ).

setup_constants_state(concrete_constants(_)).
setup_constants_state(const_and_vars(_,_)).

assert_disabled_ops_to_init_states_if_necessary :-
  % in case of no constants or properties
  % the init states will not be discovered while 
  % exploring the state space using the pge optimisation
  \+b_machine_has_constants_or_properties,
  findall(Op,disables('INITIALISATION',Op),DisOps),
  findall(Op,guaranteed('INITIALISATION',Op),EnOps),
  ltl: init_states(Init),
  member(I,Init),
  assert(disabled_operations(I,DisOps,EnOps)),
  fail.
assert_disabled_ops_to_init_states_if_necessary.

compute_disabled_enabled_operations(ActionName,DisabledOperations,EnabledOperations,NewDisabledOperations,NewEnabledOperations) :-
  (functor(ActionName,'$initialise_machine',_) ->
    % an initial state discovered
    findall(Op,disables('INITIALISATION',Op),NewDisabledOperations),
    findall(Op,guaranteed('INITIALISATION',Op),NewEnabledOperations)
  ; otherwise ->
    get_keeped_operations(DisabledOperations,ActionName,KeepDisabled),
    get_disabled_by_execution_operations(ActionName,ExecutionDisabled),
    append(KeepDisabled,ExecutionDisabled,NewDisabledOperations1),
    list_to_ord_set(NewDisabledOperations1,NewDisabledOperations),
    get_keeped_operations(EnabledOperations,ActionName,KeepEnabled),
    get_enabled_by_execution_operations(ActionName,ExecutionEnabled),
    append(KeepEnabled,ExecutionEnabled,NewEnabledOperations1),
    list_to_ord_set(NewEnabledOperations1,NewEnabledOperations)).

get_keeped_operations([],_ActionName,[]).
get_keeped_operations([Op|Ops],ActionName,KeepedOps) :-
  (keeps(ActionName,Op) ->
    KeepedOps = [Op|KeepedOpsRest]
  ; otherwise ->
    KeepedOps = KeepedOpsRest
  ),get_keeped_operations(Ops,ActionName,KeepedOpsRest).

get_disabled_by_execution_operations(ActionName,DisabledOps) :-
  findall(DisOp,disables(ActionName,DisOp),DisabledOps).

get_enabled_by_execution_operations(ActionName,EnabledOps) :-
  findall(EnabledOp,guaranteed(ActionName,EnabledOp),EnabledOps).

update_disabled_operations_l(StateIDs,NewDisabledOperations,NewEnabledOperations) :-
  maplist(update_disabled_operations(NewDisabledOperations,NewEnabledOperations),StateIDs).

update_disabled_operations(NewDisabledOperations,NewEnabledOperations,StateID) :-
  %(visited_expression(StateID,_State) ->
  (disabled_operations(StateID,DisabledOperations,EnabledOperations) ->
    ord_union(DisabledOperations,NewDisabledOperations,DisabledOperations1),
    ord_union(EnabledOperations,NewEnabledOperations,EnabledOperations1),
    retractall(disabled_operations(StateID,_,_)),
    assert(disabled_operations(StateID,DisabledOperations1,EnabledOperations1))
  ; otherwise ->
    assert(disabled_operations(StateID,NewDisabledOperations,NewEnabledOperations))
  ).
