% (c) 2012-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(ample_sets, [
      compute_ample_set2/4,
      clear_dynamic_predicates_for_POR/0,
      %reset_runtime_dynamic_predicates_POR/0, % now done via event-handling
      stutter_action/1, visible_action/1 % exported for ltl.pl for asserting and retracting stutter/visible actions
      ]).

/* SICSTUS Libraries */
:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).
:- use_module(library(ugraphs)).
%:- use_module(library(timeout)).

/* B module sources */
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,b_operation_preserves_full_invariant/1]).
:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2,occurs_in_expr/2]).
:- use_module(probsrc(specfile),[prepare_state_for_specfile_trans/2,specfile_possible_trans_name/2,specfile_trans/6]).
:- use_module(probsrc(state_space),[visited_expression/2,use_no_timeout/1]).

/* POR modules */
:- use_module(probporsrc(enable_graph)).
  % --> /* The Static Analysis' module */
:- use_module(probporsrc(static_analysis),[enable_graph/1,dependent_act/3]).

/* Importing predicates for handling the states of a respective B/Event-B machine */
:- use_module(probsrc(specfile),[state_corresponds_to_initialised_b_machine/2]).
:- use_module(probsrc(store), [empty_state/1]).

/* Other modules from ProB */
:- use_module(probsrc(debug),[debug_println/2]).
%:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(error_manager), [time_out_with_enum_warning_for_findall/3]).

% Cosider to remove this import
:- use_module(probsrc(enabling_analysis),[b_get_read_write/3]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for partial order reduction using the ample sets\' theory.').

% Dynamic predicates for keeping track of visible and stutter actions, 
% as well as of all dependent reations between the actions of the B model
:- dynamic stutter_action/1,visible_action/1,enables/2,expanded_state/1.

:- public clear_dynamic_predicates_for_POR/0.
:- public reset_static_dynamic_predicates_POR/0.
clear_dynamic_predicates_for_POR :-
    reset_static_dynamic_predicates_POR,
    reset_runtime_dynamic_predicates_POR.

:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- register_event_listener(clear_specification,clear_dynamic_predicates_for_POR,
                           'Reset Ample Sets.').
:- register_event_listener(reset_specification,reset_static_dynamic_predicates_POR,
                           'Reset Ample Sets.').

% predicates that will be asserted once on every loading of the machine
reset_static_dynamic_predicates_POR :- 
   retractall(dependent_act(_,_,_)),
   retractall(enable_graph(_)).

% predicates that will be asserted at runtime
reset_runtime_dynamic_predicates_POR :-
    retractall(stutter_action(_)),
    retractall(visible_action(_)),
    retractall(expanded_state(_)),
    retractall(enables(_,_)). % used to not track in the enable graph

/* This implementation of the ample sets approach sastisfies all four conditions of the ample sets theory (not proved yet).
   The conditions are taken from the book "Principles of Model Checking" written by C.Baier and J.P.Katoen.

For every state s in the state space, where Act(s) stands for all enabled actions in state s, the following conditions 
must be satisfied:

1. Nonemptiness condition (A1) and terminal states
   (A 1.1) ample(s) <: Act(s) (set equality is possible)
   (A 1.2) ample(s) = {} <=> Act(s) = {}

2. Dependency condition
   (A 2) Let s -a_{1}> s_{1} -a_{2}> ... -a_{n}> s_{n} -a> s' be a finite execution
         fragment in the original state space. If a depends on ample(s), then a_{i}:ample(s) for some 0<i<=n.
   
   This condition can be satisfied by the following conditions (Local Criteria):
      (A 2.1) Any action b: Act(s)\ample(s) is independent of any a:ample(s) (Here we use the dependency graph generated from the Enabling Analysis)
      (A 2.2) Any disabled action b:Act\en(s) which is dependent on ample(s) cannot become enabled by a possible sequence of actions starting with an action
              from en(s)\ample(s). (We check this condition by means of using the enable graph generated from the Enabling Analysis)

      Note that determining the local criteria's conditions above is dependent on the formalism on which we are applying the partial order reduction method.
      The conditions (A 2.1) and (A 2.2) are adjusted to the semantics of the B method/Event-B. For other formalisms such as CSP we will need to define
      the conditions differently. On the one hand because we cannot use the Constraint Based Checker of ProB (it is not implemented for CSP yet), on the other hand
      because CSP differs semantically from B/Event-B (the process structure). An idea to apply Partial Order Reduction on CSP is to make the reduction
      on the CSP operator-level ('|||', [|X|] and so on). For more detailed description of how to apply POR on CSP read the following papers:
            - H. Wehrheim. Partial order reductions for failure refinement.
            - J. Sung, Y. Liu and J.S. Dong. Model Checking CSP Revisited: Introducing a Process Analysis Toolkit.

3. Stutter condition
   (A 3) If ample(s) /= Act(s) then any action t:ample(s) is a stutter action.

4. Cycle condition
   (A 4) For any cycle s_{0}s_{1}...s_{n} in the reduced state space and actions t:Act(s_{i}), for some 
         0 < i <= n, there exists j:{1..n} such that t:ample(s_{j}). (s_{0},...,s_{n} are states)

 */

% compute_ample_set2(+CurID,+CurState,+FindInvViolations,+EnableGraph,+DisableGraph,+PGE)
%
% The main predicate for computing ample sets by respecting the four ample set conditions specified above.
% A subset of all possible enabled outgoing transitions in CurState will be computed and added to the state space. The predicate is similar
% to the compute_all_transitions/2 predicate implemented in tcltk_interface.pl.

% Arguments:
% CurID - the current ID of the state.
% CurState - the current evaluation of all global variables in the machine.
% FindInvViolations - possible values 1 (option setted) and 0 (option not setted). If FindInvViolations equals to 0
%                      the stutter condition (A 3) will be not checked.
% POROptions - the options being set for exploring the state space by means of partial order reduction
% 
compute_ample_set2(CurID,CurState,FindInvViolations,por(TYPE,UseEnableGraph,Depth,PGE)) :- %trace,
   user: reset_max_reached,
   prepare_state_for_specfile_trans(CurState,PreparedCurState),
   % getting all enabled actions at CurID
   get_all_possible_action_names_from_current_state(CurID,PreparedCurState,PGE,Actions),
   % only top level action names will be considered
   get_action_names_only(Actions,ActionNames),
   % mark state as visited (Cycle Condition (A 4))
   % we want only unique action names for analyzing and computing of ample sets
   list_to_ord_set(ActionNames,AllEnabledActions),
   %(AllEnabledActions = ['T1Calculate','T2Evaluate','T3Initiate'] -> trace; true),
   % compute a sound ample set from all actions enabled in state CurID
   compute_ample_actions(CurID,AllEnabledActions,FindInvViolations,por(TYPE,UseEnableGraph,Depth,PGE),AmpleSet),
   debug_println(20,sets(CurID,allActions(AllEnabledActions),ampleSet(AmpleSet))),
   assert(expanded_state(CurID)),
   ( (AmpleSet \= []) ->
      % compute just the actions from the ample set (case of state space reduction)
      % AmpleSet is a non-empty subset of enabled(CurID), AmpleSet can be equal to AllEnabledActions
      add_transitions(AmpleSet,CurID,Actions,FindInvViolations,PGE)
   ; otherwise ->
      true /* DEADLOCK */
   ),
   (user: max_reached ->
     user: assert_max_reached_for_node(CurID),
     MaxReached=true
   ; MaxReached=false
   ),
   (CurID==root ->
      user: save_constants(MaxReached)
   ;  user: add_new_transitions_to_cache(CurID)
   ),!. % once we have computed the ample set in CurState we don't want to backtrack

% computing a sound ample set of actions.
compute_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :- 
  % get_ample_actions/5 call only reasonable to be executed when more than one actions are enabled in CurID
  ((length(AllEnabledActions,L),L>1) -> 
    get_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
  ; otherwise ->
    AmpleSet = AllEnabledActions
  ).

/************************** ADD TRANSITIONS (BEGIN) ******************************/

add_transitions(AmpleSet,CurID,Transitions,FindInvViolations,PGE) :-
  add_transitions(no,AmpleSet,CurID,Transitions,FindInvViolations,PGE).

% first argument tests if the state has been fully explored while adding the 
% ample set transitions to the state space ('yes' means state has been fully explored, so we can stop, 
% 'no' means that the rest of the ample set transitions should be added to the state space until 
% no more ample actions in the list or until fully exploring the state)
add_transitions(yes,_AmpleSet,_CurID,_Transitions,_FindInvViolations,_PGE).
add_transitions(_FullyExplored,[],_CurID,_Transitions,_FindInvViolations,_PGE).
add_transitions(no,[ActionName|ActNames],CurID,Transitions,FindInvViolations,PGE) :-
  add_action_name_transitions(Transitions,CurID,ActionName,FindInvViolations,PGE,[],RemTransitions,FullyExplored),
  add_transitions(FullyExplored,ActNames,CurID,RemTransitions,FindInvViolations,PGE).

add_action_name_transitions([],_CurID,_ActionName,_FindInvViolations,_PGE,RemTransitions,RemTransitions,no).
add_action_name_transitions([(ActionName1,Transitions)|R],CurID,ActionName,FindInvViolations,PGE,IgnoredActions_so_far,RemTransitions,FullyExplored) :-
  (ActionName == ActionName1 ->
    % in this case we do not need to go through the rest of the transitions list
    %add_action_transitions_to_state_space(ActionName,Transitions,FindInvViolations,PGE,CycleCondition),
    add_action_transitions(ActionName,Transitions,CurID,PGE,NewIds),
    append(IgnoredActions_so_far,R,AllRemainedActions),
    (cycle_condition_check_positive_l(FindInvViolations,CurID,NewIds) ->
      debug_println(9,computing_all_other_enabled_actions(back_transition(ActionName),AllRemainedActions)),
      add_all_other_enabled_transitions(AllRemainedActions,CurID,PGE),
      debug_println(9,'*********** CYCLE DETECTED *************'),
      RemTransitions=[],
      FullyExplored = yes
    ; otherwise -> FullyExplored = no, RemTransitions = AllRemainedActions)
  ; otherwise ->
    % keeping track of all ignored actions so far, we don't want to go through the same order of actions every time we
    % want to add a transition from AmpleSet to the state space
    ord_add_element(IgnoredActions_so_far, (ActionName1,Transitions), IgnoredActions_so_far1),
    add_action_name_transitions(R,CurID,ActionName,FindInvViolations,PGE,IgnoredActions_so_far1,RemTransitions,FullyExplored)
  ).

% Adding transitions from the particular ample set to the state space.
%% add_transitions([],_CurID,_Actions,_FindInvViolations,_PGE).
%% add_transitions([ActionName|ActNames],CurID,Actions,FindInvViolations,PGE) :-
%%    add_action_name_transitions(Actions,CurID,ActionName,FindInvViolations,PGE,RemActions),
%%    add_transitions(ActNames,CurID,RemActions,FindInvViolations,PGE).

%% % An event can have parameters, hence there can be multiple actions with the same ActionName.
%% % e.g. act1(1),act1(2),act1(3),...
%% % Therefore, we must add every single action template with same action name to the state space.
%% add_action_name_transitions([],_CurID,_ActionName,_FindInvViolations,_WithPGE,RemActions,RemActions).
%% add_action_name_transitions([(ActionName1,Act,NewExpression,TransInfo,Residue)|R],CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far,RemActions) :-
%%    (ActionName == ActionName1 ->
%%         user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_),
%%         (WithPGE==with_pge -> assert(user: back_transition(NewId,ActionName,CurID)) ; true),
%%         % if successor state already visited we have to explore all actions
%%         (cycle_condition_check_positive(FindInvViolations,CurID,NewId) ->
%%             % Case in which (A 4) can be violated,
%%             % therefore we must explore all possible actions in the current state.
%%             ord_union(IgnoredActions_so_far,R,AllRemainedActions),
%%             debug_println(9,computing_all_other_enabled_actions(back_trans(ActionName),AllRemainedActions)),
%%             compute_all_other_enabled_actions(AllRemainedActions,CurID,WithPGE),
%%             %compute_all_enabled_actions(CurID,CurState,DisableGraph,WithPGE),
%%             debug_println(9,'*********** CYCLE DETECTED *************'),
%%             RemActions=[]
%%         ; otherwise ->
%%             % add only the actions from the computed ample set
%%             add_action_name_transitions(R,CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far,RemActions)
%%         )
%%    ; otherwise ->
%%         % keeping track of all ignored actions so far, we don't want to go through the same order of actions every time we
%%         % want to add a transition from AmpleSet to the state space
%%         ord_add_element(IgnoredActions_so_far, (ActionName1,Act,NewExpression,TransInfo,Residue), IgnoredActions_so_far1),
%%         add_action_name_transitions(R,CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far1,RemActions)
%%    ).

%% cycle_condition_check_positive(1,CurID,NewId) :-
%%     \+var(NewId),
%%     %visited_expression(NewId,_State),
%%     (CurID >= NewId),
%%     expanded_state(NewId).
%%     %\+not_all_transitions_added(NewId).

cycle_condition_check_positive_l(1,CurID,NewIds) :-
  member(NewId,NewIds),
  \+var(NewId),
  %visited_expression(NewId,_State),
  (CurID >= NewId),
  expanded_state(NewId).

%% The old implementation for the PGE optimisation
%% compute_all_other_enabled_actions([],_CurID,_WithPGE).
%% compute_all_other_enabled_actions([(ActionName,Act,NewExpression,TransInfo,Residue)|R],CurID,WithPGE) :-
%%     user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_),
%%     (WithPGE==with_pge -> assert(user: back_transition(NewId,ActionName,CurID)) ; true),
%%     compute_all_other_enabled_actions(R,CurID,WithPGE).

add_all_other_enabled_transitions([],_CurID,_PGE).
add_all_other_enabled_transitions([(ActionName,Transitions)|R],CurID,PGE) :-
  add_action_transitions(ActionName,Transitions,CurID,PGE,_NewIds),
  add_all_other_enabled_transitions(R,CurID,PGE).

add_action_transitions(ActionName,Transitions,CurID,PGE,NewIds) :-
  findall(NewId,
    (member((Act,NewExpression,TransInfo,Residue),Transitions), user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_)),NewIds),
  compute_disabled_actions_in_successor_states(PGE,CurID,ActionName,NewIds).

compute_disabled_actions_in_successor_states(0,_CurID,_ActionName,_NewIds).
compute_disabled_actions_in_successor_states(1,CurID,ActionName,NewIds) :-
  (pge_algo: disabled_operations(CurID,DisabledOperations) -> true; DisabledOperations = []),
  pge_algo: compute_disabled_operations(ActionName,DisabledOperations,NewDisabledOperations),
  pge_algo: update_disabled_operations_l(NewIds,NewDisabledOperations).

/************************** ADD TRANSITIONS (END) ******************************/

/* Auxiliary predicates for getting action names enabled at the current state */ 

% getting action names from the action tuples
get_action_names_only(Actions,ActionNames) :-
    maplist(get_action_name,Actions,ActionNames).

get_action_name((ActionName,_Representations),ActionName).
%get_action_name((ActionName,_A,_NewE,_TrInf,_Residue),ActionName).

% Getting all enabled actions in the current state (check performance)
get_all_possible_action_names_from_current_state(CurID,CurState,PGE,AllActions) :-
  ((PGE == 1, pge_algo: disabled_operations(CurID,DisabledOperations)) -> true; DisabledOperations =[]),
  findall((ActionName,ActionRepresentations),
  (specfile_possible_trans_name(CurState,ActionName), % getting possible action names from the loaded B/Event-B model
   findall((Act,NewExpression,TransInf,Residue),
          spec_file_trans_if_enabled(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue,DisabledOperations),ActionRepresentations),
          ActionRepresentations\=[]),
  AllActions).

spec_file_trans_if_enabled(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue,DisabledOperations) :-
  (ord_member(ActionName,DisabledOperations) ->
    debug_println(9,skipped_guard_evaluation(CurID,ActionName)),
    fail
  ; otherwise ->
    % How should we proceed with the ample computations when a time out occurs?
    specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue)
  ).

% not used: (InfoLog)
%specfile_trans_pge(CurID,DisableGraph,CurState,ActionName,Act,NewExpression,TransInf,Residue) :-
%    pge_algo: get_action_neighbors(ActionName, DisableGraph, Neighbors), % TO DO: ivo ; check this is the right module
%    action_not_disabled(ActionName,CurID,Neighbors),
%    specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue).
%
%action_not_disabled(ActionName,CurID,Neighbors) :-
%    pge_algo: action_disabled(ActionName,CurID,Neighbors), % TO DO: ivo ; check this is the right module
%    debug_println(9,action_disabled(CurID,ActionName)),
%    !,fail.
%action_not_disabled(_ActionName,_CurID,_Neighbors).

specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue) :-
    (use_no_timeout(CurID) ->
        specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue)
    ;   preferences:preference(time_out,CurTO),
        time_out_with_enum_warning_for_findall(specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue),CurTO,TimeOutRes),
        pge_algo: check_timeout_result(CurID,ActionName,CurTO,TimeOutRes),
        Residue==[]
    ).
/* Auxiliary predicates for getting action names enabled at the curent state */ 
%---------------------------------------------------------------------------------------------------------------------------------
/************************** DETERMINING POSSIBLE AMPLE SETS (BEGIN) ******************************/

/************************** PREDICATES COMPUTING AMPLE SETS WITH RESPECT TO CONDITIONS (A 2) and (A 3) *************************/

/* 

   implementation of the ample-sets approach satisfying the dependency condition (A 2)
   by using enabling and dependence analysis (see module enabling analysis)

*/

get_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :- %trace,
    get_possible_ample_actions(AllEnabledActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet).

get_possible_ample_actions([],_CurID,AllEnabledActions,_FindInvViolations,_POROptions,AllEnabledActions).
get_possible_ample_actions(Actions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :-
    % choosing of action for computing a possible ample set for CurID should be random
    random_select(Act,Actions,RemainedActions),
    get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs,IndActs), % finding set (DepActs) satisfying (A 2.1)
    ( is_enabling_dependency_condition_satisfied(IndActs,DepActs,AllEnabledActions,CurID,POROptions) -> % check whether DepActs satisfies (A 2.2)
      (check_stutter_condition(DepActs,FindInvViolations,AllEnabledActions) -> % checking (A 3)
        AmpleSet = DepActs
      ; get_possible_ample_actions(RemainedActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
      )
    ; otherwise -> % search for another valid ample set
      get_possible_ample_actions(RemainedActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
    ).

%% The old (fault) implementation
%% get_possible_ample_actions([],_CurID,AllEnabledActions,_FindInvViolations,_POROptions,AllEnabledActions).
%% get_possible_ample_actions(Actions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :-
%%     % choosing of action for computing a possible ample set for CurID should be random
%%     random_select(Act,Actions,RemainedActions),
%%     get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs), % finding set (DepActs) satisfying (A 2.1) and (A 2.2)
%%     (check_stutter_condition(DepActs,FindInvViolations,AllEnabledActions) -> % checking (A 3)
%%         AmpleSet = DepActs
%%     ; get_possible_ample_actions(RemainedActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
%%     ).

check_stutter_condition(AmpleSet,FindInvViolations,AllEnabledActions) :-
    % in case that we don't check for invariant violations, we don't have to prove the stuttering condition for AmpleSet
    ((FindInvViolations==0;ord_seteq(AmpleSet,AllEnabledActions)) ->
        true
    ; otherwise ->
        stutter_actions(AmpleSet)
    ).

% get_dependent_actions(+CurID,+POROptions,+Act,+AllEnabledActions,+EnGraph,-DepActs)
% CurID - the ID of the current state (used here only for debugging purposes)
% POROptions - partial order reduction options (e.g. UseEnableGraph, Depth)
% Act - the action with which the current dependency class will be computed
% AllEnabledActions - the set of all in CurID enabled actions and it is an ordinary set
% DepActs - the computed candidate for ample set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% EnableGraph - the enable graph for all possible enable actions' relations in the current machine. The enable graph will
%               be computed only single time (before starting with the verification of the state space).

get_dependent_actions(CurID,_POROptions,Act,AllEnabledActions,DepActs,IndActions) :-
    % finding all dependent actions to Act (A 2.1)
    % DepActions_so_far is an ordinary set because AllEnabledActions is an ordinary set and every 'dependent_act(Act1,Act)' is uniquely asserted
    get_dependency_set(Act,AllEnabledActions,CurID,DepActs),
    %% findall(Act1,(member(Act1,AllEnabledActions), is_dependent_to(Act,Act1,CurID)),DepActions_so_far),
    ord_subtract(AllEnabledActions,DepActs,IndActions).

%% get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs) :-
%%     % finding all dependent actions to Act (A 2.1)
%%     % DepActions_so_far is an ordinary set because AllEnabledActions is an ordinary set and every 'dependent_act(Act1,Act)' is uniquely asserted
%%     get_dependency_set(Act,AllEnabledActions,CurID,DepActions_so_far),
%%     %% findall(Act1,(member(Act1,AllEnabledActions), is_dependent_to(Act,Act1,CurID)),DepActions_so_far),
%%     ord_subtract(AllEnabledActions,DepActions_so_far,IndActions),
%%     (IndActions == [] ->
%%         DepActs=DepActions_so_far
%%     ; otherwise ->
%%         enable_graph(EnableGraph),
%%         vertices(EnableGraph,Vertices), % getting all possible actions that can be enabled
%%         % DisabledActions = Vertices\AllEnabledActions
%%         ord_subtract(Vertices,AllEnabledActions,DisabledActions),
%%         % finding all independent actions to Act that could enable a currently disabled action that is dependent to Act (A 2.2)
%%         findall(Act2,(member(Act2,IndActions),may_enable_dep_action(Act2,DepActions_so_far,DisabledActions,EnableGraph,CurID,POROptions)),DepEnActions),
%%         ord_union(DepActions_so_far,DepEnActions,DepActs)
%%         ,debug_println(9,ord_union(CurID,DepActions_so_far,DepEnActions,DepActs))
%%     ).

is_enabling_dependency_condition_satisfied([],_DepActs,_AllEnabledActions,_CurID,_POROptions) :- !,fail.
is_enabling_dependency_condition_satisfied(IndActs,DepActs,AllEnabledActions,CurID,POROptions) :-
  enable_graph(EnableGraph),
  vertices(EnableGraph,Vertices), % getting all possible actions that can be enabled
  % DisabledActions = Vertices\AllEnabledActions
  ord_subtract(Vertices,AllEnabledActions,DisabledActions),
  % searching for an independent action to Act that could enable a currently disabled action that is dependent to Act (A 2.2)
  \+may_enable_dep_action_list(IndActs,DepActs,DisabledActions,EnableGraph,CurID,POROptions).

may_enable_dep_action_list(IndActs,DepActs,DisabledActions,EnableGraph,CurID,POROptions) :-
  member(Act2,IndActs),
  may_enable_dep_action(Act2,DepActs,DisabledActions,EnableGraph,CurID,POROptions).


get_dependency_set(Act,AllEnabledActions,CurID,DepActions_so_far) :-
  vertices_edges_to_ugraph([],[],G),
  get_dependency_tuples_converted_to_edges(Act,AllEnabledActions,RestEnabledActions,CurID,Edges),
  add_edges(G,Edges,G1),
  get_dependency_set1(Act,G1,[],RestEnabledActions,CurID,DepActions_so_far).

get_dependency_tuples_converted_to_edges(Act,AllEnabledActions,RestEnabledActions,CurID,Res) :-
  findall(Act1,(member(Act1,AllEnabledActions), is_dependent_to(Act,Act1,CurID)),DepActs),
  findall([Act-Act1,Act1-Act],member(Act1,DepActs),Edges),
  ord_subtract(AllEnabledActions,DepActs,RestEnabledActions),
  append(Edges,Res).

get_dependency_set1(Act,G,G,_Actions,_CurID,Res) :-
  reachable(Act,G,Res).
get_dependency_set1(Act,G,_GPrime,Actions,CurID,Res) :-
  GPrime1 = G,vertices(G,Vertices),
  findall([V-Act1,Act1-V],(member(Act1,Actions),member(V,Vertices),is_dependent_to(V,Act1,CurID)),NewEdges),
  append(NewEdges,NewEdges1),
  add_edges(G,NewEdges1,G1),vertices(G1,Vertices1),
  ord_subtract(Actions,Vertices1,Actions1),
  get_dependency_set1(Act,G1,GPrime1,Actions1,CurID,Res).

is_dependent_to(Act,Act1,CurID) :-
  (dependent_act(Act,Act1,Status) ->
    (Status=predicate(Pred) -> check_enable_dependency(Act,Act1,Pred,CurID); true)
  ; otherwise ->
    fail
  ).

check_enable_dependency(Act,Act1,Pred,CurID) :-
  debug_println(9,checking_guard_dependency(Act,Act1,Pred)),
  visited_expression(CurID,RawState),
  state_corresponds_to_initialised_b_machine(RawState,State),
  empty_state(LocalState),
  % if the enable predicate Pred evaluates to false we know that (Act,Act1) is a dependent pair 
  \+enable_graph: b_test_boolean_expression_wf(Pred,LocalState,State).

/* searching for enabling path beginning from an action which is not from AmpleSet and 
   which enables action dependent on AmpleSet */
may_enable_dep_action(Act,AmpleSet,DisabledActions,EnableGraph,CurID,por(_TYPE,UseEnableGraph,Depth,_PGE)) :-
  ( UseEnableGraph = true ->
    check_enabling_path(CurID,Act,DisabledActions,AmpleSet,EnableGraph,Depth)
  ; otherwise ->
    member(DisAct,DisabledActions),
    % we check here only currently disabled actions dependent on AmpleSet that can be enabled later
    depends_on_ample_set(DisAct,AmpleSet,_DepAct),
    % is there a path that can enable Act1, which is dependent on Act
    enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph)
  ),
  !.

enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph) :-
  ( enables(Act,DisAct) ->
    % should improve performance a little bit (not applicable when using enable graphs)
    debug_println(9,enables(Act,DisAct))
  ; check_if_enabled(CurID,Act,DisAct,AmpleSet,EnableGraph) ->
    true
  ; otherwise ->
    fail
  ).

% not used (InfoLog):
%enables_pred_action(Act,DisAct,AmpleSet,CurID,EnableGraph,Depth) :-
%    % visited_expression/2 and state_corresponds_to_initialised_b_machine/2 should be called once for each state,
%    % avoid multiple calls per state
%    (Depth==0 ->
%       empty_state(ST) % empty state for 
%    ; otherwise ->
%       visited_expression(CurID,RawState),
%       state_corresponds_to_initialised_b_machine(RawState,ST)
%    ),
%    %Pred =.. [evaluate_enabling_predicate_for_por, ST, AmpleSet],
%    enable_graph: path(EnableGraph,Act,DisAct,enable_graph,
%                       evaluate_enabling_predicate_for_por(ST,AmpleSet),Depth,Path,Length),
%    debug_print_path(CurID,Path,Length).

check_if_enabled(CurID,Act,DisAct,AmpleSet,EnableGraph) :-
  min_path(Act,DisAct,EnableGraph,P,L),
  sets_disjoint(AmpleSet,P),
  debug_print_path(CurID,P,L),
  assert(enables(Act,DisAct)).

% predicates used for the enable graph feature
check_enabling_path(CurID,Act,DisabledActions,AmpleSet,EnableGraph,Depth) :-
  findall(DepDisAct,(member(DepDisAct,DisabledActions),depends_on_ample_set(DepDisAct,AmpleSet,_Act)),DependentDisabledActions),
  ( DependentDisabledActions = [] ->
    fail % in this case we know that no action dependent on AmpleSet can be enabled after executing Act
  ; otherwise ->
    enables_pred_action1(Act,DependentDisabledActions,AmpleSet,CurID,EnableGraph,Depth)).

enables_pred_action1(Act,DepDisActs,AmpleSet,CurID,EnableGraph,Depth) :-
    % visited_expression/2 and state_corresponds_to_initialised_b_machine/2 should be called once for each state,
    % avoid multiple calls per state
    (Depth==0 ->
       empty_state(ST) % empty state for 
    ; otherwise ->
       visited_expression(CurID,RawState),
       state_corresponds_to_initialised_b_machine(RawState,ST)
    ),
    %Pred =.. [evaluate_enabling_predicate_for_por, ST, AmpleSet],
    enable_graph: path_set(EnableGraph,Act,DepDisActs,enable_graph,
                           evaluate_enabling_predicate_for_por(ST,AmpleSet),Depth,Path,Length),
    debug_print_path(CurID,Path,Length).

% guaranteeing that the dependent action to AmpleSet has been enabled from an action off AmpleSet
sets_disjoint(AmpleSet,Path) :-
    list_to_ord_set(Path,OrdSet),
    \+ord_intersection(AmpleSet,OrdSet).

depends_on_ample_set(Act,AmpleSet,Act1) :-
   member(Act1,AmpleSet),
   dependent_act(Act,Act1,_Status).

debug_print_path(CurID,Path,Length) :-
   debug_println(20,'***** Enable Path *****'),
   debug_println(20,path(CurID,Path,Length)),
   debug_println(20,'***********************').

/************************** DETERMINING POSSIBLE AMPLE SETS (END) ******************************/

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (BEGIN) WITH REGARD TO THE INVARIANT ******************************/
/************************** PREDICATES WHICH SATISFY CONDITION (A3) ******************************/

stutter_actions(Actions) :-
    maplist(is_stutter_action,Actions).

% visible actions are non stutter actions
is_stutter_action(Act) :- 
   ( stutter_action(Act) -> 
       debug_println(9,stutter_action(Act))
   ; visible_action(Act) -> 
       debug_println(9,visible_action(Act)),
       fail
   ; otherwise -> 
       check_if_action_is_stutter_wrt_inv(Act)
   ),!.

% A stutter action cannot change the current invariant value.
% That's why we must check if the particular action preserves
% the full invariant in order to determine if the action is
% stutter under the following LTL formula:
% G Phi, where Phi is the invariant.
check_if_action_is_stutter_wrt_inv(Act) :- 
    b_operation_preserves_full_invariant(Act),!,
    assert(stutter_action(Act)),
    debug_println(9,stutter_according_to_proof_info(Act)).
check_if_action_is_stutter_wrt_inv(Act) :- 
   b_get_invariant_from_machine(Invariant),
   conjunction_to_list(Invariant,Conjunctions), % getting all conjunctions from the invariant
   b_get_read_write(Act,_Reads,Writes),
   ( action_does_not_modify_important_variables_in_the_inv(Writes,Conjunctions) 
     -> assert(stutter_action(Act)), 
        debug_println(9,stutter_according_to_read_write_info(Act))
   ; otherwise 
     -> assert(visible_action(Act)),fail
   ).

action_does_not_modify_important_variables_in_the_inv([],_Conjuncions).
action_does_not_modify_important_variables_in_the_inv([Var|WrVars],Conjunctions) :- 
   var_not_changing_visible_condition_in_the_inv(Var,Conjunctions),
   action_does_not_modify_important_variables_in_the_inv(WrVars,Conjunctions).

var_not_changing_visible_condition_in_the_inv(_Var,[]).
var_not_changing_visible_condition_in_the_inv(Var,[Conj|Conjs]) :- 
   ( occurs_in_expr(Var,Conj) 
     -> (\+ trivial_predicate(Var,Conj) -> fail; true)
   ; otherwise -> true
   ),
   var_not_changing_visible_condition_in_the_inv(Var,Conjs).

trivial_predicate(Var,Conj) :- 
   \+var(Conj),
   trivial_member_predicate(Var,Conj).

/* 
   
   TODO: Define formal when an action in Event-B/B can be seen as a stutter action.
   Maybe we should use here the predicate b_specialized_invariant_for_op/2 from bmachine.pl.

*/

/*trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NATURAL'),set(integer),_)),pred,_).
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NAT'),set(integer),_)),pred,_).*/
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INT'),set(integer),_)),pred,_). 
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INTEGER'),set(integer),_)),pred,_).

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (END) ******************************/
