% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(smtlib2_interpreter,[interpret_l/3]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smtlib).
:- module_info(description,'The SMT-LIB 2 Interpreter - Main Interpreter').

:- use_module(library(lists)).

:- use_module(probsrc(solver_interface), [solve_predicate/4]).
:- use_module(probsrc(atelierb_provers_interface), [disprove_predicate/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2]).
:- use_module(probsrc(btypechecker), [couplise_list/2]).
:- use_module(probsrc(eclipse_interface), [load_sets_for_cbc/1]).
:- use_module(probsrc(prob2_interface), [start_animation/0]).
:- use_module(probsrc(translate), [print_bexpr/1]).
:- use_module(smtlib_solver(smtlib2_environment)).
:- use_module(smtlib_solver(smtlib2_translation)).
:- use_module(probsrc(tools),[start_ms_timer/1, stop_ms_timer/1]).
:- use_module(probsrc(error_manager), [add_error/3, add_warning/3]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(debug), [debug_println/2, debug_mode/1]).

:- dynamic expected_result/1.
:- volatile expected_result/1.

interpret_l([],Env,Env).
interpret_l([H|T],EnvIn,EnvOut) :-
  interpret(H,EnvIn,EnvT),
  interpret_l(T,EnvT,EnvOut).

interpret(exit,E,E) :- !, throw(halt(0)).
interpret(set_logic(_X),E,E) :- !,
  % TODO: only accept supported logics.
  true.
interpret(set_info(X),E,E) :- !,
  set_info(X).
% special case for array logics, sets up an additional constraint
interpret(declare_fun(Id,[],sort('Array',[Index,Element])),EIn,EOut) :- !,
  smt_type_to_prob_type(Index,EIn,BIndex),
  smt_type_to_prob_type(Element,EIn,BElement),
  FullType = set(couple(BIndex,BElement)),
  add_identifier(EIn,Id,FullType,ETemp),
  % however, an array is a partial function. thus, we have to add a constraint
  typing_constraint(svar(Id,sort('Array',[Index,Element])),AdditionalConstraint),
  add_assertion(ETemp,AdditionalConstraint,EOut).
% special case for bit vectors, sets up an additional constraint
interpret(declare_fun(Id,[],sort('BitVec',[Length])),EIn,EOut) :- !,
  FullType = set(couple(integer,integer)),
  add_identifier(EIn,Id,FullType,ETemp),
  % however, an array is a partial function. thus, we have to add a constraint
  typing_constraint(svar(Id,sort('BitVec',[Length])),AdditionalConstraint),
  add_assertion(ETemp,AdditionalConstraint,EOut).
interpret(declare_fun(Id,[],Type),EIn,EOut) :- !,
  smt_type_to_prob_type(Type,EIn,BType),
  add_identifier(EIn,Id,BType,EOut).
interpret(declare_fun(Id,Params,Return),EIn,EOut) :- !,
  maplist(smt_type_to_prob_type_maplist(EIn),Params,BParams),
  smt_type_to_prob_type(Return,EIn,BReturn),
  couplise_list(BParams,BParamsCoupled),
  Fulltype = set(couple(BParamsCoupled,BReturn)),
  add_identifier(EIn,Id,Fulltype,ETemp),
  is_function_constraint(Id,Fulltype,Constraint),
  add_assertion(ETemp,Constraint,EOut).
interpret(declare_sort(Name,0),EIn,EOut) :- !,
  add_custom_type(EIn,Name,EOut).
interpret(define_fun(ID,Params,Return,Definition),EIn,EOut) :- !,
  smt_type_to_prob_type(Return,EIn,BReturn),
  maplist(smt_term_to_prob_term_maplist(EIn),Params,BParams),
  create_local_env(EIn,BParams,LocalEnv),
  smt_term_to_prob_term(Definition,LocalEnv,BDefinition),
  add_function(EIn,ID,BParams,BReturn,BDefinition,EOut).
interpret(assert(Term),EIn,EOut) :- !,
  smt_term_to_prob_term(Term,EIn,BTerm),
  bool_to_pred(BTerm,PredBTerm),
  add_assertion(EIn,PredBTerm,EOut).
interpret(push(Num),EIn,EOut) :- !,
  push_assertion_stack(EIn,Num,EOut).
interpret(pop(Num),EIn,EOut) :- !,
  pop_assertion_stack(EIn,Num,EOut).
interpret(set_option(Opt),EIn,EOut) :- !,
  format('Ignoring set_option(~w) !~n',[Opt]), EOut=EIn.
interpret(check_sat,E,E) :- !,
  get_assertions(E,BTerms),
  get_custom_types(E,CT),
  conjunct_predicates(BTerms,Assertions),
  QuantifiedAssertions = Assertions,
  load_sets_for_cbc(CT),
  start_animation,
  (debug_mode(on) -> print('Solving: '),nl,
      translate:print_bexpr(QuantifiedAssertions),nl
      ; true),
  start_ms_timer(Timer),
  solve_predicate(QuantifiedAssertions,_State,4,FullResult), !,
  %nl,print(State),nl,
  (debug_mode(on) -> print('ProB Solver walltime: '),stop_ms_timer(Timer) ; true),
  try_provers(FullResult,E,QuantifiedAssertions).
interpret(get_model,E,E) :- !, print_model(E). % this command is not part of the smtlib2 language anymore
interpret(X,_,_) :- !, format('missing in interpret: ~w~n',[X]), fail.

try_provers(solution(S),E,_) :- !,
   assert(stored_model(S)),
   (debug_mode(on) -> print_model(E) ; true),
   add_error_if_result_unexpected(sat),
   write('sat\n').
try_provers(contradiction_found,_,_) :- !, add_error_if_result_unexpected(unsat), write('unsat\n').
try_provers(_,_,Assertions) :-
    get_preference(try_atb_provers,true),
    % we suspect that the solver failed to identify a contradiction
    % disprove predicate might fail, i.e. if the provers are not installed
    debug_println(20,'Trying to run Atelier-B provers'),
    disprove_predicate(Assertions,Result),!,
    write_final_result(Result).
try_provers(_,_,_) :- add_error_if_result_unexpected(unknown), write('unknown\n').

write_final_result(proved) :- add_error_if_result_unexpected(unsat), write('unsat\n').
write_final_result(_) :- add_error_if_result_unexpected(unknown), write('unknown\n').

add_error_if_result_unexpected(unknown) :- % in non-strict mode: no message, we simple failed to establish a result
   get_preference(strict_raise_warnings,false),!.
add_error_if_result_unexpected(Actual) :-
    expected_result(Expected),!,
    (Actual = Expected -> true
     ; Expected = unknown -> true
     ; Actual = unknown ->
        add_warning(smtlib2_interpreter,'Result unknown; could not establish expected result: ',Expected)
     ; add_error(smtlib2_interpreter, 'Actual result differs from expected result: ','/='(Actual,Expected))).
add_error_if_result_unexpected(_). % no expected result stored

:- dynamic stored_model/1.

print_binding(E,binding(V,_,PP)) :-
    get_type(V,E,Type),
    type_for_print(Type,TypeForPrint),
    format('(define-fun ~w () ~w ~w)~n',[V,TypeForPrint,PP]).
print_model(E) :-
    stored_model(M), !,
    print('(model'), nl,
    maplist(print_binding(E),M),
    print(')'), nl.
print_model(_) :- write('no model generated\n').

type_for_print(integer,'Int') :- !.
type_for_print(X,X).

set_info(attribute(A,V)) :- !,
  set_attribute(A,V).
set_info(X) :- !, format('missing in set_info: ~w~n',[X]), fail.


set_attribute('smt-lib-version',decimal(2.0)) :- !.
set_attribute(category,string(_S)) :- !.
set_attribute(notes,string(_S)) :- !.
set_attribute(status,S) :- !,
  member(S,[unknown,sat,unsat]),
  retractall(expected_result(_)),
  assert(expected_result(S)).
set_attribute(source,string(_S)) :- !.
set_attribute(A,V) :- !,
  format('missing in set_attribute: ~w to ~w~n',[A,V]), fail.
