% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kinduction,[kinduction_symbolic_model_check/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'A finite-state symbolic model checker based on k-Induction.').

:- use_module(library(lists), [append/2,
                               maplist/3]).

:- use_module(extension('counter/counter'),[counter_init/0,
                                            new_counter/1]).

:- use_module(probsrc(before_after_predicates), [before_after_predicate_list_disjunct_with_equalities/3,
                                                 before_after_predicate_with_equalities/3,
                                                 before_after_predicate_for_operation/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,
                                 b_get_properties_from_machine/1,
                                 b_get_machine_operation/4,
                                 b_get_machine_constants/1,
                                 b_get_machine_variables/1,
                                 b_get_initialisation_from_machine/2,
                                 b_specialized_invariant_for_op/2]).
:- use_module(probsrc(btypechecker), [prime_identifiers/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                                     disjunct_predicates/2,
                                     disjunction_to_list/2,
                                     transform_bexpr/3,
                                     syntaxtransformation/5,
                                     safe_create_texpr/3,
                                     create_negation/2,
                                     get_texpr_expr/2,
                                     get_texpr_type/2,
                                     unique_typed_id/3,
                                     find_typed_identifier_uses/2]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(symbolic_model_checker(predicate_handling), [prime_predicate/2,
                                                           create_state_predicate/2,
                                                           deprime_state/2,
                                                           prime_n_times/3]).
:- use_module(symbolic_model_checker(solver_handling), [solve/2]).
:- use_module(symbolic_model_checker(bmc), [get_constraint/2,
                                            constraint_transitions/2,
                                            replay_counter_example/1]).

kinduction_symbolic_model_check(Res) :-
    counter_init, new_counter(ic3solvercalls),
    kinduction_symbolic_model_check(0,Res).

kinduction_symbolic_model_check(25,limit_reached) :- !.
kinduction_symbolic_model_check(K,Res) :-
    format('K = ~w~n', [K]),
    get_constraint(K,Base),
    (debug:debug_mode(off) -> true
     ; translate_bexpression(Base,PPBase),
       format('Base Constraint (~w): ~w~n',[K,PPBase])
    ),
    solve(Base,Result),
    Result = solution(Sol), !,
    Res = counterexample_found,
    replay_counter_example(Sol).
kinduction_symbolic_model_check(K,Res) :-
    get_step_constraint(K,Step),
    (debug:debug_mode(off) -> true
     ;  translate_bexpression(Step,PPStep),
        format('Step Constraint (~w): ~w~n',[K,PPStep])
    ),
    solve(Step,Result),
    (Result = contradiction_found(_) -> Res = property_holds ;
     Result = solution(_) -> K2 is K + 1, kinduction_symbolic_model_check(K2,Res) ;
     otherwise -> Res = solver_and_provers_too_weak).

get_step_constraint(K,Step) :-
    get_preference(use_po,false), !,
    b_get_invariant_from_machine(P),
    constraint_transitions(K,Transitions),
    step_constraint_property(K,P,Property),
    K2 is K + 1,
    prime_n_times(K2,P,PPrimed),
    create_negation(PPrimed,NegPPrimed),
    no_loop_constraint(K,NL),
    conjunct_predicates([NL,Transitions,Property,NegPPrimed],Step).
get_step_constraint(K,Step) :-
    get_preference(use_po,true), !,
    b_get_invariant_from_machine(P),
    K2 is K - 1, K3 is K + 1,
    no_loop_constraint(K3,NL),
    constraint_transitions(K2,Transitions),
    step_constraint_property(K,P,Property),
    proof_supported_step_constraint(K3,StepConstraint),
    % translate_bexpression(NL,PPNL),
    % format('No Loop Constraint: ~w~n',[PPNL]),
    conjunct_predicates([NL,Transitions,Property,StepConstraint],Step).

proof_supported_step_constraint(K,b(truth,pred,[])) :-
    K < 0, !.
proof_supported_step_constraint(N,TPred) :-
    findall(op(OpName,Pred),(before_after_predicate_for_operation(OpName,Pred)),Ops),
    maplist(op_to_step_constraint,Ops,SingleSteps),
    disjunct_predicates(SingleSteps,Step),
    N2 is N - 1,
    prime_n_times(N2,Step,TPred).

op_to_step_constraint(op(OpName,BaPred),StepConstraint) :-
    (b_specialized_invariant_for_op(OpName,UncheckedInvariant)
     -> true ; b_get_invariant_from_machine(UncheckedInvariant)),
    create_negation(UncheckedInvariant,NegUncheckedInvariant),
    prime_predicate(NegUncheckedInvariant,PrimedNegUncheckedInvariant),
    conjunct_predicates([BaPred,PrimedNegUncheckedInvariant],StepConstraint).

step_constraint_property(K,P,FullProperty) :-
    step_constraint_property_aux(K,P,Conjuncts),
    conjunct_predicates(Conjuncts,FullProperty).
step_constraint_property_aux(0,P,[P]) :- !.
step_constraint_property_aux(N,P,[P|Rest]) :-
    prime_predicate(P,PP),
    N2 is N - 1,
    step_constraint_property_aux(N2,PP,Rest).

no_loop_constraint(K,NL) :-
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    no_loop_constraint_add_primes(K,VarsAndConsts,ListOfPrimedStates),
    no_loop_constraint_aux(ListOfPrimedStates,Conjuncts),
    conjunct_predicates(Conjuncts,NL).

no_loop_constraint_aux([],[]).
no_loop_constraint_aux([PState|PStates],Conjuncts) :-
    no_loop_constraint_aux(PState,PStates,Conj1),
    no_loop_constraint_aux(PStates,Conj2),
    append([Conj1,Conj2],Conjuncts).
no_loop_constraint_aux(_State,[],[]).
no_loop_constraint_aux(PState1,[PState2|States],[NotEqualDisjunct|NotEquals]) :-
    states_not_equal(PState1,PState2,NotEqual),
    disjunct_predicates(NotEqual,NotEqualDisjunct),
    no_loop_constraint_aux(PState1,States,NotEquals).

states_not_equal(P,P,[]) :- !.
states_not_equal([V1|V1s],[V2|V2s],[NotEqual|T]) :-
    safe_create_texpr(not_equal(V1,V2),pred,NotEqual),
    states_not_equal(V1s,V2s,T).

no_loop_constraint_add_primes(0,V,[V]).
no_loop_constraint_add_primes(1,VarsAndConsts,[PVarsAndConsts,VarsAndConsts]) :-
    prime_identifiers(VarsAndConsts,PVarsAndConsts).
no_loop_constraint_add_primes(N,VAC,[VAC|VACS]) :-
    prime_identifiers(VAC,PVAC),
    N2 is N - 1,
    no_loop_constraint_add_primes(N2,PVAC,VACS).

transition_predicate(TransitionPredicate) :-
    findall(Body,b_get_machine_operation(_Name,_Results,_Parameters,Body),ListOfBodies),
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    before_after_predicate_list_disjunct_with_equalities(ListOfBodies,VarsAndConsts,TransitionPredicate).
