:- module(b_synthesis,[start_synthesis/0]).

:- use_module(synthesis(constraints)).
:- use_module(synthesis(library_setup)).
:- use_module(synthesis(user_interaction)).
:- use_module(synthesis(location_vars_to_program)).

:- use_module(probsrc(state_space)). 
:- use_module(probsrc(bmachine)). 
:- use_module(probsrc(bsyntaxtree)). 
:- use_module(probsrc(b_interpreter),[b_test_boolean_expression_wf/3]). 
:- use_module(probsrc(translate),[print_bexpr_or_subst/1]). 
:- use_module(probsrc(solver_interface),[solve_predicate/3]). 
:- use_module(probsrc(bmachine_eventb),[prime_variable/2,is_primed_id/1]).
:- use_module(probsrc(before_after_predicates),[before_after_predicate_with_equalities/3]).

:- use_module(library(lists)).
:- use_module(library(terms),[term_size/2]).
:- use_module(library(sets),[intersection/3,subtract/3]).

% invariant violated
start_synthesis :-  
    preferences:set_preference(time_out,8000) , 
    state_error(State,_Error,invariant_violated) , ! , 
    transition(_Start,Operation,_,State) , 
    start_specific_synthesis_aux(Operation,State).
% add missing transition
start_synthesis :- 
    format("~nNo invariant violation found.~nPlease give at least two examples when synthesizing a new operation.~n~n",[]) , 
    b_get_machine_variables(MachineVars) , 
    user_interaction_add_missing_transition(MachineVars,PosInput,PosOutput,VarNodes) , 
    PosInput \= [] , PosOutput \= [] , ! , 
    % search for an operation that already executes the transition or one where 
    % we only need to relax the guard
    maplist(check_transition_against_machine_operations(VarNodes),PosInput,PosOutput,Results) , 
    missing_transition_aux(Results,PosInput,[],PosOutput,[]).
start_synthesis.

% initialisation
start_specific_synthesis_aux(Operation,0) :- 
    Operation =.. ['$initialise_machine'|Initialisation] , 
    maplist(translate:translate_bvalue,Initialisation,PrettyInit) , 
    format("~nInvariant violation in initialisation state ~w",[PrettyInit]).
% specific operation that violates the invariant
start_specific_synthesis_aux(Operation,State) :- 
    State \= 0 , 
    format("~nInvariant violation in state ~w, caused by operation ~w ~n",[State,Operation]) , 
    pre_select_transitions(Operation,ExamplesAssumingYes,ExamplesAssumingNo) , 
    b_get_machine_variables(MachineVars) , 
    user_interaction(MachineVars,ExamplesAssumingNo,ExamplesAssumingYes,PosInput,NegInput,PosOutput,NegOutput,Type) , 
    start_specific_synthesis(Operation,Type,PosInput,NegInput,PosOutput,NegOutput).

% before synthesizing a new operation check if an operation already executes the desired transition
missing_transition_aux(Results,PosInput,_,PosOutput,_) :- 
    member(no_match,Results) , 
    start_specific_synthesis(action,PosInput,[],PosOutput,[]).
% done
missing_transition_aux(Results,_,_,_,_) :- 
    member((OpName,done),Results) , 
    format("~nOperation ~w already executes this state transition. ~n",[OpName]).
% synthesize new guard
missing_transition_aux(Results,PosInput,_,PosOutput,_) :- 
    member((OpName,guard),Results) , 
    format("~nStart synthesizing relaxed guard for operation ~w. ~n",[OpName]) , 
    % add some examples to cover the previous behavior
    get_valid_examples_for_operation(25,OpName,ValidExamples) , 
    append(PosInput,ValidExamples,Temp) , delete(Temp,[],Positive) , 
    delete_positive_state_from_negative(Positive,PosOutput,Negative) , 
    start_specific_synthesis(_Operation,guard,Positive,Negative,[],[]).

%% synthesize actions, invariants or guards
% we already checked the machine operations and synthesize a new operation
start_specific_synthesis(action,PosInput,_NegInput,PosOutput,_NegOutput) :- 
    PosInput = [Example|_] , 
    library_setup(action,1,Example,Library) , 
    b_get_machine_variables(MachineVars) , 
    % get current variables from an example 
    findall(VarName,(member(b(_,_,Info),Example) , member(synthesis(machinevar,VarName),Info)),CurrentVarNames) , 
    findall(VarNode,(member(VarNode,MachineVars) , VarNode = b(identifier(VarName),_,_) , 
                     member(VarName,CurrentVarNames)),CurrentVars) , 
    get_behavioral_constraint(Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) , 
    solve_predicate(BehavioralConstraint,_,Solution) , 
    PosInput = [In|_] , get_line_amount(Library,In,M) , M1 is  M - 1 , 
    length(In,ParameterAmount) , 
    preferences:set_preference(time_out,4000) ,
    synthesis_from_examples(Library,1,CurrentVars,any,action,PosInput,PosOutput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,Synthesized) , 
    print_synthesized_solution(action,Synthesized).
% user edited a transition so we first want to check if any operation already executes this transition
% or we need to synthesize a new operation 
start_specific_synthesis(_Operation,action,PosInput,_NegInput,PosOutput,_NegOutput) :- 
    b_get_machine_variables(MachineVars) , 
    maplist(check_transition_against_machine_operations(MachineVars),PosInput,PosOutput,Results) , 
    (member((OpName,done),Results)
    ->  format("~nOperation ~w already executes this state transition. ~n",[OpName])
    ;   start_specific_synthesis(action,PosInput,[],PosOutput,[])).    
% if we synthesize an invariant we reduce the examples to violating vars only
start_specific_synthesis(Operation,invariant,PosInput,NegInput,PosOutput,NegOutput) :- 
    % use PosInput because at least one invariant violating state has been graduated as valid by the user
    get_violated_and_valid_machine_invariants(Operation,PosOutput,ViolatedInvariants,ViolatingVars,ValidInvariants) , 
    findall(Name,(member(IDNode,ViolatingVars) , IDNode = b(identifier(Name),_,_)),VarNames) , 
    format("~nVariables involved in invariant violation: ~w ~n",[VarNames]) , 
   
   % for synthesizing invariants we also need output states
    append(PosInput,PosOutput,Positive) , 
    append(NegInput,NegOutput,TempNegative) , 
    delete_positive_state_from_negative(Positive,TempNegative,Negative) , 
    % don't use an example more than one time, only possible for invariant synthesis 
    % because we split every transition in two examples (input,output)
    remove_dups(Positive,Pos) , remove_dups(Negative,Neg) , 
    maplist(reduce_input_to_violating_vars(VarNames),Pos,TRPositive) , 
    maplist(reduce_input_to_violating_vars(VarNames),Neg,RNegative) , 
    
    TRPositive = [Example|_] , 
    library_setup(invariant,1,Example,Library) , 
    
    get_valid_states(5,TValidStates) , 
    maplist(reduce_input_to_violating_vars(VarNames),TValidStates,ValidStates) , 
    append(TRPositive,ValidStates,RPositive) , 
    remove_dups(RPositive,NRPositive) ,
    % first positive examples is the one that violated an invariant
    PosInput = [In|_] , PosOutput = [Out|_] , 
% TODO? find several transitions and ask the user until an undesired one is found, 
%       to make the invariant even more relaxed
    % get counter example 
    get_invariant_violating_transition(Operation,ViolatingVars,ViolatedInvariants,In,Out,Counter) , 
    start_specific_synthesis_aux(invariant,ValidInvariants,VarNames,NRPositive,RNegative,ViolatingVars,Library,Counter).

start_specific_synthesis(Operation,guard,PosInput,NegInput,_PosOutput,NegOutput) :- 
    % try to reduce input examples by checking the machine invariants 
    get_violated_and_valid_machine_invariants(Operation,NegOutput,_ViolatedInvariants,ViolatingVars,_ValidInvariants) , 
    findall(Name,member(b(identifier(Name),_,_),ViolatingVars),VarNames) , 
    format("~nVariables involved in invariant violation: ~w ~n",[VarNames]) , 

    maplist(reduce_input_to_violating_vars(VarNames),PosInput,TempPositive) , 
    maplist(reduce_input_to_violating_vars(VarNames),NegInput,Negative) , 
    delete_positive_state_from_negative(Negative,TempPositive,Positive) , 
    Positive = [Example|_] , 
    library_setup(guard,1,Example,Library) , 
    % remove dups ?
    get_behavioral_constraint_from_input(Library,Positive,Negative,BehavioralConstraint,LocationVars) , 
    solve_predicate(BehavioralConstraint,_,Solution) , 

    get_line_amount(Library,Example,M) , M1 is  M - 1 , 
    length(Example,ParameterAmount) , 
    preferences:set_preference(time_out,4000) , 
    get_guards_from_operation(Operation,Guards) , 
    synthesis_from_examples(Library,1,ViolatingVars,Guards,guard,Positive,Negative,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,Synthesized) , 
    print_synthesized_solution(guard,Synthesized).

% auxiliary predicate for synthesis of invariants to check if we can find a counter example
start_specific_synthesis_aux(invariant,ValidInvariants,VarNames,Positive,TNegative,ViolatingVars,Library,(TCounterIn,TCounterOut)) :- 
    maplist(reduce_input_to_violating_vars(VarNames),[TCounterIn,TCounterOut],[CounterIn,CounterOut]) , 
    % don't create contradiction by using the same example as positive and negative
    findall(Counter,(member(Counter,[CounterIn,CounterOut]) , \+member(Counter,Positive)),CounterStates) , 
    append(CounterStates,TNegative,Negative) , 
    get_behavioral_constraint_from_input(Library,Positive,Negative,BehavioralConstraint,LocationVars) , 
    solve_predicate(BehavioralConstraint,_,Solution) , 
    get_line_amount(Library,CounterIn,M) , M1 is  M - 1 , 
    length(CounterIn,ParameterAmount) , 
    preferences:set_preference(time_out,4000) ,
    synthesis_from_examples(Library,1,ViolatingVars,ValidInvariants,invariant,Positive,[CounterIn|Negative],BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,Synthesized) , 
    print_synthesized_solution(invariant,Synthesized).
% we can't find any invariant violating transition (counter example), therefore we just remove the violating invariants
% no invariant given
start_specific_synthesis_aux(invariant,[],_,_,_,_,_,none) :- 
    print_synthesized_solution(invariant,b(truth,pred,[])).
% print valid invariants
start_specific_synthesis_aux(invariant,ValidInvariants,_,_,_,_,_,none) :- 
    conjunct_predicates(ValidInvariants,Invariant) , 
    print_synthesized_solution(invariant,Invariant).

%% synthesis loop 
% search for further solution
synthesis_from_examples(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,Synthesized) :- 
    Solution = solution(_) ,  
    % in UsedLVars we have the used location variables from the synthesized solution, so we are able
    % to exclude those specific bindings when searching for a further solution (otherwise it would be enough to just change some deadcode)
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,UsedLVars1,_ProgramAST) , 
    find_further_solution(UsedLVars1,Solution,BehavioralConstraint,FurtherSolution,NewBehavioralConstraint) , 
    synthesis_from_examples_aux1(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Synthesized).
% expand the library if no solution can be found
synthesis_from_examples(OldLibrary,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,_,_,M1,ParameterAmount,OldSolution,Synthesized) :- 
    OldSolution \= solution(_) , 
    NewLibExpansion is LibExpansion + 1 , 
    PosInput = [Example|_] , 
    library_setup(Type,NewLibExpansion,Example,NewLibrary) , ! , 
    % library has changed 
    (NewLibrary \= OldLibrary 
    ->  format("~nExpand library.~n",[]) , 
        get_behavioral_constraint_aux(Type,NewLibrary,PosInput,NegInput,NewBehavioralConstraint,NewLocationVars) , 
        solve_predicate(NewBehavioralConstraint,_,NewSolution) , ! , 
        synthesis_from_examples(NewLibrary,NewLibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,NewBehavioralConstraint,NewLocationVars,M1,ParameterAmount,NewSolution,Synthesized)
    ;   % otherwise synthesis failed
        Synthesized = none).

% done
synthesis_from_examples_aux1(Library,_LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,_PosInput,_NegInput,_BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,none,_,Synthesized) :- 
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_UsedLVars,ProgramAST) , 
    replace_preds_by_subst_if_neccessary(Type,ProgramAST,TSolutionAST) , 
    %prime_or_unprime_variables_in_predicate(unprime,TSolutionAST,SolutionAST) , 
    conjunct_previous_invariant_or_guard_if_neccessary(Type,ValidInvariantsOrGuards,TSolutionAST,Synthesized).
% try to find a distinguishing input
synthesis_from_examples_aux1(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Synthesized) :- 
    FurtherSolution \= none , 
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_UsedLVars1,ProgramAST) , 
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,FurtherSolution,_UsedLVars2,ProgramAST2) , 
    get_distinguishing_input(Type,ValidInvariantsOrGuards,ProgramAST,ProgramAST2,Distinguishing) , 
    synthesis_from_examples_aux2(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Distinguishing,Synthesized). 

% two non equivalent programs
synthesis_from_examples_aux2(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,solution(Dist),Synthesized) :- 
    % create behavioral constraint containing the distinguishing input and start
    % synthesis again with new examples
    nl , format("Found two non equivalent ~w~n",[Type]) , 
    synthesis_from_examples_aux3(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Dist,Synthesized).  
% or we choose the one with a lower term size since we know that both ASTs are equivalent
synthesis_from_examples_aux2(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,FurtherSolution,FurtherBehavioralConstraint,Distinguishing,Synthesized) :- 
    Distinguishing \= solution(_) , 
    % no distinguishing input, no conclusion about examples possible
    % therefore we pass the "new" behavioral constraint and start synthesis again
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_UsedLVars1,ProgramAST) , 
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,FurtherSolution,_UsedLVars2,ProgramAST2) , 
    term_size(ProgramAST,TS1) , term_size(ProgramAST2,TS2) , 
    (TS1 < TS2
    ->  get_new_behavioral_constraint(LocationVars,FurtherSolution,BehavioralConstraint,NewBehavioralConstraint) , 
        Sol = Solution
    ;   NewBehavioralConstraint = FurtherBehavioralConstraint, 
        Sol = FurtherSolution) , 
    synthesis_from_examples(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,NewBehavioralConstraint,LocationVars,M1,ParameterAmount,Sol,Synthesized).

% distinguishing transition (actions)
synthesis_from_examples_aux3(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,action,Input,Output,_BehavioralConstraint,_LocationVars,M1,ParameterAmount,_Solution,_FurtherSolution,_NewBehavioralConstraint,Distinguishing,Synthesized) :- 
    %adjust_distinguishing_input(CurrentVars,Distinguishing,CompleteDist) , 
    findall(Node,(member(binding(Var,Temp,_),Distinguishing) , \+is_primed_id(Var) , 
                  value_to_ast(Temp,TNode) , add_texpr_infos(TNode,[synthesis(machinevar,Var)],Node)),DistIn) , 
    % find program with new examples from distinguishing input/state/transition
    user_interaction_distinguishing_transition(solution(Distinguishing),CurrentVars,Answer,DistOut) , 
    (Answer = none 
    ->  % synthesize guard for the action
        library_setup(guard,1,DistIn,PreLibrary) , 
        get_behavioral_constraint_from_input(PreLibrary,Input,[DistIn],PreBehavioralConstraint,PreLocationVars) , 
        solve_predicate(PreBehavioralConstraint,_,Solution) , 
        get_line_amount(PreLibrary,DistIn,PreM) , PreM1 is  PreM - 1 , 
        synthesis_from_examples(PreLibrary,LibExpansion,CurrentVars,any,guard,Input,[DistIn],PreBehavioralConstraint,PreLocationVars,PreM1,ParameterAmount,Solution,Guards) , 
        % collect so far synthesized guards
        (ValidInvariantsOrGuards = any 
        ->  NewValidInvariantsOrGuards = Guards
        ;   conjunction_to_list(ValidInvariantsOrGuards,GrdList) , 
            conjunct_previous_invariant_or_guard_if_neccessary(guard,GrdList,Guards,NewValidInvariantsOrGuards)) , 
        NewInput = Input , NewOutput = Output
    ;   NewInput = [DistIn|Input] , NewOutput = [DistOut|Output] , 
        NewValidInvariantsOrGuards = ValidInvariantsOrGuards) , 
    get_behavioral_constraint(Library,NewInput,NewOutput,DistBehavioralConstraint,DistLocationVars) , 
    solve_predicate(DistBehavioralConstraint,_,DistSolution) , 
    synthesis_from_examples(Library,LibExpansion,CurrentVars,NewValidInvariantsOrGuards,action,NewInput,NewOutput,DistBehavioralConstraint,DistLocationVars,M1,ParameterAmount,DistSolution,Synthesized).
% distinguishing input (guards, invariants)
synthesis_from_examples_aux3(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,PosInput,NegInput,_BehavioralConstraint,_LocationVars,M1,ParameterAmount,_Solution,_FurtherSolution,_NewBehavioralConstraint,Distinguishing,Synthesized) :- 
    Type \= action , 
    adjust_distinguishing_input(CurrentVars,Distinguishing,CompleteDist) , 
    findall(Value,(member(binding(Var,Temp,_),CompleteDist) , \+is_primed_id(Var) , value_to_ast(Temp,Value)),DistIn) , 
    user_interaction_distinguishing_input(Type,solution(CompleteDist),Answer) , 
    (Answer = yes
    ->  NewPosInput = [DistIn|PosInput] , 
        NewNegInput = NegInput
    ;   NewPosInput = PosInput , 
        NewNegInput = [DistIn|NegInput]) , 
    get_behavioral_constraint_from_input(Library,NewPosInput,NewNegInput,DistBehavioralConstraint,DistLocationVars) , 
    solve_predicate(DistBehavioralConstraint,_,DistSolution) , 
    synthesis_from_examples(Library,LibExpansion,CurrentVars,ValidInvariantsOrGuards,Type,NewPosInput,NewNegInput,DistBehavioralConstraint,DistLocationVars,M1,ParameterAmount,DistSolution,Synthesized).  
%%

print_synthesized_solution(_,none) :- 
    format("Synthesis failed. Maybe try again with other I/O-Examples or expand the library. ~n",[]).
print_synthesized_solution(Type,SolutionAST) :- 
    nl , format("~n~n~n~n~nSynthesized ~w: ~n",[Type]) , 
    print_bexpr_or_subst(SolutionAST).

find_further_solution(LocationVars,Solution,BehavioralConstraint,FurtherSolution,NewBehavioralConstraint) :- 
    get_new_behavioral_constraint(LocationVars,Solution,BehavioralConstraint,NewBehavioralConstraint) , 
    solve_predicate(NewBehavioralConstraint,_,TempSolution) , 
    (TempSolution = solution(_)
    ->  FurtherSolution = TempSolution 
    ;   FurtherSolution = none).

get_distinguishing_input(Type,ValidInvariantsOrGuards,AST1,AST2,Distinguishing) :- 
    b_get_invariant_from_machine(Invariant) , 
    preferences:set_preference(randomise_enumeration_order,true) , 
    % add WD-constraints for sequences
    find_typed_identifier_uses(AST1,AST1Nodes) , find_typed_identifier_uses(AST2,AST2Nodes) , 
    findall(Seq,(member(Seq,AST1Nodes) , is_seq(Seq)),Seq1) , 
    findall(Seq,(member(Seq,AST2Nodes) , is_seq(Seq)),Seq2) , 
    append(Seq1,Seq2,TSequences) , remove_dups(TSequences,Sequences) , 
    maplist(sequence_well_definedness_constraint,Sequences,SeqConstraintList) , 
    conjunct_predicates(SeqConstraintList,SeqConstraints) , 
    get_distinguishing_input_aux(Type,ValidInvariantsOrGuards,
                b(conjunct(SeqConstraints,b(conjunct(b(negation(b(equivalence(AST1,AST2),pred,[])),pred,[]),Invariant),pred,[])),pred,[]),Distinguishing) , 
    preferences:set_preference(randomise_enumeration_order,false).
% when synthesizing an action only search for a distinguishing input that doesn't violate the 
% so far synthesized guard (in case there is one)
get_distinguishing_input_aux(action,Guards,Constraint,Distinguishing) :- 
    Guards \= any , 
    ! , solve_predicate(b(conjunct(Constraint,Guards),pred,[]),_,Distinguishing).
get_distinguishing_input_aux(_,_,Constraint,Distinguishing) :- 
    solve_predicate(Constraint,_,Distinguishing).

sequence_well_definedness_constraint(SeqIDNode,WDConstraint) :- 
    % well-defined sequence: domain enumerated from 1..n
    WDConstraint = b(equal(b(domain(SeqIDNode),set(integer),[]),
        b(interval(b(integer(1),integer,[]),b(size(SeqIDNode),integer,[])),set(integer),[])),pred,[]).

% returns a list of all guards or 'any' if none is given
get_guards_from_operation(Operation,Guards) :- 
    b_get_machine_operation(Operation,_,_,OpBody) , 
    ( OpBody = b(precondition(TGrd,_),_,_)
    ; OpBody = b(rlevent(Operation,_,_,_,TGrd,_,_,_,_,_,_),_,_)
    ; OpBody = b(any(_,TGrd,_),subst,_)) , 
    ! , conjunction_to_list(TGrd,Guards).
% no given guards
get_guards_from_operation(_,any).

conjunct_previous_invariant_or_guard_if_neccessary(invariant,ValidInvariants,ProgramAST,SolutionAST) :- 
    ! , conjunct_predicates([ProgramAST|ValidInvariants],SolutionAST).
conjunct_previous_invariant_or_guard_if_neccessary(guard,Guards,ProgramAST,SolutionAST) :- 
    Guards \= any , ! , nl , 
    % search for valid guards that are already given by the machine ( NewPreCond => OldPreCond )
    find_typed_identifier_uses(ProgramAST,[],ProgIDs) , 
    findall(Pre,(member(Pre,Guards) , find_typed_identifier_uses(Pre,PreIDs) , append(ProgIDs,PreIDs,TIDs) , remove_dups(TIDs,IDs) , 
                 b_test_boolean_expression_wf(b(forall(IDs,ProgramAST,Pre),pred,[used_ids([])]),[],[])),InvalidPreConds) , 
    sets:subtract(Guards,InvalidPreConds,ValidPreConds) , 
    conjunct_predicates([ProgramAST|ValidPreConds],SolutionAST).
conjunct_previous_invariant_or_guard_if_neccessary(action,Guard,ProgramAST,SolutionAST) :- 
   \+member(Guard,[any,none]) ,  ! , 
   SolutionAST = b(precondition(Guard,ProgramAST),subst,[]).
conjunct_previous_invariant_or_guard_if_neccessary(_,_,ProgramAST,ProgramAST).

% replace equal by assign_single_id and conjunctions by parallel substitution
replace_preds_by_subst_if_neccessary(action,TSolutionAST,SolutionAST) :- 
    conjunction_to_list(TSolutionAST,EqualityList) , 
    maplist(equality_to_assignment,EqualityList,AssignmentList) , 
    SolutionAST = b(parallel(AssignmentList),subst,[]).
replace_preds_by_subst_if_neccessary(_,SolutionAST,SolutionAST).

equality_to_assignment(b(equal(Primed,Assignment),pred,[]),b(assign_single_id(Primed,Assignment),subst,[])).

% only for synthesis of invariants: 
% if we graduate [0]-->[-1] as valid and [-1]-->[-2] as invalid, 
% we don't want to assume [-1] to be a negative input from the second example
delete_positive_state_from_negative(PosInput,TempNegInput,NegInput) :- 
    findall(Duplicate,(member(Duplicate,PosInput) , member(Duplicate,TempNegInput)),Duplicates) , 
    subtract(TempNegInput,Duplicates,NegInput).    

% adjust the distinguishing input to the current variables
adjust_distinguishing_input(CurrentVars,Distinguishing,Distinguishing) :- 
    length(CurrentVars,VarAmount) , 
    length(Distinguishing,VarAmount) , !.
adjust_distinguishing_input(CurrentVars,Distinguishing,ShrunkenInput) :- 
    length(CurrentVars,VarAmount) , 
    length(Distinguishing,DistAmount) , 
    VarAmount < DistAmount , ! , 
    findall(Binding,(member(Binding,Distinguishing) , Binding = binding(VarName,_,_) , 
                     member(b(identifier(VarName),_,_),CurrentVars)),ShrunkenInput).

%complete_or_shrink_distinguishing_input(CurrentVars,Distinguishing,CompleteInput) :- 
%    findall(Name,member(b(identifier(Name),_,_),CurrentVars),VarNames) , 
%    complete_distinguishing_input(Distinguishing,VarNames,CompleteInput).

% add any matching variables to distinguishing input if neccessary 
% by comparing programs like c+d>1 and c > 0 we often get invalid one value inputs
complete_distinguishing_input([],[],[]).
complete_distinguishing_input([],[MachineVar|MT],[Binding|CompleteInput]) :- 
    any_binding_for_machine_var(MachineVar,Binding) , 
    complete_distinguishing_input([],MT,CompleteInput).
% skip primed ids
complete_distinguishing_input([binding(Var,Value,RawValue)|DT],MachineVars,[binding(Var,Value,RawValue)|CompleteInput]) :- 
    is_primed_id(Var) , ! , 
    complete_distinguishing_input(DT,MachineVars,CompleteInput).
complete_distinguishing_input([binding(Var,Value,RawValue)|DT],[Var|MT],[binding(Var,Value,RawValue)|CompleteInput]) :- 
    ! , complete_distinguishing_input(DT,MT,CompleteInput).
complete_distinguishing_input([binding(Var,Value,RawValue)|DT],[MachineVar|MT],[Binding|CompleteInput]) :- 
    any_binding_for_machine_var(MachineVar,Binding) , 
    complete_distinguishing_input([binding(Var,Value,RawValue)|DT],MT,CompleteInput).

% we can't pass anonymous variables for inputs that we don't care so just set any value
any_binding_for_machine_var(MachineVar,Binding) :- 
    b_get_machine_variables(Vars) , 
    member(b(identifier(MachineVar),Type,_),Vars) , 
    ( Type = integer
    ->  Binding = binding(MachineVar,int(0),'0')
    ; %Type = boolean
        Binding = binding(MachineVar,pred_true,'TRUE')).

% assume order that the first input is the one whose state violated the invariant
get_violated_and_valid_machine_invariants(_Operation,Input,ViolatedInvariants,ViolatingVars,ValidInvariants) :- 
    % according to specific operation
    b_get_machine_variables(TypedIDs) , 
    b_get_invariant_from_machine(MachineInvariant) , 
    conjunction_to_list(MachineInvariant,InvariantList) , 
    Input = [In|_] , 
    get_violated_and_valid_machine_invariants_aux(TypedIDs,In,InvariantList,TempViolatedInvariants) , 
    remove_dups(TempViolatedInvariants,ViolatedInvariants) , 
    conjunct_predicates(ViolatedInvariants,TempPred) , 
    constraints:find_typed_identifier_uses_with_info(TempPred,ViolatingVars) , 
    subtract(InvariantList,ViolatedInvariants,ValidInvariants).

get_violated_and_valid_machine_invariants_aux(TypedIDs,Input,InvariantList,ViolatedInvariants) :- 
    findall(Equality,(member(IDNode,TypedIDs) , member(ValueNode,Input) , 
                      nth0(I,TypedIDs,IDNode) , nth0(I,Input,ValueNode) , 
                      Equality = b(equal(IDNode,ValueNode),pred,[])),EqualityList) , 
    conjunct_predicates(EqualityList,EqualityNodes) , 
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,InvariantList,ViolatedInvariants).

get_violated_and_valid_machine_invariants_aux2(_,[],[]).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[SingleInv|T],[SingleInv|NT]) :- 
    conjunct_predicates([EqualityNodes,SingleInv],InvPred) , 
    find_identifier_uses(InvPred,[],UsedIDs) , 
    findall(Binding,(member(ID,UsedIDs) , Binding = bind(ID,_)),Bindings) , 
    \+ b_test_boolean_expression_wf(InvPred,Bindings,[]) , ! , 
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[_|T],NT) :- 
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).

% TODO: support set-expressions
% get another transition that violates the invariant, exclude given input/output example
get_invariant_violating_transition(Operation,UsedVars,ViolatedInvariants,Input,Output,Counter) :- 
    b_get_machine_operation(Operation,_,_,TOpBody) , 
    % don't create contradiction in case the guard is equivalent
    % to an invariant which we will negate in the following
    remove_pre_condition_from_body(TOpBody,OpBody) , 
    before_after_predicate_with_equalities(OpBody,UsedVars,BABody) , 
    %find_identifier_uses(OpBody,[],BodyIDs) , 
    findall(V,(member(b(identifier(V),_,_),UsedVars)),BodyIDs) , 
    b_get_machine_variables(MachineVars) , 
    create_equality_nodes_for_variables(BodyIDs,Input,Output,MachineVars,TNodeList) , 
    % negate equalities to exclude this transition
    findall(Inequality,(create_texpr(negation(Equality),pred,[],Inequality) , member(Equality,TNodeList)),NodeList) , 
    
    conjunct_predicates(ViolatedInvariants,Invariant) , 
    prime_or_unprime_variables_in_predicate(prime,Invariant,PrimedInvariant) , 
    create_texpr(negation(Invariant),pred,[],NInvariant) , 
    create_texpr(negation(PrimedInvariant),pred,[],NPrimedInvariant) , 
    conjunct_predicates([BABody,NInvariant,NPrimedInvariant|NodeList],BAExclusion) , 
    solve_predicate(BAExclusion,_,Solution) , 
    get_invariant_violating_transition_aux(Solution,Counter).

get_invariant_violating_transition_aux(Solution,none) :- 
    Solution \= solution(_).
get_invariant_violating_transition_aux(solution(Solution),(CounterInput,CounterOutput)) :- 
    % split by input and output state
    findall(Binding,(member(Binding,Solution) , Binding = binding(Var,_,_) , \+is_primed_id(Var)),Bindings) , 
    findall(Binding,(member(Binding,Solution) , Binding = binding(Var,_,_) , is_primed_id(Var)),PrimedBindings) , 
    example_bindings_to_ast_nodes(Bindings,CounterInput) , 
    example_bindings_to_ast_nodes(PrimedBindings,CounterOutput).

% rlevent/11
remove_pre_condition_from_body(b(Node,_,_),Body) :- 
    Node =.. [_,_,_,_,_,_,_,Actions|_] , 
    create_texpr(parallel(Actions),subst,[],OpBody) , 
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(b(any(_,_,TActions),subst,_),Body) :- 
    (is_list(TActions) -> Actions = TActions ; Actions = [TActions]) , 
    create_texpr(parallel(Actions),subst,[],OpBody) , 
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(PreBody,Body) :- 
    PreBody = b(precondition(_,Body),_,_).
remove_pre_condition_from_body(Body,Body).

% prime/unprime all identifiers within an ast
prime_or_unprime_variables_in_predicate(prime,b(identifier(ID),Type,Info),PrimedID) :- ! , 
    prime_variable(b(identifier(ID),Type,Info),PrimedID).
prime_or_unprime_variables_in_predicate(unprime,b(identifier(ID),Type,Info),PrimedID) :- ! , 
    (is_primed_id(ID)
    ->  user_interaction:remove_prime(b(identifier(ID),Type,Info),PrimedID)
    ;   PrimedID = b(identifier(ID),Type,Info)).
prime_or_unprime_variables_in_predicate(_,b(Node,Type,Info),b(Node,Type,Info)) :- 
    (Node = integer(_) ; Node = integer_set(_) ; Node = string(_) ; Node =.. [_]).
prime_or_unprime_variables_in_predicate(PrimeType,b(Node,Type,Info),b(PrimedNode,Type,Info)) :- 
    Node =.. [Op,In] , 
    prime_or_unprime_variables_in_predicate(PrimeType,In,PrimedIn) , 
    PrimedNode =.. [Op,PrimedIn].
prime_or_unprime_variables_in_predicate(PrimeType,b(Node,Type,Info),b(PrimedNode,Type,Info)) :- 
    Node =.. [Op,In1,In2] , 
    prime_or_unprime_variables_in_predicate(PrimeType,In1,PrimedIn1) , 
    prime_or_unprime_variables_in_predicate(PrimeType,In2,PrimedIn2) , 
    PrimedNode =.. [Op,PrimedIn1,PrimedIn2].

% get valid states for synthesis of invariants
get_valid_states(C,ValidStates) :- 
    b_get_invariant_from_machine(Invariant) , 
    % random enumeration to find more examples describing different behavior 
    preferences:set_preference(randomise_enumeration_order,true) , 
    solve_predicate(Invariant,_,solution(Solution)) ,  
    get_valid_states(C,Solution,Invariant,ValidStates) , 
    preferences:set_preference(randomise_enumeration_order,false).
get_valid_states(1,Solution,_,[ValidState]) :- 
    example_bindings_to_ast_nodes(Solution,ValidState).
get_valid_states(C,Solution,Invariant,[ValidState|NT]) :- 
    C1 is C - 1 , 
    b_get_machine_variables(MachineVars) , 
    findall(Exclusion,(member(Var,MachineVars) , Var = b(identifier(Name),_,_) , member(binding(Name,Value,_),Solution) , 
                       value_to_ast(Value,ValueNode) , Exclusion = b(not_equal(Var,ValueNode),pred,[])),ExclusionList) , 
    conjunct_predicates([Invariant|ExclusionList],NewPred) , 
    solve_predicate(NewPred,_,solution(NewSolution)) , 
    example_bindings_to_ast_nodes(Solution,ValidState) , 
    get_valid_states(C1,NewSolution,NewPred,NT).              

% InputList contains all machine variables, but for synthesizing an invariant 
% we only need to observe variables from violated invariants
reduce_input_to_violating_vars(ViolatingVars,InputExample,ReducedExample) :-  
    % find values of matching variable indices
    findall(Value,(member(Value,InputExample) , Value = b(_,_,Info) , member(synthesis(machinevar,VarName),Info) , 
                   member(VarName,ViolatingVars)),ReducedExample).

%% user adds a missing state transition 
% search for a machine operation that completely supports the missing state 
% or one where we just have to synthesize another guard
% VarNames refers to the variables used in the added state
check_transition_against_machine_operations(VarNodes,PosInput,PosOutput,Result) :- 
    findall((Op,Res,Par,Body),b_get_machine_operation(Op,Res,Par,Body),Operations) , 
    check_transition_against_machine_operations(Operations,PosInput,PosOutput,VarNodes,Result).
check_transition_against_machine_operations([],_,_,_,no_match).
% skip if the desired action's and the given operation's variables doesn't match
check_transition_against_machine_operations([(_,_,_,OpBody)|T],PosInput,PosOutput,VarNodes,Result) :- 
    find_identifier_uses(OpBody,[],BodyIDs) , 
    findall(Name,member(b(identifier(Name),_,_),VarNodes),VarNames) , 
    % find distinguishing variables 
    findall(DistName,(member(DistName,BodyIDs) , \+member(DistName,VarNames)),DistNames) , 
    DistNames \= [] , ! , 
    check_transition_against_machine_operations(T,PosInput,PosOutput,VarNodes,Result).
check_transition_against_machine_operations([(OpName,_,_,OpBody)|T],PosInput,PosOutput,VarNodes,Result) :- 
    b_get_machine_variables(MachineVars) , 
    before_after_predicate_with_equalities(OpBody,MachineVars,BABody) , 
    find_identifier_uses(OpBody,[],BodyIDs) , 
    create_equality_nodes_for_variables(BodyIDs,PosInput,PosOutput,VarNodes,NodeList) , 
    conjunct_predicates([BABody|NodeList],BABodyPred) , 
    % double length because of primed variables
    length(VarNodes,Length) , NLength is Length * 2 , length(BindList,NLength) , 
    findall(E,(member(E,BindList) , E = bind(_,_)),BindList) , 
    check_transition_against_machine_operations_aux(BABodyPred,BindList,NodeList,[(OpName,_,_,OpBody)|T],PosInput,PosOutput,VarNodes,Result).

% found operation that already executes the transition
check_transition_against_machine_operations_aux(BABodyPred,BindList,_,[(OpName,_,_,_)|_],_,_,_,(OpName,done)) :- 
    b_test_boolean_expression_wf(BABodyPred,BindList,[]) , !.
% only guard of an operation doesn't fit
check_transition_against_machine_operations_aux(_,BindList,NodeList,[(OpName,_,_,OpBody)|_],_,_,_,(OpName,guard)) :- 
    b_get_machine_variables(MachineVars) , 
    % remove guard and try if action itself supports the missing state
    OpBody = b(precondition(_PreCondition,Action),_,_) ,  
    before_after_predicate_with_equalities(Action,MachineVars,BAAction) , 
    conjunct_predicates([BAAction|NodeList],BAActionPred) , 
    b_test_boolean_expression_wf(BAActionPred,BindList,[]) , !.
% action also doesn't match so search for another operation
check_transition_against_machine_operations_aux(_,_,_,[_|T],PosInput,PosOutput,VarNodes,Result) :- 
    check_transition_against_machine_operations(T,PosInput,PosOutput,VarNodes,Result). 

% we need to set variables to values from example by creating equality nodes 
% that we can add to the before after predicate
create_equality_nodes_for_variables([],_,_,_,[]).
create_equality_nodes_for_variables([VarName|T],Input,Output,VarNodes,[Node1,Node2|NodeList]) :- 
    member(VarNode,VarNodes) , 
    VarNode = b(identifier(VarName),_,_) , 
    member(InputNode,Input) , InputNode = b(_,_,InfoIn) , member(synthesis(machinevar,VarName),InfoIn) , 
    member(OutputNode,Output) , OutputNode = b(_,_,InfoOut) , member(synthesis(machinevar,VarName),InfoOut) , 
    Node1 = b(equal(VarNode,InputNode),pred,[]) , 
    prime_variable(VarNode,PrimedVarNode) , 
    Node2 = b(equal(PrimedVarNode,OutputNode),pred,[]) , 
    create_equality_nodes_for_variables(T,Input,Output,VarNodes,NodeList).
%%

% get valid inputs from visited expressions during model checking 
get_valid_examples_for_operation(Amount,OpName,ValidExamples) :- 
    findall(SFrom,(transition(From,OpName,_To) , visited_expression(From,SFrom)),Exprs) , 
    maplist(get_pretty_value_list,Exprs,TempValidExamples) , 
    % randomized selection of examples 
    get_n_random_elements(Amount,TempValidExamples,ValidExamples). 

get_n_random_elements(_,[],[]).
get_n_random_elements(1,[H|_],[H]).
get_n_random_elements(C,List,[E|Elements]) :- 
    C1 is C - 1 , 
    random:random_select(E,List,Rest) , 
    get_n_random_elements(C1,Rest,Elements).

get_pretty_value_list(L,PL) :- 
    findall(Value,member(bind(_,Value),L),Values) , 
    maplist(translate_bvalue,Values,PL).

% split examples from model checking (check if invariant violated or valid) to 
% give assumptions to the user
pre_select_transitions(Operation,ExamplesAssumingYes,ExamplesAssumingNo) :- 
    findall((From,To),
        (transition(From,Operation,_,To) , \+ invariant_not_yet_checked(To)),Transitions) , 
    findall((SF,ST),
        (member((F,T),Transitions) , visited_expression(F,SF) , visited_expression(T,ST) , 
            invariant_violated(T)),ExamplesAssumingNo) , 
    findall((SF,ST),
        (member((F,T),Transitions) , visited_expression(F,SF) , visited_expression(T,ST) , 
            \+invariant_violated(T)),TExamplesAssumingYes), ! , 
    % take at most 5 positive examples
    length(TExamplesAssumingYes,L) , 
    (L > 5
    ->  sublist(TExamplesAssumingYes,ExamplesAssumingYes,_,5,_)
    ;   ExamplesAssumingYes = TExamplesAssumingYes).