:- module(constraints,[get_behavioral_constraint_from_input/5, 
                       get_behavioral_constraint/5, 
                       get_behavioral_constraint_aux/6, 
                       get_new_behavioral_constraint/4, 
                       get_line_amount/3, 
                       delete_numbers_from_atom/2, 
                       find_typed_identifier_uses_with_info/2,
                       value_to_ast/2]). 

:- use_module(library(lists)). 
:- use_module(library(avl),[list_to_avl/2,avl_to_list/2]). 
:- use_module(probsrc(translate)).
:- use_module(probsrc(bsyntaxtree)). 
:- use_module(probsrc(tools),[flatten/2]).

% exclude a solution and add this node to the behavioral constraint 
get_new_behavioral_constraint(LocationVars,solution(Solution),BehavioralConstraint,NewBehavioralConstraint) :- 
    exclude_solutions(LocationVars,Solution,StatePredicate) , 
    create_negation(StatePredicate,NegStatePredicate) , 
    conjunct_predicates([BehavioralConstraint,NegStatePredicate],NewBehavioralConstraint). 

exclude_solutions(LocationVars,Solution,Exclusion) :- 
    findall(Equal,(member(ID,LocationVars) , get_texpr_id(ID,IDName) , 
                   member(binding(IDName,Value,_),Solution) , Equal = b(equal(ID,b(value(Value),integer,[])),pred,[])),EqualList) , 
    conjunct_predicates(EqualList,Exclusion).

get_behavioral_constraint_aux(action,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :- 
    get_behavioral_constraint(Library,PosInput,PosOutput,BehavioralConstraint,LocationVars).
get_behavioral_constraint_aux(_,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :- 
    get_behavioral_constraint_from_input(Library,PosInput,PosOutput,BehavioralConstraint,LocationVars).
    
% create outputs and get behavioral constraint 
get_behavioral_constraint_from_input(Library,PosInput,NegInput,BehavioralConstraint,LocationVars) :- 
    % get output values
    length(PosInput,PosLength) , length(NegInput,NegLength) , 
    % create empty lists 
    length(TempPosOutput,PosLength) , length(TempNegOutput,NegLength) , 
    % for our task to repair invariant violations output is either "true" or "false"
    findall(In,(member(In,TempPosOutput) , In = ['TRUE']),PosOutput) , 
    findall(In,(member(In,TempNegOutput) , In = ['FALSE']),NegOutput) , 
    append(PosInput,NegInput,Input) , 
    append(PosOutput,NegOutput,Output) , 
    get_behavioral_constraint(Library,Input,Output,BehavioralConstraint,LocationVars).

% get the functional constraint for each Input-Output Example 
% it is sufficient to create one well-definedness constraint surrounding the behavioral constraint 
% but the library constraint needs to be defined for every examples functional constraint 
get_behavioral_constraint(Library,InputExamples,OutputExamples,BehavioralConstraint,LocationVars) :- 
    % lists of currently used components, necessary if a library is used several times
    assert(usedOperators([])) , assert(usedArithmetic([])) , 
    % create library constraint separated from the funtional constraint 
    % to make sure to just obtain one set of location variables 
    get_library_constraint(Library,LibConstraint,InputVars,OutputVars) , 
    % set the temporary location variables with prefix "l" 
    maplist(get_location_var_from_ast_node,InputVars,LocationVarsIn) , 
    maplist(get_location_var_from_ast_node,OutputVars,LocationVarsOut) , 
    % get some information needed for well-definedness constraint 
    InputExamples = [InputValues|_] , 
    get_line_amount(Library,InputValues,M) , 
    length(InputValues,VarAmount) , 
    % create well-definedness constraint separated from function constraints 
    get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) , 
    get_range_of_integer_nodes(InputExamples,InMin,InMax) , 
    maplist(get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars),InputExamples,OutputExamples,FunctionConstraints) , 
    conjunct_predicates([WellDefinednessConstraint|FunctionConstraints],BehavioralConstraint) , 
    append(InputVars,OutputVars,Vars) , 
    maplist(get_location_var_from_ast_node,Vars,LocationVars) , 
    retract(usedOperators(_)) , retract(usedArithmetic(_)). 

get_range_of_integer_nodes(InputExamples,InMin,InMax) :- 
    % get range of used integers to support the constraint solver
    flatten(InputExamples,InputSet) , 
    findall(Num,member(b(integer(Num),integer,_),InputSet),IntegerNodes) , 
    IntegerNodes \= [] , ! , 
    min_member(TInMin,IntegerNodes) , max_member(TInMax,IntegerNodes) , 
    InMin is TInMin - 10 , InMax is TInMax + 10.
get_range_of_integer_nodes(_,0,1).

get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars,InputValues,OutputValues,FunctionConstraint) :-  
    get_line_amount(Library,InputValues,M) , 
    % enumerate input/output values
    transition_values_to_ast(InputValues,input,InputValueNodes) , 
    transition_values_to_ast(OutputValues,output,OutputValueNodes) , 
    get_connectivity_constraint(InputValueNodes,OutputValueNodes,InputVars,OutputVars,ConnectivityConstraint) , 
    set_input_location_vars(InputValueNodes,InputLocationVars) , 
    set_output_location_vars(OutputValueNodes,M,OutputLocationVars) , 
    conjunct_predicates([OutputLocationVars,InputLocationVars,LibConstraint,ConnectivityConstraint],TempFunctionConstraint) , 
    % get all used identifier as ast nodes
    find_typed_identifier_uses_with_info(TempFunctionConstraint,ListOfIDNodes) , 
    % all temporary library variables in exists node except 'constant'
    findall(ID,(member(ID,ListOfIDNodes) , get_texpr_id(ID,Name) ,  
                atom_codes(Name,[Char|_]) , Char \= 108 , 
                \+atom_concat(constant,_,Name)),TTemporaryIDs) , 
    remove_dup_id_nodes(TTemporaryIDs,TemporaryIDs) , 
    b_ast_cleanup:find_identifier_uses(TempFunctionConstraint,[],TUsedIDs) , 
    remove_dups(TUsedIDs,UsedIDs) , 
    % create upper and lower bound for input values (kodkod warning)
    % expand interval
    Min is InMin - 5*abs(InMin) , 
    Max is InMax + 5*abs(InMax) , 
    create_texpr(integer(Min),integer,[],LowerBound) , 
    create_texpr(integer(Max),integer,[],UpperBound) , 
    findall(InputRestriction,(member(In,InputVars) , In = b(_,integer,_) , InputRestriction = b(member(
                In,b(interval(LowerBound,UpperBound),set(integer),[])),pred,[])),InputRestrictions) , 
    conjunct_predicates(InputRestrictions,InRestrictions) , 

    %findall(Restr,(member(Node,TemporaryIDs) , Node = b(identifier(Name),set(_),_) , atom_codes(Name,[Char|_]) , Char \= 108 ,  Restr = b(member(Node,b(pow_subset(b(integer_set('NATURAL'),set(integer),[])),set(set(integer)),[])),pred,[])),TSetIDRestrs) , 
    %conjunct_predicates(TSetIDRestrs,SetIDRestrs) , 

    conjunct_predicates([TempFunctionConstraint,InRestrictions],ExistsBody) , 
    FunctionConstraint = b(exists(TemporaryIDs,ExistsBody),pred,[used_ids(UsedIDs)]).

% we have duplicate nodes because of undefined types of sets
remove_dup_id_nodes(Nodes,Distinct) :- 
    remove_dup_id_nodes(Nodes,[],Distinct).
remove_dup_id_nodes([],Acc,Acc).
remove_dup_id_nodes([Node|T],Acc,Distinct) :- 
    Node = b(identifier(Name),_,_) , 
    \+member(b(identifier(Name),_,_),Acc) , ! , 
    remove_dup_id_nodes(T,[Node|Acc],Distinct).
remove_dup_id_nodes([_|T],Acc,Distinct) :- 
    remove_dup_id_nodes(T,Acc,Distinct).

% define connections of temporary inputs and outputs
get_connectivity_constraint(InputExampleNodes,OutputExampleNodes,InputVars,OutputVars,ConnectivityConstraint) :-  
    append([InputVars,OutputVars,InputExampleNodes,OutputExampleNodes],SetOfVars) , 
    findall((X,Y),
        (member(X,SetOfVars) , member(Y,SetOfVars) , 
         X \= Y , nth0(I1,SetOfVars,X) , nth0(I2,SetOfVars,Y) , I1 < I2),ListOfVarPairs) , 
    maplist(get_single_connectivity_constraint,ListOfVarPairs,ConstraintList) , 
    conjunct_predicates(ConstraintList,ConnectivityConstraint).

get_single_connectivity_constraint((Var1,Var2),ConnectivityConstraint) :- 
    get_location_var_from_ast_node(Var1,LocationVar1) , 
    get_location_var_from_ast_node(Var2,LocationVar2) , 
    % only connect two variables if they have the same type  
    (valid_connection(Var1,Var2) 
    ->  LHS = b(equal(LocationVar1,LocationVar2),pred,[]) , 
        RHS = b(equal(Var1,Var2),pred,[]) , 
        create_implication(LHS,RHS,ConnectivityConstraint)
    % prevent equality of such incompatible location variables 
    ;   ConnectivityConstraint = b(not_equal(LocationVar1,LocationVar2),pred,[])).  

valid_connection(Var1,Var2) :- 
    get_texpr_info(Var1,Info1) , 
    get_texpr_info(Var2,Info2) , 
    % same type
    valid_type_connection(Var1,Var2) , 
    % no predicates within boolean operators (boolean_true /= pred_true), i.e. avoid (c>0) = d 
    valid_connection_aux(Info1,Info2).

valid_connection_aux(Info1,Info2) :- 
    member(synthesis(noPred),Info1) , 
    (member(synthesis(input(_)),Info2) ;  
    (member(synthesis(Comp2,_),Info2) , atom_concat(constant,_,Comp2))).
valid_connection_aux(Info1,Info2) :- 
    member(synthesis(noPred),Info2) , 
    (member(synthesis(input(_)),Info1) ; 
    (member(synthesis(Comp1,_),Info1) , atom_concat(constant,_,Comp1))) .
valid_connection_aux(Info1,Info2) :- 
    member(synthesis(noInput),Info1) , 
    \+member(synthesis(input(_)),Info2). 
valid_connection_aux(Info1,Info2) :- 
    member(synthesis(noInput),Info2) ,  
    \+member(synthesis(input(_)),Info1).
valid_connection_aux(Info1,Info2) :- 
    (\+member(synthesis(noPred),Info1) , \+member(synthesis(noPred),Info2)).

valid_type_connection(b(_,Type,_),b(_,Type,_)). 
valid_type_connection(b(_,Type1,_),b(_,Type2,_)) :- 
    (Type1 = seq(InnerType) ; Type1 = set(couple(integer,InnerType))) , 
    (Type2 = seq(InnerType) ; Type2 = set(couple(integer,InnerType))).

get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) :- 
    get_consistency_constraint(LocationVarsOut,ConsistencyConstraint) , 
    get_acyclic_constraint(LocationVarsIn,LocationVarsOut,AcyclicConstraint) , 
    NewM is M - 1 , % minus one to use an interval
    get_location_constraint(LocationVarsIn,0,NewM,InputRestriction) , 
    get_location_constraint(LocationVarsOut,VarAmount,NewM,OutputRestriction) , 
    conjunct_predicates([InputRestriction,OutputRestriction,ConsistencyConstraint,AcyclicConstraint],WellDefinednessConstraint).

% create restriction nodes (membership to an interval) to the domain 
% of each location variable
get_location_constraint(LocationVars,A,B,RestrictionNode) :- 
    maplist(get_single_location_constraint(A,B),LocationVars,RestrictionList) , 
    conjunct_predicates(RestrictionList,RestrictionNode).
get_single_location_constraint(A,B,LocationVar,RestrictionNode) :- 
    RestrictionNode = b(member(LocationVar,
        b(interval(b(integer(A),integer,[]),b(integer(B),integer,[])),set(integer),[])),pred,[]).

% consistency constraint guarantees that there is only one output in each line
get_consistency_constraint(LocationVarsOut,Constraint) :- 
    % find all pairs of location variables (not redundant)
    % and create "not_equal"-Node for each pair
    findall(InequalityNode,
        (member(L1,LocationVarsOut) , member(L2,LocationVarsOut) , 
         L1 \= L2 , nth0(I1,LocationVarsOut,L1) , nth0(I2,LocationVarsOut,L2) , 
         I1 < I2 , create_texpr(not_equal(L1,L2),pred,[],InequalityNode)),InequalityList) , 
    conjunct_predicates(InequalityList,Constraint).

% acyclic constraint guarantees that every operators input is defined before 
% its corresponding output
get_acyclic_constraint(LocationVarsIn,LocationVarsOut,Constraint) :- 
    get_acyclic_constraint_list(LocationVarsIn,LocationVarsOut,ConstraintList) , 
    conjunct_predicates(ConstraintList,Constraint).
get_acyclic_constraint_list(_,[],[]).  
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :- 
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) , 
    delete_numbers_from_atom(ComponentID,Component) , 
    % constant has no input 
    \+member(Component,[constant,negation]) ,  
    member(In1,LocationVarsIn) , 
    In1 = b(_,_,Info1) , 
    member(synthesis(ComponentID,left_input),Info1) , ! , 
    member(In2,LocationVarsIn) , 
    In2 = b(_,_,Info2) ,
    member(synthesis(ComponentID,right_input),Info2) ,  
    create_texpr(less(In1,Out),pred,[],Node1) , 
    create_texpr(less(In2,Out),pred,[],Node2) , 
    conjunct_predicates([Node1,Node2],Constraint) , 
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :- 
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) , 
    delete_numbers_from_atom(ComponentID,Component) , 
    % negation just one input 
    Component = negation , ! , 
    member(In,LocationVarsIn) , 
    In = b(_,_,Info) , 
    member(synthesis(ComponentID,input),Info) ,  
    create_texpr(less(In,Out),pred,[],Constraint) , 
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
% skip constants 
get_acyclic_constraint_list(LocationVarsIn,[_|ROut],ConstraintList) :- 
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).

% construct the library constraint by encoding the given library components as logical formulas
get_library_constraint((Operators,Arithmetic),Constraint,InputVars,OutputVars) :- 
    maplist(get_single_operator_constraint,Operators,ConstraintOpList) , 
    maplist(get_single_arithmetic_constraint,Arithmetic,ConstraintArithList) , 
    append(ConstraintOpList,ConstraintArithList,ConstraintList) , 
    conjunct_predicates(ConstraintList,TConstraint) , 
    find_typed_identifier_uses_with_info(TConstraint,ListOfIDNodes) , 
    % split used identifier by input and output
    findall(I,(member(I,ListOfIDNodes) , get_texpr_id(I,IName) , atom_concat(i,_Suffix,IName)),InputVars) , 
    findall(O,(member(O,ListOfIDNodes) , get_texpr_id(O,OName) , atom_concat(o,_Suffix,OName)),OutputVars) , 
    get_wd_constraints(ListOfIDNodes,[],WDConstraintList) , 
    conjunct_predicates([TConstraint|WDConstraintList],Constraint).

% create well-definedness constraints to prevent division by zero or max/min of empty sets
get_wd_constraints([],Acc,Acc).
get_wd_constraints([b(Node,Type,Info)|T],Acc,WDConstraintList) :- 
    member(synthesis(OpName,_),Info) , 
    delete_numbers_from_atom(OpName,div) , ! , 
    get_wd_constraints(T,[b(not_equal(b(Node,Type,Info),b(integer(0),integer,[])),pred,[])|Acc],WDConstraintList).
get_wd_constraints([b(Node,Type,Info)|T],Acc,WDConstraintList) :- 
    member(synthesis(OpName,_),Info) , 
    delete_numbers_from_atom(OpName,Op) , 
    (Op = min ; Op = max) , ! , 
    get_wd_constraints(T,[b(greater(b(card(b(Node,Type,Info)),integer,[]),b(integer(0),integer,[])),pred,[])|Acc],WDConstraintList).
get_wd_constraints([_|T],Acc,WDConstraintList) :- 
    get_wd_constraints(T,Acc,WDConstraintList).

% library components represented as logical formulas
% (BOOLEAN,BOOLEAN) :: BOOLEAN
get_single_operator_constraint(Operator,Constraint) :- 
    member(Operator,[conjunct,disjunct]) , ! , 
    unique_typed_id("i",boolean,TempIn1) , 
    unique_typed_id("i",boolean,TempIn2) , 
    unique_typed_id("o",boolean,TempOut) , 
    get_component_id(op,Operator,OpID) , 
    atom_number_concat(Operator,OpID,OperatorName) , 
    % set information of identifier nodes 
    add_texpr_infos(TempIn1,[synthesis(OperatorName,left_input)],In1) , 
    add_texpr_infos(TempIn2,[synthesis(OperatorName,right_input)],In2) , 
    add_texpr_infos(TempOut,[synthesis(OperatorName,output)],Out) , 
    get_single_operator_constraint_aux2(Operator,In1,In2,Out,Truth,Falsity) , 
    disjunct_predicates([Truth,Falsity],Constraint).
% (BOOLEAN :: BOOLEAN)
get_single_operator_constraint(negation,Constraint) :- ! , 
    unique_typed_id("i",boolean,TempIn) , 
    unique_typed_id("o",boolean,TempOut) ,  
    get_component_id(op,negation,OpID) , 
    atom_number_concat(negation,OpID,Negation) , 
    % set information of identifier nodes 
    add_texpr_infos(TempIn,[synthesis(Negation,input),synthesis(noInput)],In) , 
    add_texpr_infos(TempOut,[synthesis(Negation,output)],Out) , 
    Truth = b(conjunct(b(equal(In,b(boolean_false,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) , 
    Falsity = b(conjunct(b(equal(In,b(boolean_true,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]) , 
    disjunct_predicates([Truth,Falsity],Constraint).
% constant :: INTEGER || constant :: BOOLEAN || constant :: SET
get_single_operator_constraint(Operator,Constraint) :- 
    get_single_operator_constraint_aux1(Operator,TempOut,Constant) , ! , 
    get_texpr_id(Constant,ConstantName) , 
    % set information of identifier node
    add_texpr_infos(TempOut,[synthesis(ConstantName,output)],Out) , 
    Constraint = b(equal(Constant,Out),pred,[]).
% set-expressions 
% SET :: INTEGER
get_single_operator_constraint(Operator-SetType,Constraint) :- 
    member(Operator,[card,min,max]) , 
    unique_typed_id("i",SetType,TempIn) , 
    unique_typed_id("o",integer,TempOut) , 
    get_component_id(op,Operator,OpID) , 
    atom_number_concat(Operator,OpID,NOperator) , 
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) , 
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) , 
    Node =.. [Operator,In] , 
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).
% (SET,SET) :: SET
get_single_operator_constraint(Operator-SetType,Constraint) :- 
    member(Operator,[union,intersection,concat,insert_front,insert_tail,restrict_front,restrict_tail,overwrite,set_subtraction,cartesian_product,image,domain_restriction,domain_subtraction,range_restriction,range_subtraction]) , 
    get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) , 
    get_component_id(op,Operator,OpID) , 
    atom_number_concat(Operator,OpID,NOperator) , 
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input)],In1) , 
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input)],In2) , 
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) , 
    Node =.. [Operator,In1,In2] , 
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_operator_constraint(Operator-SetType,Constraint) :- 
    member(Operator,[first,last,front,tail,general_union,general_intersection,reverse,range,domain]) , 
    get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) , 
    get_component_id(op,Operator,OpID) , 
    atom_number_concat(Operator,OpID,NOperator) , 
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) , 
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) , 
    Node =.. [Operator,In] , 
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_operator_constraint(size-SetType,Constraint) :- 
    get_fn_types(SetType,integer,Type) , 
    unique_typed_id("i",seq(Type),TempIn) , 
    unique_typed_id("o",integer,TempOut) , 
    get_component_id(op,size,OpID) , 
    atom_number_concat(size,OpID,NOperator) , 
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) , 
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) , 
    Constraint = b(equal(b(size(In),integer,[]),Out),pred,[]).

% (INTEGER,INTEGER) :: BOOLEAN || (BOOLEAN,BOOLEAN) :: BOOLEAN || (SET,SET) :: BOOLEAN
get_single_operator_constraint(Operator,Constraint) :- 
    get_single_operator_constraint_aux2(Operator,TempIn1,TempIn2,TempOut,Op,FurtherInfo) , 
    get_component_id(op,Op,OpID) , 
    atom_number_concat(Op,OpID,NOperator) , 
    % set information of identifier nodes 
    % further info 'noPred' to avoid predicates within boolean equality node later on 
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input)|FurtherInfo],In1) , 
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input)|FurtherInfo],In2) , 
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) , 
    TempTrueNode =.. [Op,In1,In2] , 
    TrueNode = b(TempTrueNode,pred,[]) , 
    BoolNodeTrue = b(equal(Out,b(boolean_true,boolean,[])),pred,[]) , 
    inverse_operator(Op,InverseOp) , 
    TempFalseNode =.. [InverseOp,In1,In2] , 
    FalseNode = b(TempFalseNode,pred,[]) , 
    BoolNodeFalse = b(equal(Out,b(boolean_false,boolean,[])),pred,[]) , 
    create_equivalence(TrueNode,BoolNodeTrue,Truth) , 
    create_equivalence(FalseNode,BoolNodeFalse,Falsity) , 
    conjunct_predicates([Truth,Falsity],Constraint).

get_single_operator_constraint_aux1(constant-Type,TempOut,Constant) :- 
    unique_typed_id("o",Type,TempOut) , 
    unique_typed_id("constant",Type,Constant).
% operators 
get_single_operator_constraint_aux2(equal-Type,TempIn1,TempIn2,TempOut,equal,Info) :- 
    unique_typed_id("i",Type,TempIn1) , 
    unique_typed_id("i",Type,TempIn2) , 
    unique_typed_id("o",boolean,TempOut) , 
    (Type = boolean 
    ->  Info = [synthesis(noPred)]
    ;   Info = []).
get_single_operator_constraint_aux2(conjunct,In1,In2,Out,Truth,Falsity) :- 
    Truth = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) , 
    Falsity = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_single_operator_constraint_aux2(disjunct,In1,In2,Out,Truth,Falsity) :- 
    Truth = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) , 
    Falsity = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_single_operator_constraint_aux2(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[subset,subset_strict]) , ! , 
    unique_typed_id("i",SetType,TempIn1) , 
    unique_typed_id("i",SetType,TempIn2) , 
    unique_typed_id("o",boolean,TempOut).
get_single_operator_constraint_aux2(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[member,not_member]) , ! , 
    (SetType = set(Type) ; SetType = seq(Type)) , 
    unique_typed_id("i",Type,TempIn1) , 
    unique_typed_id("i",SetType,TempIn2) , 
    unique_typed_id("o",boolean,TempOut).
get_single_operator_constraint_aux2(Operator,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    unique_typed_id("i",integer,TempIn1) , 
    unique_typed_id("i",integer,TempIn2) , 
    unique_typed_id("o",boolean,TempOut).
% set- and seq-expressions
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :- 
    member(Operator,[general_union,general_intersection]) , 
    unique_typed_id("i",SetType,TempIn) , 
    SetType = set(InnerType) , 
    unique_typed_id("o",InnerType,TempOut).
get_single_operator_constraint_aux3(image-SetType,TempIn1,TempIn2,TempOut,image,[]) :- 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(couple(A,B)),TempIn1) , 
    unique_typed_id("i",set(A),TempIn2) , 
    unique_typed_id("o",set(B),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[domain_restriction,domain_subtraction]) , 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(A),TempIn1) , 
    unique_typed_id("i",set(couple(A,B)),TempIn2) , 
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[range_restriction,range_subtraction]) , 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(couple(A,B)),TempIn1) , 
    unique_typed_id("i",set(B),TempIn2) , 
    unique_typed_id("o",set(couple(A,B)),TempOut).
% function inverse ~
get_single_operator_constraint_aux3(reverse-SetType,TempIn,_,TempOut,reverse,[]) :- 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(couple(A,B)),TempIn) , 
    unique_typed_id("o",set(couple(B,A)),TempOut).
get_single_operator_constraint_aux3(domain-SetType,TempIn,_,TempOut,domain,[]) :- 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(couple(A,B)),TempIn) , 
    unique_typed_id("o",set(A),TempOut).
get_single_operator_constraint_aux3(range-SetType,TempIn,_,TempOut,range,[]) :- 
    get_fn_types(SetType,A,B) , 
    unique_typed_id("i",set(couple(A,B)),TempIn) , 
    unique_typed_id("o",set(B),TempOut).
% .. 
get_single_operator_constraint_aux3(cartesian_product-_,TempIn1,TempIn2,TempOut,cartesian_product,[]) :- 
    unique_typed_id("i",set(A),TempIn1) , 
    unique_typed_id("i",set(B),TempIn2) , 
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[union,intersection,set_subtraction]) , 
    unique_typed_id("i",SetType,TempIn1) , 
    unique_typed_id("i",SetType,TempIn2) , 
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(overwrite-SetType,TempIn1,TempIn2,TempOut,overwrite,[]) :- 
    get_fn_types(SetType,integer,Type) , 
    unique_typed_id("i",seq(Type),TempIn1) , 
    unique_typed_id("i",set(couple(integer,Type)),TempIn2) , 
    unique_typed_id("o",seq(Type),TempOut).
get_single_operator_constraint_aux3(concat-SetType,TempIn1,TempIn2,TempOut,concat,[]) :- 
    get_fn_types(SetType,integer,Type) , 
    unique_typed_id("i",seq(Type),TempIn1) , 
    unique_typed_id("i",seq(Type),TempIn2) , 
    unique_typed_id("o",seq(Type),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :- 
    member(Operator,[first,last]) , 
    unique_typed_id("i",SetType,TempIn) , 
    SetType = set(InnerType) , 
    unique_typed_id("o",InnerType,TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :- 
    member(Operator,[front,tail]) , 
    unique_typed_id("i",SetType,TempIn) , 
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(insert_front-SetType,TempIn1,TempIn2,TempOut,insert_front,[]) :- 
    SetType = set(couple(integer,InnerType)) , 
    unique_typed_id("i",InnerType,TempIn1) , 
    unique_typed_id("i",SetType,TempIn2) , 
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(insert_tail-SetType,TempIn1,TempIn2,TempOut,insert_tail,[]) :- 
    SetType = set(couple(integer,InnerType)) , 
    unique_typed_id("i",SetType,TempIn1) , 
    unique_typed_id("i",InnerType,TempIn2) , 
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :- 
    member(Operator,[restrict_front,restrict_tail]) ,     
    unique_typed_id("i",SetType,TempIn1) , 
    unique_typed_id("i",integer,TempIn2) , 
    unique_typed_id("o",SetType,TempOut).

% (INTEGER,INTEGER) :: INTEGER
get_single_arithmetic_constraint(ArithmeticOp,Constraint) :- 
    unique_typed_id("i",integer,TempIn1) , 
    unique_typed_id("i",integer,TempIn2) , 
    unique_typed_id("o",integer,TempOut) , 
    get_component_id(arit,ArithmeticOp,ArithID) , 
    atom_number_concat(ArithmeticOp,ArithID,NArithmetic) , 
    % set information of identifier nodes 
    add_texpr_infos(TempIn1,[synthesis(NArithmetic,left_input)],In1) , 
    add_texpr_infos(TempIn2,[synthesis(NArithmetic,right_input)],In2) , 
    add_texpr_infos(TempOut,[synthesis(NArithmetic,output)],Out) , 
    Node =.. [ArithmeticOp,In1,In2] , 
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).

get_fn_types(seq(Type),integer,Type).
get_fn_types(set(couple(A,B)),A,B).

% add prefix "l" to Var for connectivity constraint and create information for 
% input variables or adopt the temporary variables information (these are created during 
% operator definition and encode whether temporary variable is left/right input or output)
get_location_var_from_ast_node(b(identifier(Name),Type,Info),b(identifier(LocationName),integer,TypedInfo)) :- 
    ! , atom_concat(l,Name,LocationName) , 
    append(Info,[synthesis(type,Type)],TypedInfo).
get_location_var_from_ast_node(b(Node,Type,Info),b(identifier(LocationName),integer,TypedInfo)) :- 
    get_location_var_from_ast_node_aux(Node,Type,ValueCode) , 
    % set input/output enumeration as prefix to get unique name
    ( member(synthesis(input(C)),Info)
    ->  number_codes(C,CCode)
    ;   (member(synthesis(output(C)),Info) , number_codes(C,[TCCode]) , CCode = [111,TCCode])) , 
    append(Info,[synthesis(type,Type)],TypedInfo) , 
    append(CCode,ValueCode,Codes) , 
    atom_codes(LocationName,[108|Codes]).

get_location_var_from_ast_node_aux(integer(Value),integer,ValueCode) :- 
    number_codes(Value,ValueCode).
get_location_var_from_ast_node_aux(_,set(_),ValueCode) :- 
    atom_codes(set,ValueCode).
get_location_var_from_ast_node_aux(_,seq(_),ValueCode) :- 
    atom_codes(seq,ValueCode).
get_location_var_from_ast_node_aux(Bool,boolean,ValueCode) :- 
    atom_codes(Bool,ValueCode).
    
% enumerate used components for identifier nodes information so we definitely 
% know which input/output locations belong together if a component is used multiple times
get_component_id(op,Operator,OpID) :- 
    usedOperators(UsedOps) , 
    retract(usedOperators(_)) , 
    findall(Operator,member(Operator,UsedOps),OpList) , 
    length(OpList,OpAmount) , 
    OpID is OpAmount + 1 , 
    assert(usedOperators([Operator|UsedOps])).
get_component_id(arit,Operator,OpID) :- 
    usedArithmetic(UsedOps) , 
    retract(usedArithmetic(_)) , 
    findall(Operator,member(Operator,UsedOps),OpList) , 
    length(OpList,OpAmount) , 
    OpID is OpAmount + 1 , 
    assert(usedArithmetic([Operator|UsedOps])).

% evaluate M := |I| + N
% the value of M determines the amount of lines in the program 
get_line_amount((Operators,Arithmetic),InputVars,M) :-
    length(InputVars,AmountIn) , 
    length(Operators,N1) , 
    length(Arithmetic,N2) , 
    N is N1 + N2 , 
    M is AmountIn + N.

% set the location variables of the program parameters and output given by I/O-Examples 
% each input parameters are defined in the first lines of the program
set_input_location_vars(InputValueNodes,InputLocations) :- 
    set_input_location_vars(InputValueNodes,0,[],InputLocationsList) , 
    conjunct_predicates(InputLocationsList,InputLocations).
set_input_location_vars([],_,Acc,Acc).
set_input_location_vars([InputValueNode|R],C,Acc,InputLocations) :- 
    get_location_var_from_ast_node(InputValueNode,InputLocationNode) , 
    InputLocation = b(equal(InputLocationNode,b(integer(C),integer,[])),pred,[]) , 
    C1 is C + 1 , 
    set_input_location_vars(R,C1,[InputLocation|Acc],InputLocations).
% we dispose output values in descending order starting at the last line of the program
set_output_location_vars(OutputValueNodes,M,OutputLocations) :- 
    M1 is M - 1 , 
    set_output_location_vars(OutputValueNodes,M1,[],OutputLocationsList) , 
    conjunct_predicates(OutputLocationsList,OutputLocations).
set_output_location_vars([],_,Acc,Acc).
set_output_location_vars([OutputValueNode|R],C,Acc,OutputLocations) :- 
    get_location_var_from_ast_node(OutputValueNode,OutputLocationNode) , 
    OutputLocation = b(equal(OutputLocationNode,b(integer(C),integer,[])),pred,[]) , 
    C1 is C - 1 , 
    set_output_location_vars(R,C1,[OutputLocation|Acc],OutputLocations).

% enumerate input values in info list, variable names are set during conversion to ast (see location_vars_to_program)
transition_values_to_ast(InputValues,StateType,InputNodes) :-  
    transition_values_to_ast(InputValues,StateType,0,InputNodes).
transition_values_to_ast([],_,_,[]).
transition_values_to_ast([Value|R],input,C,[b(Node,Type,[synthesis(input(C))|Info])|T]) :- 
    value_to_ast(Value,b(Node,Type,Info)) , 
    C1 is C + 1 , 
    transition_values_to_ast(R,input,C1,T).
transition_values_to_ast([Value|R],output,C,[b(Node,Type,[synthesis(output(C))|Info])|T]) :- 
    value_to_ast(Value,b(Node,Type,Info)) , 
    C1 is C + 1 , 
    transition_values_to_ast(R,output,C1,T).

value_to_ast(b(Node,Type,Info),b(Node,Type,Info)) :- !.
value_to_ast(Value,AST) :- 
    value_to_ast_aux(Value,AST).

value_to_ast_aux((Value1,Value2),b(couple(Node1,Node2),couple(Type1,Type2),[])) :- 
    value_to_ast_aux(Value1,Node1) , 
    value_to_ast_aux(Value2,Node2) , 
    Node1 = b(_,Type1,_) , Node2 = b(_,Type2,_).
value_to_ast_aux(Value,b(integer(Value),integer,[])) :- 
    integer(Value).
value_to_ast_aux(int(Value),b(integer(Value),integer,[])).
value_to_ast_aux(Bool,b(boolean_true,boolean,[])) :- 
    Bool = 'TRUE' ; Bool = pred_true.
value_to_ast_aux(Bool,b(boolean_false,boolean,[])) :- 
    Bool = 'FALSE' ; Bool = pred_false.
value_to_ast_aux(string(S),b(string(S),string,[])).
value_to_ast_aux(Atom,b(string(Atom),string,[])) :- 
    Atom \= [] , atom(Atom).
value_to_ast_aux(ListOrSet,b(value([]),set(_),[])) :- 
    (ListOrSet = [] ; ListOrSet = {}).
value_to_ast_aux(avl_set(Tree),b(value(avl_set(Tree)),set(Type),[])) :- 
    Tree = node(_,_,_,_,_) , 
    avl_to_list(Tree,[Value-_|_]) , 
    value_to_ast(Value,b(_,Type,_)).
value_to_ast_aux(List,b(set_extension(NodeList),set(Type),[])) :- 
    is_list(List) , 
    maplist(value_to_ast,List,NodeList) , 
    NodeList = [b(_,Type,_)|_].
% ...

inverse_operator(greater,less_equal).
inverse_operator(equal,not_equal).
inverse_operator(not_equal,equal).
inverse_operator(greater_equal,less).
inverse_operator(subset,not_subset).
inverse_operator(subset_strict,not_subset_strict).

atom_number_concat(Atom,Number,Concatenation) :- 
    number_codes(Number,NumberCodes) , 
    atom_codes(Atom,AtomCodes) , 
    append(AtomCodes,NumberCodes,ConcatCodes) , 
    atom_codes(Concatenation,ConcatCodes). 

delete_numbers_from_atom(NumberAtom,Atom) :- 
    atom_codes(NumberAtom,Codes) , 
    delete_numbers_from_codelist(Codes,NewCodes) , 
    atom_codes(Atom,NewCodes).
delete_numbers_from_codelist([],[]).
delete_numbers_from_codelist([Code|R],[Code|NR]) :- 
    (Code > 57 ; Code < 48) , ! , 
    delete_numbers_from_codelist(R,NR).
delete_numbers_from_codelist([_|R],NR) :- 
    delete_numbers_from_codelist(R,NR).

% search identifier nodes within an ast and keep each nodes information
find_typed_identifier_uses_with_info(AST,UsedIDs) :- 
    find_typed_identifier_uses_with_info2(AST,TempUsedIDs) , 
    remove_dups(TempUsedIDs,UsedIDs).

find_typed_identifier_uses_with_info2(List,IDs) :- 
    is_list(List) , 
    findall(ID,(member(ID,List) , ID = b(identifier(_),_,_)),IDs).
find_typed_identifier_uses_with_info2(b(identifier(IDName),Type,Info),[b(identifier(IDName),Type,Info)]).
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :- 
    Node =.. [_,Arg] , 
    find_typed_identifier_uses_with_info2(Arg,UsedIDs).
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :- 
    Node =.. [_,Arg1,Arg2] ,
    find_typed_identifier_uses_with_info2(Arg1,UsedIDs1) , 
    find_typed_identifier_uses_with_info2(Arg2,UsedIDs2) , 
    append(UsedIDs1,UsedIDs2,UsedIDs).
% rlevent 
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :- 
    Node =.. [_,_,_,_,_,Body|_] ,
    find_typed_identifier_uses_with_info2(Body,UsedIDs).
find_typed_identifier_uses_with_info2(_,[]).