% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(tools_strings,
                 [string_concatenate/3,
                  ajoin/2, ajoin_with_limit/3, ajoin_with_sep/3,
                  safe_name/2,
                  predicate_functor/3,
                  atom_codes_with_limit/2, atom_codes_with_limit/3,
                  truncate_atom/3
                  ]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'A few utilities on strings/atoms seperated out from tools.pl to avoid cyclic module dependencies.').


:- mode string_concatenate(i,i,o).
string_concatenate(X,Y,XY) :- atom(X),atom(Y),!, atom_concat(X,Y,XY).
string_concatenate(X,Y,XY) :- 
    safe_name(X,Xs),safe_name(Y,Ys),append(Xs,Ys,XYs),atom_codes(XY,XYs).

% tests are stored in tools.pl to avoid cyclic module dependencies
%:- assert_must_succeed((tools_strings: ajoin_with_sep([link,a,xa],'.',Text), Text == 'link.a.xa')).
%:- assert_must_succeed((tools_strings: ajoin_with_sep([link],'.',Text), Text == 'link')).

ajoin_with_sep(List,Sep,Text) :- ajoin_with_sep_aux(List,'',Sep,Text).

ajoin_with_sep_aux([],Txt,_Sep,Text) :- Text=Txt.
ajoin_with_sep_aux([H|T],Txt,Sep,Text) :- 
   ( Txt == '' -> toAtom(H,TTxt)
   ; ajoin([Txt,Sep,H],TTxt)), ajoin_with_sep_aux(T,TTxt,Sep,Text).


/* Concats a list of atoms, but checks if numbers or compound terms are
   involved, which are converted to simple atoms */
ajoin([],Text) :- Text=''.
ajoin([A|T],Result) :- toAtom(A,Atom),
    ajoin_aux(T,Atom,Result).
ajoin_aux([],Acc,Result) :- Result=Acc.
ajoin_aux([A|T],Acc,Result) :- toAtom(A,Atom),
    safe_atom_concat(Acc,Atom,NewAcc,ExcOcc),
    (ExcOcc==true -> Result = NewAcc ; ajoin_aux(T,NewAcc,Result)).

:- use_module(library(codesio),[write_to_codes/2]).

toAtom(Number,Atom) :- number(Number),!,number_chars(Number,C),atom_chars(Atom,C).
toAtom(Atom,Atom) :- atomic(Atom),!.
toAtom(Term,Atom) :- write_to_codes(Term,Codes),safe_atom_codes(Atom,Codes).

% a copy of safe_atom_codes/2 from tools to avoid module dependency on error_manager
safe_atom_codes(V,C) :- var(V),var(C),!,
   print_error('Variables in call: '),print_error(safe_atom_codes(V,C)),
   C='$VARIABLE$'.
safe_atom_codes(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),atom_codes(A,C),
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).

% will concatenate until the Limit is reached or exceeded; it may produce atoms longer than Limit 
% (if first atom already longer than limit + it adds ...
%:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],100,Text), Text == 'ABCD')).
%:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],2,Text), Text == 'AB...')).

ajoin_with_limit([],_,Text) :- 
        !,Text=''.
ajoin_with_limit([A|T],Limit,Result) :- toAtom(A,Atom), atom_length(Atom,Len),
    ajoin_with_limit_aux(T,Atom,Len,Limit,Result).
ajoin_with_limit_aux([],Acc,_,_,Result) :- !,Result=Acc.
ajoin_with_limit_aux([A|T],Acc,Len,Limit,Result) :- toAtom(A,Atom),
    atom_length(Atom,AL), L1 is Len+AL,
    (L1 > Limit -> safe_atom_concat(Acc,'...',Result,_)
      ; safe_atom_concat(Acc,Atom,NewAcc,ExcOcc),
        (ExcOcc==true -> Result = NewAcc ; ajoin_with_limit_aux(T,NewAcc,L1,Limit,Result))).


safe_atom_concat(A,B,C,ExceptionOcc) :-
    on_exception(error(representation_error(max_atom_length),_),atom_concat(A,B,C),
       (print(exception(max_atom_length)),nl,A=C,ExceptionOcc=true)).


:- use_module(tools_printing,[print_error/1]).
safe_name(X,N) :- atomic(X),!, name(X,N).
safe_name(X,N) :- var(X),!,atom_codes(var,N).  /* ensure that "5.10" is converted to '5.10' and not 5.1 */
safe_name(lambda_res(X),[114,101,115,95|N]) :- !, name(X,N).
safe_name(X,N) :- functor(X,F,_),name(F,N), print_error(non_atomic_in_safe_name(X)).


predicate_functor(X,F,N) :- var(X),!, print_error(var_in_predicate_functor),F='$VAR',N=0.
predicate_functor(_Module:Pred,F,N) :- !,predicate_functor(Pred,F,N).
predicate_functor(P,F,N) :- functor(P,F,N).



atom_codes_with_limit(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),
                  atom_codes(A,C),
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).


atom_codes_with_limit(A,Limit,Codes) :- var(A), Limit >= 0, !,
    truncate_codes(Codes,Limit,TCodes,_),
    atom_codes(A,TCodes).
atom_codes_with_limit(A,Limit,Codes) :- Limit < 1, !, atom_codes(A,Codes).
atom_codes_with_limit(A,Limit,Codes) :- atom_codes(A,Codes1),
    truncate_codes(Codes,Limit,Codes1,_).


truncate_codes([],_,[],false).
truncate_codes([H|T],Count,Res,Trunc) :-
   Count<1 -> Res = [46,46,46],Trunc=true /* '...' */
           ; Res = [H|TT], C1 is Count-1, truncate_codes(T,C1,TT,Trunc).
           
%:- assert_must_succeed((tools_strings:truncate_atom(abcd,100,Text), Text == 'abcd')).
%:- assert_must_succeed((tools_strings:truncate_atom(abcd,2,Text), Text == 'ab...')).
%:- assert_must_succeed((tools_strings:truncate_atom(abcd,0,Text), Text == '...')).
truncate_atom(Atom,Limit,NewAtom) :-
   atom_codes(Atom,Codes),
   truncate_codes(Codes,Limit,TCodes,Trunc),
   (Trunc=true -> atom_codes(NewAtom,TCodes) ; NewAtom=Atom).
